/*==========================================================================
 *
 *  Copyright (C) 1996 Brian King
 *
 *  File:       win32gui.c
 *  Content:    Win32-specific gui features for UAE port.
 *
 ***************************************************************************/

#define FRONTEND 0

#define _WIN32_WINNT 0x600

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <winspool.h>
#include <winuser.h>
#include <mmsystem.h>
#include <commctrl.h>
#include <commdlg.h>
#include <dlgs.h>
#include <process.h>
#include <prsht.h>
#include <richedit.h>
#include <shellapi.h>
#include <Shlobj.h>
#include <shlwapi.h>
#include <ddraw.h>
#include <shobjidl.h>
#include <dbt.h>

#include "resource.h"
#include "sysconfig.h"
#include "sysdeps.h"
#include "gui.h"
#include "options.h"
#include "memory.h"
#include "custom.h"
#include "events.h"
#include "newcpu.h"
#include "traps.h"
#include "disk.h"
#include "uae.h"
#include "threaddep/thread.h"
#include "filesys.h"
#include "autoconf.h"
#include "inputdevice.h"
#include "xwin.h"
#include "keyboard.h"
#include "zfile.h"
#include "parallel.h"
#include "audio.h"
#include "arcadia.h"
#include "fsdb.h"

#include "dxwrap.h"
#include "win32.h"
#include "registry.h"
#include "picasso96_win.h"
#include "win32gui.h"
#include "win32gfx.h"
#include "sounddep/sound.h"
#include "od-win32/parser.h"
#include "od-win32/ahidsound.h"
#include "target.h"
#include "savestate.h"
#include "avioutput.h"
#include "opengl.h"
#include "direct3d.h"
#include "akiko.h"
#include "cdtv.h"
#include "gfxfilter.h"
#include "driveclick.h"
#ifdef PROWIZARD
#include "moduleripper.h"
#endif
#include "catweasel.h"
#include "lcd.h"
#include "uaeipc.h"
#include "crc32.h"
#include "rp.h"

#define ARCHIVE_STRING "*.zip;*.7z;*.rar;*.lha;*.lzh;*.lzx"

#define DISK_FORMAT_STRING "(*.adf;*.adz;*.gz;*.dms;*.fdi;*.ipf;*.exe)\0*.adf;*.adz;*.gz;*.dms;*.fdi;*.ipf;*.exe;*.ima;*.wrp;*.dsq;" ARCHIVE_STRING "\0"
#define ROM_FORMAT_STRING "(*.rom;*.roz)\0*.rom;*.roz;" ARCHIVE_STRING "\0"
#define USS_FORMAT_STRING_RESTORE "(*.uss)\0*.uss;*.gz;"  ARCHIVE_STRING "\0"
#define USS_FORMAT_STRING_SAVE "(*.uss)\0*.uss\0"
#define HDF_FORMAT_STRING "(*.hdf;*.rdf;*.hdz;*.rdz)\0*.hdf;*.rdf;*.hdz;*.rdz\0"
#define INP_FORMAT_STRING "(*.inp)\0*.inp\0"
#define CONFIG_HOST "Host"
#define CONFIG_HARDWARE "Hardware"

static char szNone[MAX_DPATH];

static int allow_quit;
static int restart_requested;
static int full_property_sheet = 1;
static struct uae_prefs *pguiprefs;
struct uae_prefs workprefs;
static int currentpage;
static int qs_request_reset;
static int qs_override;
int gui_active;

extern HWND (WINAPI *pHtmlHelp)(HWND, LPCSTR, UINT, LPDWORD);

#undef HtmlHelp
#ifndef HH_DISPLAY_TOPIC
#define HH_DISPLAY_TOPIC 0
#endif
#define HtmlHelp(a,b,c,d) if(pHtmlHelp) (*pHtmlHelp)(a,b,c,(LPDWORD)d); else \
{ char szMessage[MAX_DPATH]; WIN32GUI_LoadUIString(IDS_NOHELP, szMessage, MAX_DPATH); gui_message(szMessage); }

extern HWND hAmigaWnd;
extern char help_file[MAX_DPATH];

extern int mouseactive;

char config_filename[MAX_DPATH] = "";

#define Error(x) MessageBox(NULL, (x), "WinUAE Error", MB_OK)

void WIN32GUI_LoadUIString(DWORD id, char *string, DWORD dwStringLen)
{
    if(LoadString(hUIDLL ? hUIDLL : hInst, id, string, dwStringLen) == 0)
	LoadString(hInst, id, string, dwStringLen);
}

static int C_PAGES;
#define MAX_C_PAGES 30
static int LOADSAVE_ID = -1, MEMORY_ID = -1, KICKSTART_ID = -1, CPU_ID = -1,
    DISPLAY_ID = -1, HW3D_ID = -1, CHIPSET_ID = -1, CHIPSET2_ID = -1, SOUND_ID = -1, FLOPPY_ID = -1, DISK_ID = -1,
    HARDDISK_ID = -1, PORTS_ID = -1, INPUT_ID = -1, MISC1_ID = -1, MISC2_ID = -1, AVIOUTPUT_ID = -1,
    PATHS_ID = -1, QUICKSTART_ID = -1, ABOUT_ID = -1, FRONTEND_ID = -1;
static HWND pages[MAX_C_PAGES];
#define MAX_IMAGETOOLTIPS 10
static HWND guiDlg, panelDlg, ToolTipHWND;
static HACCEL hAccelTable;
struct ToolTipHWNDS {
    WNDPROC proc;
    HWND hwnd;
    int imageid;
};
static struct ToolTipHWNDS ToolTipHWNDS2[MAX_IMAGETOOLTIPS + 1];

static void ew (HWND hDlg, DWORD id, int enable)
{
    HWND w = GetDlgItem(hDlg, id);
    if (!w)
	return;
    if (!enable && w == GetFocus())
	SendMessage(hDlg, WM_NEXTDLGCTL, 0, FALSE);
    EnableWindow(w, !!enable);
}

void write_disk_history (void)
{
    int i, j;
    char tmp[16];
    UAEREG *fkey;

    fkey = regcreatetree (NULL, "DiskImageMRUList");
    if (fkey == NULL)
	return;
    j = 1;
    for (i = 0; i <= MAX_PREVIOUS_FLOPPIES; i++) {
	char *s = DISK_history_get(i);
	if (s == 0 || strlen(s) == 0)
	    continue;
	sprintf (tmp, "Image%02d", j);
	regsetstr (fkey, tmp, s);
	j++;
   }
    while (j <= MAX_PREVIOUS_FLOPPIES) {
	char *s = "";
	sprintf (tmp, "Image%02d", j);
	regsetstr (fkey, tmp, s);
	j++;
    }
    regclosetree (fkey);
}

void reset_disk_history (void)
{
    int i;

    for (i = 0; i < MAX_PREVIOUS_FLOPPIES; i++)
	DISK_history_add (NULL, i);
    write_disk_history();
}

UAEREG *read_disk_history (void)
{
    static int regread;
    char tmp2[1000];
    DWORD size2;
    int idx, idx2;
    UAEREG *fkey;
    char tmp[1000];
    DWORD size;

    fkey = regcreatetree (NULL, "DiskImageMRUList");
    if (fkey == NULL || regread)
	return fkey;

    idx = 0;
    for (;;) {
	size = sizeof (tmp);
	size2 = sizeof (tmp2);
	if (!regenumstr (fkey, idx, tmp, &size, tmp2, &size2))
	    break;
	if (strlen (tmp) == 7) {
	    idx2 = atol (tmp + 5) - 1;
	    if (idx2 >= 0)
		DISK_history_add (tmp2, idx2);
	}
	idx++;
    }
    regread = 1;
    return fkey;
}

void exit_gui (int ok)
{
    if (!gui_active)
	return;
    if (guiDlg == NULL)
	return;
    SendMessage (guiDlg, WM_COMMAND, ok ? IDOK : IDCANCEL, 0);
}

static int getcbn (HWND hDlg, int v, char *out, int len)
{
    LRESULT val = SendDlgItemMessage (hDlg, v, CB_GETCURSEL, 0, 0L);
    out[0] = 0;
    if (val == CB_ERR) {
	SendDlgItemMessage (hDlg, v, WM_GETTEXT, (WPARAM)len, (LPARAM)out);
	return 1;
    } else {
	val = SendDlgItemMessage (hDlg, v, CB_GETLBTEXT, (WPARAM)val, (LPARAM)out);
	return 0;
    }
}

/* base Drag'n'Drop code borrowed from http://www.codeproject.com/listctrl/jianghong.asp */

static int bDragging = 0;
static HIMAGELIST hDragImageList;
static int drag_start (HWND hWnd, HWND hListView, LPARAM lParam)
{
    POINT p, pt;
    int bFirst, iPos, iHeight;
    HIMAGELIST hOneImageList, hTempImageList;
    IMAGEINFO imf;

    // You can set your customized cursor here
    p.x = 8;
    p.y = 8;
    // Ok, now we create a drag-image for all selected items
    bFirst = TRUE;
    iPos = ListView_GetNextItem(hListView, -1, LVNI_SELECTED);
    while (iPos != -1) {
        if (bFirst) {
	    // For the first selected item,
	    // we simply create a single-line drag image
	    hDragImageList = ListView_CreateDragImage(hListView, iPos, &p);
	    ImageList_GetImageInfo(hDragImageList, 0, &imf);
	    iHeight = imf.rcImage.bottom;
	    bFirst = FALSE;
	} else {
	    // For the rest selected items,
	    // we create a single-line drag image, then
	    // append it to the bottom of the complete drag image
	    hOneImageList = ListView_CreateDragImage(hListView, iPos, &p);
	    hTempImageList = ImageList_Merge(hDragImageList, 0, hOneImageList, 0, 0, iHeight);
	    ImageList_Destroy(hDragImageList);
	    ImageList_Destroy(hOneImageList);
	    hDragImageList = hTempImageList;
	    ImageList_GetImageInfo(hDragImageList, 0, &imf);
	    iHeight = imf.rcImage.bottom;
	}
	iPos = ListView_GetNextItem(hListView, iPos, LVNI_SELECTED);
    }

    // Now we can initialize then start the drag action
    ImageList_BeginDrag(hDragImageList, 0, 0, 0);

    pt = ((NM_LISTVIEW*) ((LPNMHDR)lParam))->ptAction;
    ClientToScreen(hListView, &pt);

    ImageList_DragEnter(NULL, pt.x, pt.y);

    bDragging = TRUE;

    // Don't forget to capture the mouse
    SetCapture (hWnd);

    return 1;
}

static int drag_end (HWND hWnd, HWND hListView, LPARAM lParam, int **draggeditems)
{
    int iPos, cnt;
    LVHITTESTINFO lvhti;
    LVITEM  lvi;

    *draggeditems = NULL;
    if (!bDragging)
	return -1;
    // End the drag-and-drop process
    bDragging = FALSE;
    ImageList_DragLeave(hListView);
    ImageList_EndDrag();
    ImageList_Destroy(hDragImageList);
    ReleaseCapture();

    // Determine the dropped item
    lvhti.pt.x = LOWORD(lParam);
    lvhti.pt.y = HIWORD(lParam);
    ClientToScreen(hWnd, &lvhti.pt);
    ScreenToClient(hListView, &lvhti.pt);
    ListView_HitTest(hListView, &lvhti);

    // Out of the ListView?
    if (lvhti.iItem == -1)
	return -1;
    // Not in an item?
    if ((lvhti.flags & LVHT_ONITEMLABEL) == 0 &&  (lvhti.flags & LVHT_ONITEMSTATEICON) == 0)
	return -1;
    // Dropped item is selected?
    lvi.iItem = lvhti.iItem;
    lvi.iSubItem = 0;
    lvi.mask = LVIF_STATE;
    lvi.stateMask = LVIS_SELECTED;
    ListView_GetItem(hListView, &lvi);
    if (lvi.state & LVIS_SELECTED)
	return -1;
    // Rearrange the items
    iPos = ListView_GetNextItem(hListView, -1, LVNI_SELECTED);
    cnt = 0;
    while (iPos != -1) {
	iPos = ListView_GetNextItem(hListView, iPos, LVNI_SELECTED);
	cnt++;
    }
    if (cnt == 0)
	return -1;
    *draggeditems = xmalloc (sizeof (int) * (cnt + 1));
    iPos = ListView_GetNextItem(hListView, -1, LVNI_SELECTED);
    cnt = 0;
    while (iPos != -1) {
	(*draggeditems)[cnt++] = iPos;
	iPos = ListView_GetNextItem(hListView, iPos, LVNI_SELECTED);
    }
    (*draggeditems)[cnt] = -1;
    return lvhti.iItem;
}
static int drag_move (HWND hWnd, LPARAM lParam)
{
    POINT p;

    if (!bDragging)
	return 0;
    p.x = LOWORD(lParam);
    p.y = HIWORD(lParam);
    ClientToScreen(hWnd, &p);
    ImageList_DragMove(p.x, p.y);
    return 1;
}

static HWND cachedlist = NULL;

#define MIN_CHIP_MEM 0
#define MAX_CHIP_MEM 5
#define MIN_FAST_MEM 0
#define MAX_FAST_MEM 4
#define MIN_SLOW_MEM 0
#define MAX_SLOW_MEM 4
#define MIN_Z3_MEM 0
#define MAX_Z3_MEM ((max_z3fastmem >> 20) < 512 ? 9 : ((max_z3fastmem >> 20) < 1024 ? 10 : ((max_z3fastmem >> 20) < 2048) ? 11 : 12))
#define MIN_P96_MEM 0
#define MAX_P96_MEM ((max_z3fastmem >> 20) < 512 ? 8 : ((max_z3fastmem >> 20) < 1024 ? 9 : ((max_z3fastmem >> 20) < 2048) ? 10 : 11))
#define MIN_MB_MEM 0
#define MAX_MB_MEM 7

#define MIN_M68K_PRIORITY 1
#define MAX_M68K_PRIORITY 16
#define MIN_CACHE_SIZE 0
#define MAX_CACHE_SIZE 8
#define MIN_REFRESH_RATE 1
#define MAX_REFRESH_RATE 10
#define MIN_SOUND_MEM 0
#define MAX_SOUND_MEM 6

struct romscandata {
    UAEREG *fkey;
    int got;
};

static struct romdata *scan_single_rom_2 (struct zfile *f)
{
    uae_u8 buffer[20] = { 0 };
    uae_u8 *rombuf;
    int cl = 0, size;
    struct romdata *rd = 0;

    zfile_fseek (f, 0, SEEK_END);
    size = zfile_ftell (f);
    zfile_fseek (f, 0, SEEK_SET);
    if (size > 524288 * 2)  {/* don't skip KICK disks or 1M ROMs */
	write_log ("'%s': too big %d, ignored\n", zfile_getname(f), size);
	return 0;
    }
    zfile_fread (buffer, 1, 11, f);
    if (!memcmp (buffer, "KICK", 4)) {
	zfile_fseek (f, 512, SEEK_SET);
	if (size > 262144)
	    size = 262144;
    } else if (!memcmp (buffer, "AMIROMTYPE1", 11)) {
	cl = 1;
	size -= 11;
    } else {
	zfile_fseek (f, 0, SEEK_SET);
    }
    rombuf = xcalloc (size, 1);
    if (!rombuf)
	return 0;
    zfile_fread (rombuf, 1, size, f);
    if (cl > 0) {
	if (decode_cloanto_rom_do (rombuf, size, size))
	    cl = 0;
    }
    if (!cl)
	rd = getromdatabydata (rombuf, size);
    if (!rd) {
	;//write_log ("Unknown: Size=%d, Name='%s'\nCRC32=%08X SHA1=%s\n",
	 //  size, zfile_getname(f), get_crc32(rombuf, size), get_sha1_txt(rombuf, size));
    }
    free (rombuf);
    return rd;
}

static struct romdata *scan_single_rom (char *path)
{
    struct zfile *z;
    char tmp[MAX_DPATH];
    struct romdata *rd;

    strcpy (tmp, path);
    rd = scan_arcadia_rom (tmp, 0);
    if (rd)
	return rd;
    rd = getromdatabypath(path);
    if (rd && rd->crc32 == 0xffffffff)
	return rd;
    z = zfile_fopen (path, "rb");
    if (!z)
	return 0;
    return scan_single_rom_2 (z);
}

static int addrom (UAEREG *fkey, struct romdata *rd, char *name)
{
    char tmp1[MAX_DPATH], tmp2[MAX_DPATH];

    sprintf (tmp1, "ROM%03d", rd->id);
    if (rd->group) {
	char *p = tmp1 + strlen (tmp1);
	sprintf (p, "_%02d_%02d", rd->group >> 16, rd->group & 65535);
    }
    if (regexists (fkey, tmp1))
	return 0;
    tmp2[0] = 0;
    if (name)
	strcpy (tmp2, name);
    if (rd->crc32 == 0xffffffff) {
	if (rd->configname)
	    sprintf(tmp2, ":%s", rd->configname);
	else
	    sprintf(tmp2, ":ROM%03d", rd->id);
    }
    if (!regsetstr (fkey, tmp1, tmp2))
	return 0;
    return 1;
}

static int isromext(char *path)
{
    char *ext;
    int i;

    if (!path)
	return 0;
    ext = strrchr (path, '.');
    if (!ext)
	return 0;
    ext++;

    if (!stricmp (ext, "rom") ||  !stricmp (ext, "adf") || !stricmp (ext, "key")
	|| !stricmp (ext, "a500") || !stricmp (ext, "a1200") || !stricmp (ext, "a4000"))
	    return 1;
    for (i = 0; uae_archive_extensions[i]; i++) {
	if (!stricmp (ext, uae_archive_extensions[i]))
	    return 1;
    }
    return 0;
}

static int scan_rom_2 (struct zfile *f, struct romscandata *rsd)
{
    char *path = zfile_getname(f);
    struct romdata *rd;

    if (!isromext(path))
	return 0;
    rd = scan_single_rom_2 (f);
    if (rd) {
	addrom (rsd->fkey, rd, path);
	rsd->got = 1;
    }
    return 0;
}

static int scan_rom (char *path, UAEREG *fkey)
{
    struct romscandata rsd = { fkey, 0 };
    struct romdata *rd;
    int cnt = 0;

    if (!isromext(path)) {
	//write_log("ROMSCAN: skipping file '%s', unknown extension\n", path);
	return 0;
    }
    for (;;) {
	char tmp[MAX_DPATH];
	strcpy (tmp, path);
	rd = scan_arcadia_rom (tmp, cnt++);
	if (rd) {
	    if (!addrom (fkey, rd, tmp))
		return 1;
	    continue;
	}
	break;
    }
    zfile_zopen (path, scan_rom_2, &rsd);
    return rsd.got;
}

static int listrom (int *roms)
{
    int i;

    i = 0;
    while (roms[i] >= 0) {
	struct romdata *rd = getromdatabyid (roms[i]);
	if (rd && romlist_get (rd))
	    return 1;
	i++;
    }
    return 0;
}

static void show_rom_list (void)
{
    char *p;
    char unavail[MAX_DPATH], avail[MAX_DPATH];
    char *p1, *p2;
    int *rp;
    int romtable[] = {
	5, 4, -1, -1, // A500 1.2
	6, 32, -1, -1, // A500 1.3
	7, -1, -1, // A500+
	8, 9, 10, -1, -1, // A600
	23, 24, -1, -1, // A1000
	11, 31, 15, -1, -1, // A1200
	59, 71, 61, -1, -1, // A3000
	16, 46, 31, 13, 12, -1, -1, // A4000
	18, -1, 19, -1, -1, // CD32
	20, 21, 22, -1, 6, 32, -1, -1, // CDTV
	49, 50, 51, -1, 5, 4, -1, -1, // ARCADIA
	46, 16, 17, 31, 13, 12, -1, -1, // highend, any 3.x A4000
	53, 54, 55, -1, -1, // A590/A2091
	//56, 57, -1, -1, // A4091
	0, 0, 0
    };

    WIN32GUI_LoadUIString (IDS_ROM_AVAILABLE, avail, sizeof (avail));
    WIN32GUI_LoadUIString (IDS_ROM_UNAVAILABLE, unavail, sizeof (avail));
    strcat (avail, "\n");
    strcat (unavail, "\n");
    p1 = "A500 Boot ROM 1.2\0A500 Boot ROM 1.3\0A500+\0A600\0A1000\0A1200\0A3000\0A4000\0\nCD32\0CDTV\0Arcadia Multi Select\0High end WinUAE\0\nA590/A2091 SCSI Boot ROM\0\0";

    p = malloc (100000);
    if (!p)
	return;
    WIN32GUI_LoadUIString (IDS_ROMSCANEND, p, 100);
    strcat (p, "\n\n");

    rp = romtable;
    while(rp[0]) {
	int ok = 0;
	p2 = p1 + strlen(p1) + 1;
	strcat (p, " ");
	strcat (p, p1); strcat (p, ": ");
	if (listrom (rp))
	    ok = 1;
	while(*rp++ != -1);
	if (*rp != -1) {
	    if (ok) {
		ok = 0;
		if (listrom (rp))
		    ok = 1;
	    }
	    while(*rp++ != -1);
	}
	rp++;
	if (ok)
	    strcat (p, avail); else strcat (p, unavail);
	p1 = p2;
    }

    pre_gui_message (p);
    free (p);
}

static int scan_roms_2 (UAEREG *fkey, char *path)
{
    char buf[MAX_DPATH];
    WIN32_FIND_DATA find_data;
    HANDLE handle;
    int ret;

    if (!path)
	return 0;
    write_log("ROM scan directory '%s'\n", path);
    strcpy (buf, path);
    strcat (buf, "*.*");
    ret = 0;
    handle = FindFirstFile (buf, &find_data);
    if (handle == INVALID_HANDLE_VALUE)
	return 0;
    for (;;) {
	char tmppath[MAX_DPATH];
	strcpy (tmppath, path);
	strcat (tmppath, find_data.cFileName);
	if (!(find_data.dwFileAttributes & (FILE_ATTRIBUTE_DIRECTORY |FILE_ATTRIBUTE_SYSTEM)) && find_data.nFileSizeLow < 10000000) {
	    if (scan_rom (tmppath, fkey))
		ret = 1;
	}
	if (FindNextFile (handle, &find_data) == 0) {
	    FindClose (handle);
	    break;
	}
    }
    return ret;
}

#define MAX_ROM_PATHS 10

static int scan_roms_3(void *fkey, char **paths, char *path)
{
    int i, ret;
    char pathp[MAX_DPATH];

    ret = 0;
    GetFullPathName(path, MAX_DPATH, pathp, NULL);
    for (i = 0; i < MAX_ROM_PATHS; i++) {
	if (paths[i] && !strcmpi(paths[i], pathp))
	    return ret;
    }
    ret = scan_roms_2 (fkey, pathp);
    for (i = 0; i < MAX_ROM_PATHS; i++) {
	if (!paths[i]) {
	    paths[i] = my_strdup(pathp);
	    break;
	}
    }
    return ret;
}

extern int get_rom_path(char *out, int mode);

int scan_roms (int show)
{
    char path[MAX_DPATH];
    static int recursive;
    int id, i, ret, keys, cnt;
    UAEREG *fkey, *fkey2;
    char *paths[MAX_ROM_PATHS];

    if (recursive)
	return 0;
    recursive++;

    regdeletetree (NULL, "DetectedROMs");
    fkey = regcreatetree (NULL, "DetectedROMs");
    if (fkey == NULL)
	goto end;

    cnt = 0;
    ret = 0;
    for (i = 0; i < MAX_ROM_PATHS; i++)
	paths[i] = NULL;
    for (;;) {
	keys = get_keyring();
	fetch_path ("KickstartPath", path, sizeof path);
	cnt += scan_roms_3 (fkey, paths, path);
	if (1) {
	    for(i = 0; i < MAX_ROM_PATHS; i++) {
		ret = get_rom_path (path, i);
		if (ret < 0)
		    break;
		cnt += scan_roms_3 (fkey, paths, path);
	    }
	    if (get_keyring() > keys) { /* more keys detected in previous scan? */
		write_log ("ROM scan: more keys found, restarting..\n");
		for (i = 0; i < MAX_ROM_PATHS; i++) {
		    xfree (paths[i]);
		    paths[i] = NULL;
		}
		continue;
	    }
	}
	break;
    }
    if (cnt == 0)
	scan_roms_3 (fkey, paths, workprefs.path_rom);

    for (i = 0; i < MAX_ROM_PATHS; i++)
	xfree(paths[i]);

    fkey2 = regcreatetree (NULL, "DetectedROMS");
    if (fkey2) {
	id = 1;
	for (;;) {
	    struct romdata *rd = getromdatabyid(id);
	    if (!rd)
		break;
	    if (rd->crc32 == 0xffffffff)
		addrom(fkey, rd, NULL);
	    id++;
	}
	regclosetree (fkey2);
    }

end:
    read_rom_list ();
    if (show)
	show_rom_list ();

    regclosetree (fkey);
    recursive--;
    return ret;
}

struct ConfigStruct {
    char Name[MAX_DPATH];
    char Path[MAX_DPATH];
    char Fullpath[MAX_DPATH];
    char HostLink[MAX_DPATH];
    char HardwareLink[MAX_DPATH];
    char Description[CFG_DESCRIPTION_LENGTH];
    int Type, Directory;
    struct ConfigStruct *Parent, *Child;
    int host, hardware;
    HTREEITEM item;
};

static char *configreg[] = { "ConfigFile", "ConfigFileHardware", "ConfigFileHost" };
static char *configreg2[] = { "", "ConfigFileHardware_Auto", "ConfigFileHost_Auto" };
static struct ConfigStruct **configstore;
static int configstoresize, configstoreallocated, configtype, configtypepanel;

static struct ConfigStruct *getconfigstorefrompath (char *path, char *out, int type)
{
    int i;
    for (i = 0; i < configstoresize; i++) {
	if (((configstore[i]->Type == 0 || configstore[i]->Type == 3) && type == 0) || (configstore[i]->Type == type)) {
	    char path2[MAX_DPATH];
	    strcpy (path2, configstore[i]->Path);
	    strncat (path2, configstore[i]->Name, MAX_DPATH);
	    if (!strcmp (path, path2)) {
		strcpy (out, configstore[i]->Fullpath);
		strncat (out, configstore[i]->Name, MAX_DPATH);
		return configstore[i];
	    }
	}
    }
    return 0;
}

int target_cfgfile_load (struct uae_prefs *p, char *filename, int type, int isdefault)
{
    int v, i, type2;
    DWORD ct, ct2, size;
    char tmp1[MAX_DPATH], tmp2[MAX_DPATH];
    char fname[MAX_DPATH];

    strcpy (fname, filename);
    if (!zfile_exists (fname)) {
	fetch_configurationpath (fname, sizeof (fname));
	if (memcmp(fname, filename, strlen(fname)))
	    strcat (fname, filename);
	else
	    strcpy (fname, filename);
    }

    if (isdefault)
	qs_override = 1;
    if (type < 0) {
	type = 0;
	cfgfile_get_description (fname, NULL, NULL, NULL, &type);
    }
    if (type == 0 || type == 1) {
	discard_prefs (p, 0);
    }
    type2 = type;
    if (type == 0)
	default_prefs (p, type);
    regqueryint (NULL, "ConfigFile_NoAuto", &ct2);
    v = cfgfile_load (p, fname, &type2, ct2);
    if (!v)
	return v;
    if (type > 0)
	return v;
    for (i = 1; i <= 2; i++) {
	if (type != i) {
	    size = sizeof (ct);
	    ct = 0;
	    regqueryint (NULL, configreg2[i], &ct);
	    if (ct && ((i == 1 && p->config_hardware_path[0] == 0) || (i == 2 && p->config_host_path[0] == 0) || ct2)) {
		size = sizeof (tmp1);
		regquerystr (NULL, configreg[i], tmp1, &size);
		fetch_path ("ConfigurationPath", tmp2, sizeof (tmp2));
		strcat (tmp2, tmp1);
		v = i;
		cfgfile_load (p, tmp2, &v, 1);
	    }
	}
    }
    v = 1;
    return v;
}

static int gui_width = 640, gui_height = 480;

static int mm = 0;
static void m(void)
{
    write_log ("%d:0: %dx%d %dx%d %dx%d\n", mm, currprefs.gfx_size.width, currprefs.gfx_size.height,
	workprefs.gfx_size.width, workprefs.gfx_size.height, changed_prefs.gfx_size.width, changed_prefs.gfx_size.height);
    write_log ("%d:1: %dx%d %dx%d %dx%d\n", mm, currprefs.gfx_size_fs.width, currprefs.gfx_size_fs.height,
	workprefs.gfx_size_fs.width, workprefs.gfx_size_fs.height, changed_prefs.gfx_size_fs.width, changed_prefs.gfx_size_fs.height);
    mm++;
}

/* if drive is -1, show the full GUI, otherwise file-requester for DF[drive] */
void gui_display(int shortcut)
{
    static int here;
    HRESULT hr;
    int w, h;

    if (here)
	return;
    here++;
    gui_active++;
    screenshot_prepare();
#ifdef D3D
    D3D_guimode (TRUE);
#endif
    setpaused();
    inputdevice_unacquire ();
    clearallkeys ();
    setmouseactive (0);

    w = h = -1;
    if (!WIN32GFX_IsPicassoScreen() && currprefs.gfx_afullscreen && (currprefs.gfx_size.width < gui_width || currprefs.gfx_size.height < gui_height)) {
	w = currprefs.gfx_size.width;
	h = currprefs.gfx_size.height;
    }
    if (WIN32GFX_IsPicassoScreen() && currprefs.gfx_pfullscreen && (picasso96_state.Width < gui_width || picasso96_state.Height < gui_height)) {
	w = currprefs.gfx_size.width;
	h = currprefs.gfx_size.height;
    }
    scaleresource_setmaxsize (-1, -1);
    if (w > 0 && h > 0)
	scaleresource_setmaxsize (w, h);
    WIN32GFX_ClearPalette();
    manual_painting_needed++; /* So that WM_PAINT will refresh the display */

    if (isfullscreen () > 0) {
	hr = DirectDraw_FlipToGDISurface();
	if (FAILED(hr))
	    write_log ("FlipToGDISurface failed, %s\n", DXError (hr));
    }

    if (shortcut == -1) {
	int ret;
	ret = GetSettings (0, hAmigaWnd);
	if (!ret) {
	    savestate_state = 0;
	}
    } else if (shortcut >= 0 && shortcut < 4) {
	DiskSelection(hAmigaWnd, IDC_DF0+shortcut, 0, &changed_prefs, 0);
    } else if (shortcut == 5) {
	if (DiskSelection(hAmigaWnd, IDC_DOSAVESTATE, 9, &changed_prefs, 0))
	    save_state (savestate_fname, "Description!");
    } else if (shortcut == 4) {
	if (DiskSelection(hAmigaWnd, IDC_DOLOADSTATE, 10, &changed_prefs, 0))
	    savestate_state = STATE_DORESTORE;
    }
    manual_painting_needed--; /* So that WM_PAINT doesn't need to use custom refreshing */
    manual_palette_refresh_needed = 1;
    resumepaused();
    inputdevice_copyconfig (&changed_prefs, &currprefs);
    inputdevice_config_change_test ();
    clearallkeys ();
    inputdevice_acquire ();
    setmouseactive (1);
#ifdef D3D
    D3D_guimode (FALSE);
#endif
#ifdef AVIOUTPUT
    AVIOutput_Begin ();
#endif
    fpscounter_reset ();
    WIN32GFX_SetPalette();
#ifdef PICASSO96
    DX_SetPalette (0, 256);
#endif
    screenshot_free();
    write_disk_history();
    gui_active--;
    here--;
}

static void prefs_to_gui (struct uae_prefs *p)
{
    workprefs = *p;
    /* filesys hack */
    workprefs.mountitems = currprefs.mountitems;
    memcpy(&workprefs.mountconfig, &currprefs.mountconfig, MOUNT_CONFIG_SIZE * sizeof (struct uaedev_config_info));

    updatewinfsmode (&workprefs);
#if 0
#ifdef _DEBUG
    if (workprefs.gfx_framerate < 5)
	workprefs.gfx_framerate = 5;
#endif
#endif
}

static void gui_to_prefs (void)
{
    /* Always copy our prefs to changed_prefs, ... */
    changed_prefs = workprefs;
    /* filesys hack */
    currprefs.mountitems = changed_prefs.mountitems;
    memcpy(&currprefs.mountconfig, &changed_prefs.mountconfig, MOUNT_CONFIG_SIZE * sizeof (struct uaedev_config_info));

    updatewinfsmode (&changed_prefs);
}

int DirectorySelection(HWND hDlg, int flag, char *path)
{
    BROWSEINFO bi;
    LPITEMIDLIST pidlBrowse;
    char buf[MAX_DPATH];

    buf[0] = 0;
    bi.hwndOwner = hDlg;
    bi.pidlRoot = NULL;
    bi.pszDisplayName = buf;
    bi.lpszTitle = "Select folder";
    bi.ulFlags = 0;
    bi.lpfn = NULL;
    bi.lParam = 0;

    // Browse for a folder and return its PIDL.
    pidlBrowse = SHBrowseForFolder(&bi);
    if (pidlBrowse != NULL) {
	if (SHGetPathFromIDList(pidlBrowse, buf)) {
	    strcpy (path, buf);
	    return 1;
	}
    }
    return 0;
}

// Common routine for popping up a file-requester
// flag - 0 for floppy loading, 1 for floppy creation, 2 for loading hdf, 3 for saving hdf
// flag - 4 for loading .uae config-files, 5 for saving .uae config-files
// flag = 6 for loading .rom files, 7 for loading .key files
// flag = 8 for loading configurations
// flag = 9 for saving snapshots
// flag = 10 for loading snapshots
// flag = 11 for selecting flash files
// flag = 12 for loading anything
// flag = 13 for selecting path
// flag = 14 for loading filesystem
// flag = 15 for loading input
// flag = 16 for recording input
int DiskSelection_2 (HWND hDlg, WPARAM wParam, int flag, struct uae_prefs *prefs, char *path_out, int *multi)
{
    static int statefile_previousfilter;
    OPENFILENAME openFileName;
    char full_path[MAX_DPATH] = "";
    char full_path2[MAX_DPATH];
    char file_name[MAX_DPATH] = "";
    char init_path[MAX_DPATH] = "";
    BOOL result = FALSE;
    char *amiga_path = NULL;
    char description[CFG_DESCRIPTION_LENGTH] = "";
    char *p, *nextp;
    int all = 1;
    int next;
    int filterindex = 0;

    char szTitle[MAX_DPATH] = { 0 };
    char szFormat[MAX_DPATH];
    char szFilter[MAX_DPATH] = { 0 };

    memset (&openFileName, 0, sizeof (OPENFILENAME));

    strncpy (init_path, start_path_data, MAX_DPATH);
    switch (flag)
    {
	case 0:
	case 1:
	    fetch_path ("FloppyPath", init_path, sizeof (init_path));
	break;
	case 2:
	case 3:
	    fetch_path ("hdfPath", init_path, sizeof (init_path));
	break;
	case 6:
	case 7:
	case 11:
	    fetch_path ("KickstartPath", init_path, sizeof (init_path));
	break;
	case 4:
	case 5:
	case 8:
	    fetch_path ("ConfigurationPath", init_path, sizeof (init_path));
	break;
	case 9:
	case 10:
	    fetch_path ("StatefilePath", init_path, sizeof (init_path));
	break;
	case 15:
	case 16:
	    fetch_path ("InputPath", init_path, sizeof (init_path));
	break;

    }

    openFileName.lStructSize = os_winnt ? sizeof (OPENFILENAME) : OPENFILENAME_SIZE_VERSION_400;
    openFileName.hwndOwner = hDlg;
    openFileName.hInstance = hInst;

    switch (flag) {
    case 0:
	WIN32GUI_LoadUIString(IDS_SELECTADF, szTitle, MAX_DPATH);
	WIN32GUI_LoadUIString(IDS_ADF, szFormat, MAX_DPATH);
	sprintf(szFilter, "%s ", szFormat);
	memcpy(szFilter + strlen(szFilter), DISK_FORMAT_STRING, sizeof(DISK_FORMAT_STRING) + 1);

	openFileName.lpstrDefExt = "ADF";
	openFileName.lpstrFilter = szFilter;
	break;
    case 1:
	WIN32GUI_LoadUIString(IDS_CHOOSEBLANK, szTitle, MAX_DPATH);
	WIN32GUI_LoadUIString(IDS_ADF, szFormat, MAX_DPATH);
	sprintf(szFilter, "%s ", szFormat);
	memcpy(szFilter + strlen(szFilter), "(*.adf)\0*.adf\0", 15);

	openFileName.lpstrDefExt = "ADF";
	openFileName.lpstrFilter = szFilter;
	break;
    case 2:
    case 3:
	WIN32GUI_LoadUIString(IDS_SELECTHDF, szTitle, MAX_DPATH);
	WIN32GUI_LoadUIString(IDS_HDF, szFormat, MAX_DPATH);
	sprintf(szFilter, "%s ", szFormat);
	memcpy(szFilter + strlen(szFilter),  HDF_FORMAT_STRING, sizeof (HDF_FORMAT_STRING) + 1);

	openFileName.lpstrDefExt = "HDF";
	openFileName.lpstrFilter = szFilter;
	break;
    case 4:
    case 5:
	WIN32GUI_LoadUIString(IDS_SELECTUAE, szTitle, MAX_DPATH);
	WIN32GUI_LoadUIString(IDS_UAE, szFormat, MAX_DPATH );
	sprintf(szFilter, "%s ", szFormat);
	memcpy(szFilter + strlen(szFilter), "(*.uae)\0*.uae\0", 15);

	openFileName.lpstrDefExt = "UAE";
	openFileName.lpstrFilter = szFilter;
	break;
    case 6:
	WIN32GUI_LoadUIString(IDS_SELECTROM, szTitle, MAX_DPATH);
	WIN32GUI_LoadUIString(IDS_ROM, szFormat, MAX_DPATH);
	sprintf(szFilter, "%s ", szFormat);
	memcpy(szFilter + strlen(szFilter), ROM_FORMAT_STRING, sizeof (ROM_FORMAT_STRING) + 1);

	openFileName.lpstrDefExt = "ROM";
	openFileName.lpstrFilter = szFilter;
	break;
    case 7:
	WIN32GUI_LoadUIString(IDS_SELECTKEY, szTitle, MAX_DPATH);
	WIN32GUI_LoadUIString(IDS_KEY, szFormat, MAX_DPATH);
	sprintf(szFilter, "%s ", szFormat);
	memcpy(szFilter + strlen(szFilter), "(*.key)\0*.key\0", 15);

	openFileName.lpstrDefExt = "KEY";
	openFileName.lpstrFilter = szFilter;
	break;
    case 15:
    case 16:
	WIN32GUI_LoadUIString(flag == 15 ? IDS_RESTOREINP : IDS_SAVEINP, szTitle, MAX_DPATH);
	WIN32GUI_LoadUIString(IDS_INP, szFormat, MAX_DPATH);
	sprintf(szFilter, "%s ", szFormat);
	memcpy(szFilter + strlen(szFilter), INP_FORMAT_STRING, sizeof (INP_FORMAT_STRING) + 1);

	openFileName.lpstrDefExt = "INP";
	openFileName.lpstrFilter = szFilter;
	break;
    case 9:
    case 10:
	WIN32GUI_LoadUIString(flag == 10 ? IDS_RESTOREUSS : IDS_SAVEUSS, szTitle, MAX_DPATH);
	WIN32GUI_LoadUIString(IDS_USS, szFormat, MAX_DPATH);
	sprintf(szFilter, "%s ", szFormat);
	if (flag == 10) {
	    memcpy(szFilter + strlen(szFilter), USS_FORMAT_STRING_RESTORE, sizeof (USS_FORMAT_STRING_RESTORE) + 1);
	    all = 1;
	} else {
	    char tmp[MAX_DPATH];
	    memcpy(szFilter + strlen(szFilter), USS_FORMAT_STRING_SAVE, sizeof (USS_FORMAT_STRING_SAVE) + 1);
	    p = szFilter;
	    while (p[0] != 0 || p[1] !=0 ) p++;
	    p++;
	    WIN32GUI_LoadUIString(IDS_STATEFILE_UNCOMPRESSED, tmp, sizeof(tmp));
	    strcat(p, tmp);
	    strcat(p, " (*.uss)");
	    p += strlen(p) + 1;
	    strcpy(p, "*.uss");
	    p += strlen(p) + 1;
	    WIN32GUI_LoadUIString(IDS_STATEFILE_RAMDUMP, tmp, sizeof(tmp));
	    strcat(p, tmp);
	    strcat(p, " (*.dat)");
	    p += strlen(p) + 1;
	    strcpy (p, "*.dat");
	    p += strlen(p) + 1;
	    WIN32GUI_LoadUIString(IDS_STATEFILE_WAVE, tmp, sizeof(tmp));
	    strcat(p, tmp);
	    strcat(p, " (*.wav)");
	    p += strlen(p) + 1;
	    strcpy (p, "*.wav");
	    p += strlen(p) + 1;
	    *p = 0;
	    all = 0;
	    filterindex = statefile_previousfilter;
	}
	openFileName.lpstrDefExt = "USS";
	openFileName.lpstrFilter = szFilter;
	break;
    case 11:
	WIN32GUI_LoadUIString(IDS_SELECTFLASH, szTitle, MAX_DPATH);
	WIN32GUI_LoadUIString(IDS_FLASH, szFormat, MAX_DPATH);
	sprintf(szFilter, "%s ", szFormat);
	memcpy(szFilter + strlen(szFilter), "(*.nvr)\0*.nvr\0", 15);

	openFileName.lpstrDefExt = "NVR";
	openFileName.lpstrFilter = szFilter;
	break;
    case 8:
    default:
	WIN32GUI_LoadUIString(IDS_SELECTINFO, szTitle, MAX_DPATH);

	openFileName.lpstrFilter = NULL;
	openFileName.lpstrDefExt = NULL;
	break;
    case 12:
	WIN32GUI_LoadUIString(IDS_SELECTFS, szTitle, MAX_DPATH);

	openFileName.lpstrFilter = NULL;
	openFileName.lpstrDefExt = NULL;
	openFileName.lpstrInitialDir = path_out;
	break;
    case 13:
	WIN32GUI_LoadUIString(IDS_SELECTINFO, szTitle, MAX_DPATH);

	openFileName.lpstrFilter = NULL;
	openFileName.lpstrDefExt = NULL;
	openFileName.lpstrInitialDir = path_out;
	break;
    case 14:
	strcpy (szTitle, "Select supported archive file");
	sprintf (szFilter, "%s (%s)", "Archive", ARCHIVE_STRING);
	strcpy (szFilter + strlen(szFilter) + 1, ARCHIVE_STRING);

	openFileName.lpstrFilter = NULL;
	openFileName.lpstrFilter = szFilter;
	openFileName.lpstrInitialDir = path_out;
	break;
    }
    if (all) {
	p = szFilter;
	while (p[0] != 0 || p[1] !=0) p++;
	p++;
	strcpy (p, "All files (*.*)");
	p += strlen(p) + 1;
	strcpy (p, "*.*");
	p += strlen(p) + 1;
	*p = 0;
    }
    openFileName.Flags = OFN_EXPLORER | OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST |
	OFN_LONGNAMES | OFN_HIDEREADONLY | OFN_NOCHANGEDIR;
    openFileName.lpstrCustomFilter = NULL;
    openFileName.nMaxCustFilter = 0;
    openFileName.nFilterIndex = filterindex;
    openFileName.lpstrFile = full_path;
    openFileName.nMaxFile = MAX_DPATH;
    openFileName.lpstrFileTitle = file_name;
    openFileName.nMaxFileTitle = MAX_DPATH;
    openFileName.lpstrInitialDir = init_path;
    openFileName.lpfnHook = NULL;
    openFileName.lpTemplateName = NULL;
    openFileName.lCustData = 0;
    openFileName.lpstrTitle = szTitle;

    if (multi)
	openFileName.Flags |= OFN_ALLOWMULTISELECT;
    if (flag == 1 || flag == 3 || flag == 5 || flag == 9 || flag == 11 || flag == 16) {
	if (!(result = GetSaveFileName (&openFileName)))
	    write_log ("GetSaveFileName() failed, err=%d.\n", GetLastError());
    } else {
	if (!(result = GetOpenFileName (&openFileName)))
	    write_log ("GetOpenFileName() failed, err=%d.\n", GetLastError());
    }
    memcpy (full_path2, full_path, sizeof (full_path));
    next = 0;
    nextp = full_path2 + openFileName.nFileOffset;
    if (path_out) {
	if (multi) {
	    while (nextp[0])
		nextp += strlen (nextp) + 1;
	    memcpy (path_out, full_path2, nextp - full_path2 + 1);
	} else {
	    strcpy (path_out, full_path2);
	}
    }
    nextp = full_path2 + openFileName.nFileOffset;
    if (nextp[strlen(nextp) + 1] == 0)
	multi = 0;
    while (result && next >= 0)
    {
	next = -1;
	if (multi) {
	    if (nextp[0] == 0)
		break;
	    sprintf (full_path, "%s\\%s", full_path2, nextp);
	    nextp += strlen (nextp) + 1;
	}
	switch (wParam)
	{
	case IDC_PATH_NAME:
	case IDC_PATH_FILESYS:
	    if (flag == 8) {
		if(strstr(full_path, "Configurations\\")) {
		    strcpy(full_path, init_path);
		    strcat(full_path, file_name);
		}
	    }
	    SetDlgItemText (hDlg, wParam, full_path);
	    break;
	case IDC_DF0:
	case IDC_DF0QQ:
	    SetDlgItemText (hDlg, IDC_DF0TEXT, full_path);
	    strcpy(prefs->df[0], full_path);
	    DISK_history_add (full_path, -1);
	    next = IDC_DF1;
	    break;
	case IDC_DF1:
	case IDC_DF1QQ:
	    SetDlgItemText (hDlg, IDC_DF1TEXT, full_path);
	    strcpy(prefs->df[1], full_path);
	    DISK_history_add (full_path, -1);
	    next = IDC_DF2;
	    break;
	case IDC_DF2:
	    SetDlgItemText (hDlg, IDC_DF2TEXT, full_path);
	    strcpy(prefs->df[2], full_path);
	    DISK_history_add (full_path, -1);
	    next = IDC_DF3;
	    break;
	case IDC_DF3:
	    SetDlgItemText (hDlg, IDC_DF3TEXT, full_path);
	    strcpy(prefs->df[3], full_path);
	    DISK_history_add (full_path, -1);
	    break;
	case IDC_DOSAVESTATE:
	case IDC_DOLOADSTATE:
	    statefile_previousfilter = openFileName.nFilterIndex;
	    savestate_initsave (full_path, openFileName.nFilterIndex);
	    break;
	case IDC_CREATE:
	    {
		char disk_name[32];
		disk_name[0] = 0; disk_name[31] = 0;
		GetDlgItemText(hDlg, IDC_CREATE_NAME, disk_name, 30);
		disk_creatediskfile(full_path, 0, SendDlgItemMessage(hDlg, IDC_FLOPPYTYPE, CB_GETCURSEL, 0, 0L), disk_name);
	    }
	    break;
	case IDC_CREATE_RAW:
	    disk_creatediskfile(full_path, 1, SendDlgItemMessage(hDlg, IDC_FLOPPYTYPE, CB_GETCURSEL, 0, 0L), NULL);
	    break;
	case IDC_LOAD:
	    if (target_cfgfile_load(&workprefs, full_path, 0, 0) == 0) {
		char szMessage[MAX_DPATH];
		WIN32GUI_LoadUIString (IDS_COULDNOTLOADCONFIG, szMessage, MAX_DPATH);
		pre_gui_message (szMessage);
	    } else {
		SetDlgItemText (hDlg, IDC_EDITDESCRIPTION, workprefs.description);
		SetDlgItemText (hDlg, IDC_EDITNAME, full_path);
		SetDlgItemText (hDlg, IDC_CONFIGLINK, workprefs.config_host_path);
	    }
	    break;
	case IDC_SAVE:
	    SetDlgItemText (hDlg, IDC_EDITNAME, full_path);
	    cfgfile_save (&workprefs, full_path, 0);
	    break;
	case IDC_ROMFILE:
	    strcpy (workprefs.romfile, full_path);
	    break;
	case IDC_ROMFILE2:
	    strcpy (workprefs.romextfile, full_path);
	    break;
	case IDC_FLASHFILE:
	    strcpy (workprefs.flashfile, full_path);
	    break;
	case IDC_CARTFILE:
	    strcpy (workprefs.cartfile, full_path);
	    break;
	case IDC_INPREC_PLAY:
	    inprec_open(full_path, IsDlgButtonChecked(hDlg, IDC_INPREC_PLAYMODE) == BST_CHECKED ? -1 : -2);
	    break;
	case IDC_INPREC_RECORD:
	    inprec_open(full_path, 1);
	    break;
	}
	if (flag == 0 || flag == 1) {
	    amiga_path = strstr (openFileName.lpstrFile, openFileName.lpstrFileTitle);
	    if (amiga_path && amiga_path != openFileName.lpstrFile) {
		*amiga_path = 0;
		regsetstr (NULL, "FloppyPath", openFileName.lpstrFile);
	    }
	} else if (flag == 2 || flag == 3) {
	    amiga_path = strstr (openFileName.lpstrFile, openFileName.lpstrFileTitle);
	    if (amiga_path && amiga_path != openFileName.lpstrFile) {
		*amiga_path = 0;
	        regsetstr (NULL, "hdfPath", openFileName.lpstrFile);
	    }
	}
	if (!multi)
	    next = -1;
	if (next >= 0)
	    wParam = next;
    }
    return result;
}
int DiskSelection(HWND hDlg, WPARAM wParam, int flag, struct uae_prefs *prefs, char *path_out)
{
    return DiskSelection_2 (hDlg, wParam, flag, prefs, path_out, NULL);
}
int MultiDiskSelection(HWND hDlg, WPARAM wParam, int flag, struct uae_prefs *prefs, char *path_out)
{
    int multi = 0;
    return DiskSelection_2 (hDlg, wParam, flag, prefs, path_out, &multi);
}
static int loopmulti (char *s, char *out)
{
    static int index;

    if (!out) {
	index = strlen (s) + 1;
	return 1;
    }
    if (index < 0)
	return 0;
    if (!s[index]) {
	if (s[strlen (s) + 1] == 0) {
	    strcpy (out, s);
	    index = -1;
	    return 1;
	}
	return 0;
    }
    sprintf (out, "%s\\%s", s, s + index);
    index += strlen (s + index) + 1;
    return 1;
}

static BOOL CreateHardFile (HWND hDlg, UINT hfsizem, char *dostype)
{
    HANDLE hf;
    int i = 0;
    BOOL result = FALSE;
    LONG highword = 0;
    DWORD ret, written;
    char init_path[MAX_DPATH] = "";
    uae_u64 hfsize;
    uae_u32 dt;
    uae_u8 b;

    hfsize = (uae_u64)hfsizem * 1024 * 1024;
    DiskSelection (hDlg, IDC_PATH_NAME, 3, &workprefs, 0);
    GetDlgItemText (hDlg, IDC_PATH_NAME, init_path, MAX_DPATH);
    if (*init_path && hfsize) {
	SetCursor (LoadCursor(NULL, IDC_WAIT));
	if ((hf = CreateFile (init_path, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL) ) != INVALID_HANDLE_VALUE) {
	    if (hfsize >= 0x80000000) {
		highword = (DWORD)(hfsize >> 32);
		ret = SetFilePointer (hf, (DWORD)hfsize, &highword, FILE_BEGIN);
	    } else {
		ret = SetFilePointer (hf, (DWORD)hfsize, NULL, FILE_BEGIN);
	    }
	    if (ret == INVALID_SET_FILE_POINTER && GetLastError() != NO_ERROR)
		write_log ("SetFilePointer() failure for %s to posn %ud\n", init_path, hfsize);
	    else
		result = SetEndOfFile (hf);
	    SetFilePointer (hf, 0, NULL, FILE_BEGIN);
	    b = 0;
	    WriteFile (hf, &b, 1, &written, NULL);
	    WriteFile (hf, &b, 1, &written, NULL);
	    WriteFile (hf, &b, 1, &written, NULL);
	    WriteFile (hf, &b, 1, &written, NULL);
	    if (sscanf (dostype, "%x", &dt) > 0) {
		SetFilePointer (hf, 0, NULL, FILE_BEGIN);
		b = dt >> 24;
		WriteFile (hf, &b, 1, &written, NULL);
		b = dt >> 16;
		WriteFile (hf, &b, 1, &written, NULL);
		b = dt >> 8;
		WriteFile (hf, &b, 1, &written, NULL);
		b = dt >> 0;
		WriteFile (hf, &b, 1, &written, NULL);
	    }
	    CloseHandle (hf);
	} else {
	    write_log ("CreateFile() failed to create %s\n", init_path);
	}
	SetCursor (LoadCursor (NULL, IDC_ARROW));
    }
    return result;
}

static const char *memsize_names[] = {
/* 0 */ szNone,
/* 1 */ "256 K",
/* 2 */ "512 K",
/* 3 */ "1 MB",
/* 4 */ "2 MB",
/* 5 */ "4 MB",
/* 6 */ "8 MB",
/* 7 */ "16 MB",
/* 8 */ "32 MB",
/* 9 */ "64 MB",
/* 10*/ "128 MB",
/* 11*/ "256 MB",
/* 12*/ "512 MB",
/* 13*/ "1 GB",
/* 14*/ "1.5MB",
/* 15*/ "1.8MB",
/* 16*/ "2 GB",
};

static unsigned long memsizes[] = {
/* 0 */ 0,
/* 1 */ 0x00040000, /* 256-K */
/* 2 */ 0x00080000, /* 512-K */
/* 3 */ 0x00100000, /* 1-meg */
/* 4 */ 0x00200000, /* 2-meg */
/* 5 */ 0x00400000, /* 4-meg */
/* 6 */ 0x00800000, /* 8-meg */
/* 7 */ 0x01000000, /* 16-meg */
/* 8 */ 0x02000000, /* 32-meg */
/* 9 */ 0x04000000, /* 64-meg */
/* 10*/ 0x08000000, //128 Meg
/* 11*/ 0x10000000, //256 Meg
/* 12*/ 0x20000000, //512 Meg The correct size is set in mman.c
/* 13*/ 0x40000000, //1GB
/* 14*/ 0x00180000, //1.5MB
/* 15*/ 0x001C0000, //1.8MB
/* 16*/ 0x80000000, //2GB
};

static int msi_chip[] = { 1, 2, 3, 4, 5, 6 };
static int msi_bogo[] = { 0, 2, 3, 14, 15 };
static int msi_fast[] = { 0, 3, 4, 5, 6 };
static int msi_z3fast[] = { 0, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 16 };
static int msi_gfx[] = { 0, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13 };

static int CalculateHardfileSize (HWND hDlg)
{
    BOOL Translated = FALSE;
    UINT mbytes = 0;

    mbytes = GetDlgItemInt(hDlg, IDC_HF_SIZE, &Translated, FALSE);
    if (mbytes <= 0)
	mbytes = 0;
    if( !Translated )
	mbytes = 0;
    return mbytes;
}

static const char *nth[] = {
    "", "second ", "third ", "fourth ", "fifth ", "sixth ", "seventh ", "eighth ", "ninth ", "tenth "
};

static void setguititle (HWND phwnd)
{
    static char title[200];
    char title2[1000];
    char *name;
    static HWND hwnd;

    if (phwnd)
	hwnd = phwnd;
    if (hwnd && !title[0]) {
        GetWindowText (hwnd, title, sizeof (title));
	if (WINUAEBETA > 0) {
	    strcat (title, BetaStr);
	    if (strlen(WINUAEEXTRA) > 0) {
    		strcat (title, " ");
		strcat (title, WINUAEEXTRA);
	    }
	}
    }
    title2[0] = 0;
    name = config_filename;
    if (name && strlen (name) > 0) {
	strcat (title2, "[");
	strcat (title2, name);
	strcat (title2, "] - ");
    }
    strcat (title2, title);
    SetWindowText (hwnd, title2);
}


static void GetConfigPath (char *path, struct ConfigStruct *parent, int noroot)
{
    if (parent == 0) {
	path[0] = 0;
	if (!noroot) {
	    fetch_path ("ConfigurationPath", path, MAX_DPATH);
	}
	return;
    }
    if (parent) {
	GetConfigPath (path, parent->Parent, noroot);
	strncat (path, parent->Name, MAX_DPATH);
	strncat (path, "\\", MAX_DPATH);
    }
}

void FreeConfigStruct (struct ConfigStruct *config)
{
    free (config);
}
struct ConfigStruct *AllocConfigStruct (void)
{
    struct ConfigStruct *config;

    config = xcalloc (sizeof (struct ConfigStruct), 1);
    return config;
}

static struct ConfigStruct *GetConfigs (struct ConfigStruct *configparent, int usedirs, int *level)
{
    DWORD num_bytes = 0;
    char path[MAX_DPATH];
    char path2[MAX_DPATH];
    char shortpath[MAX_DPATH];
    WIN32_FIND_DATA find_data;
    struct ConfigStruct *config, *first;
    HANDLE handle;

    first = NULL;
    GetConfigPath (path, configparent, FALSE);
    GetConfigPath (shortpath, configparent, TRUE);
    strcpy (path2, path);
    strncat (path2, "*.*", MAX_DPATH);
    handle = FindFirstFile(path2, &find_data );
    if (handle == INVALID_HANDLE_VALUE) {
#ifndef SINGLEFILE
	// Either the directory has no .CFG files, or doesn't exist.
	// Create the directory, even if it already exists.  No harm, and don't check return codes, because
	// we may be doing this on a read-only media like CD-ROM.
	if (configparent == NULL) {
	    GetConfigPath (path, NULL, FALSE);
	    CreateDirectory (path, NULL);
	}
#endif
	return NULL;
    }
    for (;;) {
	config = NULL;
	if (strcmp (find_data.cFileName, ".") && strcmp (find_data.cFileName, "..")) {
	    int ok = 0;
	    config = AllocConfigStruct ();
	    strcpy (config->Path, shortpath);
	    strcpy (config->Fullpath, path);
	    if ((find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) && usedirs) {
		if ((*level) < 2) {
		    struct ConfigStruct *child;
		    strcpy (config->Name, find_data.cFileName);
		    strcpy (config->Path, shortpath);
		    strcat (config->Path, config->Name);
		    strcat (config->Path, "\\");
		    strcpy (config->Fullpath, path);
		    strcat (config->Fullpath, config->Name);
		    strcat (config->Fullpath, "\\");
		    config->Directory = 1;
		    (*level)++;
		    config->Parent = configparent;
		    if (!stricmp (config->Name, CONFIG_HOST))
			config->host = 1;
		    if (!stricmp (config->Name, CONFIG_HARDWARE))
			config->hardware = 1;
		    child = GetConfigs (config, usedirs, level);
		    (*level)--;
		    if (child)
			config->Child = child;
		    ok = 1;
		}
	    } else if (!(find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
		char path3[MAX_DPATH];
		if (strlen (find_data.cFileName) > 4 && !strcasecmp (find_data.cFileName + strlen (find_data.cFileName) - 4, ".uae")) {
		    strcpy (path3, path);
		    strncat (path3, find_data.cFileName, MAX_DPATH);
		    if (cfgfile_get_description (path3, config->Description, config->HostLink, config->HardwareLink, &config->Type)) {
			strcpy (config->Name, find_data.cFileName);
			ok = 1;
		    }
		}
	    }
	    if (!ok) {
		FreeConfigStruct (config);
		config = NULL;
	    }
	}
	if (config) {
	    if (configparent) {
		config->host = configparent->host;
		config->hardware = configparent->hardware;
	    }
	    config->Parent = configparent;
	    if (configstore == NULL || configstoreallocated == configstoresize) {
		configstoreallocated += 100;
		configstore = realloc (configstore, sizeof (struct ConfigStruct*) * configstoreallocated);
	    }
	    configstore[configstoresize++] = config;
	    if (first == NULL)
		first = config;
	}
	if(FindNextFile (handle, &find_data) == 0) {
	    FindClose(handle);
	    break;
	}
    }
    return first;
}

static void FreeConfigStore (void)
{
    int i;
    for (i = 0; i < configstoresize; i++)
	FreeConfigStruct (configstore[i]);
    free (configstore);
    configstore = 0;
    configstoresize = configstoreallocated = 0;
}
static struct ConfigStruct *CreateConfigStore (struct ConfigStruct *oldconfig)
{
    int level = 0, i;
    char path[MAX_DPATH], name[MAX_DPATH];
    struct ConfigStruct *cs;

    if (oldconfig) {
	strcpy (path, oldconfig->Path);
	strcpy (name, oldconfig->Name);
    }
    FreeConfigStore ();
    GetConfigs (NULL, 1, &level);
    if (oldconfig) {
	for (i = 0; i < configstoresize; i++) {
	    cs = configstore[i];
	    if (!cs->Directory && !strcmp (path, cs->Path) && !strcmp (name, cs->Name))
		return cs;
	}
    }
    return 0;
}

static char *HandleConfiguration (HWND hDlg, int flag, struct ConfigStruct *config)
{
    char name[MAX_DPATH], desc[MAX_DPATH];
    char path[MAX_DPATH];
    static char full_path[MAX_DPATH];

    full_path[0] = 0;
    GetDlgItemText (hDlg, IDC_EDITNAME, name, MAX_DPATH);
    strcpy (config_filename, name);
    if (flag == CONFIG_SAVE_FULL || flag == CONFIG_SAVE) {
	if (strlen (name) < 4 || strcasecmp (name + strlen (name) - 4, ".uae")) {
	    strcat (name, ".uae");
	    SetDlgItemText (hDlg, IDC_EDITNAME, name);
	}
	if (config)
	    strcpy (config->Name, name);
    }
    GetDlgItemText (hDlg, IDC_EDITDESCRIPTION, desc, MAX_DPATH);
    if (config) {
	strcpy (path, config->Fullpath);
    } else {
	fetch_configurationpath (path, sizeof (path));
    }
    strncat (path, name, MAX_DPATH);
    strcpy (full_path, path);
    switch (flag)
    {
	case CONFIG_SAVE_FULL:
	    DiskSelection(hDlg, IDC_SAVE, 5, &workprefs, 0);
	break;

	case CONFIG_LOAD_FULL:
	    DiskSelection(hDlg, IDC_LOAD, 4, &workprefs, 0);
	    EnableWindow(GetDlgItem (hDlg, IDC_VIEWINFO), workprefs.info[0]);
	break;

	case CONFIG_SAVE:
	    if (strlen (name) == 0 || strcmp (name, ".uae") == 0) {
		char szMessage[MAX_DPATH];
		WIN32GUI_LoadUIString(IDS_MUSTENTERNAME, szMessage, MAX_DPATH);
		pre_gui_message (szMessage);
	    } else {
		strcpy (workprefs.description, desc);
		cfgfile_save (&workprefs, path, configtypepanel);
	    }
	    break;

	case CONFIG_LOAD:
	    if (strlen (name) == 0) {
		char szMessage[MAX_DPATH];
		WIN32GUI_LoadUIString(IDS_MUSTSELECTCONFIG, szMessage, MAX_DPATH);
		pre_gui_message (szMessage);
	    } else {
		if (target_cfgfile_load (&workprefs, path, configtypepanel, 0) == 0) {
		    char szMessage[MAX_DPATH];
		    WIN32GUI_LoadUIString (IDS_COULDNOTLOADCONFIG, szMessage, MAX_DPATH);
		    pre_gui_message (szMessage);
		    config_filename[0] = 0;
		} else {
		    ew (hDlg, IDC_VIEWINFO, workprefs.info[0]);
		}
	    break;

	    case CONFIG_DELETE:
		if (strlen (name) == 0) {
		    char szMessage[MAX_DPATH];
		    WIN32GUI_LoadUIString(IDS_MUSTSELECTCONFIGFORDELETE, szMessage, MAX_DPATH);
		    pre_gui_message (szMessage);
		} else {
		    char szMessage[MAX_DPATH];
		    char szTitle[MAX_DPATH];
		    WIN32GUI_LoadUIString (IDS_DELETECONFIGCONFIRMATION, szMessage, MAX_DPATH);
		    WIN32GUI_LoadUIString (IDS_DELETECONFIGTITLE, szTitle, MAX_DPATH );
		    if (MessageBox (hDlg, szMessage, szTitle,
			MB_YESNO | MB_ICONWARNING | MB_APPLMODAL | MB_SETFOREGROUND) == IDYES) {
			cfgfile_backup (path);
			DeleteFile (path);
			write_log ("deleted config '%s'\n", path);
			config_filename[0] = 0;
		    }
		}
	    break;
	}
    }
    setguititle (NULL);
    return full_path;
}


static int disk_in_drive (int entry)
{
    int i;
    for (i = 0; i < 4; i++) {
	if (strlen (workprefs.dfxlist[entry]) > 0 && !strcmp (workprefs.dfxlist[entry], workprefs.df[i]))
	    return i;
    }
    return -1;
}

static int disk_swap (int entry, int mode)
{
    int drv, i, drvs[4] = { -1, -1, -1, -1 };

    for (i = 0; i < MAX_SPARE_DRIVES; i++) {
	drv = disk_in_drive (i);
	if (drv >= 0)
	    drvs[drv] = i;
    }
    if ((drv = disk_in_drive (entry)) >= 0) {
	if (mode < 0) {
	    workprefs.df[drv][0] = 0;
	    return 1;
	}

	if (strcmp (workprefs.df[drv], currprefs.df[drv])) {
	    strcpy (workprefs.df[drv], currprefs.df[drv]);
	    disk_insert (drv, workprefs.df[drv]);
	} else {
	    workprefs.df[drv][0] = 0;
	}
	if (drvs[0] < 0 || drvs[1] < 0 || drvs[2] < 0 || drvs[3] < 0) {
	    drv++;
	    while (drv < 4 && drvs[drv] >= 0)
		drv++;
	    if (drv < 4 && workprefs.dfxtype[drv] >= 0) {
		strcpy (workprefs.df[drv], workprefs.dfxlist[entry]);
		disk_insert (drv, workprefs.df[drv]);
	    }
	}
	return 1;
    }
    for (i = 0; i < 4; i++) {
	if (drvs[i] < 0 && workprefs.dfxtype[i] >= 0) {
	    strcpy (workprefs.df[i], workprefs.dfxlist[entry]);
	    disk_insert (i, workprefs.df[i]);
	    return 1;
	}
    }
    strcpy (workprefs.df[0], workprefs.dfxlist[entry]);
    disk_insert (0, workprefs.df[0]);
    return 1;
}

static int input_selected_device, input_selected_widget, input_total_devices;
static int input_selected_event, input_selected_sub_num;

static void set_lventry_input (HWND list, int index)
{
    int flags, i, sub;
    char name[256];
    char custom[MAX_DPATH];
    char af[10];

    inputdevice_get_mapped_name (input_selected_device, index, &flags, name, custom, input_selected_sub_num);
    if (flags & IDEV_MAPPED_AUTOFIRE_SET)
	WIN32GUI_LoadUIString (IDS_YES, af, sizeof (af));
    else if (flags & IDEV_MAPPED_AUTOFIRE_POSSIBLE)
	WIN32GUI_LoadUIString (IDS_NO, af, sizeof (af));
    else
	strcpy (af,"-");
    ListView_SetItemText(list, index, 1, custom[0] ? custom : name);
    ListView_SetItemText(list, index, 2, af);
    sub = 0;
    for (i = 0; i < MAX_INPUT_SUB_EVENT; i++) {
	if (inputdevice_get_mapped_name (input_selected_device, index, &flags, name, custom, i) || custom[0])
	    sub++;
    }
    sprintf (name, "%d", sub);
    ListView_SetItemText(list, index, 3, name);
}

static void update_listview_input (HWND hDlg)
{
    int i;
    if (!input_total_devices)
	return;
    for (i = 0; i < inputdevice_get_widget_num (input_selected_device); i++)
	set_lventry_input (GetDlgItem (hDlg, IDC_INPUTLIST), i);
}

static int clicked_entry = -1;

#define LOADSAVE_COLUMNS 2
#define INPUT_COLUMNS 4
#define HARDDISK_COLUMNS 8
#define DISK_COLUMNS 3
#define MAX_COLUMN_HEADING_WIDTH 20

#define LV_LOADSAVE 1
#define LV_HARDDISK 2
#define LV_INPUT 3
#define LV_DISK 4

static int listview_num_columns;
static int listview_column_width[HARDDISK_COLUMNS];

void InitializeListView (HWND hDlg)
{
    int lv_type;
    HWND list;
    LV_ITEM lvstruct;
    LV_COLUMN lvcolumn;
    RECT rect;
    char column_heading[HARDDISK_COLUMNS][MAX_COLUMN_HEADING_WIDTH];
    char blocksize_str[6] = "";
    char readwrite_str[10] = "";
    char size_str[32] = "";
    char volname_str[MAX_DPATH] = "";
    char devname_str[MAX_DPATH] = "";
    char bootpri_str[6] = "";
    int width = 0;
    int items = 0, result = 0, i, j, entry = 0, temp = 0;
    char tmp[10], tmp2[MAX_DPATH];

    if (hDlg == pages[HARDDISK_ID]) {
	listview_num_columns = HARDDISK_COLUMNS;
	lv_type = LV_HARDDISK;
	strcpy (column_heading[0], "*");
	WIN32GUI_LoadUIString(IDS_DEVICE, column_heading[1], MAX_COLUMN_HEADING_WIDTH);
	WIN32GUI_LoadUIString(IDS_VOLUME, column_heading[2], MAX_COLUMN_HEADING_WIDTH);
	WIN32GUI_LoadUIString(IDS_PATH, column_heading[3], MAX_COLUMN_HEADING_WIDTH);
	WIN32GUI_LoadUIString(IDS_RW, column_heading[4], MAX_COLUMN_HEADING_WIDTH);
	WIN32GUI_LoadUIString(IDS_BLOCKSIZE, column_heading[5], MAX_COLUMN_HEADING_WIDTH);
	WIN32GUI_LoadUIString(IDS_HFDSIZE, column_heading[6], MAX_COLUMN_HEADING_WIDTH);
	WIN32GUI_LoadUIString(IDS_BOOTPRI, column_heading[7], MAX_COLUMN_HEADING_WIDTH);
	list = GetDlgItem(hDlg, IDC_VOLUMELIST);
    } else if (hDlg == pages[INPUT_ID]) {
	listview_num_columns = INPUT_COLUMNS;
	lv_type = LV_INPUT;
	WIN32GUI_LoadUIString(IDS_INPUTHOSTWIDGET, column_heading[0], MAX_COLUMN_HEADING_WIDTH);
	WIN32GUI_LoadUIString(IDS_INPUTAMIGAEVENT, column_heading[1], MAX_COLUMN_HEADING_WIDTH);
	WIN32GUI_LoadUIString(IDS_INPUTAUTOFIRE, column_heading[2], MAX_COLUMN_HEADING_WIDTH);
	strcpy (column_heading[3], "#");
	list = GetDlgItem(hDlg, IDC_INPUTLIST);
    } else {
	listview_num_columns = DISK_COLUMNS;
	lv_type = LV_DISK;
	strcpy (column_heading[0], "#");
	WIN32GUI_LoadUIString(IDS_DISK_IMAGENAME, column_heading[1], MAX_COLUMN_HEADING_WIDTH);
	WIN32GUI_LoadUIString(IDS_DISK_DRIVENAME, column_heading[2], MAX_COLUMN_HEADING_WIDTH);
	list = GetDlgItem (hDlg, IDC_DISK);
    }
    cachedlist = list;

    ListView_DeleteAllItems(list);

    for(i = 0; i < listview_num_columns; i++)
	listview_column_width[i] = ListView_GetStringWidth(list, column_heading[i]) + 15;

    // If there are no columns, then insert some
    lvcolumn.mask = LVCF_WIDTH;
    if(ListView_GetColumn(list, 1, &lvcolumn) == FALSE)
    {
	for(i = 0; i < listview_num_columns; i++)
	{
	    lvcolumn.mask     = LVCF_FMT | LVCF_TEXT | LVCF_WIDTH | LVCF_SUBITEM;
	    lvcolumn.iSubItem = i;
	    lvcolumn.fmt      = LVCFMT_LEFT;
	    lvcolumn.pszText  = column_heading[i];
	    lvcolumn.cx       = listview_column_width[i];
	    ListView_InsertColumn(list, i, &lvcolumn);
	}
    }
    if (lv_type == LV_INPUT)
    {
	for (i = 0; input_total_devices && i < inputdevice_get_widget_num (input_selected_device); i++) {
	    char name[100];
	    inputdevice_get_widget_type (input_selected_device, i, name);
	    lvstruct.mask     = LVIF_TEXT | LVIF_PARAM;
	    lvstruct.pszText  = name;
	    lvstruct.lParam   = 0;
	    lvstruct.iItem    = i;
	    lvstruct.iSubItem = 0;
	    result = ListView_InsertItem(list, &lvstruct);
	    width = ListView_GetStringWidth(list, lvstruct.pszText) + 15;
	    if( width > listview_column_width[0])
		listview_column_width[0] = width;
	    entry++;
	}
	listview_column_width [1] = 260;
	listview_column_width [2] = 65;
	listview_column_width [3] = 30;
	update_listview_input (hDlg);
    }
    else if (lv_type == LV_DISK)
    {
	for (i = 0; i < MAX_SPARE_DRIVES; i++) {
	    int drv;
	    sprintf (tmp, "%d", i + 1);
	    lvstruct.mask     = LVIF_TEXT | LVIF_PARAM;
	    lvstruct.pszText  = tmp;
	    lvstruct.lParam   = 0;
	    lvstruct.iItem    = i;
	    lvstruct.iSubItem = 0;
	    result = ListView_InsertItem (list, &lvstruct);
	    strcpy (tmp2, workprefs.dfxlist[i]);
	    j = strlen (tmp2) - 1;
	    if (j < 0)
		j = 0;
	    while (j > 0) {
		if ((tmp2[j - 1] == '\\' || tmp2[j - 1] == '/')) {
		    if (!(j >= 5 && (tmp2[j - 5] == '.' || tmp2[j - 4] == '.')))
			break;
		}
		j--;
	    }
	    ListView_SetItemText (list, result, 1, tmp2 + j);
	    drv = disk_in_drive (i);
	    tmp[0] = 0;
	    if (drv >= 0)
		sprintf (tmp, "DF%d:", drv);
	    ListView_SetItemText (list, result, 2, tmp);
	    width = ListView_GetStringWidth(list, lvstruct.pszText) + 15;
	    if (width > listview_column_width[0])
		listview_column_width[0] = width;
	    entry++;
	}
	listview_column_width[0] = 30;
	listview_column_width[1] = 336;
	listview_column_width[2] = 50;

    }
    else if (lv_type == LV_HARDDISK)
    {
#ifdef FILESYS
	for(i = 0; i < workprefs.mountitems; i++)
	{
	    struct uaedev_config_info *uci = &workprefs.mountconfig[i];
	    int nosize = 0, type;
	    struct mountedinfo mi;

	    type = get_filesys_unitconfig (&workprefs, i, &mi);
	    if (type < 0) {
		type = uci->ishdf ? FILESYS_HARDFILE : FILESYS_VIRTUAL;
		nosize = 1;
	    }

	    if (nosize)
		strcpy (size_str, "n/a");
	    else if (mi.size >= 1024 * 1024 * 1024)
		sprintf (size_str, "%.1fG", ((double)(uae_u32)(mi.size / (1024 * 1024))) / 1024.0);
	    else
		sprintf (size_str, "%.1fM", ((double)(uae_u32)(mi.size / (1024))) / 1024.0);

	    if (uci->controller >= HD_CONTROLLER_IDE0 && uci->controller <= HD_CONTROLLER_IDE3) {
		sprintf (blocksize_str, "%d", uci->blocksize);
		sprintf (devname_str, "*IDE%d*", uci->controller - HD_CONTROLLER_IDE0);
		strcpy (volname_str, "n/a");
		strcpy (bootpri_str, "n/a");
	    } else if (uci->controller >= HD_CONTROLLER_SCSI0 && uci->controller <= HD_CONTROLLER_SCSI6) {
		sprintf (blocksize_str, "%d", uci->blocksize);
		sprintf (devname_str, "*SCSI%d*", uci->controller - HD_CONTROLLER_SCSI0);
		strcpy (volname_str, "n/a");
		strcpy (bootpri_str, "n/a");
	    } else if (uci->controller == HD_CONTROLLER_PCMCIA_SRAM) {
		strcpy (blocksize_str, "n/a");
		strcpy(devname_str, "*SCSRAM*");
		strcpy (volname_str, "n/a");
		strcpy (bootpri_str, "n/a");
	    } else if (type == FILESYS_HARDFILE) {
		sprintf (blocksize_str, "%d", uci->blocksize);
		strcpy (devname_str, uci->devname);
		strcpy (volname_str, "n/a");
		sprintf (bootpri_str, "%d", uci->bootpri);
	    } else if (type == FILESYS_HARDFILE_RDB || type == FILESYS_HARDDRIVE || uci->controller) {
		sprintf (blocksize_str, "%d", uci->blocksize);
		strcpy (devname_str, "n/a");
		strcpy (volname_str, "n/a");
		strcpy (bootpri_str, "n/a");
	    } else {
		strcpy (blocksize_str, "n/a");
		strcpy (devname_str, uci->devname);
		strcpy (volname_str, uci->volname);
		strcpy (size_str, "n/a");
		sprintf (bootpri_str, "%d", uci->bootpri);
	    }
	    if (!mi.ismedia) {
		strcpy (blocksize_str, "n/a");
		strcpy (size_str, "n/a");
	    }
	    WIN32GUI_LoadUIString (uci->readonly ? IDS_NO : IDS_YES, readwrite_str, sizeof (readwrite_str));

	    lvstruct.mask     = LVIF_TEXT | LVIF_PARAM;
	    lvstruct.pszText  = mi.ismedia == 0 ? "E" : (nosize ? "X" : (mi.ismounted ? "*" : " "));
	    if (uci->controller)
		lvstruct.pszText = " ";
	    lvstruct.lParam   = 0;
	    lvstruct.iItem    = i;
	    lvstruct.iSubItem = 0;
	    result = ListView_InsertItem (list, &lvstruct);
	    if (result != -1) {

		listview_column_width[0] = 15;

		ListView_SetItemText(list, result, 1, devname_str);
		width = ListView_GetStringWidth(list, devname_str) + 10;
		if(width > listview_column_width[1])
		    listview_column_width[1] = width;

		ListView_SetItemText(list, result, 2, volname_str);
		width = ListView_GetStringWidth(list, volname_str) + 10;
		if(width > listview_column_width[2])
		    listview_column_width[2] = width;

		listview_column_width [3] = 150;
		ListView_SetItemText(list, result, 3, uci->rootdir);
		width = ListView_GetStringWidth(list, uci->rootdir) + 10;
		if(width > listview_column_width[3])
		    listview_column_width[3] = width;

		ListView_SetItemText(list, result, 4, readwrite_str);
		width = ListView_GetStringWidth(list, readwrite_str) + 10;
		if(width > listview_column_width[4])
		    listview_column_width[4] = width;

		ListView_SetItemText(list, result, 5, blocksize_str);
		width = ListView_GetStringWidth(list, blocksize_str) + 10;
		if(width > listview_column_width[5])
		    listview_column_width[5] = width;

		ListView_SetItemText(list, result, 6, size_str);
		width = ListView_GetStringWidth(list, size_str) + 10;
		if(width > listview_column_width[6])
		    listview_column_width[6] = width;

		ListView_SetItemText(list, result, 7, bootpri_str);
		width = ListView_GetStringWidth(list, bootpri_str) + 10;
		if(width > listview_column_width[7] )
		    listview_column_width[7] = width;
	    }
	}
#endif
    }

    if(result != -1) {
	if(GetWindowRect(list, &rect)) {
	    ScreenToClient(hDlg, (LPPOINT)&rect);
	    ScreenToClient(hDlg, (LPPOINT)&rect.right);
	    if(listview_num_columns == 2) {
		if((temp = rect.right - rect.left - listview_column_width[0] - 4) > listview_column_width[1])
		    listview_column_width[1] = temp;
	    }
	}

	// Adjust our column widths so that we can see the contents...
	for(i = 0; i < listview_num_columns; i++) {
	    ListView_SetColumnWidth(list, i, listview_column_width[i]);
	}

	// Turn on full-row-select option
	ListView_SetExtendedListViewStyle(list, LVS_EX_FULLROWSELECT);

	// Redraw the items in the list...
	items = ListView_GetItemCount(list);
	ListView_RedrawItems(list, 0, items);
    }
}

static int listview_find_selected (HWND list)
{
    int i, items;
    items = ListView_GetItemCount (list);
    for (i = 0; i < items; i++) {
	if (ListView_GetItemState (list, i, LVIS_SELECTED) == LVIS_SELECTED)
	    return i;
    }
    return -1;
}

static int listview_entry_from_click (HWND list, int *column)
{
    POINT point;
    POINTS p;
    DWORD pos = GetMessagePos ();
    int items, entry;

    p = MAKEPOINTS (pos);
    point.x = p.x;
    point.y = p.y;
    ScreenToClient (list, &point);
    entry = ListView_GetTopIndex (list);
    items = entry + ListView_GetCountPerPage (list);
    if (items > ListView_GetItemCount (list))
	items = ListView_GetItemCount (list);

    while (entry <= items) {
	RECT rect;
	/* Get the bounding rectangle of an item. If the mouse
	 * location is within the bounding rectangle of the item,
	 * you know you have found the item that was being clicked.  */
	if (ListView_GetItemRect (list, entry, &rect, LVIR_BOUNDS)) {
	    if (PtInRect (&rect, point)) {
		POINT ppt;
		int i, x;
		UINT flag = LVIS_SELECTED | LVIS_FOCUSED;

		ListView_GetItemPosition(list, entry, &ppt);
		x = ppt.x;
		ListView_SetItemState (list, entry, flag, flag);
		for (i = 0; i < listview_num_columns && column; i++) {
		    if (x < point.x && x + listview_column_width[i] > point.x) {
			*column = i;
			break;
		    }
		    x += listview_column_width[i];
		}
		return entry;
	    }
	}
	entry++;
    }
    return -1;
}

static INT_PTR CALLBACK InfoSettingsProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int recursive = 0;

    switch (msg)
    {
	case WM_INITDIALOG:
	    recursive++;
	    SetDlgItemText (hDlg, IDC_PATH_NAME, workprefs.info);
	    recursive--;
	return TRUE;

	case WM_COMMAND:
	    if (recursive)
		break;
	    recursive++;

	    switch(wParam)
	    {
		case IDC_SELECTOR:
		    DiskSelection (hDlg, IDC_PATH_NAME, 8, &workprefs, 0);
		break;
		case IDOK:
		    EndDialog (hDlg, 1);
		break;
		case IDCANCEL:
		    EndDialog (hDlg, 0);
		break;
	    }

	    GetDlgItemText(hDlg, IDC_PATH_NAME, workprefs.info, sizeof workprefs.info);
	    recursive--;
	break;
    }
    return FALSE;
}

static HTREEITEM AddConfigNode (HWND hDlg, struct ConfigStruct *config, char *name, char *desc, char *path, int isdir, int expand, HTREEITEM parent)
{
    TVINSERTSTRUCT is;
    HWND TVhDlg;
    char s[MAX_DPATH] = "";
    char file_name[MAX_DPATH], file_path[MAX_DPATH];

    GetDlgItemText (hDlg, IDC_EDITNAME, file_name, MAX_DPATH);
    GetDlgItemText (hDlg, IDC_EDITPATH, file_path, MAX_DPATH);
    TVhDlg = GetDlgItem(hDlg, IDC_CONFIGTREE);
    memset (&is, 0, sizeof (is));
    is.hInsertAfter = isdir < 0 ? TVI_ROOT : TVI_SORT;
    is.hParent = parent;
    is.itemex.mask = TVIF_TEXT | TVIF_STATE | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_PARAM;
    if (!strcmp (file_name, name) && !strcmp (file_path, path)) {
	is.itemex.state |= TVIS_SELECTED;
	is.itemex.stateMask |= TVIS_SELECTED;
    }
    if (isdir) {
	strcat (s, " ");
	is.itemex.state |= TVIS_BOLD;
	is.itemex.stateMask |= TVIS_BOLD;
    }
    if (expand) {
	is.itemex.state |= TVIS_EXPANDED;
	is.itemex.stateMask |= TVIS_EXPANDED;
    }
    strcat (s, name);
    if (strlen (s) > 4 && !stricmp (s + strlen (s) - 4, ".uae"))
	s[strlen(s) - 4] = 0;
    if (desc && strlen(desc) > 0) {
	strcat (s, " (");
	strcat (s, desc);
	strcat (s, ")");
    }
    is.itemex.pszText = s;
    is.itemex.iImage = is.itemex.iSelectedImage = isdir > 0 ? 0 : (isdir < 0) ? 2 : 1;
    is.itemex.lParam = (LPARAM)config;
    return TreeView_InsertItem (TVhDlg, &is);
}

static int LoadConfigTreeView (HWND hDlg, int idx, HTREEITEM parent)
{
    struct ConfigStruct *cparent, *config;
    int cnt = 0;

    if (configstoresize == 0)
	return cnt;
    if (idx < 0) {
	idx = 0;
	for (;;) {
	    config = configstore[idx];
	    if (config->Parent == NULL)
		break;
	    idx++;
	    if (idx >= configstoresize)
		return cnt;
	}
    }
    cparent = configstore[idx]->Parent;
    idx = 0;
    while (idx < configstoresize) {
	config = configstore[idx];
	if ((configtypepanel == 1 && !config->hardware) || (configtypepanel == 2 && !config->host) || (configtypepanel == 0 && (config->host || config->hardware))) {
	    idx++;
	    continue;
	}
	if (config->Parent == cparent) {
	    if (config->Directory) {
		HTREEITEM par = AddConfigNode (hDlg, config, config->Name, NULL, config->Path, 1, config->hardware || config->host, parent);
		int idx2 = 0;
		for (;;) {
		    if (configstore[idx2] == config->Child) {
			config->item = par;
			if (LoadConfigTreeView (hDlg, idx2, par) == 0) {
			    if (!config->hardware && !config->host)
				TreeView_DeleteItem (GetDlgItem(hDlg, IDC_CONFIGTREE), par);
			}
			break;
		    }
		    idx2++;
		    if (idx2 >= configstoresize)
			break;
		}
	    } else if (!config->Directory) {
		if (((config->Type == 0 || config->Type == 3) && configtype == 0) || (config->Type == configtype)) {
		    config->item = AddConfigNode (hDlg, config, config->Name, config->Description, config->Path, 0, 0, parent);
		    cnt++;
		}
	    }
	}
	idx++;
    }
    return cnt;
}

static void InitializeConfig (HWND hDlg, struct ConfigStruct *config)
{
    int i, j, idx1, idx2;

    if (config == NULL) {
	SetDlgItemText (hDlg, IDC_EDITNAME, "");
	SetDlgItemText (hDlg, IDC_EDITDESCRIPTION, "");
    } else {
	SetDlgItemText (hDlg, IDC_EDITNAME, config->Name);
	SetDlgItemText (hDlg, IDC_EDITDESCRIPTION, config->Description);
    }
    SendDlgItemMessage (hDlg, IDC_CONFIGLINK, CB_RESETCONTENT, 0, 0L);
    SendDlgItemMessage (hDlg, IDC_CONFIGLINK, CB_ADDSTRING, 0, (LPARAM)"");
    idx1 = 1;
    idx2 = 0;
    for (j = 0; j < 2; j++) {
	for (i = 0; i < configstoresize; i++) {
	    struct ConfigStruct *cs = configstore[i];
	    if ((j == 0 && cs->Type == CONFIG_TYPE_HOST) || (j == 1 && cs->Type == CONFIG_TYPE_HARDWARE)) {
		char tmp2[MAX_DPATH];
		strcpy (tmp2, configstore[i]->Path);
		strncat (tmp2, configstore[i]->Name, MAX_DPATH);
		SendDlgItemMessage (hDlg, IDC_CONFIGLINK, CB_ADDSTRING, 0, (LPARAM)tmp2);
		if (config && (!strcmpi (tmp2, config->HardwareLink) || !strcmpi (tmp2, config->HostLink)))
		    idx2 = idx1;
		idx1++;
	    }
	}
    }
    SendDlgItemMessage (hDlg, IDC_CONFIGLINK, CB_SETCURSEL, idx2, 0);
}

static HTREEITEM InitializeConfigTreeView (HWND hDlg)
{
    HIMAGELIST himl = ImageList_Create (16, 16, ILC_COLOR8 | ILC_MASK, 3, 0);
    HWND TVhDlg = GetDlgItem(hDlg, IDC_CONFIGTREE);
    HTREEITEM parent;
    char path[MAX_DPATH];
    int i;

    if (himl) {
	HICON icon;
	icon = LoadIcon (hInst, (LPCSTR)MAKEINTRESOURCE(IDI_FOLDER));
	ImageList_AddIcon (himl, icon);
	icon = LoadIcon (hInst, (LPCSTR)MAKEINTRESOURCE(IDI_CONFIGFILE));
	ImageList_AddIcon (himl, icon);
	icon = LoadIcon (hInst, (LPCSTR)MAKEINTRESOURCE(IDI_ROOT));
	ImageList_AddIcon (himl, icon);
	TreeView_SetImageList (TVhDlg, himl, TVSIL_NORMAL);
    }
    for (i = 0; i < configstoresize; i++)
	configstore[i]->item = NULL;
    TreeView_DeleteAllItems (TVhDlg);
    GetConfigPath (path, NULL, FALSE);
    parent = AddConfigNode (hDlg, NULL, path, NULL, NULL, 0, 1, NULL);
    LoadConfigTreeView (hDlg, -1, parent);
    return parent;
}

static void ConfigToRegistry (struct ConfigStruct *config, int type)
{
    if (config) {
	char path[MAX_DPATH];
	strcpy (path, config->Path);
	strncat (path, config->Name, MAX_DPATH);
	regsetstr (NULL, configreg[type], path);
    }
}
static void ConfigToRegistry2 (DWORD ct, int type, DWORD noauto)
{
    if (type > 0)
	regsetint (NULL, configreg2[type], ct);
    if (noauto == 0 || noauto == 1)
	regsetint (NULL, "ConfigFile_NoAuto", noauto);
}

static void checkautoload (HWND	hDlg, struct ConfigStruct *config)
{
    int ct = 0;

    if (configtypepanel > 0)
	regqueryint (NULL, configreg2[configtypepanel], &ct);
    if (!config || config->Directory) {
	ct = 0;
	ConfigToRegistry2 (ct, configtypepanel, -1);
    }
    CheckDlgButton(hDlg, IDC_CONFIGAUTO, ct ? BST_CHECKED : BST_UNCHECKED);
    ew (hDlg, IDC_CONFIGAUTO, configtypepanel > 0 && config && !config->Directory ? TRUE : FALSE);
    regqueryint (NULL, "ConfigFile_NoAuto", &ct);
    CheckDlgButton(hDlg, IDC_CONFIGNOLINK, ct ? BST_CHECKED : BST_UNCHECKED);
}

static struct ConfigStruct *fixloadconfig (HWND hDlg, struct ConfigStruct *config)
{
    int i;

    if (config && configtypepanel == 0 && (config->host || config->hardware))
	return NULL;
    if ((!config && configtypepanel) || (config && (configtypepanel == 2 && !config->host) || (configtypepanel == 1 && !config->hardware))) {
	for (i = 0; i < configstoresize; i++) {
	    struct ConfigStruct *cs = configstore[i];
	    if (cs->Directory && ((configtypepanel == 1 && cs->hardware) || (configtypepanel == 2 && cs->host))) {
		config = cs;
		SetDlgItemText (hDlg, IDC_EDITPATH, config->Path);
		break;
	    }
	}
    }
    return config;
}

static struct ConfigStruct *initloadsave (HWND hDlg, struct ConfigStruct *config)
{
    HTREEITEM root;
    char name_buf[MAX_DPATH];
    DWORD dwRFPsize = sizeof (name_buf);
    char path[MAX_DPATH];

    EnableWindow (GetDlgItem (hDlg, IDC_VIEWINFO), workprefs.info[0]);
    SetDlgItemText (hDlg, IDC_EDITPATH, "");
    SetDlgItemText (hDlg, IDC_EDITDESCRIPTION, workprefs.description);
    root = InitializeConfigTreeView (hDlg);
    if (regquerystr (NULL, configreg[configtypepanel], name_buf, &dwRFPsize)) {
        struct ConfigStruct *config2 = getconfigstorefrompath (name_buf, path, configtypepanel);
        if (config2)
	    config = config2;
	checkautoload (hDlg, config);
    }
    config = fixloadconfig (hDlg, config);
    if (config && config->item)
	TreeView_SelectItem (GetDlgItem(hDlg, IDC_CONFIGTREE), config->item);
    else
	TreeView_SelectItem (GetDlgItem(hDlg, IDC_CONFIGTREE), root);
    EnableWindow (GetDlgItem(hDlg, IDC_CONFIGAUTO), configtypepanel > 0);
    EnableWindow (GetDlgItem(hDlg, IDC_CONFIGLINK), configtypepanel == 0);
    EnableWindow (GetDlgItem(hDlg, IDC_CONFIGNOLINK), configtypepanel == 0);
    return config;
}

static INT_PTR CALLBACK LoadSaveDlgProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    char *cfgfile;
    static int recursive;
    static struct ConfigStruct *config;

    switch (msg)
    {
    case WM_INITDIALOG:
	recursive++;
	if (!configstore) {
	    CreateConfigStore (NULL);
	    config = NULL;
	}
	pages[LOADSAVE_ID] = hDlg;
	currentpage = LOADSAVE_ID;
	config = initloadsave (hDlg, config);
	recursive--;
	return TRUE;

    case WM_USER:
	break;

    case WM_COMMAND:
    {
	recursive++;
	switch (LOWORD (wParam))
	{
	    case IDC_SAVE:
		HandleConfiguration (hDlg, CONFIG_SAVE_FULL, config);
		config = CreateConfigStore (config);
		config = fixloadconfig (hDlg, config);
		ConfigToRegistry (config, configtypepanel);
		InitializeConfigTreeView (hDlg);
		InitializeConfig (hDlg, config);
	    break;
	    case IDC_QUICKSAVE:
		HandleConfiguration (hDlg, CONFIG_SAVE, config);
		config = CreateConfigStore (config);
		config = fixloadconfig (hDlg, config);
		ConfigToRegistry (config, configtypepanel);
		InitializeConfigTreeView (hDlg);
		InitializeConfig (hDlg, config);
	    break;
	    case IDC_QUICKLOAD:
		cfgfile = HandleConfiguration (hDlg, CONFIG_LOAD, config);
		ConfigToRegistry (config, configtypepanel);
		InitializeConfig (hDlg, config);
		if (full_property_sheet) {
		    inputdevice_updateconfig (&workprefs);
		} else {
		    uae_restart (-1, cfgfile);
		    exit_gui(1);
		}
	    break;
	    case IDC_LOAD:
		cfgfile = HandleConfiguration (hDlg, CONFIG_LOAD_FULL, config);
		ConfigToRegistry (config, configtypepanel);
		InitializeConfig (hDlg, config);
		if (full_property_sheet) {
		    inputdevice_updateconfig (&workprefs);
		} else {
		    uae_restart (-1, cfgfile);
		    exit_gui(1);
		}
	    break;
	    case IDC_DELETE:
		HandleConfiguration (hDlg, CONFIG_DELETE, config);
		config = CreateConfigStore (config);
		config = fixloadconfig (hDlg, config);
		InitializeConfigTreeView (hDlg);
	    break;
	    case IDC_VIEWINFO:
		if (workprefs.info[0]) {
		    char name_buf[MAX_DPATH];
		    if (strstr (workprefs.info, "Configurations\\"))
			sprintf (name_buf, "%s\\%s", start_path_data, workprefs.info);
		    else
			strcpy (name_buf, workprefs.info);
		    ShellExecute (NULL, NULL, name_buf, NULL, NULL, SW_SHOWNORMAL);
		}
	    break;
	    case IDC_SETINFO:
		if (CustomDialogBox(IDD_SETINFO, hDlg, InfoSettingsProc))
		    EnableWindow( GetDlgItem( hDlg, IDC_VIEWINFO ), workprefs.info[0] );
	    break;
	    case IDC_CONFIGAUTO:
	    if (configtypepanel > 0) {
		int ct = IsDlgButtonChecked (hDlg, IDC_CONFIGAUTO) == BST_CHECKED ? 1 : 0;
		ConfigToRegistry2 (ct, configtypepanel, -1);
	    }
	    break;
	    case IDC_CONFIGNOLINK:
	    if (configtypepanel == 0) {
		int ct = IsDlgButtonChecked (hDlg, IDC_CONFIGNOLINK) == BST_CHECKED ? 1 : 0;
		ConfigToRegistry2 (-1, -1, ct);
	    }
	    break;
	    case IDC_CONFIGLINK:
	    if (HIWORD (wParam) == CBN_SELCHANGE || HIWORD (wParam) == CBN_KILLFOCUS)  {
		LRESULT val;
		char tmp[MAX_DPATH];
		tmp[0] = 0;
		val = SendDlgItemMessage (hDlg, IDC_CONFIGLINK, CB_GETCURSEL, 0, 0L);
		if (val == CB_ERR)
		    SendDlgItemMessage (hDlg, IDC_CONFIGLINK, WM_GETTEXT, (WPARAM)sizeof(tmp), (LPARAM)tmp);
		else
		    SendDlgItemMessage (hDlg, IDC_CONFIGLINK, CB_GETLBTEXT, (WPARAM)val, (LPARAM)tmp);
		strcpy (workprefs.config_host_path, tmp);
	    }
	    break;
	}
	recursive++;
	break;
    }

    case WM_NOTIFY:
    {
	LPNMHDR nm = (LPNMHDR)lParam;
	if (nm->hwndFrom == GetDlgItem (hDlg, IDC_CONFIGTREE)) {
	    switch (nm->code)
	    {
		case NM_DBLCLK:
		{
		    HTREEITEM ht = TreeView_GetSelection (GetDlgItem(hDlg, IDC_CONFIGTREE));
		    if (ht != NULL) {
			TVITEMEX pitem;
			memset (&pitem, 0, sizeof (pitem));
			pitem.mask = TVIF_HANDLE | TVIF_PARAM;
			pitem.hItem = ht;
			if (TreeView_GetItem (GetDlgItem(hDlg, IDC_CONFIGTREE), &pitem)) {
			    struct ConfigStruct *config = (struct ConfigStruct*)pitem.lParam;
			    if (config && !config->Directory) {
				cfgfile = HandleConfiguration (hDlg, CONFIG_LOAD, config);
				ConfigToRegistry (config, configtypepanel);
				if (!full_property_sheet)
				    uae_restart (0, cfgfile);
				exit_gui (1);
			    }
			}
		    }
		    return TRUE;
		}
		break;
		case TVN_SELCHANGING:
		return FALSE;
		case TVN_SELCHANGED:
		{
		    LPNMTREEVIEW tv = (LPNMTREEVIEW)lParam;
		    struct ConfigStruct *c = (struct ConfigStruct*)tv->itemNew.lParam;
		    if (c) {
			config = c;
			if (!config->Directory) {
			    InitializeConfig (hDlg, config);
			} else {
			    InitializeConfig (hDlg, NULL);
			}
			SetDlgItemText (hDlg, IDC_EDITPATH, config->Path);
		    }
		    if (configtypepanel > 0) {
			if (c && !c->Directory) {
			    ConfigToRegistry (config, configtypepanel);
			    InitializeConfig (hDlg, config);
			}
			checkautoload (hDlg, c);
		    }
		    return TRUE;
		}
		break;
	    }
	}
	break;
    }
    }

    return FALSE;
}

#define MAX_CONTRIBUTORS_LENGTH 2048

static INT_PTR CALLBACK ContributorsProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    CHARFORMAT CharFormat;
    char szContributors1[MAX_CONTRIBUTORS_LENGTH];
    char szContributors2[MAX_CONTRIBUTORS_LENGTH];
    char szContributors[MAX_CONTRIBUTORS_LENGTH * 2];

    switch (msg) {
     case WM_COMMAND:
	if (wParam == ID_OK) {
	    EndDialog (hDlg, 1);
	    return TRUE;
	}
	break;
     case WM_INITDIALOG:
	CharFormat.cbSize = sizeof (CharFormat);

	WIN32GUI_LoadUIString(IDS_CONTRIBUTORS1, szContributors1, MAX_CONTRIBUTORS_LENGTH);
	WIN32GUI_LoadUIString(IDS_CONTRIBUTORS2, szContributors2, MAX_CONTRIBUTORS_LENGTH);
	sprintf(szContributors, "%s%s", szContributors1, szContributors2);

	SetDlgItemText (hDlg, IDC_CONTRIBUTORS, szContributors );
	SendDlgItemMessage (hDlg, IDC_CONTRIBUTORS, EM_GETCHARFORMAT, 0, (LPARAM) & CharFormat);
	CharFormat.dwMask |= CFM_SIZE | CFM_FACE;
	CharFormat.yHeight = 8 * 20; /* height in twips, where a twip is 1/20th of a point - for a pt.size of 18 */

	strcpy (CharFormat.szFaceName, os_vista ? "Segoe UI" : (os_winnt ? "Tahoma" : "Times New Roman"));
	SendDlgItemMessage (hDlg, IDC_CONTRIBUTORS, EM_SETCHARFORMAT, SCF_ALL, (LPARAM) & CharFormat);
	return TRUE;
    }
    return FALSE;
}

static void DisplayContributors (HWND hDlg)
{
    CustomDialogBox(IDD_CONTRIBUTORS, hDlg, ContributorsProc);
}

typedef struct url_info
{
    int   id;
    BOOL  state;
    char *display;
    char *url;
} urlinfo;

static urlinfo urls[] =
{
    {IDC_CLOANTOHOME, FALSE, "Cloanto's Amiga Forever", "http://www.amigaforever.com/"},
    {IDC_AMIGAHOME, FALSE, "Amiga Inc.", "http://www.amiga.com"},
    {IDC_PICASSOHOME, FALSE, "Picasso96 Home Page", "http://www.picasso96.cogito.de/"},
    {IDC_UAEHOME, FALSE, "UAE Home Page", "http://uae.coresystems.de/"},
    {IDC_WINUAEHOME, FALSE, "WinUAE Home Page", "http://www.winuae.net/"},
    {IDC_AIABHOME, FALSE, "AIAB", "http://www.amigainabox.co.uk/"},
    {IDC_THEROOTS, FALSE, "Back To The Roots", "http://www.back2roots.org/"},
    {IDC_ABIME, FALSE, "abime.net", "http://www.abime.net/"},
    {IDC_CAPS, FALSE, "SPS", "http://www.softpres.org/"},
    {IDC_AMIGASYS, FALSE, "AmigaSYS", "http://amigasys.extra.hu/"},
    {IDC_AMIKIT, FALSE, "AmiKit", "http://amikit.amiga.sk/"},
    { -1, FALSE, NULL, NULL }
};

static void SetupRichText(HWND hDlg, urlinfo *url)
{
    CHARFORMAT CharFormat;
    CharFormat.cbSize = sizeof (CharFormat);

    SetDlgItemText(hDlg, url->id, url->display);
    SendDlgItemMessage(hDlg, url->id, EM_GETCHARFORMAT, 0, (LPARAM)&CharFormat);
    CharFormat.dwMask   |= CFM_UNDERLINE | CFM_SIZE | CFM_FACE | CFM_COLOR;
    CharFormat.dwEffects = url->state ? CFE_UNDERLINE : 0;
    CharFormat.yHeight = 10 * 20; /* height in twips, where a twip is 1/20th of a point - for a pt.size of 18 */

    CharFormat.crTextColor = GetSysColor(COLOR_ACTIVECAPTION);
    strcpy(CharFormat.szFaceName, "Tahoma" );
    SendDlgItemMessage(hDlg, url->id, EM_SETCHARFORMAT, SCF_ALL, (LPARAM)&CharFormat);
    SendDlgItemMessage(hDlg, url->id, EM_SETBKGNDCOLOR, 0, GetSysColor(COLOR_3DFACE));
}

static void url_handler(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int last_rectangle = -1;
    int i;
    BOOL found = FALSE;
    HCURSOR m_hCursor = NULL;
    POINT point;
    point.x = LOWORD (lParam);
    point.y = HIWORD (lParam);

    for (i = 0; urls[i].id >= 0; i++)
    {
	RECT rect;
	GetWindowRect(GetDlgItem(hDlg, urls[i].id), &rect);
	ScreenToClient(hDlg, (POINT*)&rect);
	ScreenToClient(hDlg, (POINT*)&rect.right);
	if(PtInRect(&rect, point))
	{
	    if(msg == WM_LBUTTONDOWN)
	    {
		ShellExecute (NULL, NULL, urls[i].url , NULL, NULL, SW_SHOWNORMAL);
		SetCursor(LoadCursor(NULL, MAKEINTRESOURCE(IDC_ARROW)));
	    }
	    else
	    {
		if(i != last_rectangle)
		{
		    // try and load the system hand (Win2000+)
		    m_hCursor = LoadCursor(NULL, MAKEINTRESOURCE(IDC_HAND) );
		    if (!m_hCursor)
		    {
			// retry with our fallback hand
			m_hCursor = LoadCursor(hInst, MAKEINTRESOURCE(IDC_MYHAND) );
		    }
		    SetCursor(m_hCursor);
		    urls[i].state = TRUE;
		    SetupRichText(hDlg, &urls[i]);

		    if(last_rectangle != -1)
		    {
			urls[last_rectangle].state = FALSE;
			SetupRichText(hDlg, &urls[last_rectangle]);
		    }
		}
	    }
	    last_rectangle = i;
	    found = TRUE;
	    break;
	}
    }

    if(!found && last_rectangle >= 0)
    {
	SetCursor(LoadCursor(NULL, MAKEINTRESOURCE(IDC_ARROW)));
	urls[last_rectangle].state = FALSE;
	SetupRichText(hDlg, &urls[last_rectangle]);
	last_rectangle = -1;
    }
}

static void setac (HWND hDlg, int id)
{
    SHAutoComplete (GetDlgItem (hDlg, id), SHACF_FILESYSTEM | SHACF_AUTOAPPEND_FORCE_ON | SHACF_AUTOSUGGEST_FORCE_ON | SHACF_USETAB);
}
static void setautocomplete (HWND hDlg, int id)
{
    HWND item = FindWindowEx(GetDlgItem (hDlg, id), NULL, "Edit", NULL);
    if (item)
	SHAutoComplete (item, SHACF_FILESYSTEM | SHACF_AUTOAPPEND_FORCE_ON | SHACF_AUTOSUGGEST_FORCE_ON | SHACF_USETAB);
}
static void setmultiautocomplete (HWND hDlg, int *ids)
{
    int i;
    for (i = 0; ids[i] >= 0; i++)
	setautocomplete (hDlg, ids[i]);
}

static void setpath (HWND hDlg, char *name, DWORD d, char *def)
{
    char tmp[MAX_DPATH];

    strcpy (tmp, def);
    fetch_path (name, tmp, sizeof (tmp));
    SetDlgItemText (hDlg, d, tmp);
}

static void values_to_pathsdialog (HWND hDlg)
{
    setpath (hDlg, "KickstartPath", IDC_PATHS_ROM, "Roms");
    setpath (hDlg, "ConfigurationPath", IDC_PATHS_CONFIG, "Configurations");
    setpath (hDlg, "ScreenshotPath", IDC_PATHS_SCREENSHOT, "ScreenShots");
    setpath (hDlg, "StatefilePath", IDC_PATHS_SAVESTATE, "Savestates");
    setpath (hDlg, "SaveimagePath", IDC_PATHS_SAVEIMAGE, "SaveImages");
    setpath (hDlg, "VideoPath", IDC_PATHS_AVIOUTPUT, "Videos");
}

static void resetregistry (void)
{
    regdeletetree (NULL, "DetectedROMs");
    regdelete (NULL, "QuickStartMode");
    regdelete (NULL, "ConfigFile");
    regdelete (NULL, "ConfigFileHardware");
    regdelete (NULL, "ConfigFileHost");
    regdelete (NULL, "ConfigFileHardware_Auto");
    regdelete (NULL, "ConfigFileHost_Auto");
    regdelete (NULL, "ConfigurationPath");
    regdelete (NULL, "SaveimagePath");
    regdelete (NULL, "ScreenshotPath");
    regdelete (NULL, "StatefilePath");
    regdelete (NULL, "VideoPath");
    regdelete (NULL, "QuickStartModel");
    regdelete (NULL, "QuickStartConfiguration");
    regdelete (NULL, "QuickStartCompatibility");
    regdelete (NULL, "QuickStartHostConfig");
}

int path_type;
static INT_PTR CALLBACK PathsDlgProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    void create_afnewdir(int);
    static int recursive;
    static int ptypes[3], numtypes;
    int val, selpath = 0;
    char tmp[MAX_DPATH], pathmode[32];

    switch (msg)
    {
	case WM_INITDIALOG:
	recursive++;
	pages[PATHS_ID] = hDlg;
	setac (hDlg, IDC_PATHS_ROM);
	setac (hDlg, IDC_PATHS_CONFIG);
	setac (hDlg, IDC_PATHS_SCREENSHOT);
	setac (hDlg, IDC_PATHS_SAVESTATE);
	setac (hDlg, IDC_PATHS_SAVEIMAGE);
	setac (hDlg, IDC_PATHS_AVIOUTPUT);
	currentpage = PATHS_ID;
	ShowWindow (GetDlgItem (hDlg, IDC_RESETREGISTRY), FALSE);
	numtypes = 0;
	SendDlgItemMessage (hDlg, IDC_PATHS_DEFAULTTYPE, CB_RESETCONTENT, 0, 0L);
	if (af_path_2005 & 1) {
	    WIN32GUI_LoadUIString(IDS_DEFAULT_AF2005, tmp, sizeof tmp);
	    SendDlgItemMessage (hDlg, IDC_PATHS_DEFAULTTYPE, CB_ADDSTRING, 0, (LPARAM)tmp);
	    if (path_type == PATH_TYPE_NEWAF)
		selpath = numtypes;
	    ptypes[numtypes++] = PATH_TYPE_NEWAF;
	}
	if (start_path_new1[0]) {
	    WIN32GUI_LoadUIString(IDS_DEFAULT_NEWWINUAE, tmp, sizeof tmp);
	    SendDlgItemMessage (hDlg, IDC_PATHS_DEFAULTTYPE, CB_ADDSTRING, 0, (LPARAM)tmp);
	    if (path_type == PATH_TYPE_NEWWINUAE)
		selpath = numtypes;
	    ptypes[numtypes++] = PATH_TYPE_NEWWINUAE;
	}
	if (af_path_2005 & 2) {
	    SendDlgItemMessage (hDlg, IDC_PATHS_DEFAULTTYPE, CB_ADDSTRING, 0, (LPARAM)"AmigaForeverData");
	    if (path_type == PATH_TYPE_AMIGAFOREVERDATA)
		selpath = numtypes;
	    ptypes[numtypes++] = PATH_TYPE_AMIGAFOREVERDATA;
	}
	if (af_path_old) {
	    WIN32GUI_LoadUIString(IDS_DEFAULT_AF, tmp, sizeof tmp);
	    SendDlgItemMessage (hDlg, IDC_PATHS_DEFAULTTYPE, CB_ADDSTRING, 0, (LPARAM)tmp);
	    if (path_type == PATH_TYPE_OLDAF)
		selpath = numtypes;
	    ptypes[numtypes++] = PATH_TYPE_OLDAF;
	}
	WIN32GUI_LoadUIString(IDS_DEFAULT_WINUAE, tmp, sizeof tmp);
	SendDlgItemMessage (hDlg, IDC_PATHS_DEFAULTTYPE, CB_ADDSTRING, 0, (LPARAM)tmp);
	if (path_type == 0)
	    selpath = numtypes;
	ptypes[numtypes++] = 0;
	SendDlgItemMessage (hDlg, IDC_PATHS_DEFAULTTYPE, CB_SETCURSEL, selpath, 0);
	EnableWindow(GetDlgItem (hDlg, IDC_PATHS_DEFAULTTYPE), numtypes > 0 ? TRUE : FALSE);
	values_to_pathsdialog (hDlg);
	recursive--;
	return TRUE;

	case WM_COMMAND:
	if (recursive > 0)
	    break;
	recursive++;
	switch (LOWORD (wParam))
	{
	    case IDC_PATHS_ROMS:
	    fetch_path ("KickstartPath", tmp, sizeof (tmp));
	    if (DirectorySelection (hDlg, 0, tmp)) {
		load_keyring(&workprefs, NULL);
		set_path ("KickstartPath", tmp);
		if (!scan_roms (1))
		    gui_message_id (IDS_ROMSCANNOROMS);
		values_to_pathsdialog (hDlg);
	    }
	    break;
	    case IDC_PATHS_ROM:
	    GetWindowText (GetDlgItem (hDlg, IDC_PATHS_ROM), tmp, sizeof (tmp));
	    set_path ("KickstartPath", tmp);
	    break;
	    case IDC_PATHS_CONFIGS:
	    fetch_path ("ConfigurationPath", tmp, sizeof (tmp));
	    if (DirectorySelection (hDlg, 0, tmp)) {
		set_path ("ConfigurationPath", tmp);
		values_to_pathsdialog (hDlg);
		FreeConfigStore ();
	    }
	    break;
	    case IDC_PATHS_CONFIG:
	    GetWindowText (GetDlgItem (hDlg, IDC_PATHS_CONFIG), tmp, sizeof (tmp));
	    set_path ("ConfigurationPath", tmp);
	    FreeConfigStore ();
	    break;
	    case IDC_PATHS_SCREENSHOTS:
	    fetch_path ("ScreenshotPath", tmp, sizeof (tmp));
	    if (DirectorySelection (hDlg, 0, tmp)) {
		set_path ("ScreenshotPath", tmp);
		values_to_pathsdialog (hDlg);
	    }
	    break;
	    case IDC_PATHS_SCREENSHOT:
	    GetWindowText (GetDlgItem (hDlg, IDC_PATHS_SCREENSHOT), tmp, sizeof (tmp));
	    set_path ("ScreenshotPath", tmp);
	    break;
	    case IDC_PATHS_SAVESTATES:
	    fetch_path ("StatefilePath", tmp, sizeof (tmp));
	    if (DirectorySelection (hDlg, 0, tmp)) {
		set_path ("StatefilePath", tmp);
		values_to_pathsdialog (hDlg);
	    }
	    break;
	    case IDC_PATHS_SAVESTATE:
	    GetWindowText (GetDlgItem (hDlg, IDC_PATHS_SAVESTATE), tmp, sizeof (tmp));
	    set_path ("StatefilePath", tmp);
	    break;
	    case IDC_PATHS_SAVEIMAGES:
	    fetch_path ("SaveimagePath", tmp, sizeof (tmp));
	    if (DirectorySelection (hDlg, 0, tmp)) {
		set_path ("SaveimagePath", tmp);
		values_to_pathsdialog (hDlg);
	    }
	    break;
	    case IDC_PATHS_SAVEIMAGE:
	    GetWindowText (GetDlgItem (hDlg, IDC_PATHS_SAVEIMAGE), tmp, sizeof (tmp));
	    set_path ("SaveimagePath", tmp);
	    break;
	    case IDC_PATHS_AVIOUTPUTS:
	    fetch_path ("VideoPath", tmp, sizeof (tmp));
	    if (DirectorySelection (hDlg, 0, tmp)) {
		set_path ("VideoPath", tmp);
		values_to_pathsdialog (hDlg);
	    }
	    break;
	    case IDC_PATHS_AVIOUTPUT:
	    GetWindowText (GetDlgItem (hDlg, IDC_PATHS_AVIOUTPUT), tmp, sizeof (tmp));
	    set_path ("VideoPath", tmp);
	    break;
	    case IDC_PATHS_DEFAULT:
	    val = SendDlgItemMessage (hDlg, IDC_PATHS_DEFAULTTYPE, CB_GETCURSEL, 0, 0L);
	    if (val != CB_ERR && val >= 0 && val < numtypes) {
		val = ptypes[val];
		if (val == PATH_TYPE_WINUAE) {
		    strcpy (start_path_data, start_path_exe);
		    path_type = PATH_TYPE_WINUAE;
		    strcpy (pathmode, "WinUAE");
		} else if (val == PATH_TYPE_NEWWINUAE && start_path_new1[0]) {
		    strcpy (start_path_data, start_path_new1);
		    path_type = PATH_TYPE_NEWWINUAE;
		    create_afnewdir(0);
		    strcpy (pathmode, "WinUAE_2");
		} else if (val == PATH_TYPE_OLDAF && start_path_af[0]) {
		    strcpy (start_path_data, start_path_af);
		    strcpy (pathmode, "AF");
		    path_type = PATH_TYPE_OLDAF;
		} else if (val == PATH_TYPE_NEWAF && start_path_new1[0]) {
		    strcpy (pathmode, "AF2005");
		    path_type = PATH_TYPE_NEWAF;
		    create_afnewdir(0);
		    strcpy (start_path_data, start_path_new1);
		} else if (val == PATH_TYPE_AMIGAFOREVERDATA && start_path_new2[0]) {
		    strcpy (pathmode, "AMIGAFOREVERDATA");
		    path_type = PATH_TYPE_AMIGAFOREVERDATA;
		    strcpy (start_path_data, start_path_new1);
		}
		SetCurrentDirectory (start_path_data);
		regsetstr (NULL, "PathMode", pathmode);
		set_path ("KickstartPath", NULL);
		set_path ("ConfigurationPath", NULL);
		set_path ("ScreenshotPath", NULL);
		set_path ("StatefilePath", NULL);
		set_path ("SaveimagePath", NULL);
		set_path ("VideoPath", NULL);
		values_to_pathsdialog (hDlg);
		FreeConfigStore ();
	    }
	    break;
	    case IDC_ROM_RESCAN:
	    scan_roms (1);
	    break;
	    case IDC_RESETREGISTRY:
	    resetregistry ();
	    break;
	    case IDC_RESETDISKHISTORY:
	    reset_disk_history ();
	    break;
	}
	recursive--;
    }
    return FALSE;
}

struct amigamodels {
    int compalevels;
    int id;
};
static struct amigamodels amodels[] = {
    { 4, IDS_QS_MODEL_A500 }, // "Amiga 500"
    { 4, IDS_QS_MODEL_A500P }, // "Amiga 500+"
    { 4, IDS_QS_MODEL_A600 }, // "Amiga 600"
    { 4, IDS_QS_MODEL_A1000 }, // "Amiga 1000"
    { 3, IDS_QS_MODEL_A1200 }, // "Amiga 1200"
    { 1, IDS_QS_MODEL_A3000 }, // "Amiga 3000"
    { 1, IDS_QS_MODEL_A4000 }, // "Amiga 4000"
    { 0, }, //{ 1, IDS_QS_MODEL_A4000T }, // "Amiga 4000T"
    { 3, IDS_QS_MODEL_CD32 }, // "CD32"
    { 4, IDS_QS_MODEL_CDTV }, // "CDTV"
    { 4, IDS_QS_MODEL_ARCADIA }, // "Arcadia"
    { 1, IDS_QS_MODEL_UAE }, // "Expanded UAE example configuration"
    { -1 }
};

static DWORD quickstart_model = 0, quickstart_conf = 0, quickstart_compa = 1, quickstart_floppy = 1;
static int quickstart_ok, quickstart_ok_floppy;
static void addfloppytype (HWND hDlg, int n);
static void addfloppyhistory (HWND hDlg);

static void enable_for_quickstart (HWND hDlg)
{
    int v = quickstart_ok && quickstart_ok_floppy ? TRUE : FALSE;
    ew (guiDlg, IDC_RESETAMIGA, !full_property_sheet ? TRUE : FALSE);
    ShowWindow (GetDlgItem (hDlg, IDC_QUICKSTART_SETCONFIG), quickstart ? SW_HIDE : SW_SHOW);
}

static void load_quickstart (HWND hDlg, int romcheck)
{
    ew (guiDlg, IDC_RESETAMIGA, FALSE);
    workprefs.nr_floppies = quickstart_floppy;
    quickstart_ok = built_in_prefs (&workprefs, quickstart_model, quickstart_conf, quickstart_compa, romcheck);
    enable_for_quickstart (hDlg);
    addfloppytype (hDlg, 0);
    addfloppytype (hDlg, 1);
    addfloppyhistory (hDlg);
    config_filename[0] = 0;
    setguititle (NULL);
}

static void quickstarthost (HWND hDlg, char *name)
{
    int type = CONFIG_TYPE_HOST;
    char tmp[MAX_DPATH];

    if (getconfigstorefrompath (name, tmp, CONFIG_TYPE_HOST)) {
	if (cfgfile_load (&workprefs, tmp, &type, 1))
	    workprefs.start_gui = 1;
    }
}

static void init_quickstartdlg_tooltip (HWND hDlg, char *tt)
{
    TOOLINFO ti;

    ti.cbSize = sizeof(TOOLINFO);
    ti.uFlags = TTF_SUBCLASS | TTF_IDISHWND;
    ti.hwnd = hDlg;
    ti.hinst = hInst;
    ti.uId = (UINT_PTR)GetDlgItem (hDlg, IDC_QUICKSTART_CONFIGURATION);
    ti.lpszText = tt;
    SendMessage (ToolTipHWND, TTM_DELTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);
    if (!tt)
	return;
    SendMessage (ToolTipHWND, TTM_ADDTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);
}

static void init_quickstartdlg (HWND hDlg)
{
    static int firsttime;
    int i, j, idx, idx2, qssize;
    char tmp1[2 * MAX_DPATH], tmp2[MAX_DPATH], hostconf[MAX_DPATH];
    char *p1, *p2;

    qssize = sizeof (tmp1);
    regquerystr (NULL, "QuickStartHostConfig", hostconf, &qssize);
    if (firsttime == 0) {
	regqueryint (NULL, "QuickStartModel", &quickstart_model);
	regqueryint (NULL, "QuickStartConfiguration", &quickstart_conf);
	regqueryint (NULL, "QuickStartCompatibility", &quickstart_compa);
	regqueryint (NULL, "QuickStartFloppies", &quickstart_floppy);
	if (quickstart) {
	    workprefs.df[0][0] = 0;
	    workprefs.df[1][0] = 0;
	    workprefs.df[2][0] = 0;
	    workprefs.df[3][0] = 0;
	    load_quickstart (hDlg, 1);
	    quickstarthost (hDlg, hostconf);
	}
       firsttime = 1;
    }

    CheckDlgButton (hDlg, IDC_QUICKSTARTMODE, quickstart);

    WIN32GUI_LoadUIString (IDS_QS_MODELS, tmp1, sizeof (tmp1));
    strcat (tmp1, "\n");
    p1 = tmp1;
    SendDlgItemMessage (hDlg, IDC_QUICKSTART_MODEL, CB_RESETCONTENT, 0, 0L);
    idx = idx2 = 0;
    i = 0;
    while (amodels[i].compalevels >= 0) {
	if (amodels[i].compalevels > 0) {
	    p2 = strchr (p1, '\n');
	    if (p2 && strlen (p2) > 0) {
		*p2++ = 0;
		SendDlgItemMessage (hDlg, IDC_QUICKSTART_MODEL, CB_ADDSTRING, 0, (LPARAM)p1);
		p1 = p2;
	    }
	    if (i == quickstart_model)
		idx2 = idx;
	    idx++;
	}
	i++;
    }
    SendDlgItemMessage (hDlg, IDC_QUICKSTART_MODEL, CB_SETCURSEL, idx2, 0);

    WIN32GUI_LoadUIString (amodels[quickstart_model].id, tmp1, sizeof (tmp1));
    strcat (tmp1, "\n");
    p1 = tmp1;
    init_quickstartdlg_tooltip (hDlg, 0);
    SendDlgItemMessage (hDlg, IDC_QUICKSTART_CONFIGURATION, CB_RESETCONTENT, 0, 0L);
    i = 0;
    for (;;) {
	p2 = strchr (p1, '\n');
	if (!p2)
	    break;
	*p2++= 0;
	SendDlgItemMessage (hDlg, IDC_QUICKSTART_CONFIGURATION, CB_ADDSTRING, 0, (LPARAM)p1);
	p1 = p2;
	p2 = strchr (p1, '\n');
	if (!p2)
	    break;
	*p2++= 0;
	if (quickstart_conf == i && strlen (p1) > 0)
	    init_quickstartdlg_tooltip (hDlg, p1);
	p1 = p2;
	i++;
    }
    if (quickstart_conf >= i)
	quickstart_conf = 0;
    SendDlgItemMessage (hDlg, IDC_QUICKSTART_CONFIGURATION, CB_SETCURSEL, quickstart_conf, 0);

    if (quickstart_compa >= amodels[quickstart_model].compalevels)
	quickstart_compa = 1;
    if (quickstart_compa >= amodels[quickstart_model].compalevels)
	quickstart_compa = 0;
    i = amodels[quickstart_model].compalevels;
    ew (hDlg, IDC_QUICKSTART_COMPATIBILITY, i > 1);
    SendDlgItemMessage (hDlg, IDC_QUICKSTART_COMPATIBILITY, TBM_SETRANGE, TRUE, MAKELONG (0, i > 1 ? i - 1 : 1));
    SendDlgItemMessage (hDlg, IDC_QUICKSTART_COMPATIBILITY, TBM_SETPAGESIZE, 0, 1);
    SendDlgItemMessage( hDlg, IDC_QUICKSTART_COMPATIBILITY, TBM_SETPOS, TRUE, quickstart_compa);

    SendDlgItemMessage (hDlg, IDC_QUICKSTART_HOSTCONFIG, CB_RESETCONTENT, 0, 0L);
    WIN32GUI_LoadUIString (IDS_DEFAULT_HOST, tmp1, sizeof (tmp1));
    SendDlgItemMessage (hDlg, IDC_QUICKSTART_HOSTCONFIG, CB_ADDSTRING, 0, (LPARAM)tmp1);
    idx = 0;
    j = 1;
    for (i = 0; i < configstoresize; i++) {
	if (configstore[i]->Type == CONFIG_TYPE_HOST) {
	    strcpy (tmp2, configstore[i]->Path);
	    strncat (tmp2, configstore[i]->Name, MAX_DPATH);
	    if (!strcmp (tmp2, hostconf))
		idx = j;
	    SendDlgItemMessage (hDlg, IDC_QUICKSTART_HOSTCONFIG, CB_ADDSTRING, 0, (LPARAM)tmp2);
	    j++;
	}
    }
    SendDlgItemMessage (hDlg, IDC_QUICKSTART_HOSTCONFIG, CB_SETCURSEL, idx, 0);
    regsetint (NULL, "QuickStartModel", quickstart_model);
    regsetint (NULL, "QuickStartConfiguration", quickstart_conf);
    regsetint (NULL, "QuickStartCompatibility", quickstart_compa);
}

static void floppytooltip (HWND hDlg, int num, uae_u32 crc32);
static void testimage (HWND hDlg, int num)
{
    int ret;
    int reload = 0;
    uae_u32 crc32;
    int messageid = -1;
    char tmp[MAX_DPATH];

    floppytooltip (hDlg, num, 0);
    quickstart_ok_floppy = 0;
    if (workprefs.dfxtype[0] < 0) {
	quickstart_ok_floppy = 1;
	return;
    }
    if (!workprefs.df[num][0])
	return;
    ret = DISK_examine_image (&workprefs, num, &crc32);
    if (!ret)
	return;
    floppytooltip (hDlg, num, crc32);
    if (num > 0)
	return;
    if (!full_property_sheet)
	return;
    switch (ret)
    {
	case 10:
	quickstart_ok_floppy = 1;
	break;
	case 11:
	quickstart_ok_floppy = 1;
	if (quickstart_model != 1 && quickstart_model != 2 && quickstart_model != 4) {
	    quickstart_model = 4;
	    messageid = IDS_IMGCHK_KS2;
	    reload = 1;
	}
	break;
	case 12:
	quickstart_ok_floppy = 1;
	if (quickstart_model != 4) {
	    quickstart_model = 4;
	    messageid = IDS_IMGCHK_KS3;
	    reload = 1;
	}
	break;
	case 4:
	messageid = IDS_IMGCHK_BOOTBLOCKNO;
	break;
	case 3:
	messageid = IDS_IMGCHK_BOOTBLOCKCRCERROR;
	break;
	case 2:
	messageid = IDS_IMGCHK_DAMAGED;
	break;
    }
    if (messageid > 0) {
	WIN32GUI_LoadUIString (messageid, tmp, sizeof (tmp));
	gui_message (tmp);
    }
    if (reload && quickstart) {
	load_quickstart (hDlg, 1);
	init_quickstartdlg (hDlg);
    }
}

static INT_PTR CALLBACK FloppyDlgProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam);

static INT_PTR CALLBACK QuickstartDlgProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int recursive;
    int ret = FALSE, i;
    char tmp[MAX_DPATH];
    static char df0[MAX_DPATH];
    static char df1[MAX_DPATH];
    static int dfxtype[2] = { -1, -1 };
    static int doinit;
    LRESULT val;

    switch(msg)
    {
	case WM_INITDIALOG:
	{
	    int ids[] = { IDC_DF0TEXTQ, IDC_DF1TEXTQ, -1 };
	    pages[QUICKSTART_ID] = hDlg;
	    currentpage = QUICKSTART_ID;
	    enable_for_quickstart (hDlg);
	    strcpy (df0, workprefs.df[0]);
	    strcpy (df1, workprefs.df[1]);
	    setmultiautocomplete (hDlg, ids);
	    doinit = 1;
	    break;
	}
	case WM_NULL:
	    if (recursive > 0)
		break;
	    recursive++;
	    if (doinit) {
		addfloppytype (hDlg, 0);
		addfloppytype (hDlg, 1);
		addfloppyhistory (hDlg);
		init_quickstartdlg (hDlg);
	    }
	    doinit = 0;
	    recursive--;
	break;

	case WM_COMMAND:
	if (recursive > 0)
	    break;
	recursive++;
	if (HIWORD (wParam) == CBN_SELCHANGE || HIWORD (wParam) == CBN_KILLFOCUS)  {
	    switch (LOWORD (wParam))
	    {
		case IDC_QUICKSTART_MODEL:
		val = SendDlgItemMessage (hDlg, IDC_QUICKSTART_MODEL, CB_GETCURSEL, 0, 0L);
		if (val != CB_ERR) {
		    i = 0;
		    while (amodels[i].compalevels >= 0) {
			if (amodels[i].compalevels > 0)
			    val--;
			if (val < 0)
			    break;
			i++;
		    }
		    if (i != quickstart_model) {
			quickstart_model = i;
			init_quickstartdlg (hDlg);
			if (quickstart)
			    load_quickstart (hDlg, 1);
			if (quickstart && !full_property_sheet)
			    qs_request_reset = 2;
		    }
		}
		break;
		case IDC_QUICKSTART_CONFIGURATION:
		val = SendDlgItemMessage (hDlg, IDC_QUICKSTART_CONFIGURATION, CB_GETCURSEL, 0, 0L);
		if (val != CB_ERR && val != quickstart_conf) {
		    quickstart_conf = val;
		    init_quickstartdlg (hDlg);
		    if (quickstart)
			load_quickstart (hDlg, 1);
		    if (quickstart && !full_property_sheet)
			qs_request_reset = 2;
		}
		break;
		case IDC_QUICKSTART_HOSTCONFIG:
		val = SendDlgItemMessage (hDlg, IDC_QUICKSTART_HOSTCONFIG, CB_GETCURSEL, 0, 0);
		if (val != CB_ERR) {
		    SendDlgItemMessage (hDlg, IDC_QUICKSTART_HOSTCONFIG, CB_GETLBTEXT, (WPARAM)val, (LPARAM)tmp);
		    regsetstr (NULL, "QuickStartHostConfig", tmp);
		    quickstarthost (hDlg, tmp);
		    if (val == 0 && quickstart)
			load_quickstart (hDlg, 0);
		}
		break;
	    }
	} else {
	    switch (LOWORD (wParam))
	    {
		case IDC_QUICKSTARTMODE:
		quickstart = IsDlgButtonChecked (hDlg, IDC_QUICKSTARTMODE);
		regsetint (NULL, "QuickStartMode", quickstart);
		if (quickstart) {
		    init_quickstartdlg (hDlg);
		    load_quickstart (hDlg, 0);
		}
		enable_for_quickstart (hDlg);
		break;
	    }
	}
	switch (LOWORD (wParam))
	{
	    case IDC_DF0TEXTQ:
	    case IDC_DF0WPQ:
	    case IDC_EJECT0Q:
	    case IDC_DF0QQ:
	    case IDC_DF1TEXTQ:
	    case IDC_DF1WPQ:
	    case IDC_EJECT1Q:
	    case IDC_DF1QQ:
	    case IDC_DF0QENABLE:
	    case IDC_DF1QENABLE:
	    if (currentpage == QUICKSTART_ID)
		ret = FloppyDlgProc (hDlg, msg, wParam, lParam);
	    break;
	    case IDC_QUICKSTART_SETCONFIG:
	    load_quickstart (hDlg, 1);
	    break;
	}
	recursive--;
	case WM_HSCROLL:
	if (recursive > 0)
	    break;
	recursive++;
	if ((HWND)lParam == GetDlgItem (hDlg, IDC_QUICKSTART_COMPATIBILITY)) {
	    val = SendMessage ((HWND)lParam, TBM_GETPOS, 0, 0);
	    if (val >= 0 && val != quickstart_compa) {
		quickstart_compa = val;
		init_quickstartdlg (hDlg);
		if (quickstart)
		    load_quickstart (hDlg, 0);
	    }
	}
	recursive--;
	break;
    }
    if (recursive == 0 && quickstart) {
	recursive++;
	if (strcmp (workprefs.df[0], df0) || workprefs.dfxtype[0] != dfxtype[0]) {
	    strcpy (df0, workprefs.df[0]);
	    dfxtype[0] = workprefs.dfxtype[0];
	    testimage (hDlg, 0);
	    enable_for_quickstart (hDlg);
	}
	if (strcmp (workprefs.df[1], df1) || workprefs.dfxtype[1] != dfxtype[1]) {
	    strcpy (df1, workprefs.df[1]);
	    dfxtype[1] = workprefs.dfxtype[1];
	    testimage (hDlg, 1);
	}
	recursive--;
    }
    return ret;
}

static void init_aboutdlg (HWND hDlg)
{
    CHARFORMAT CharFormat;
    int i;

    CharFormat.cbSize = sizeof (CharFormat);

    SetDlgItemText (hDlg, IDC_RICHEDIT1, "WinUAE");
    SendDlgItemMessage (hDlg, IDC_RICHEDIT1, EM_GETCHARFORMAT, 0, (LPARAM) & CharFormat);
    CharFormat.dwMask |= CFM_BOLD | CFM_SIZE | CFM_FACE;
    CharFormat.dwEffects = CFE_BOLD;
    CharFormat.yHeight = 18 * 20; /* height in twips, where a twip is 1/20th of a point - for a pt.size of 18 */

    strcpy (CharFormat.szFaceName, "Times New Roman");
    SendDlgItemMessage (hDlg, IDC_RICHEDIT1, EM_SETCHARFORMAT, SCF_ALL, (LPARAM) & CharFormat);
    SendDlgItemMessage (hDlg, IDC_RICHEDIT1, EM_SETBKGNDCOLOR, 0, GetSysColor (COLOR_3DFACE));

    SetDlgItemText (hDlg, IDC_RICHEDIT2, VersionStr );
    SendDlgItemMessage (hDlg, IDC_RICHEDIT2, EM_GETCHARFORMAT, 0, (LPARAM) & CharFormat);
    CharFormat.dwMask |= CFM_SIZE | CFM_FACE;
    CharFormat.yHeight = 10 * 20;
    strcpy (CharFormat.szFaceName, "Times New Roman");
    SendDlgItemMessage (hDlg, IDC_RICHEDIT2, EM_SETCHARFORMAT, SCF_ALL, (LPARAM) & CharFormat);
    SendDlgItemMessage (hDlg, IDC_RICHEDIT2, EM_SETBKGNDCOLOR, 0, GetSysColor (COLOR_3DFACE));

    for(i = 0; urls[i].id >= 0; i++)
	SetupRichText(hDlg, &urls[i]);
}

static INT_PTR CALLBACK AboutDlgProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch( msg )
    {
	case WM_INITDIALOG:
	    pages[ABOUT_ID] = hDlg;
	    currentpage = ABOUT_ID;
	    init_aboutdlg (hDlg);
	    break;

	case WM_COMMAND:
	    if (wParam == IDC_CONTRIBUTORS)
		DisplayContributors (hDlg);
	    break;
	case WM_SETCURSOR:
	    return TRUE;
	    break;
	case WM_LBUTTONDOWN:
	case WM_MOUSEMOVE:
	    url_handler( hDlg, msg, wParam, lParam );
	    break;
    }

    return FALSE;
}

static void enable_for_displaydlg (HWND hDlg)
{
    int rtg = ! workprefs.address_space_24;
#ifndef PICASSO96
    rtg = FALSE;
#endif
    ew (hDlg, IDC_SCREENMODE_RTG, rtg);
    ew (hDlg, IDC_XCENTER, TRUE);
    ew (hDlg, IDC_YCENTER, TRUE);
    ew (hDlg, IDC_LM_SCANLINES, TRUE);
    ew (hDlg, IDC_FRAMERATE2, !workprefs.gfx_avsync);
    ew (hDlg, IDC_FRAMERATE, !workprefs.cpu_cycle_exact);
    ew (hDlg, IDC_LORES, !workprefs.gfx_autoresolution);
    ew (hDlg, IDC_LM_NORMAL, !workprefs.gfx_autoresolution);
    ew (hDlg, IDC_LM_DOUBLED, !workprefs.gfx_autoresolution);
    ew (hDlg, IDC_LM_SCANLINES, !workprefs.gfx_autoresolution);
}

static void enable_for_chipsetdlg (HWND hDlg)
{
    int enable = workprefs.cpu_cycle_exact ? FALSE : TRUE;

#if !defined (CPUEMU_12)
    ew (hDlg, IDC_CYCLEEXACT, FALSE);
#endif
    ew (hDlg, IDC_FASTCOPPER, enable);
    ew (hDlg, IDC_GENLOCK, full_property_sheet);
    ew (hDlg, IDC_BLITIMM, enable);
    if (enable == FALSE) {
	workprefs.immediate_blits = 0;
	CheckDlgButton (hDlg, IDC_FASTCOPPER, FALSE);
	CheckDlgButton (hDlg, IDC_BLITIMM, FALSE);
    }
    ew (hDlg, IDC_CS_EXT, workprefs.cs_compatible ? TRUE : FALSE);
}

static DWORD idnth[] = { IDS_SECOND, IDS_THIRD, IDS_FOURTH, IDS_FIFTH, IDS_SIXTH, IDS_SEVENTH, IDS_EIGHTH, IDS_NINTH, IDS_TENTH, -1 };
static void LoadNthString( DWORD value, char *nth, DWORD dwNthMax )
{
    nth[0] = 0;
    if (value >= 1 && value <= 9)
	WIN32GUI_LoadUIString(idnth[value - 1], nth, dwNthMax);
}

static int fakerefreshrates[] = { 50, 60, 100, 120, 0 };
static int storedrefreshrates[MAX_REFRESH_RATES + 1];

static void init_frequency_combo (HWND hDlg, int dmode)
{
    int i, j, freq, tmp;
    char hz[20], hz2[20], txt[100];
    LRESULT index;

    i = 0; index = 0;
    while ((freq = DisplayModes[dmode].refresh[i]) > 0 && index < MAX_REFRESH_RATES) {
	storedrefreshrates[index++] = freq;
	i++;
    }
    i = 0;
    while ((freq = fakerefreshrates[i]) > 0 && index < MAX_REFRESH_RATES) {
	for (j = 0; j < index; j++) {
	    if (storedrefreshrates[j] == freq) break;
	}
	if (j == index)
	    storedrefreshrates[index++] = -freq;
	i++;
    }
    storedrefreshrates[index] = 0;
    for (i = 0; i < index; i++) {
	for (j = i + 1; j < index; j++) {
	    if (abs(storedrefreshrates[i]) >= abs(storedrefreshrates[j])) {
		tmp = storedrefreshrates[i];
		storedrefreshrates[i] = storedrefreshrates[j];
		storedrefreshrates[j] = tmp;
	    }
	}
    }

    hz[0] = hz2[0] = 0;
    SendDlgItemMessage(hDlg, IDC_REFRESHRATE, CB_RESETCONTENT, 0, 0);
    WIN32GUI_LoadUIString (IDS_VSYNC_DEFAULT, txt, sizeof (txt));
    SendDlgItemMessage(hDlg, IDC_REFRESHRATE, CB_ADDSTRING, 0, (LPARAM)txt);
    for (i = 0; i < index; i++) {
	freq = storedrefreshrates[i];
	if (freq < 0) {
	    freq = -freq;
	    sprintf (hz, "(%dHz)", freq);
	} else {
	    sprintf (hz, "%dHz", freq);
	}
	if (freq == 50 || freq == 100)
	    strcat (hz, " PAL");
	if (freq == 60 || freq == 120)
	    strcat (hz, " NTSC");
	if (abs(workprefs.gfx_refreshrate) == freq)
	    strcpy (hz2, hz);
	SendDlgItemMessage (hDlg, IDC_REFRESHRATE, CB_ADDSTRING, 0, (LPARAM)hz);
    }
    index = CB_ERR;
    if (hz2[0] >= 0)
	index = SendDlgItemMessage (hDlg, IDC_REFRESHRATE, CB_SELECTSTRING, 0, (LPARAM)hz2);
    if (index == CB_ERR) {
	WIN32GUI_LoadUIString (IDS_VSYNC_DEFAULT, txt, sizeof (txt));
	SendDlgItemMessage(hDlg, IDC_REFRESHRATE, CB_SELECTSTRING, i, (LPARAM)txt);
	workprefs.gfx_refreshrate = 0;
    }
}

#define MAX_FRAMERATE_LENGTH 40
#define MAX_NTH_LENGTH 20

static int display_mode_index(uae_u32 x, uae_u32 y, uae_u32 d)
{
    int i;

    i = 0;
    while (DisplayModes[i].depth >= 0) {
	if (DisplayModes[i].res.width == x &&
	    DisplayModes[i].res.height == y &&
	    DisplayModes[i].depth == d)
	    break;
	i++;
    }
    if(DisplayModes[i].depth < 0)
	i = -1;
    return i;
}

static int da_mode_selected;

static int *getp_da (void)
{
    int *p = 0;
    switch (da_mode_selected)
    {
	case 0:
	p = &workprefs.gfx_luminance;
	break;
	case 1:
	p = &workprefs.gfx_contrast;
	break;
	case 2:
	p = &workprefs.gfx_gamma;
	break;
    }
    return p;
}

static void update_da(void)
{
    currprefs.gfx_gamma = workprefs.gfx_gamma;
    currprefs.gfx_luminance = workprefs.gfx_luminance;
    currprefs.gfx_contrast = workprefs.gfx_contrast;
    init_colors ();
    init_custom();
    updatedisplayarea ();
    WIN32GFX_WindowMove ();
}
static void handle_da (HWND hDlg)
{
    int *p;
    int v;

    p = getp_da ();
    if (!p)
	return;
    v = SendDlgItemMessage (hDlg, IDC_DA_SLIDER, TBM_GETPOS, 0, 0) * 10;
    if (v == *p)
	return;
    *p = v;
    update_da();
}

void init_da (HWND hDlg)
{
    int *p;
    SendDlgItemMessage(hDlg, IDC_DA_MODE, CB_RESETCONTENT, 0, 0);
    SendDlgItemMessage(hDlg, IDC_DA_MODE, CB_ADDSTRING, 0, (LPARAM)"Brightness");
    SendDlgItemMessage(hDlg, IDC_DA_MODE, CB_ADDSTRING, 0, (LPARAM)"Contrast");
    SendDlgItemMessage(hDlg, IDC_DA_MODE, CB_ADDSTRING, 0, (LPARAM)"Gamma");
    if (da_mode_selected == CB_ERR)
	da_mode_selected = 0;
    SendDlgItemMessage (hDlg, IDC_DA_MODE, CB_SETCURSEL, da_mode_selected, 0);
    SendDlgItemMessage (hDlg, IDC_DA_SLIDER, TBM_SETPAGESIZE, 0, 1);
    SendDlgItemMessage (hDlg, IDC_DA_SLIDER, TBM_SETRANGE, TRUE, MAKELONG (-99, 99));
    p = getp_da ();
    if (p)
	SendDlgItemMessage (hDlg, IDC_DA_SLIDER, TBM_SETPOS, TRUE, (*p) / 10);
}

static int gui_display_depths[3];
static void init_display_mode (HWND hDlg)
{
   int d, d2, index;

   switch (workprefs.color_mode)
    {
    case 2:
	d = 16;
	break;
    case 5:
	d = 32;
	break;
    default:
	d = 8;
	break;
    }

    if (workprefs.gfx_afullscreen) {
	d2 = d;
	if ((index = WIN32GFX_AdjustScreenmode(&workprefs.gfx_size_fs.width, &workprefs.gfx_size_fs.height, &d2)) >= 0) {
	    switch (d2)
	    {
	    case 15:
		workprefs.color_mode = 1;
		d = 2;
		break;
	    case 16:
		workprefs.color_mode = 2;
		d = 2;
		break;
	    case 32:
		workprefs.color_mode = 5;
		d = 4;
		break;
	    default:
		workprefs.color_mode = 0;
		d = 1;
		break;
	    }
	}
    } else {
	d = d / 8;
    }

    if ((index = display_mode_index (workprefs.gfx_size_fs.width, workprefs.gfx_size_fs.height, d)) >= 0) {
	int i, cnt;
	SendDlgItemMessage (hDlg, IDC_RESOLUTION, CB_SETCURSEL, DisplayModes[index].residx, 0);
	SendDlgItemMessage(hDlg, IDC_RESOLUTIONDEPTH, CB_RESETCONTENT, 0, 0);
	cnt = 0;
	gui_display_depths[0] = gui_display_depths[1] = gui_display_depths[2] = -1;
	for (i = 0; DisplayModes[i].depth >= 0; i++) {
	    if (DisplayModes[i].residx == DisplayModes[index].residx) {
		char tmp[64];
		sprintf (tmp, "%d", DisplayModes[i].depth * 8);
		SendDlgItemMessage(hDlg, IDC_RESOLUTIONDEPTH, CB_ADDSTRING, 0, (LPARAM)tmp);
		if (DisplayModes[i].depth == d)
		    SendDlgItemMessage (hDlg, IDC_RESOLUTIONDEPTH, CB_SETCURSEL, cnt, 0);
		gui_display_depths[cnt] = DisplayModes[i].depth;
		cnt++;
	    }
	}
	init_frequency_combo (hDlg, index);
    }
}

static int display_toselect(int fs, int vsync, int p96)
{
    if (p96)
	return fs *  2 + vsync;
    if (fs == 2)
	return 3;
    if (!vsync)
	return fs;
    if (fs == 1 && vsync)
	return 2;
    return fs;
}
static void display_fromselect(int val, int *fs, int *vsync, int p96)
{
    if (val == CB_ERR)
	return;
    *fs = 0;
    *vsync = 0;
    if (p96) {
	*fs = val / 2;
	*vsync = val & 1;
	return;
    }
    switch (val)
    {
	case 0:
	*fs = 0;
	break;
	case 1:
	*fs = 1;
	break;
	case 2:
	*fs = 1;
	*vsync = 1;
	break;
	case 3:
	*fs = 2;
	break;
    }
}

static void values_to_displaydlg (HWND hDlg)
{
    char buffer[MAX_DPATH], buffer2[MAX_DPATH];
    char Nth[MAX_NTH_LENGTH];
    LPSTR blah[1] = { Nth };
    LPTSTR string = NULL;
    int v;

    init_display_mode (hDlg);

    SetDlgItemInt (hDlg, IDC_XSIZE, workprefs.gfx_size_win.width, FALSE);
    SetDlgItemInt (hDlg, IDC_YSIZE, workprefs.gfx_size_win.height, FALSE);

    v = workprefs.chipset_refreshrate;
    if (v == 0)
	v = currprefs.ntscmode ? 60 : 50;
    SendDlgItemMessage (hDlg, IDC_FRAMERATE2, TBM_SETPOS, TRUE, v);
    sprintf (buffer, "%d", v);
    SetDlgItemText (hDlg, IDC_RATE2TEXT, buffer);

    v = workprefs.cpu_cycle_exact ? 1 : workprefs.gfx_framerate;
    SendDlgItemMessage (hDlg, IDC_FRAMERATE, TBM_SETPOS, TRUE, v);
    WIN32GUI_LoadUIString(IDS_FRAMERATE, buffer, sizeof buffer);
    LoadNthString (v - 1, Nth, MAX_NTH_LENGTH);
    if(FormatMessage(FORMAT_MESSAGE_FROM_STRING | FORMAT_MESSAGE_ARGUMENT_ARRAY | FORMAT_MESSAGE_ALLOCATE_BUFFER,
	    buffer, 0, 0, (LPTSTR)&string, MAX_FRAMERATE_LENGTH + MAX_NTH_LENGTH, (va_list *)blah ) == 0)
    {
	DWORD dwLastError = GetLastError();
	sprintf (buffer, "Every %s Frame", nth[v - 1]);
	SetDlgItemText(hDlg, IDC_RATETEXT, buffer);
    } else {
	SetDlgItemText( hDlg, IDC_RATETEXT, string);
	LocalFree(string);
    }

    CheckRadioButton (hDlg, IDC_LM_NORMAL, IDC_LM_SCANLINES, IDC_LM_NORMAL + workprefs.gfx_linedbl);

    SendDlgItemMessage(hDlg, IDC_SCREENMODE_NATIVE, CB_RESETCONTENT, 0, 0);
    WIN32GUI_LoadUIString(IDS_SCREEN_WINDOWED, buffer, sizeof buffer);
    SendDlgItemMessage(hDlg, IDC_SCREENMODE_NATIVE, CB_ADDSTRING, 0, (LPARAM)buffer);
    WIN32GUI_LoadUIString(IDS_SCREEN_FULLSCREEN, buffer, sizeof buffer);
    SendDlgItemMessage(hDlg, IDC_SCREENMODE_NATIVE, CB_ADDSTRING, 0, (LPARAM)buffer);
    WIN32GUI_LoadUIString(IDS_SCREEN_VSYNC, buffer2, sizeof buffer2);
    sprintf (buffer + strlen(buffer), " + %s", buffer2);
    SendDlgItemMessage(hDlg, IDC_SCREENMODE_NATIVE, CB_ADDSTRING, 0, (LPARAM)buffer);
    WIN32GUI_LoadUIString(IDS_SCREEN_FULLWINDOW, buffer, sizeof buffer);
    SendDlgItemMessage(hDlg, IDC_SCREENMODE_NATIVE, CB_ADDSTRING, 0, (LPARAM)buffer);
    SendDlgItemMessage(hDlg, IDC_SCREENMODE_NATIVE, CB_SETCURSEL, display_toselect(workprefs.gfx_afullscreen, workprefs.gfx_avsync, 0), 0);

    SendDlgItemMessage(hDlg, IDC_SCREENMODE_RTG, CB_RESETCONTENT, 0, 0);
    WIN32GUI_LoadUIString(IDS_SCREEN_WINDOWED, buffer, sizeof buffer);
    SendDlgItemMessage(hDlg, IDC_SCREENMODE_RTG, CB_ADDSTRING, 0, (LPARAM)buffer);
    sprintf (buffer + strlen(buffer), " + %s", buffer2);
    SendDlgItemMessage(hDlg, IDC_SCREENMODE_RTG, CB_ADDSTRING, 0, (LPARAM)buffer);
    WIN32GUI_LoadUIString(IDS_SCREEN_FULLSCREEN, buffer, sizeof buffer);
    SendDlgItemMessage(hDlg, IDC_SCREENMODE_RTG, CB_ADDSTRING, 0, (LPARAM)buffer);
    WIN32GUI_LoadUIString(IDS_SCREEN_VSYNC, buffer2, sizeof buffer2);
    sprintf (buffer + strlen(buffer), " + %s", buffer2);
    SendDlgItemMessage(hDlg, IDC_SCREENMODE_RTG, CB_ADDSTRING, 0, (LPARAM)buffer);
    WIN32GUI_LoadUIString(IDS_SCREEN_FULLWINDOW, buffer, sizeof buffer);
    SendDlgItemMessage(hDlg, IDC_SCREENMODE_RTG, CB_ADDSTRING, 0, (LPARAM)buffer);
    sprintf (buffer + strlen(buffer), " + %s", buffer2);
    SendDlgItemMessage(hDlg, IDC_SCREENMODE_RTG, CB_ADDSTRING, 0, (LPARAM)buffer);
    SendDlgItemMessage(hDlg, IDC_SCREENMODE_RTG, CB_SETCURSEL, display_toselect(workprefs.gfx_pfullscreen, workprefs.gfx_pvsync, 1), 0);

    SendDlgItemMessage(hDlg, IDC_LORES, CB_RESETCONTENT, 0, 0);
    SendDlgItemMessage(hDlg, IDC_LORES, CB_ADDSTRING, 0, (LPARAM)"Lores");
    SendDlgItemMessage(hDlg, IDC_LORES, CB_ADDSTRING, 0, (LPARAM)"Hires (normal)");
    SendDlgItemMessage(hDlg, IDC_LORES, CB_ADDSTRING, 0, (LPARAM)"SuperHires");
    SendDlgItemMessage (hDlg, IDC_LORES, CB_SETCURSEL, workprefs.gfx_resolution, 0);

    CheckDlgButton (hDlg, IDC_ASPECT, workprefs.gfx_correct_aspect);
    CheckDlgButton (hDlg, IDC_LORES_SMOOTHED, workprefs.gfx_lores_mode);

    CheckDlgButton (hDlg, IDC_XCENTER, workprefs.gfx_xcenter);
    CheckDlgButton (hDlg, IDC_YCENTER, workprefs.gfx_ycenter);

    init_da (hDlg);
}

static void init_resolution_combo (HWND hDlg)
{
    int i = 0, idx = -1;
    char tmp[64];

    SendDlgItemMessage(hDlg, IDC_RESOLUTION, CB_RESETCONTENT, 0, 0);
    while (DisplayModes[i].depth >= 0) {
	if (DisplayModes[i].residx != idx) {
	    sprintf (tmp, "%dx%d", DisplayModes[i].res.width, DisplayModes[i].res.height);
	    SendDlgItemMessage(hDlg, IDC_RESOLUTION, CB_ADDSTRING, 0, (LPARAM)tmp);
	    idx = DisplayModes[i].residx;
	}
	i++;
    }
}
static void init_displays_combo (HWND hDlg)
{
    int i = 0;
    SendDlgItemMessage(hDlg, IDC_DISPLAYSELECT, CB_RESETCONTENT, 0, 0);
    while (Displays[i].name) {
	SendDlgItemMessage(hDlg, IDC_DISPLAYSELECT, CB_ADDSTRING, 0, (LPARAM)Displays[i].name);
	i++;
    }
    if (workprefs.gfx_display >= i)
	workprefs.gfx_display = 0;
    SendDlgItemMessage(hDlg, IDC_DISPLAYSELECT, CB_SETCURSEL, workprefs.gfx_display, 0);
}

static void values_from_displaydlg (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    BOOL success = FALSE;
    int i, j;
    int gfx_width = workprefs.gfx_size_win.width;
    int gfx_height = workprefs.gfx_size_win.height;
    LRESULT posn;

    display_fromselect(SendDlgItemMessage (hDlg, IDC_SCREENMODE_NATIVE, CB_GETCURSEL, 0, 0),
	&workprefs.gfx_afullscreen, &workprefs.gfx_avsync, 0);
    display_fromselect(SendDlgItemMessage (hDlg, IDC_SCREENMODE_RTG, CB_GETCURSEL, 0, 0),
	&workprefs.gfx_pfullscreen, &workprefs.gfx_pvsync, 1);

    workprefs.gfx_lores_mode     = IsDlgButtonChecked (hDlg, IDC_LORES_SMOOTHED);
    workprefs.gfx_correct_aspect = IsDlgButtonChecked (hDlg, IDC_ASPECT);
    workprefs.gfx_linedbl = (IsDlgButtonChecked(hDlg, IDC_LM_SCANLINES) ? 2 :
			      IsDlgButtonChecked(hDlg, IDC_LM_DOUBLED) ? 1 : 0);

    workprefs.gfx_framerate = SendDlgItemMessage (hDlg, IDC_FRAMERATE, TBM_GETPOS, 0, 0);
    workprefs.chipset_refreshrate = SendDlgItemMessage (hDlg, IDC_FRAMERATE2, TBM_GETPOS, 0, 0);

    {
	char buffer[MAX_FRAMERATE_LENGTH];
	char Nth[MAX_NTH_LENGTH];
	LPSTR blah[1] = { Nth };
	LPTSTR string = NULL;

	WIN32GUI_LoadUIString(IDS_FRAMERATE, buffer, MAX_FRAMERATE_LENGTH);
	LoadNthString(workprefs.gfx_framerate - 1, Nth, MAX_NTH_LENGTH);
	if(FormatMessage(FORMAT_MESSAGE_FROM_STRING | FORMAT_MESSAGE_ARGUMENT_ARRAY | FORMAT_MESSAGE_ALLOCATE_BUFFER,
			   buffer, 0, 0, (LPTSTR)&string, MAX_FRAMERATE_LENGTH + MAX_NTH_LENGTH, (va_list *)blah ) == 0)
	{
	    DWORD dwLastError = GetLastError();
	    sprintf (buffer, "Every %s Frame", nth[workprefs.gfx_framerate - 1]);
	    SetDlgItemText(hDlg, IDC_RATETEXT, buffer);
	}
	else
	{
	    SetDlgItemText(hDlg, IDC_RATETEXT, string);
	    LocalFree(string);
	}
	sprintf (buffer, "%d", workprefs.chipset_refreshrate);
	SetDlgItemText (hDlg, IDC_RATE2TEXT, buffer);
	workprefs.gfx_size_win.width  = GetDlgItemInt(hDlg, IDC_XSIZE, &success, FALSE);
	if(!success)
	    workprefs.gfx_size_win.width = 800;
	workprefs.gfx_size_win.height = GetDlgItemInt(hDlg, IDC_YSIZE, &success, FALSE);
	if(!success)
	    workprefs.gfx_size_win.height = 600;
    }
    if (workprefs.chipset_refreshrate == (currprefs.ntscmode ? 60 : 50))
	workprefs.chipset_refreshrate = 0;
    workprefs.gfx_xcenter = (IsDlgButtonChecked (hDlg, IDC_XCENTER) ? 2 : 0 ); /* Smart centering */
    workprefs.gfx_ycenter = (IsDlgButtonChecked (hDlg, IDC_YCENTER) ? 2 : 0 ); /* Smart centering */

    if (msg == WM_COMMAND && HIWORD (wParam) == CBN_SELCHANGE)
    {
	if (LOWORD (wParam) == IDC_DISPLAYSELECT) {
	    posn = SendDlgItemMessage (hDlg, IDC_DISPLAYSELECT, CB_GETCURSEL, 0, 0);
	    if (posn != CB_ERR && posn != workprefs.gfx_display) {
		if (Displays[posn].disabled)
		    posn = 0;
		workprefs.gfx_display = posn;
		DisplayModes = Displays[workprefs.gfx_display].DisplayModes;
		init_resolution_combo (hDlg);
		init_display_mode (hDlg);
	    }
	    return;
	} else if (LOWORD (wParam) == IDC_LORES) {
	    posn = SendDlgItemMessage (hDlg, IDC_LORES, CB_GETCURSEL, 0, 0);
	    if (posn != CB_ERR)
		workprefs.gfx_resolution = posn;
	} else if (LOWORD (wParam) == IDC_RESOLUTION || LOWORD(wParam) == IDC_RESOLUTIONDEPTH) {
	    LRESULT posn1, posn2;
	    posn1 = SendDlgItemMessage (hDlg, IDC_RESOLUTION, CB_GETCURSEL, 0, 0);
	    if (posn1 == CB_ERR)
		return;
	    posn2 = SendDlgItemMessage (hDlg, IDC_RESOLUTIONDEPTH, CB_GETCURSEL, 0, 0);
	    if (posn2 == CB_ERR)
		return;
	    for (i = 0; DisplayModes[i].depth >= 0; i++) {
		if (DisplayModes[i].residx == posn1)
		    break;
	    }
	    if (DisplayModes[i].depth < 0)
		return;
	    j = i;
	    while (DisplayModes[i].residx == posn1) {
		if (DisplayModes[i].depth == gui_display_depths[posn2])
		    break;
		i++;
	    }
	    if (DisplayModes[i].residx != posn1)
		i = j;
	    workprefs.gfx_size_fs.width  = DisplayModes[i].res.width;
	    workprefs.gfx_size_fs.height = DisplayModes[i].res.height;
	    switch(DisplayModes[i].depth)
	    {
	    case 2:
		workprefs.color_mode = 2;
		break;
	    case 3:
	    case 4:
		workprefs.color_mode = 5;
		break;
	    default:
		workprefs.color_mode = 0;
		break;
	    }
	    /* Set the Int boxes */
	    SetDlgItemInt (hDlg, IDC_XSIZE, workprefs.gfx_size_win.width, FALSE);
	    SetDlgItemInt (hDlg, IDC_YSIZE, workprefs.gfx_size_win.height, FALSE);
	    init_frequency_combo (hDlg, i);
	} else if (LOWORD (wParam) == IDC_REFRESHRATE) {
	    LRESULT posn1, posn2;
	    posn1 = SendDlgItemMessage (hDlg, IDC_REFRESHRATE, CB_GETCURSEL, 0, 0);
	    if (posn1 == CB_ERR)
		return;
	    posn2 = SendDlgItemMessage (hDlg, IDC_RESOLUTION, CB_GETCURSEL, 0, 0);
	    if (posn2 == CB_ERR)
		return;
	    if (posn1 == 0) {
		workprefs.gfx_refreshrate = 0;
	    } else {
		posn1--;
		workprefs.gfx_refreshrate = storedrefreshrates[posn1];
	    }
	    values_to_displaydlg (hDlg);
	} else if (LOWORD (wParam) == IDC_DA_MODE) {
	    da_mode_selected = SendDlgItemMessage (hDlg, IDC_DA_MODE, CB_GETCURSEL, 0, 0);
	    init_da (hDlg);
	    handle_da (hDlg);
	}
    }

    updatewinfsmode (&workprefs);
}

static int hw3d_changed;

static INT_PTR CALLBACK DisplayDlgProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int recursive = 0;

    switch (msg)
    {
    case WM_INITDIALOG:
	pages[DISPLAY_ID] = hDlg;
	currentpage = DISPLAY_ID;
	SendDlgItemMessage (hDlg, IDC_FRAMERATE, TBM_SETPAGESIZE, 0, 1);
	SendDlgItemMessage (hDlg, IDC_FRAMERATE, TBM_SETRANGE, TRUE, MAKELONG (MIN_REFRESH_RATE, MAX_REFRESH_RATE));
	SendDlgItemMessage (hDlg, IDC_FRAMERATE2, TBM_SETPAGESIZE, 0, 1);
	SendDlgItemMessage (hDlg, IDC_FRAMERATE2, TBM_SETRANGE, TRUE, MAKELONG (1, 99));
	init_displays_combo (hDlg);
	init_resolution_combo (hDlg);
	init_da (hDlg);

    case WM_USER:
	recursive++;
	values_to_displaydlg (hDlg);
	enable_for_displaydlg (hDlg);
	recursive--;
	break;

    case WM_HSCROLL:
    case WM_COMMAND:
	if (recursive > 0)
	    break;
	recursive++;
	if (LOWORD (wParam) == IDC_DA_RESET) {
	    int *p;
	    da_mode_selected = SendDlgItemMessage (hDlg, IDC_DA_MODE, CB_GETCURSEL, 0, 0);
	    p = getp_da();
	    if (p)
		*p = 0;
	    init_da(hDlg);
	    update_da();
	}
	handle_da (hDlg);
	values_from_displaydlg (hDlg, msg, wParam, lParam);
	enable_for_displaydlg (hDlg);
	recursive--;
	break;

    }
    if (hw3d_changed && recursive == 0) {
	recursive++;
	enable_for_displaydlg (hDlg);
	values_to_displaydlg (hDlg);
	hw3d_changed = 0;
	recursive--;
    }
    return FALSE;
}

static void values_to_chipsetdlg (HWND hDlg)
{
    char Nth[MAX_NTH_LENGTH];
    LPSTR blah[1] = { Nth };
    LPTSTR string = NULL;
    int which_button;

    switch(workprefs.chipset_mask)
    {
    case 0:
	CheckRadioButton(hDlg, IDC_OCS, IDC_AGA, IDC_OCS + 0);
	break;
    case CSMASK_ECS_AGNUS:
	CheckRadioButton(hDlg, IDC_OCS, IDC_AGA, IDC_OCS + 1);
	break;
    case CSMASK_ECS_DENISE:
	CheckRadioButton(hDlg, IDC_OCS, IDC_AGA, IDC_OCS + 2);
	break;
    case CSMASK_ECS_AGNUS | CSMASK_ECS_DENISE:
	CheckRadioButton(hDlg, IDC_OCS, IDC_AGA, IDC_OCS + 3);
	break;
    case CSMASK_AGA:
    case CSMASK_ECS_AGNUS | CSMASK_ECS_DENISE | CSMASK_AGA:
	CheckRadioButton(hDlg, IDC_OCS, IDC_AGA, IDC_OCS + 4);
	break;
    }
    CheckDlgButton (hDlg, IDC_NTSC, workprefs.ntscmode);
    CheckDlgButton (hDlg, IDC_GENLOCK, workprefs.genlock);
    CheckDlgButton (hDlg, IDC_BLITIMM, workprefs.immediate_blits);
    CheckDlgButton (hDlg, IDC_FASTERRTG, workprefs.picasso96_nocustom);
    CheckRadioButton (hDlg, IDC_COLLISION0, IDC_COLLISION3, IDC_COLLISION0 + workprefs.collision_level);
    CheckDlgButton (hDlg, IDC_CYCLEEXACT, workprefs.cpu_cycle_exact);
    switch (workprefs.produce_sound) {
     case 0: which_button = IDC_CS_SOUND0; break;
     case 1: case 2: which_button = IDC_CS_SOUND1; break;
     default: which_button = IDC_CS_SOUND2; break;
    }
    CheckRadioButton (hDlg, IDC_CS_SOUND0, IDC_CS_SOUND2, which_button);
    SendDlgItemMessage (hDlg, IDC_CS_EXT, CB_RESETCONTENT, 0, 0);
    SendDlgItemMessage (hDlg, IDC_CS_EXT, CB_ADDSTRING, 0, (LPARAM)"");
    SendDlgItemMessage (hDlg, IDC_CS_EXT, CB_ADDSTRING, 0, (LPARAM)"Generic");
    SendDlgItemMessage (hDlg, IDC_CS_EXT, CB_ADDSTRING, 0, (LPARAM)"CDTV");
    SendDlgItemMessage (hDlg, IDC_CS_EXT, CB_ADDSTRING, 0, (LPARAM)"CD32");
    SendDlgItemMessage (hDlg, IDC_CS_EXT, CB_ADDSTRING, 0, (LPARAM)"A500");
    SendDlgItemMessage (hDlg, IDC_CS_EXT, CB_ADDSTRING, 0, (LPARAM)"A500+");
    SendDlgItemMessage (hDlg, IDC_CS_EXT, CB_ADDSTRING, 0, (LPARAM)"A600");
    SendDlgItemMessage (hDlg, IDC_CS_EXT, CB_ADDSTRING, 0, (LPARAM)"A1000");
    SendDlgItemMessage (hDlg, IDC_CS_EXT, CB_ADDSTRING, 0, (LPARAM)"A1200");
    SendDlgItemMessage (hDlg, IDC_CS_EXT, CB_ADDSTRING, 0, (LPARAM)"A2000");
    SendDlgItemMessage (hDlg, IDC_CS_EXT, CB_ADDSTRING, 0, (LPARAM)"A3000");
    SendDlgItemMessage (hDlg, IDC_CS_EXT, CB_ADDSTRING, 0, (LPARAM)"A3000T");
    SendDlgItemMessage (hDlg, IDC_CS_EXT, CB_ADDSTRING, 0, (LPARAM)"A4000");
    SendDlgItemMessage (hDlg, IDC_CS_EXT, CB_ADDSTRING, 0, (LPARAM)"A4000T");
    SendDlgItemMessage (hDlg, IDC_CS_EXT, CB_SETCURSEL, workprefs.cs_compatible, 0);
}

static void values_from_chipsetdlg (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    BOOL success = FALSE;
    int n, snd;

    workprefs.genlock = IsDlgButtonChecked (hDlg, IDC_GENLOCK);
    workprefs.immediate_blits = IsDlgButtonChecked (hDlg, IDC_BLITIMM);
    workprefs.picasso96_nocustom = IsDlgButtonChecked (hDlg, IDC_FASTERRTG);
    n = IsDlgButtonChecked (hDlg, IDC_CYCLEEXACT) ? 1 : 0;
    if (workprefs.cpu_cycle_exact != n) {
	workprefs.cpu_cycle_exact = workprefs.blitter_cycle_exact = n;
	if (n) {
	    if (workprefs.cpu_model == 68000) {
		workprefs.cpu_compatible = 1;
		workprefs.m68k_speed = 0;
	    }
	    workprefs.immediate_blits = 0;
	    workprefs.gfx_framerate = 1;
	}
    }
    workprefs.collision_level = IsDlgButtonChecked (hDlg, IDC_COLLISION0) ? 0
				 : IsDlgButtonChecked (hDlg, IDC_COLLISION1) ? 1
				 : IsDlgButtonChecked (hDlg, IDC_COLLISION2) ? 2 : 3;
    workprefs.chipset_mask = IsDlgButtonChecked (hDlg, IDC_OCS) ? 0
			      : IsDlgButtonChecked (hDlg, IDC_ECS_AGNUS) ? CSMASK_ECS_AGNUS
			      : IsDlgButtonChecked (hDlg, IDC_ECS_DENISE) ? CSMASK_ECS_DENISE
			      : IsDlgButtonChecked (hDlg, IDC_ECS) ? CSMASK_ECS_AGNUS | CSMASK_ECS_DENISE
			      : CSMASK_AGA | CSMASK_ECS_AGNUS | CSMASK_ECS_DENISE;
    n = IsDlgButtonChecked (hDlg, IDC_NTSC) ? 1 : 0;
    if (workprefs.ntscmode != n) {
	workprefs.ntscmode = n;
#ifdef AVIOUTPUT
	avioutput_fps = n ? VBLANK_HZ_NTSC : VBLANK_HZ_PAL;
#endif
    }
    snd = IsDlgButtonChecked (hDlg, IDC_CS_SOUND0) ? 0
	: IsDlgButtonChecked (hDlg, IDC_CS_SOUND1) ? 2 : 3;
    if (snd == 0 || snd == 3) {
	workprefs.produce_sound = snd;
    } else if (snd == 2) {
	if (workprefs.produce_sound == 0)
	    workprefs.produce_sound = 2;
	else if (workprefs.produce_sound >= 2)
	    workprefs.produce_sound = 2;
    }
    n = SendDlgItemMessage (hDlg, IDC_CS_EXT, CB_GETCURSEL, 0, 0);
    if (n != CB_ERR) {
	workprefs.cs_compatible = n;
	built_in_chipset_prefs (&workprefs);
    }
}

static INT_PTR CALLBACK ChipsetDlgProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int recursive = 0;

    switch (msg) {
    case WM_INITDIALOG:
	pages[CHIPSET_ID] = hDlg;
	currentpage = CHIPSET_ID;
#ifndef	AGA
	ew (hDlg, IDC_AGA, FALSE);
#endif

    case WM_USER:
	recursive++;
	values_to_chipsetdlg (hDlg);
	enable_for_chipsetdlg (hDlg);
	recursive--;
	break;

    case WM_HSCROLL:
    case WM_COMMAND:
	if (recursive > 0)
	    break;
	recursive++;
	values_from_chipsetdlg (hDlg, msg, wParam, lParam);
	enable_for_chipsetdlg( hDlg );
	recursive--;
	break;
    }
    return FALSE;
}

static void values_to_chipsetdlg2 (HWND hDlg)
{
    char txt[32];
    uae_u32 rev;

    switch(workprefs.cs_ciaatod)
    {
	case 0:
	CheckRadioButton(hDlg, IDC_CS_CIAA_TOD1, IDC_CS_CIAA_TOD3, IDC_CS_CIAA_TOD1);
	break;
	case 1:
	CheckRadioButton(hDlg, IDC_CS_CIAA_TOD1, IDC_CS_CIAA_TOD3, IDC_CS_CIAA_TOD2);
	break;
	case 2:
	CheckRadioButton(hDlg, IDC_CS_CIAA_TOD1, IDC_CS_CIAA_TOD3, IDC_CS_CIAA_TOD3);
	break;
    }
    switch(workprefs.cs_rtc)
    {
	case 0:
	CheckRadioButton(hDlg, IDC_CS_RTC1, IDC_CS_RTC3, IDC_CS_RTC1);
	break;
	case 1:
	CheckRadioButton(hDlg, IDC_CS_RTC1, IDC_CS_RTC3, IDC_CS_RTC2);
	break;
	case 2:
	CheckRadioButton(hDlg, IDC_CS_RTC1, IDC_CS_RTC3, IDC_CS_RTC3);
	break;
    }
    CheckDlgButton (hDlg, IDC_CS_COMPATIBLE, workprefs.cs_compatible);
    CheckDlgButton (hDlg, IDC_CS_RESETWARNING, workprefs.cs_resetwarning);
    CheckDlgButton (hDlg, IDC_CS_NOEHB, workprefs.cs_denisenoehb);
    CheckDlgButton (hDlg, IDC_CS_BLITTERBUG, workprefs.cs_agnusbltbusybug);
    CheckDlgButton (hDlg, IDC_CS_KSMIRROR_E0, workprefs.cs_ksmirror_e0);
    CheckDlgButton (hDlg, IDC_CS_KSMIRROR_A8, workprefs.cs_ksmirror_a8);
    CheckDlgButton (hDlg, IDC_CS_CIAOVERLAY, workprefs.cs_ciaoverlay);
    CheckDlgButton (hDlg, IDC_CS_DF0IDHW, workprefs.cs_df0idhw);
    CheckDlgButton (hDlg, IDC_CS_CD32CD, workprefs.cs_cd32cd);
    CheckDlgButton (hDlg, IDC_CS_CD32C2P, workprefs.cs_cd32c2p);
    CheckDlgButton (hDlg, IDC_CS_CD32NVRAM, workprefs.cs_cd32nvram);
    CheckDlgButton (hDlg, IDC_CS_CDTVCD, workprefs.cs_cdtvcd);
    CheckDlgButton (hDlg, IDC_CS_CDTVRAM, workprefs.cs_cdtvram);
    CheckDlgButton (hDlg, IDC_CS_CDTVRAMEXP, workprefs.cs_cdtvcard);
    CheckDlgButton (hDlg, IDC_CS_A1000RAM, workprefs.cs_a1000ram);
    CheckDlgButton (hDlg, IDC_CS_RAMSEY, workprefs.cs_ramseyrev >= 0);
    CheckDlgButton (hDlg, IDC_CS_FATGARY, workprefs.cs_fatgaryrev >= 0);
    CheckDlgButton (hDlg, IDC_CS_AGNUS, workprefs.cs_agnusrev >= 0);
    CheckDlgButton (hDlg, IDC_CS_DENISE, workprefs.cs_deniserev >= 0);
    CheckDlgButton (hDlg, IDC_CS_DMAC, workprefs.cs_mbdmac == 1);
    CheckDlgButton (hDlg, IDC_CS_DMAC2, workprefs.cs_mbdmac == 2);
    CheckDlgButton (hDlg, IDC_CS_A2091, workprefs.cs_a2091 > 0);
    CheckDlgButton (hDlg, IDC_CS_A4091, workprefs.cs_a4091 > 0);
    CheckDlgButton (hDlg, IDC_CS_CDTVSCSI, workprefs.cs_cdtvscsi > 0);
    CheckDlgButton (hDlg, IDC_CS_SCSIMODE, workprefs.scsi == 2);
    CheckDlgButton (hDlg, IDC_CS_PCMCIA, workprefs.cs_pcmcia > 0);
    CheckDlgButton (hDlg, IDC_CS_SLOWISFAST, workprefs.cs_slowmemisfast > 0);
    CheckDlgButton (hDlg, IDC_CS_IDE1, workprefs.cs_ide > 0 && (workprefs.cs_ide & 1));
    CheckDlgButton (hDlg, IDC_CS_IDE2, workprefs.cs_ide > 0 && (workprefs.cs_ide & 2));
    txt[0] = 0;
    sprintf (txt, "%d", workprefs.cs_rtc_adjust);
    SetDlgItemText(hDlg, IDC_CS_RTCADJUST, txt);
    txt[0] = 0;
    if (workprefs.cs_fatgaryrev >= 0)
	sprintf (txt, "%02.2X", workprefs.cs_fatgaryrev);
    SetDlgItemText(hDlg, IDC_CS_FATGARYREV, txt);
    txt[0] = 0;
    if (workprefs.cs_ramseyrev >= 0)
	sprintf (txt, "%02.2X", workprefs.cs_ramseyrev);
    SetDlgItemText(hDlg, IDC_CS_RAMSEYREV, txt);
    txt[0] = 0;
    if (workprefs.cs_agnusrev >= 0) {
	rev = workprefs.cs_agnusrev;
	sprintf (txt, "%02.2X", rev);
    } else if (workprefs.cs_compatible) {
	rev = 0;
	if (workprefs.ntscmode)
	    rev |= 0x10;
	rev |= (workprefs.chipset_mask & CSMASK_AGA) ? 0x23 : 0;
	rev |= (currprefs.chipset_mask & CSMASK_ECS_AGNUS) ? 0x20 : 0;
	if (workprefs.chipmem_size > 1024 * 1024 && (workprefs.chipset_mask & CSMASK_ECS_AGNUS))
	    rev |= 0x21;
	sprintf (txt, "%02.2X", rev);
    }
    SetDlgItemText(hDlg, IDC_CS_AGNUSREV, txt);
    txt[0] = 0;
    if (workprefs.cs_deniserev >= 0) {
	rev = workprefs.cs_deniserev;
	sprintf (txt, "%01.1X", rev);
    } else if (workprefs.cs_compatible) {
	rev = 0xf;
	if (workprefs.chipset_mask & CSMASK_ECS_DENISE)
	    rev = 0xc;
	if (workprefs.chipset_mask & CSMASK_AGA)
	    rev = 0x8;
	sprintf (txt, "%01.1X", rev);
    }
    SetDlgItemText(hDlg, IDC_CS_DENISEREV, txt);
}

static void values_from_chipsetdlg2 (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    char txt[32], *p;
    int v;

    workprefs.cs_compatible = IsDlgButtonChecked (hDlg, IDC_CS_COMPATIBLE);
    workprefs.cs_resetwarning = IsDlgButtonChecked (hDlg, IDC_CS_RESETWARNING);
    workprefs.cs_denisenoehb = IsDlgButtonChecked (hDlg, IDC_CS_NOEHB);
    workprefs.cs_agnusbltbusybug = IsDlgButtonChecked (hDlg, IDC_CS_BLITTERBUG);
    workprefs.cs_ksmirror_e0 = IsDlgButtonChecked (hDlg, IDC_CS_KSMIRROR_E0);
    workprefs.cs_ksmirror_a8 = IsDlgButtonChecked (hDlg, IDC_CS_KSMIRROR_A8);
    workprefs.cs_ciaoverlay = IsDlgButtonChecked (hDlg, IDC_CS_CIAOVERLAY);
    workprefs.cs_df0idhw = IsDlgButtonChecked (hDlg, IDC_CS_DF0IDHW);
    workprefs.cs_cd32cd = IsDlgButtonChecked (hDlg, IDC_CS_CD32CD);
    workprefs.cs_cd32c2p = IsDlgButtonChecked (hDlg, IDC_CS_CD32C2P);
    workprefs.cs_cd32nvram = IsDlgButtonChecked (hDlg, IDC_CS_CD32NVRAM);
    workprefs.cs_cdtvcd = IsDlgButtonChecked (hDlg, IDC_CS_CDTVCD);
    workprefs.cs_cdtvram = IsDlgButtonChecked (hDlg, IDC_CS_CDTVRAM);
    workprefs.cs_cdtvcard = IsDlgButtonChecked (hDlg, IDC_CS_CDTVRAMEXP) ? 64 : 0;
    workprefs.cs_a1000ram = IsDlgButtonChecked (hDlg, IDC_CS_A1000RAM);
    workprefs.cs_ramseyrev = IsDlgButtonChecked (hDlg, IDC_CS_RAMSEY) ? 0x0f : -1;
    workprefs.cs_fatgaryrev = IsDlgButtonChecked (hDlg, IDC_CS_FATGARY) ? 0x00 : -1;
    workprefs.cs_mbdmac = IsDlgButtonChecked (hDlg, IDC_CS_DMAC) ? 1 : 0;
    if (workprefs.cs_mbdmac == 0)
	workprefs.cs_mbdmac = IsDlgButtonChecked (hDlg, IDC_CS_DMAC2) ? 2 : 0;
    workprefs.cs_a2091 = IsDlgButtonChecked (hDlg, IDC_CS_A2091) ? 1 : 0;
    workprefs.cs_a4091 = IsDlgButtonChecked (hDlg, IDC_CS_A4091) ? 1 : 0;
#if 0
    if (msg == WM_COMMAND && LOWORD(wParam) == IDC_CS_SCSIMODE)
	workprefs.scsi = IsDlgButtonChecked (hDlg, IDC_CS_SCSIMODE) ? 2 : 0;
#endif
    workprefs.cs_cdtvscsi = IsDlgButtonChecked (hDlg, IDC_CS_CDTVSCSI) ? 1 : 0;
    workprefs.cs_pcmcia = IsDlgButtonChecked (hDlg, IDC_CS_PCMCIA) ? 1 : 0;
    workprefs.cs_slowmemisfast = IsDlgButtonChecked (hDlg, IDC_CS_SLOWISFAST) ? 1 : 0;
    workprefs.cs_ide = IsDlgButtonChecked (hDlg, IDC_CS_IDE1) ? 1 : (IsDlgButtonChecked (hDlg, IDC_CS_IDE2) ? 2 : 0);
    workprefs.cs_ciaatod = IsDlgButtonChecked (hDlg, IDC_CS_CIAA_TOD1) ? 0
	: (IsDlgButtonChecked (hDlg, IDC_CS_CIAA_TOD2) ? 1 : 2);
    workprefs.cs_rtc = IsDlgButtonChecked (hDlg, IDC_CS_RTC1) ? 0
	: (IsDlgButtonChecked (hDlg, IDC_CS_RTC2) ? 1 : 2);

    if (workprefs.cs_rtc) {
	txt[0] = 0;
	SendDlgItemMessage (hDlg, IDC_CS_RTCADJUST, WM_GETTEXT, (WPARAM)sizeof(txt), (LPARAM)txt);
	workprefs.cs_rtc_adjust = atol(txt);
    }
    if (workprefs.cs_fatgaryrev >= 0) {
	txt[0] = 0;
	SendDlgItemMessage (hDlg, IDC_CS_FATGARYREV, WM_GETTEXT, (WPARAM)sizeof(txt), (LPARAM)txt);
	v = strtol (txt, &p, 16);
	if (v >= 0 && v <= 255)
	    workprefs.cs_fatgaryrev = v;
    }
    if (workprefs.cs_ramseyrev >= 0) {
	txt[0] = 0;
	SendDlgItemMessage (hDlg, IDC_CS_RAMSEYREV, WM_GETTEXT, (WPARAM)sizeof(txt), (LPARAM)txt);
	v = strtol (txt, &p, 16);
	if (v >= 0 && v <= 255)
	    workprefs.cs_ramseyrev = v;
    }
    if (workprefs.cs_agnusrev >= 0) {
	txt[0] = 0;
	SendDlgItemMessage (hDlg, IDC_CS_AGNUSREV, WM_GETTEXT, (WPARAM)sizeof(txt), (LPARAM)txt);
	v = strtol (txt, &p, 16);
	if (v >= 0 && v <= 255)
	    workprefs.cs_agnusrev = v;
    }
    if (workprefs.cs_deniserev >= 0) {
	txt[0] = 0;
	SendDlgItemMessage (hDlg, IDC_CS_DENISEREV, WM_GETTEXT, (WPARAM)sizeof(txt), (LPARAM)txt);
	v = strtol (txt, &p, 16);
	if (v >= 0 && v <= 15)
	    workprefs.cs_deniserev = v;
    }

}

static void enable_for_chipsetdlg2 (HWND hDlg)
{
    int e = workprefs.cs_compatible ? FALSE : TRUE;

    ew (hDlg, IDC_CS_FATGARY, e);
    ew (hDlg, IDC_CS_RAMSEY, e);
    ew (hDlg, IDC_CS_AGNUS, e);
    ew (hDlg, IDC_CS_DENISE, e);
    ew (hDlg, IDC_CS_FATGARYREV, e);
    ew (hDlg, IDC_CS_RAMSEYREV, e);
    ew (hDlg, IDC_CS_AGNUSREV, e);
    ew (hDlg, IDC_CS_DENISEREV, e);
    ew (hDlg, IDC_CS_IDE1, e);
    ew (hDlg, IDC_CS_IDE2, e);
    ew (hDlg, IDC_CS_DMAC, e);
    ew (hDlg, IDC_CS_DMAC2, e);
    ew (hDlg, IDC_CS_A2091, e);
    ew (hDlg, IDC_CS_A4091, e);
    ShowWindow (GetDlgItem(hDlg, IDC_CS_SCSIMODE), SW_HIDE);
    ew (hDlg, IDC_CS_SCSIMODE, FALSE);
    ew (hDlg, IDC_CS_CDTVSCSI, e);
    ew (hDlg, IDC_CS_PCMCIA, e);
    ew (hDlg, IDC_CS_SLOWISFAST, e);
    ew (hDlg, IDC_CS_CD32CD, e);
    ew (hDlg, IDC_CS_CD32NVRAM, e);
    ew (hDlg, IDC_CS_CD32C2P, e);
    ew (hDlg, IDC_CS_CDTVCD, e);
    ew (hDlg, IDC_CS_CDTVRAM, e);
    ew (hDlg, IDC_CS_CDTVRAMEXP, e);
    ew (hDlg, IDC_CS_RESETWARNING, e);
    ew (hDlg, IDC_CS_NOEHB, e);
    ew (hDlg, IDC_CS_BLITTERBUG, e);
    ew (hDlg, IDC_CS_KSMIRROR_E0, e);
    ew (hDlg, IDC_CS_KSMIRROR_A8, e);
    ew (hDlg, IDC_CS_CIAOVERLAY, e);
    ew (hDlg, IDC_CS_A1000RAM, e);
    ew (hDlg, IDC_CS_DF0IDHW, e);
    ew (hDlg, IDC_CS_CIAA_TOD1, e);
    ew (hDlg, IDC_CS_CIAA_TOD2, e);
    ew (hDlg, IDC_CS_CIAA_TOD3, e);
    ew (hDlg, IDC_CS_RTC1, e);
    ew (hDlg, IDC_CS_RTC2, e);
    ew (hDlg, IDC_CS_RTC3, e);
    ew (hDlg, IDC_CS_RTCADJUST, e);
}

static INT_PTR CALLBACK ChipsetDlgProc2 (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int recursive = 0;

    switch (msg) {
    case WM_INITDIALOG:
	pages[CHIPSET2_ID] = hDlg;
	currentpage = CHIPSET2_ID;
    case WM_USER:
	recursive++;
	values_to_chipsetdlg2 (hDlg);
	enable_for_chipsetdlg2 (hDlg);
	recursive--;
	break;
    case WM_HSCROLL:
    case WM_COMMAND:
	if (recursive > 0)
	    break;
	recursive++;
	values_from_chipsetdlg2 (hDlg, msg, wParam, lParam);
	enable_for_chipsetdlg2 (hDlg);
	recursive--;
	break;
    }
    return FALSE;
}

static void enable_for_memorydlg (HWND hDlg)
{
    int z3 = ! workprefs.address_space_24;
    int fast = workprefs.chipmem_size <= 0x200000;

#ifndef AUTOCONFIG
    z3 = FALSE;
    fast = FALSE;
#endif
    ew (hDlg, IDC_Z3TEXT, z3);
    ew (hDlg, IDC_Z3FASTRAM, z3);
    ew (hDlg, IDC_Z3FASTMEM, z3);
    ew (hDlg, IDC_FASTMEM, fast);
    ew (hDlg, IDC_FASTRAM, fast);
    ew (hDlg, IDC_FASTTEXT, fast);
    ew (hDlg, IDC_GFXCARDTEXT, z3);
    ew (hDlg, IDC_P96RAM, z3);
    ew (hDlg, IDC_P96MEM, z3);
    ew (hDlg, IDC_MBRAM1, z3);
    ew (hDlg, IDC_MBMEM1, z3);
    ew (hDlg, IDC_MBRAM2, z3);
    ew (hDlg, IDC_MBMEM2, z3);
}

static void values_to_memorydlg (HWND hDlg)
{
    uae_u32 mem_size = 0;

    switch (workprefs.chipmem_size) {
     case 0x00040000: mem_size = 0; break;
     case 0x00080000: mem_size = 1; break;
     case 0x00100000: mem_size = 2; break;
     case 0x00200000: mem_size = 3; break;
     case 0x00400000: mem_size = 4; break;
     case 0x00800000: mem_size = 5; break;
    }
    SendDlgItemMessage (hDlg, IDC_CHIPMEM, TBM_SETPOS, TRUE, mem_size);
    SetDlgItemText (hDlg, IDC_CHIPRAM, memsize_names[msi_chip[mem_size]]);

    mem_size = 0;
    switch (workprefs.fastmem_size) {
     case 0x00000000: mem_size = 0; break;
     case 0x00100000: mem_size = 1; break;
     case 0x00200000: mem_size = 2; break;
     case 0x00400000: mem_size = 3; break;
     case 0x00800000: mem_size = 4; break;
     case 0x01000000: mem_size = 5; break;
    }
    SendDlgItemMessage (hDlg, IDC_FASTMEM, TBM_SETPOS, TRUE, mem_size);
    SetDlgItemText (hDlg, IDC_FASTRAM, memsize_names[msi_fast[mem_size]]);

    mem_size = 0;
    switch (workprefs.bogomem_size) {
     case 0x00000000: mem_size = 0; break;
     case 0x00080000: mem_size = 1; break;
     case 0x00100000: mem_size = 2; break;
     case 0x00180000: mem_size = 3; break;
     case 0x001C0000: mem_size = 4; break;
    }
    SendDlgItemMessage (hDlg, IDC_SLOWMEM, TBM_SETPOS, TRUE, mem_size);
    SetDlgItemText (hDlg, IDC_SLOWRAM, memsize_names[msi_bogo[mem_size]]);

    mem_size = 0;
    switch (workprefs.z3fastmem_size) {
     case 0x00000000: mem_size = 0; break; /*   0-megs */
     case 0x00100000: mem_size = 1; break; /*   1-megs */
     case 0x00200000: mem_size = 2; break; /*   2-megs */
     case 0x00400000: mem_size = 3; break; /*   4-megs */
     case 0x00800000: mem_size = 4; break; /*   8-megs */
     case 0x01000000: mem_size = 5; break; /*  16-megs */
     case 0x02000000: mem_size = 6; break; /*  32-megs */
     case 0x04000000: mem_size = 7; break; /*  64-megs */
     case 0x08000000: mem_size = 8; break; /* 128-megs */
     case 0x10000000: mem_size = 9; break; /* 256-megs */
     case 0x20000000: mem_size = 10; break; /* 512-megs */
     case 0x40000000: mem_size = 11; break; /* 1 GB */
     case 0x80000000: mem_size = 12; break; /* 2 GB */

    }
    SendDlgItemMessage (hDlg, IDC_Z3FASTMEM, TBM_SETPOS, TRUE, mem_size);
    SetDlgItemText (hDlg, IDC_Z3FASTRAM, memsize_names[msi_z3fast[mem_size]]);

    mem_size = 0;
    switch (workprefs.gfxmem_size) {
     case 0x00000000: mem_size = 0; break;
     case 0x00100000: mem_size = 1; break;
     case 0x00200000: mem_size = 2; break;
     case 0x00400000: mem_size = 3; break;
     case 0x00800000: mem_size = 4; break;
     case 0x01000000: mem_size = 5; break;
     case 0x02000000: mem_size = 6; break;
     case 0x04000000: mem_size = 7; break;
     case 0x08000000: mem_size = 8; break;
     case 0x10000000: mem_size = 9; break;
     case 0x20000000: mem_size = 10; break;
     case 0x40000000: mem_size = 11; break;
    }
    SendDlgItemMessage (hDlg, IDC_P96MEM, TBM_SETPOS, TRUE, mem_size);
    SetDlgItemText (hDlg, IDC_P96RAM, memsize_names[msi_gfx[mem_size]]);

    mem_size = 0;
    switch (workprefs.mbresmem_low_size) {
     case 0x00000000: mem_size = 0; break;
     case 0x00100000: mem_size = 1; break;
     case 0x00200000: mem_size = 2; break;
     case 0x00400000: mem_size = 3; break;
     case 0x00800000: mem_size = 4; break;
     case 0x01000000: mem_size = 5; break;
     case 0x02000000: mem_size = 6; break;
     case 0x04000000: mem_size = 7; break;
    }
    SendDlgItemMessage (hDlg, IDC_MBMEM1, TBM_SETPOS, TRUE, mem_size);
    SetDlgItemText (hDlg, IDC_MBRAM1, memsize_names[msi_gfx[mem_size]]);

    mem_size = 0;
    switch (workprefs.mbresmem_high_size) {
     case 0x00000000: mem_size = 0; break;
     case 0x00100000: mem_size = 1; break;
     case 0x00200000: mem_size = 2; break;
     case 0x00400000: mem_size = 3; break;
     case 0x00800000: mem_size = 4; break;
     case 0x01000000: mem_size = 5; break;
     case 0x02000000: mem_size = 6; break;
     case 0x04000000: mem_size = 7; break;
    }
    SendDlgItemMessage (hDlg, IDC_MBMEM2, TBM_SETPOS, TRUE, mem_size);
    SetDlgItemText (hDlg, IDC_MBRAM2, memsize_names[msi_gfx[mem_size]]);
}

static void fix_values_memorydlg (void)
{
    if (workprefs.chipmem_size > 0x200000)
	workprefs.fastmem_size = 0;
}

static INT_PTR CALLBACK MemoryDlgProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int recursive = 0;

    switch (msg)
    {
    case WM_INITDIALOG:
	pages[MEMORY_ID] = hDlg;
	currentpage = MEMORY_ID;

	SendDlgItemMessage (hDlg, IDC_CHIPMEM, TBM_SETRANGE, TRUE, MAKELONG (MIN_CHIP_MEM, MAX_CHIP_MEM));
	SendDlgItemMessage (hDlg, IDC_FASTMEM, TBM_SETRANGE, TRUE, MAKELONG (MIN_FAST_MEM, MAX_FAST_MEM));
	SendDlgItemMessage (hDlg, IDC_SLOWMEM, TBM_SETRANGE, TRUE, MAKELONG (MIN_SLOW_MEM, MAX_SLOW_MEM));
	SendDlgItemMessage (hDlg, IDC_Z3FASTMEM, TBM_SETRANGE, TRUE, MAKELONG (MIN_Z3_MEM, MAX_Z3_MEM));
	SendDlgItemMessage (hDlg, IDC_P96MEM, TBM_SETRANGE, TRUE, MAKELONG (MIN_P96_MEM, MAX_P96_MEM));
	SendDlgItemMessage (hDlg, IDC_MBMEM1, TBM_SETRANGE, TRUE, MAKELONG (MIN_MB_MEM, MAX_MB_MEM));
	SendDlgItemMessage (hDlg, IDC_MBMEM2, TBM_SETRANGE, TRUE, MAKELONG (MIN_MB_MEM, MAX_MB_MEM));

    case WM_USER:
	recursive++;
	fix_values_memorydlg ();
	values_to_memorydlg (hDlg);
	enable_for_memorydlg (hDlg);
	recursive--;
	break;

    case WM_HSCROLL:
	workprefs.chipmem_size = memsizes[msi_chip[SendMessage (GetDlgItem (hDlg, IDC_CHIPMEM), TBM_GETPOS, 0, 0)]];
	workprefs.bogomem_size = memsizes[msi_bogo[SendMessage (GetDlgItem (hDlg, IDC_SLOWMEM), TBM_GETPOS, 0, 0)]];
	workprefs.fastmem_size = memsizes[msi_fast[SendMessage (GetDlgItem (hDlg, IDC_FASTMEM), TBM_GETPOS, 0, 0)]];
	workprefs.z3fastmem_size = memsizes[msi_z3fast[SendMessage (GetDlgItem (hDlg, IDC_Z3FASTMEM), TBM_GETPOS, 0, 0)]];
	workprefs.gfxmem_size = memsizes[msi_gfx[SendMessage (GetDlgItem (hDlg, IDC_P96MEM), TBM_GETPOS, 0, 0)]];
	workprefs.mbresmem_low_size = memsizes[msi_gfx[SendMessage (GetDlgItem (hDlg, IDC_MBMEM1), TBM_GETPOS, 0, 0)]];
	workprefs.mbresmem_high_size = memsizes[msi_gfx[SendMessage (GetDlgItem (hDlg, IDC_MBMEM2), TBM_GETPOS, 0, 0)]];
	fix_values_memorydlg ();
	values_to_memorydlg (hDlg);
	enable_for_memorydlg (hDlg);
	break;

    case WM_COMMAND:
	if (recursive > 0)
	    break;
	recursive++;
	values_to_memorydlg (hDlg);
	recursive--;
	break;

    }
    return FALSE;
}

static void addromfiles (UAEREG *fkey, HWND hDlg, DWORD d, char *path, int type)
{
    int idx, idx2;
    char tmp[MAX_DPATH];
    char tmp2[MAX_DPATH];
    char seltmp[MAX_DPATH];
    struct romdata *rdx;

    rdx = scan_single_rom (path);
    SendDlgItemMessage(hDlg, d, CB_RESETCONTENT, 0, 0);
    SendDlgItemMessage(hDlg, d, CB_ADDSTRING, 0, (LPARAM)"");
    idx = 0;
    seltmp[0] = 0;
    for (;fkey;) {
	DWORD size = sizeof (tmp);
	DWORD size2 = sizeof (tmp2);
	if (!regenumstr (fkey, idx, tmp, &size, tmp2, &size2))
	    break;
	if (strlen (tmp) == 6) {
	    idx2 = atol (tmp + 3);
	    if (idx2 >= 0) {
		struct romdata *rd = getromdatabyid (idx2);
		if (rd && (rd->type & type)) {
		    getromname (rd, tmp);
		    SendDlgItemMessage(hDlg, d, CB_ADDSTRING, 0, (LPARAM)tmp);
		    if (rd == rdx)
			strcpy (seltmp, tmp);
		}
	    }
	}
	idx++;
    }
    if (seltmp[0])
	SendDlgItemMessage (hDlg, d, CB_SELECTSTRING, (WPARAM)-1, (LPARAM)seltmp);
    else
	SetDlgItemText(hDlg, d, path);
}

static void getromfile (HWND hDlg, DWORD d, char *path, int size)
{
    LRESULT val = SendDlgItemMessage (hDlg, d, CB_GETCURSEL, 0, 0L);
    if (val == CB_ERR) {
	SendDlgItemMessage (hDlg, d, WM_GETTEXT, (WPARAM)size, (LPARAM)path);
    } else {
	char tmp1[MAX_DPATH];
	struct romdata *rd;
	SendDlgItemMessage (hDlg, d, CB_GETLBTEXT, (WPARAM)val, (LPARAM)tmp1);
	path[0] = 0;
	rd = getromdatabyname (tmp1);
	if (rd) {
	    struct romlist *rl = getromlistbyromdata(rd);
	    if (rd->configname)
		sprintf (path, ":%s", rd->configname);
	    else if (rl)
		strncpy (path, rl->path, size);
	}
    }
}

static void values_from_kickstartdlg (HWND hDlg)
{
    getromfile (hDlg, IDC_ROMFILE, workprefs.romfile, sizeof (workprefs.romfile));
    getromfile (hDlg, IDC_ROMFILE2, workprefs.romextfile, sizeof (workprefs.romextfile));
    getromfile (hDlg, IDC_CARTFILE, workprefs.cartfile, sizeof (workprefs.cartfile));
}

static void values_to_kickstartdlg (HWND hDlg)
{
    UAEREG *fkey;

    fkey = regcreatetree (NULL, "DetectedROMs");
    load_keyring(&workprefs, NULL);
    addromfiles (fkey, hDlg, IDC_ROMFILE, workprefs.romfile,
        ROMTYPE_KICK | ROMTYPE_KICKCD32);
    addromfiles (fkey, hDlg, IDC_ROMFILE2, workprefs.romextfile,
        ROMTYPE_EXTCD32 | ROMTYPE_EXTCDTV | ROMTYPE_ARCADIABIOS);
    addromfiles (fkey, hDlg, IDC_CARTFILE, workprefs.cartfile,
        ROMTYPE_AR | ROMTYPE_SUPERIV | ROMTYPE_NORDIC | ROMTYPE_XPOWER | ROMTYPE_ARCADIAGAME | ROMTYPE_HRTMON | ROMTYPE_CD32CART);
    regclosetree (fkey);

    SetDlgItemText(hDlg, IDC_FLASHFILE, workprefs.flashfile);
    CheckDlgButton(hDlg, IDC_KICKSHIFTER, workprefs.kickshifter);
    CheckDlgButton(hDlg, IDC_MAPROM, workprefs.maprom);
    //CheckDlgButton(hDlg, IDC_HRTMON, workprefs.cart_internal == 1);
}

static void init_kickstart (HWND hDlg)
{
#if !defined(AUTOCONFIG)
    ew (hDlg, IDC_MAPROM), FALSE);
#endif
#if !defined (CDTV) && !defined (CD32)
    ew (hDlg, IDC_FLASHFILE), FALSE);
    ew (hDlg, IDC_ROMFILE2), FALSE);
#endif
#if !defined (ACTION_REPLAY)
    ew (hDlg, IDC_CARTFILE), FALSE);
#endif
#if defined (UAE_MINI)
    ew (hDlg, IDC_KICKSHIFTER), FALSE);
    ew (hDlg, IDC_ROMCHOOSER2), FALSE);
    ew (hDlg, IDC_CARTCHOOSER), FALSE);
    ew (hDlg, IDC_FLASHCHOOSER), FALSE);
#endif
    if (!regexiststree (NULL , "DetectedROMs"))
	scan_roms (1);
}

static INT_PTR CALLBACK KickstartDlgProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int recursive;
    char tmp[MAX_DPATH];

    switch( msg )
    {
    case WM_INITDIALOG:
    {
	int ids[] = { IDC_ROMFILE, IDC_ROMFILE2, IDC_CARTFILE, -1 };
	pages[KICKSTART_ID] = hDlg;
	currentpage = KICKSTART_ID;
	init_kickstart (hDlg);
	values_to_kickstartdlg (hDlg);
	setmultiautocomplete (hDlg, ids);
	setac (hDlg, IDC_FLASHFILE);
	return TRUE;
    }
    case WM_COMMAND:
	if (recursive > 0)
	    break;
	recursive++;
	if (HIWORD (wParam) == CBN_SELCHANGE || HIWORD (wParam) == CBN_KILLFOCUS)  {
	    switch (LOWORD (wParam))
	    {
		case IDC_ROMFILE:
		case IDC_ROMFILE2:
		case IDC_CARTFILE:
		values_from_kickstartdlg (hDlg);
		break;
	    }
	}
	switch (LOWORD (wParam))
	{
	case IDC_KICKCHOOSER:
	    DiskSelection(hDlg, IDC_ROMFILE, 6, &workprefs, 0);
	    values_to_kickstartdlg (hDlg);
	    break;

	case IDC_ROMCHOOSER2:
	    DiskSelection(hDlg, IDC_ROMFILE2, 6, &workprefs, 0);
	    values_to_kickstartdlg (hDlg);
	    break;

	case IDC_FLASHCHOOSER:
	    DiskSelection(hDlg, IDC_FLASHFILE, 11, &workprefs, 0);
	    values_to_kickstartdlg (hDlg);
	    break;
	case IDC_FLASHFILE:
	    GetWindowText (GetDlgItem (hDlg, IDC_FLASHFILE), tmp, sizeof (tmp));
	    strcpy (workprefs.flashfile, tmp);
	    break;

	case IDC_CARTCHOOSER:
	    DiskSelection(hDlg, IDC_CARTFILE, 6, &workprefs, 0);
	    values_to_kickstartdlg (hDlg);
	    break;

	case IDC_KICKSHIFTER:
	    workprefs.kickshifter = IsDlgButtonChecked(hDlg, IDC_KICKSHIFTER);
	    break;

	case IDC_MAPROM:
	    workprefs.maprom = IsDlgButtonChecked(hDlg, IDC_MAPROM) ? 0xe00000 : 0;
	    break;
#if 0
	case IDC_HRTMON:
	    workprefs.cart_internal = IsDlgButtonChecked(hDlg, IDC_HRTMON) ? 1 : 0;
	    ew (hDlg, IDC_CARTFILE, workprefs.cart_internal == 1 ? FALSE : TRUE);
	    break;
#endif
	}
	recursive--;
	break;
    }
    return FALSE;
}

static void enable_for_miscdlg (HWND hDlg)
{
    if (!full_property_sheet) {
	ew (hDlg, IDC_JULIAN, TRUE);
	ew (hDlg, IDC_CTRLF11, TRUE);
	ew (hDlg, IDC_SOCKETS, FALSE);
	ew (hDlg, IDC_SHOWGUI, FALSE);
	ew (hDlg, IDC_NOSPEED, TRUE);
	ew (hDlg, IDC_NOSPEEDPAUSE, TRUE);
	ew (hDlg, IDC_NOSOUND, TRUE);
	ew (hDlg, IDC_NOOVERLAY, TRUE);
	ew (hDlg, IDC_DOSAVESTATE, TRUE);
	ew (hDlg, IDC_SCSIMODE, FALSE);
	ew (hDlg, IDC_SCSIDEVICE, FALSE);
	ew (hDlg, IDC_CLOCKSYNC, FALSE);
	ew (hDlg, IDC_CATWEASEL, FALSE);
	ew (hDlg, IDC_STATE_CAPTURE, FALSE);
	ew (hDlg, IDC_STATE_RATE, FALSE);
	ew (hDlg, IDC_STATE_BUFFERSIZE, FALSE);
	ew (hDlg, IDC_LANGUAGE, FALSE);
    } else {
#if !defined (BSDSOCKET)
	EnableWindow (GetDlgItem(hDlg, IDC_SOCKETS), FALSE);
#endif
#if !defined (SCSIEMU)
	EnableWindow (GetDlgItem(hDlg, IDC_SCSIDEVICE), FALSE);
	EnableWindow (GetDlgItem(hDlg, IDC_SCSIMODE), TRUE);
#endif
	ew (hDlg, IDC_DOSAVESTATE, FALSE);
	ew (hDlg, IDC_STATE_RATE, workprefs.statecapture ? TRUE : FALSE);
	ew (hDlg, IDC_STATE_BUFFERSIZE, workprefs.statecapture ? TRUE : FALSE);
    }
}

static void misc_kbled (HWND hDlg, int v, int nv)
{
    char *defname = v == IDC_KBLED1 ? "(NumLock)" : v == IDC_KBLED2 ? "(CapsLock)" : "(ScrollLock)";
    SendDlgItemMessage (hDlg, v, CB_RESETCONTENT, 0, 0L);
    SendDlgItemMessage (hDlg, v, CB_ADDSTRING, 0, (LPARAM)defname);
    SendDlgItemMessage (hDlg, v, CB_ADDSTRING, 0, (LPARAM)"POWER");
    SendDlgItemMessage (hDlg, v, CB_ADDSTRING, 0, (LPARAM)"DF0");
    SendDlgItemMessage (hDlg, v, CB_ADDSTRING, 0, (LPARAM)"DF1");
    SendDlgItemMessage (hDlg, v, CB_ADDSTRING, 0, (LPARAM)"DF2");
    SendDlgItemMessage (hDlg, v, CB_ADDSTRING, 0, (LPARAM)"DF3");
    SendDlgItemMessage (hDlg, v, CB_ADDSTRING, 0, (LPARAM)"HD");
    SendDlgItemMessage (hDlg, v, CB_ADDSTRING, 0, (LPARAM)"CD");
    SendDlgItemMessage (hDlg, v, CB_SETCURSEL, nv, 0);
}

static void misc_getkbled (HWND hDlg, int v, int n)
{
    LRESULT nv = SendDlgItemMessage(hDlg, v, CB_GETCURSEL, 0, 0L);
    if (nv != CB_ERR) {
	workprefs.keyboard_leds[n] = nv;
	misc_kbled (hDlg, v, nv);
    }
    workprefs.keyboard_leds_in_use = workprefs.keyboard_leds[0] | workprefs.keyboard_leds[1] | workprefs.keyboard_leds[2];
}

static void misc_getpri (HWND hDlg, int v, int *n)
{
    LRESULT nv = SendDlgItemMessage(hDlg, v, CB_GETCURSEL, 0, 0L);
    if (nv != CB_ERR)
	*n = nv;
}

static void misc_addpri (HWND hDlg, int v, int pri)
{
    int i;
    SendDlgItemMessage (hDlg, v, CB_RESETCONTENT, 0, 0L);
    i = 0;
    while (priorities[i].name) {
	SendDlgItemMessage (hDlg, v, CB_ADDSTRING, 0, (LPARAM)priorities[i].name);
	i++;
    }
    SendDlgItemMessage (hDlg, v, CB_SETCURSEL, pri, 0);
}

extern char *get_aspi_path(int);

static void misc_scsi(HWND hDlg)
{
    SendDlgItemMessage (hDlg, IDC_SCSIMODE, CB_RESETCONTENT, 0, 0);
    SendDlgItemMessage (hDlg, IDC_SCSIMODE, CB_ADDSTRING, 0, (LPARAM)(os_winnt ? "SPTI" : "(SPTI)"));
    SendDlgItemMessage (hDlg, IDC_SCSIMODE, CB_ADDSTRING, 0, (LPARAM)(os_winnt ? "SPTI + SCSI SCAN" : "(SPTI + SCSI SCAN)"));
    SendDlgItemMessage (hDlg, IDC_SCSIMODE, CB_ADDSTRING, 0, (LPARAM)((get_aspi_path(0)) ? "AdaptecASPI" : "(AdaptecASPI)"));
    SendDlgItemMessage (hDlg, IDC_SCSIMODE, CB_ADDSTRING, 0, (LPARAM)((get_aspi_path(1)) ? "NeroASPI" : "(NeroASPI)"));
    SendDlgItemMessage (hDlg, IDC_SCSIMODE, CB_ADDSTRING, 0, (LPARAM)((get_aspi_path(2)) ? "FrogASPI" : "(FrogASPI)"));
    SendDlgItemMessage (hDlg, IDC_SCSIMODE, CB_SETCURSEL, workprefs.win32_uaescsimode - 1, 0);
}

static void misc_lang(HWND hDlg)
{
    int i, idx = 0, cnt = 0, lid;
    WORD langid = -1;

    if (regqueryint (NULL, "Language", &lid))
	langid = (WORD)lid;
    SendDlgItemMessage (hDlg, IDC_LANGUAGE, CB_RESETCONTENT, 0, 0);
    SendDlgItemMessage (hDlg, IDC_LANGUAGE, CB_ADDSTRING, 0, (LPARAM)"Autodetect");
    SendDlgItemMessage (hDlg, IDC_LANGUAGE, CB_ADDSTRING, 0, (LPARAM)"English (built-in)");
    if (langid == 0)
	idx = 1;
    cnt = 2;
    for (i = 0; langs[i].name; i++) {
	HMODULE hm = language_load (langs[i].id);
	if (hm) {
	    FreeLibrary (hm);
	    SendDlgItemMessage (hDlg, IDC_LANGUAGE, CB_ADDSTRING, 0, (LPARAM)langs[i].name);
	    if (langs[i].id == langid)
		idx = cnt;
	    cnt++;
	}
    }
    SendDlgItemMessage (hDlg, IDC_LANGUAGE, CB_SETCURSEL, idx, 0);
}
static void misc_setlang(int v)
{
    int i;
    WORD langid = 0;
    v-=2;
    if (v >= 0) {
	for (i = 0; langs[i].name; i++) {
	    HMODULE hm = language_load (langs[i].id);
	    if (hm) {
		FreeLibrary(hm);
		if (v == 0) {
		    langid = langs[i].id;
		    break;
		}
		v--;
	    }
	}
    }
    if (v == -2)
	langid = -1;
    regsetint (NULL, "Language", langid);
    FreeLibrary(hUIDLL);
    hUIDLL = NULL;
    if (langid >= 0)
	hUIDLL = language_load(langid);
    restart_requested = 1;
    exit_gui(0);
}

static void values_to_miscdlg (HWND hDlg)
{
    char txt[100];
    int cw;

    if (currentpage == MISC1_ID) {

	CheckDlgButton (hDlg, IDC_SOCKETS, workprefs.socket_emu);
	CheckDlgButton (hDlg, IDC_ILLEGAL, workprefs.illegal_mem);
	CheckDlgButton (hDlg, IDC_SHOWGUI, workprefs.start_gui);
	CheckDlgButton (hDlg, IDC_JULIAN, workprefs.win32_middle_mouse);
	CheckDlgButton (hDlg, IDC_CREATELOGFILE, workprefs.win32_logfile);
	CheckDlgButton (hDlg, IDC_CTRLF11, workprefs.win32_ctrl_F11_is_quit);
	CheckDlgButton (hDlg, IDC_NOOVERLAY, workprefs.win32_no_overlay);
	CheckDlgButton (hDlg, IDC_SHOWLEDS, workprefs.leds_on_screen);
	CheckDlgButton (hDlg, IDC_SCSIDEVICE, workprefs.scsi == 1);
	CheckDlgButton (hDlg, IDC_SANA2, workprefs.sana2);
	CheckDlgButton (hDlg, IDC_NOTASKBARBUTTON, workprefs.win32_notaskbarbutton);
	CheckDlgButton (hDlg, IDC_ALWAYSONTOP, workprefs.win32_alwaysontop);
	CheckDlgButton (hDlg, IDC_CLOCKSYNC, workprefs.tod_hack);
	CheckDlgButton (hDlg, IDC_POWERSAVE, workprefs.win32_powersavedisabled);
	cw = catweasel_detect();
	ew (hDlg, IDC_CATWEASEL, cw);
	if (!cw && workprefs.catweasel < 100)
	    workprefs.catweasel = 0;
	CheckDlgButton (hDlg, IDC_CATWEASEL, workprefs.catweasel);
	CheckDlgButton (hDlg, IDC_STATE_CAPTURE, workprefs.statecapture);
	CheckDlgButton (hDlg, IDC_MOUSETRICK, workprefs.win32_outsidemouse);

	misc_kbled (hDlg, IDC_KBLED1, workprefs.keyboard_leds[0]);
	misc_kbled (hDlg, IDC_KBLED2, workprefs.keyboard_leds[1]);
	misc_kbled (hDlg, IDC_KBLED3, workprefs.keyboard_leds[2]);
	CheckDlgButton (hDlg, IDC_KBLED_USB, workprefs.win32_kbledmode);

	SendDlgItemMessage (hDlg, IDC_STATE_RATE, CB_RESETCONTENT, 0, 0);
	SendDlgItemMessage (hDlg, IDC_STATE_RATE, CB_ADDSTRING, 0, (LPARAM)"1");
	SendDlgItemMessage (hDlg, IDC_STATE_RATE, CB_ADDSTRING, 0, (LPARAM)"5");
	SendDlgItemMessage (hDlg, IDC_STATE_RATE, CB_ADDSTRING, 0, (LPARAM)"10");
	SendDlgItemMessage (hDlg, IDC_STATE_RATE, CB_ADDSTRING, 0, (LPARAM)"20");
	SendDlgItemMessage (hDlg, IDC_STATE_RATE, CB_ADDSTRING, 0, (LPARAM)"30");
	sprintf (txt, "%d", workprefs.statecapturerate / 50);
	SendDlgItemMessage( hDlg, IDC_STATE_RATE, WM_SETTEXT, 0, (LPARAM)txt);

	SendDlgItemMessage (hDlg, IDC_STATE_BUFFERSIZE, CB_RESETCONTENT, 0, 0);
	SendDlgItemMessage (hDlg, IDC_STATE_BUFFERSIZE, CB_ADDSTRING, 0, (LPARAM)"5");
	SendDlgItemMessage (hDlg, IDC_STATE_BUFFERSIZE, CB_ADDSTRING, 0, (LPARAM)"10");
	SendDlgItemMessage (hDlg, IDC_STATE_BUFFERSIZE, CB_ADDSTRING, 0, (LPARAM)"20");
	SendDlgItemMessage (hDlg, IDC_STATE_BUFFERSIZE, CB_ADDSTRING, 0, (LPARAM)"50");
	SendDlgItemMessage (hDlg, IDC_STATE_BUFFERSIZE, CB_ADDSTRING, 0, (LPARAM)"100");
	sprintf (txt, "%d", workprefs.statecapturebuffersize / (1024 * 1024));
	SendDlgItemMessage( hDlg, IDC_STATE_BUFFERSIZE, WM_SETTEXT, 0, (LPARAM)txt);

	misc_scsi(hDlg);
	misc_lang(hDlg);

    } else if (currentpage == MISC2_ID) {

	CheckDlgButton (hDlg, IDC_INACTIVE_PAUSE, workprefs.win32_inactive_pause);
	CheckDlgButton (hDlg, IDC_INACTIVE_NOSOUND, workprefs.win32_inactive_nosound || workprefs.win32_inactive_pause);
	CheckDlgButton (hDlg, IDC_MINIMIZED_PAUSE, workprefs.win32_iconified_pause);
	CheckDlgButton (hDlg, IDC_MINIMIZED_NOSOUND, workprefs.win32_iconified_nosound || workprefs.win32_iconified_pause);
	misc_addpri (hDlg, IDC_ACTIVE_PRIORITY, workprefs.win32_active_priority);
	misc_addpri (hDlg, IDC_INACTIVE_PRIORITY, workprefs.win32_inactive_priority);
	misc_addpri (hDlg, IDC_MINIMIZED_PRIORITY, workprefs.win32_iconified_priority);

    }
}

static INT_PTR MiscDlgProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    char txt[100];
    int v;
    static int recursive;

    if (recursive)
	return FALSE;
    recursive++;

    switch (msg)
    {

    case WM_USER:
	values_to_miscdlg (hDlg);
	enable_for_miscdlg (hDlg);
	recursive--;
	return TRUE;

    case WM_COMMAND:

	if (currentpage == MISC1_ID) {
	    if (HIWORD (wParam) == CBN_SELENDOK || HIWORD (wParam) == CBN_KILLFOCUS || HIWORD (wParam) == CBN_EDITCHANGE)  {
		switch (LOWORD (wParam))
		{
		    case IDC_KBLED1:
		    misc_getkbled (hDlg, IDC_KBLED1, 0);
		    break;
		    case IDC_KBLED2:
		    misc_getkbled (hDlg, IDC_KBLED2, 1);
		    break;
		    case IDC_KBLED3:
		    misc_getkbled (hDlg, IDC_KBLED3, 2);
		    break;
		    case IDC_STATE_RATE:
		    getcbn (hDlg, IDC_STATE_RATE, txt, sizeof (txt));
		    workprefs.statecapturerate = atol (txt) * 50;
		    break;
		    case IDC_STATE_BUFFERSIZE:
		    getcbn (hDlg, IDC_STATE_BUFFERSIZE, txt, sizeof (txt));
		    break;
		    case IDC_SCSIMODE:
		    v = SendDlgItemMessage(hDlg, IDC_SCSIMODE, CB_GETCURSEL, 0, 0L);
		    if (v != CB_ERR)
			workprefs.win32_uaescsimode = v + 1;
		    break;
		    case IDC_LANGUAGE:
		    if (HIWORD (wParam) == CBN_SELENDOK) {
			v = SendDlgItemMessage(hDlg, IDC_LANGUAGE, CB_GETCURSEL, 0, 0L);
			if (v != CB_ERR)
			    misc_setlang(v);
		    }
		    break;
		}
	    }
	} else if (currentpage == MISC2_ID) {
	    misc_getpri (hDlg, IDC_ACTIVE_PRIORITY, &workprefs.win32_active_priority);
	    misc_getpri (hDlg, IDC_INACTIVE_PRIORITY, &workprefs.win32_inactive_priority);
	    misc_getpri (hDlg, IDC_MINIMIZED_PRIORITY, &workprefs.win32_iconified_priority);
	}

	switch(wParam)
	{
	case IDC_DOSAVESTATE:
	    if (DiskSelection(hDlg, wParam, 9, &workprefs, 0))
		save_state (savestate_fname, "Description!");
	    break;
	case IDC_DOLOADSTATE:
	    if (DiskSelection(hDlg, wParam, 10, &workprefs, 0))
		savestate_state = STATE_DORESTORE;
	    break;
	case IDC_STATE_CAPTURE:
	    workprefs.statecapture = IsDlgButtonChecked (hDlg, IDC_STATE_CAPTURE);
	    enable_for_miscdlg (hDlg);
	    break;
	case IDC_SOCKETS:
	    workprefs.socket_emu = IsDlgButtonChecked (hDlg, IDC_SOCKETS);
	    break;
	case IDC_ILLEGAL:
	    workprefs.illegal_mem = IsDlgButtonChecked (hDlg, IDC_ILLEGAL);
	    break;
	case IDC_JULIAN:
	    workprefs.win32_middle_mouse = IsDlgButtonChecked(hDlg, IDC_JULIAN);
	    break;
	case IDC_NOOVERLAY:
	    workprefs.win32_no_overlay = IsDlgButtonChecked(hDlg, IDC_NOOVERLAY);
	    break;
	case IDC_SHOWLEDS:
	    workprefs.leds_on_screen = IsDlgButtonChecked(hDlg, IDC_SHOWLEDS);
	    break;
	case IDC_SHOWGUI:
	    workprefs.start_gui = IsDlgButtonChecked (hDlg, IDC_SHOWGUI);
	    break;
	case IDC_CREATELOGFILE:
	    workprefs.win32_logfile = IsDlgButtonChecked (hDlg, IDC_CREATELOGFILE);
	    enable_for_miscdlg(hDlg);
	    break;
	case IDC_POWERSAVE:
	    workprefs.win32_powersavedisabled = IsDlgButtonChecked (hDlg, IDC_POWERSAVE);
	    break;
	case IDC_INACTIVE_NOSOUND:
	    if (!IsDlgButtonChecked (hDlg, IDC_INACTIVE_NOSOUND))
		CheckDlgButton (hDlg, IDC_INACTIVE_PAUSE, BST_UNCHECKED);
	case IDC_INACTIVE_PAUSE:
	    workprefs.win32_inactive_pause = IsDlgButtonChecked (hDlg, IDC_INACTIVE_PAUSE);
	    if (workprefs.win32_inactive_pause)
		CheckDlgButton (hDlg, IDC_INACTIVE_NOSOUND, BST_CHECKED);
	    workprefs.win32_inactive_nosound = IsDlgButtonChecked (hDlg, IDC_INACTIVE_NOSOUND);
	    break;
	case IDC_MINIMIZED_NOSOUND:
	    if (!IsDlgButtonChecked (hDlg, IDC_MINIMIZED_NOSOUND))
		CheckDlgButton (hDlg, IDC_MINIMIZED_PAUSE, BST_UNCHECKED);
	case IDC_MINIMIZED_PAUSE:
	    workprefs.win32_iconified_pause = IsDlgButtonChecked (hDlg, IDC_MINIMIZED_PAUSE);
	    if (workprefs.win32_iconified_pause)
		CheckDlgButton (hDlg, IDC_MINIMIZED_NOSOUND, BST_CHECKED);
	    workprefs.win32_iconified_nosound = IsDlgButtonChecked (hDlg, IDC_MINIMIZED_NOSOUND);
	    break;
	case IDC_CTRLF11:
	    workprefs.win32_ctrl_F11_is_quit = IsDlgButtonChecked (hDlg, IDC_CTRLF11);
	    break;
	case IDC_SCSIDEVICE:
	    workprefs.scsi = IsDlgButtonChecked (hDlg, IDC_SCSIDEVICE) ? 1 : 0;
	    enable_for_miscdlg (hDlg);
	    break;
	case IDC_SANA2:
	    workprefs.sana2 = IsDlgButtonChecked (hDlg, IDC_SANA2) ? 1 : 0;
	    break;
	case IDC_CLOCKSYNC:
	    workprefs.tod_hack = IsDlgButtonChecked (hDlg, IDC_CLOCKSYNC);
	    break;
	case IDC_CATWEASEL:
	    workprefs.catweasel = IsDlgButtonChecked (hDlg, IDC_CATWEASEL) ? -1 : 0;
	    break;
	case IDC_MOUSETRICK:
	    workprefs.win32_outsidemouse = IsDlgButtonChecked (hDlg, IDC_MOUSETRICK) ? -1 : 0;
	    break;
	case IDC_NOTASKBARBUTTON:
	    workprefs.win32_notaskbarbutton = IsDlgButtonChecked (hDlg, IDC_NOTASKBARBUTTON);
	    break;
	case IDC_ALWAYSONTOP:
	    workprefs.win32_alwaysontop = IsDlgButtonChecked (hDlg, IDC_ALWAYSONTOP);
	    break;
	case IDC_KBLED_USB:
	    workprefs.win32_kbledmode = IsDlgButtonChecked (hDlg, IDC_KBLED_USB) ? 1 : 0;
	    break;
	}
	recursive--;
	return TRUE;
    }
    recursive--;
    return FALSE;
}

static INT_PTR CALLBACK MiscDlgProc1 (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    currentpage = MISC1_ID;
    if (msg == WM_INITDIALOG) {
	pages[MISC1_ID] = hDlg;
	values_to_miscdlg (hDlg);
	enable_for_miscdlg (hDlg);
	return TRUE;
    }
    return MiscDlgProc (hDlg, msg, wParam, lParam);
}

static INT_PTR CALLBACK MiscDlgProc2 (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    currentpage = MISC2_ID;
    if (msg == WM_INITDIALOG) {
	pages[MISC2_ID] = hDlg;
	values_to_miscdlg (hDlg);
	enable_for_miscdlg (hDlg);
	return TRUE;
    }
    return MiscDlgProc (hDlg, msg, wParam, lParam);
}

static int cpu_ids[]   = { IDC_CPU0, IDC_CPU1, IDC_CPU2, IDC_CPU3, IDC_CPU4, IDC_CPU5 };
static int fpu_ids[]   = { IDC_FPU0, IDC_FPU1, IDC_FPU2, IDC_FPU3 };
static int trust_ids[] = { IDC_TRUST0, IDC_TRUST1, IDC_TRUST1, IDC_TRUST2 };

static void enable_for_cpudlg (HWND hDlg)
{
    BOOL enable = FALSE, enable2 = FALSE;
    BOOL cpu_based_enable = FALSE;
    BOOL fpu;

    /* These four items only get enabled when adjustable CPU style is enabled */
    ew (hDlg, IDC_SPEED, workprefs.m68k_speed > 0);
    ew (hDlg, IDC_COMPATIBLE24, workprefs.cpu_model == 68020);
    ew (hDlg, IDC_CS_CPU_TEXT, (!workprefs.cpu_cycle_exact || workprefs.cpu_model > 68000) && workprefs.m68k_speed > 0);
    ew (hDlg, IDC_CS_CHIPSET_TEXT, (!workprefs.cpu_cycle_exact || workprefs.cpu_model > 68000) && workprefs.m68k_speed > 0);
    ew (hDlg, IDC_CS_HOST, !workprefs.cpu_cycle_exact || workprefs.cpu_model > 68000);
    ew (hDlg, IDC_CS_68000, !workprefs.cpu_cycle_exact || workprefs.cpu_model > 68000);
    ew (hDlg, IDC_CS_ADJUSTABLE, !workprefs.cpu_cycle_exact || workprefs.cpu_model > 68000);
    ew (hDlg, IDC_CPUTEXT, workprefs.m68k_speed > 0 );
    ew (hDlg, IDC_CPUIDLE, workprefs.m68k_speed != 0 ? TRUE : FALSE);
#if !defined(CPUEMU_0) || defined(CPUEMU_68000_ONLY)
    ew (hDlg, IDC_CPU1, FALSE);
    ew (hDlg, IDC_CPU2, FALSE);
    ew (hDlg, IDC_CPU3, FALSE);
    ew (hDlg, IDC_CPU4, FALSE);
    ew (hDlg, IDC_CPU5, FALSE);
#endif

    cpu_based_enable = workprefs.cpu_model >= 68020 &&
		       workprefs.address_space_24 == 0;

    enable = cpu_based_enable && workprefs.cachesize;
#ifndef JIT
    enable = FALSE;
#endif
    enable2 = enable && workprefs.compforcesettings && candirect;

    ew (hDlg, IDC_TRUST0, enable2);
    ew (hDlg, IDC_TRUST1, enable2);
    ew (hDlg, IDC_TRUST2, enable2);
    ew (hDlg, IDC_HARDFLUSH, enable2);
    ew (hDlg, IDC_CONSTJUMP, enable2);
    ew (hDlg, IDC_JITFPU, enable2);
    ew (hDlg, IDC_NOFLAGS, enable2);
    ew (hDlg, IDC_CS_CACHE_TEXT, cpu_based_enable && workprefs.cachesize);
    ew (hDlg, IDC_CACHE, cpu_based_enable && workprefs.cachesize);
    ew (hDlg, IDC_CACHETEXT, cpu_based_enable && workprefs.cachesize);
    ew (hDlg, IDC_FORCE, enable);
    ew (hDlg, IDC_JITENABLE, cpu_based_enable);
    ew (hDlg, IDC_COMPATIBLE, !workprefs.cpu_cycle_exact && !workprefs.cachesize);
    ew (hDlg, IDC_COMPATIBLE_FPU, workprefs.fpu_model > 0);

    fpu = TRUE;
    if (workprefs.cpu_model < 68020 || workprefs.cpu_model > 68030)
	fpu = FALSE;
    ew (hDlg, IDC_FPU1, fpu);
    ew (hDlg, IDC_FPU2, fpu);
    ew (hDlg, IDC_FPU3, workprefs.cpu_model >= 68040);

#ifdef JIT
    if(enable) {
	if (!canbang) {
	    workprefs.compforcesettings = TRUE;
	    workprefs.comptrustbyte = 1;
	    workprefs.comptrustword = 1;
	    workprefs.comptrustlong = 1;
	    workprefs.comptrustnaddr= 1;
	}
    } else {
	workprefs.cachesize = 0; // Disable JIT
    }
#endif
}

static void values_to_cpudlg (HWND hDlg)
{
    char cache[8] = "";
    BOOL enable = FALSE;
    BOOL cpu_based_enable = FALSE;
    int cpu;

    SendDlgItemMessage (hDlg, IDC_SPEED, TBM_SETPOS, TRUE, workprefs.m68k_speed <= 0 ? 1 : workprefs.m68k_speed / CYCLE_UNIT );
    SetDlgItemInt( hDlg, IDC_CPUTEXT, workprefs.m68k_speed <= 0 ? 1 : workprefs.m68k_speed / CYCLE_UNIT, FALSE );
    CheckDlgButton (hDlg, IDC_COMPATIBLE, workprefs.cpu_compatible);
    CheckDlgButton (hDlg, IDC_COMPATIBLE24, workprefs.address_space_24);
    CheckDlgButton (hDlg, IDC_COMPATIBLE_FPU, workprefs.fpu_strict);
    SendDlgItemMessage (hDlg, IDC_CPUIDLE, TBM_SETPOS, TRUE, workprefs.cpu_idle == 0 ? 0 : 12 - workprefs.cpu_idle / 15);
    cpu = (workprefs.cpu_model - 68000) / 10;
    if (cpu >= 5)
	cpu--;
    CheckRadioButton (hDlg, IDC_CPU0, IDC_CPU5, cpu_ids[cpu]);
    CheckRadioButton (hDlg, IDC_FPU0, IDC_FPU3, fpu_ids[workprefs.fpu_model == 0 ? 0 : (workprefs.fpu_model == 68881 ? 1 : (workprefs.fpu_model == 68882 ? 2 : 3))]);

    if (workprefs.m68k_speed == -1)
	CheckRadioButton(hDlg, IDC_CS_HOST, IDC_CS_ADJUSTABLE, IDC_CS_HOST);
    else if (workprefs.m68k_speed == 0)
	CheckRadioButton(hDlg, IDC_CS_HOST, IDC_CS_ADJUSTABLE, IDC_CS_68000);
    else
	CheckRadioButton(hDlg, IDC_CS_HOST, IDC_CS_ADJUSTABLE, IDC_CS_ADJUSTABLE);

    cpu_based_enable = (workprefs.cpu_model >= 68020) && (workprefs.address_space_24 == 0);

    enable = cpu_based_enable && workprefs.cachesize;

#ifdef JIT
    if(enable) {
	if (!canbang || !candirect) {
	    workprefs.compforcesettings = TRUE;
	    workprefs.comptrustbyte = 1;
	    workprefs.comptrustword = 1;
	    workprefs.comptrustlong = 1;
	    workprefs.comptrustnaddr = 1;
	}
    } else {
#endif
	workprefs.cachesize = 0; // Disable JIT
#ifdef JIT
    }
#endif

    if(!workprefs.compforcesettings) {
	workprefs.comptrustbyte = 0;
	workprefs.comptrustword = 0;
	workprefs.comptrustlong = 0;
	workprefs.comptrustnaddr= 0;
    }

    CheckRadioButton(hDlg, IDC_TRUST0, IDC_TRUST2, trust_ids[workprefs.comptrustbyte]);

    SendDlgItemMessage(hDlg, IDC_CACHE, TBM_SETPOS, TRUE, workprefs.cachesize / 1024);
    sprintf(cache, "%d MB", workprefs.cachesize / 1024 );
    SetDlgItemText( hDlg, IDC_CACHETEXT, cache );

    CheckDlgButton(hDlg, IDC_FORCE, workprefs.compforcesettings);
    CheckDlgButton(hDlg, IDC_NOFLAGS, workprefs.compnf);
    CheckDlgButton(hDlg, IDC_JITFPU, workprefs.compfpu);
    CheckDlgButton(hDlg, IDC_HARDFLUSH, workprefs.comp_hardflush);
    CheckDlgButton(hDlg, IDC_CONSTJUMP, workprefs.comp_constjump);
    CheckDlgButton(hDlg, IDC_JITENABLE, workprefs.cachesize > 0);
}

static void values_from_cpudlg (HWND hDlg)
{
    int newcpu, newfpu, newtrust, oldcache, jitena;
    static int cachesize_prev, comptrust_prev, compforce_prev;

    workprefs.cpu_compatible = workprefs.cpu_cycle_exact | (IsDlgButtonChecked (hDlg, IDC_COMPATIBLE) ? 1 : 0);
    workprefs.fpu_strict = IsDlgButtonChecked (hDlg, IDC_COMPATIBLE_FPU) ? 1 : 0;
    workprefs.address_space_24 = IsDlgButtonChecked (hDlg, IDC_COMPATIBLE24) ? 1 : 0;
    workprefs.m68k_speed = IsDlgButtonChecked (hDlg, IDC_CS_HOST) ? -1
	: IsDlgButtonChecked (hDlg, IDC_CS_68000) ? 0
	: SendMessage (GetDlgItem (hDlg, IDC_SPEED), TBM_GETPOS, 0, 0) * CYCLE_UNIT;

    newcpu = IsDlgButtonChecked (hDlg, IDC_CPU0) ? 68000
	: IsDlgButtonChecked (hDlg, IDC_CPU1) ? 68010
	: IsDlgButtonChecked (hDlg, IDC_CPU2) ? 68020
	: IsDlgButtonChecked (hDlg, IDC_CPU3) ? 68030
	: IsDlgButtonChecked (hDlg, IDC_CPU4) ? 68040
	: IsDlgButtonChecked (hDlg, IDC_CPU5) ? 68060 : 0;
    newfpu = IsDlgButtonChecked (hDlg, IDC_FPU0) ? 0
	: IsDlgButtonChecked (hDlg, IDC_FPU1) ? 1
	: IsDlgButtonChecked (hDlg, IDC_FPU2) ? 2
	: IsDlgButtonChecked (hDlg, IDC_FPU3) ? 3 : 0;

    /* When switching away from 68000, disable 24 bit addressing.  */
    workprefs.cpu_model = newcpu;
    switch(newcpu)
    {
	case 68000:
	case 68010:
	    workprefs.fpu_model = 0;
	    workprefs.address_space_24 = 1;
	    if (newcpu == 0 && workprefs.cpu_cycle_exact)
		workprefs.m68k_speed = 0;
	break;
	case 68020:
	    workprefs.fpu_model = newfpu == 0 ? 0 : (newfpu == 2 ? 68882 : 68881);
	break;
	case 68030:
	    workprefs.address_space_24 = 0;
	    workprefs.fpu_model = newfpu == 0 ? 0 : (newfpu == 2 ? 68882 : 68881);
	break;
	case 68040:
	    workprefs.fpu_model = newfpu ? 68040 : 0;
	    workprefs.address_space_24 = 0;
	    if (workprefs.fpu_model)
		workprefs.fpu_model = 68040;
	break;
	case 68060:
	    workprefs.fpu_model = newfpu ? 68060 : 0;
	    workprefs.address_space_24 = 0;
	break;
    }
    newtrust = (IsDlgButtonChecked(hDlg, IDC_TRUST0) ? 0
	: IsDlgButtonChecked(hDlg, IDC_TRUST1) ? 1 : 3);
    workprefs.comptrustbyte = newtrust;
    workprefs.comptrustword = newtrust;
    workprefs.comptrustlong = newtrust;
    workprefs.comptrustnaddr= newtrust;

    workprefs.compforcesettings = IsDlgButtonChecked(hDlg, IDC_FORCE);
    workprefs.compnf            = IsDlgButtonChecked(hDlg, IDC_NOFLAGS);
    workprefs.compfpu           = IsDlgButtonChecked(hDlg, IDC_JITFPU);
    workprefs.comp_hardflush    = IsDlgButtonChecked(hDlg, IDC_HARDFLUSH);
    workprefs.comp_constjump    = IsDlgButtonChecked(hDlg, IDC_CONSTJUMP);

#ifdef JIT
    oldcache = workprefs.cachesize;
    jitena = IsDlgButtonChecked (hDlg, IDC_JITENABLE) ? 1 : 0;
    workprefs.cachesize = SendMessage(GetDlgItem(hDlg, IDC_CACHE), TBM_GETPOS, 0, 0) * 1024;
    if (!jitena) {
	cachesize_prev = workprefs.cachesize;
	comptrust_prev = workprefs.comptrustbyte;
	compforce_prev = workprefs.compforcesettings;
	workprefs.cachesize = 0;
    } else if (jitena && !oldcache) {
	workprefs.cachesize = 8192;
	if (cachesize_prev) {
	    workprefs.cachesize = cachesize_prev;
	    workprefs.comptrustbyte = comptrust_prev;
	    workprefs.comptrustword = comptrust_prev;
	    workprefs.comptrustlong = comptrust_prev;
	    workprefs.comptrustnaddr = comptrust_prev;
	    workprefs.compforcesettings = compforce_prev;
	}
    }
    if (oldcache == 0 && workprefs.cachesize > 0)
	canbang = 1;
    if (!candirect)
	canbang = 0;

#endif
    workprefs.cpu_idle = SendMessage(GetDlgItem(hDlg, IDC_CPUIDLE), TBM_GETPOS, 0, 0);
    if (workprefs.cpu_idle > 0)
	workprefs.cpu_idle = (12 - workprefs.cpu_idle) * 15;

    if (workprefs.cachesize > 0)
	workprefs.cpu_compatible = 0;
    if (pages[KICKSTART_ID])
	SendMessage(pages[KICKSTART_ID], WM_USER, 0, 0 );
    if (pages[DISPLAY_ID])
	SendMessage(pages[DISPLAY_ID], WM_USER, 0, 0 );
    if (pages[MEMORY_ID])
	SendMessage(pages[MEMORY_ID], WM_USER, 0, 0 );
}

static INT_PTR CALLBACK CPUDlgProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int recursive = 0;

    switch (msg) {
    case WM_INITDIALOG:
	pages[CPU_ID] = hDlg;
	currentpage = CPU_ID;
	SendDlgItemMessage (hDlg, IDC_SPEED, TBM_SETRANGE, TRUE, MAKELONG (MIN_M68K_PRIORITY, MAX_M68K_PRIORITY));
	SendDlgItemMessage (hDlg, IDC_SPEED, TBM_SETPAGESIZE, 0, 1);
	SendDlgItemMessage (hDlg, IDC_CACHE, TBM_SETRANGE, TRUE, MAKELONG (MIN_CACHE_SIZE, MAX_CACHE_SIZE));
	SendDlgItemMessage (hDlg, IDC_CACHE, TBM_SETPAGESIZE, 0, 1);
	SendDlgItemMessage (hDlg, IDC_CPUIDLE, TBM_SETRANGE, TRUE, MAKELONG (0, 10));
	SendDlgItemMessage (hDlg, IDC_CPUIDLE, TBM_SETPAGESIZE, 0, 1);

    case WM_USER:
	recursive++;
	values_to_cpudlg (hDlg);
	enable_for_cpudlg (hDlg);
	recursive--;
	return TRUE;

    case WM_COMMAND:
	if (recursive > 0)
	    break;
	recursive++;
	values_from_cpudlg (hDlg);
	values_to_cpudlg (hDlg);
	enable_for_cpudlg (hDlg);
	recursive--;
	break;

    case WM_HSCROLL:
	recursive++;
	values_from_cpudlg( hDlg );
	values_to_cpudlg( hDlg );
	enable_for_cpudlg( hDlg );
	recursive--;
	break;
    }
    return FALSE;
}

static void enable_for_sounddlg (HWND hDlg)
{
    int numdevs;

    enumerate_sound_devices (&numdevs);
    if( numdevs == 0 )
	ew (hDlg, IDC_SOUNDCARDLIST, FALSE);
    else
	ew (hDlg, IDC_SOUNDCARDLIST, workprefs.produce_sound);

    ew (hDlg, IDC_FREQUENCY, workprefs.produce_sound);
    ew (hDlg, IDC_SOUNDFREQ, workprefs.produce_sound ? TRUE : FALSE);
    ew (hDlg, IDC_SOUNDSTEREO, workprefs.produce_sound);
    ew (hDlg, IDC_SOUNDINTERPOLATION, workprefs.produce_sound);
    ew (hDlg, IDC_SOUNDVOLUME, workprefs.produce_sound);
    ew (hDlg, IDC_SOUNDVOLUME2, workprefs.produce_sound);
    ew (hDlg, IDC_SOUNDSTEREOSEP, workprefs.sound_stereo > 0 && workprefs.produce_sound);
    ew (hDlg, IDC_SOUNDSTEREOMIX, workprefs.sound_stereo > 0 && workprefs.produce_sound);

    ew (hDlg, IDC_SOUNDBUFFERMEM, workprefs.produce_sound);
    ew (hDlg, IDC_SOUNDBUFFERRAM, workprefs.produce_sound);
    ew (hDlg, IDC_SOUNDADJUST, workprefs.produce_sound);
    ew (hDlg, IDC_SOUNDADJUSTNUM, workprefs.produce_sound);
    ew (hDlg, IDC_SOUNDBUFFERTEXT, workprefs.produce_sound);

    ew (hDlg, IDC_SOUNDDRIVE, workprefs.produce_sound);
    ew (hDlg, IDC_SOUNDDRIVESELECT, workprefs.produce_sound);
    ew (hDlg, IDC_SOUNDDRIVEVOLUME, workprefs.produce_sound);
    ew (hDlg, IDC_SOUNDDRIVEVOLUME2, workprefs.produce_sound);
    ew (hDlg, IDC_AUDIOSYNC, workprefs.produce_sound);
    ew (hDlg, IDC_SOUNDFILTER, workprefs.produce_sound);
    ew (hDlg, IDC_SOUNDSWAP, workprefs.produce_sound);

    ew (hDlg, IDC_SOUNDCALIBRATE, workprefs.produce_sound && full_property_sheet);
}

static int exact_log2 (int v)
{
    int l = 0;
    while ((v >>= 1) != 0)
	l++;
    return l;
}

static char *drivesounds;

static void sound_loaddrivesamples (void)
{
    WIN32_FIND_DATA fd;
    HANDLE h;
    char *p;
    int len = 0;
    char dirname[1024];

    free (drivesounds);
    p = drivesounds = 0;
    sprintf (dirname, "%s\\uae_data\\*.wav", start_path_data);
    h = FindFirstFile (dirname, &fd);
    if (h == INVALID_HANDLE_VALUE)
	return;
    for (;;) {
	if (!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
	    char *name = fd.cFileName;
	    if (strlen (name) > strlen (DS_NAME_CLICK) + 4 && !strncmp (name, DS_NAME_CLICK, strlen (DS_NAME_CLICK))) {
		if (p - drivesounds < 1000) {
		    char *oldp = p;
		    len += 2000;
		    drivesounds = p = realloc (drivesounds, len);
		    if (oldp) {
			do {
			    p = p + strlen (p) + 1;
			} while (p[0]);
		    }
		}
		strcpy (p, name + strlen (DS_NAME_CLICK));
		p[strlen(name + strlen (DS_NAME_CLICK)) - 4] = 0;
		p += strlen (p);
		*p++ = 0;
		*p = 0;
	    }
	}
	if (!FindNextFile (h, &fd))
	    break;
    }
    FindClose (h);
}

extern int soundpercent;

static void update_soundgui (HWND hDlg)
{
    int bufsize;
    char txt[20];

    bufsize = exact_log2 (workprefs.sound_maxbsiz / 1024);
    sprintf (txt, "%d", bufsize);
    SetDlgItemText (hDlg, IDC_SOUNDBUFFERMEM, txt);

    SendDlgItemMessage( hDlg, IDC_SOUNDVOLUME, TBM_SETPOS, TRUE, 100 - workprefs.sound_volume);
    sprintf (txt, "%d%%", 100 - workprefs.sound_volume);
    SetDlgItemText (hDlg, IDC_SOUNDVOLUME2, txt);

    SendDlgItemMessage( hDlg, IDC_SOUNDDRIVEVOLUME, TBM_SETPOS, TRUE, 100 - workprefs.dfxclickvolume);
    sprintf (txt, "%d%%", 100 - workprefs.dfxclickvolume);
    SetDlgItemText (hDlg, IDC_SOUNDDRIVEVOLUME2, txt);
}

static int soundfreqs[] = { 11025, 15000, 22050, 32000, 44100, 48000, 0 };

static void values_to_sounddlg (HWND hDlg)
{
    int which_button;
    int sound_freq = workprefs.sound_freq;
    int produce_sound = workprefs.produce_sound;
    int stereo = workprefs.sound_stereo;
    char txt[100], txt2[100], *p;
    int i, selected;
    LRESULT idx;

    if (workprefs.sound_maxbsiz & (workprefs.sound_maxbsiz - 1))
	workprefs.sound_maxbsiz = DEFAULT_SOUND_MAXB;

    SendDlgItemMessage(hDlg, IDC_SOUNDFILTER, CB_RESETCONTENT, 0, 0);
    WIN32GUI_LoadUIString (IDS_SOUND_FILTER_OFF, txt, sizeof (txt));
    SendDlgItemMessage(hDlg, IDC_SOUNDFILTER, CB_ADDSTRING, 0, (LPARAM)txt);
    WIN32GUI_LoadUIString (IDS_SOUND_FILTER_EMULATED, txt, sizeof (txt));
    SendDlgItemMessage(hDlg, IDC_SOUNDFILTER, CB_ADDSTRING, 0, (LPARAM)txt);
    WIN32GUI_LoadUIString (IDS_SOUND_FILTER_EMULATED_E, txt, sizeof (txt));
    SendDlgItemMessage(hDlg, IDC_SOUNDFILTER, CB_ADDSTRING, 0, (LPARAM)txt);
    WIN32GUI_LoadUIString (IDS_SOUND_FILTER_ON_AGA, txt, sizeof (txt));
    SendDlgItemMessage(hDlg, IDC_SOUNDFILTER, CB_ADDSTRING, 0, (LPARAM)txt);
    WIN32GUI_LoadUIString (IDS_SOUND_FILTER_ON_A500, txt, sizeof (txt));
    SendDlgItemMessage(hDlg, IDC_SOUNDFILTER, CB_ADDSTRING, 0, (LPARAM)txt);
    i = 0;
    switch (workprefs.sound_filter)
    {
	case 0:
	i = 0;
	break;
	case 1:
	i = workprefs.sound_filter_type ? 2 : 1;
	break;
	case 2:
	i = workprefs.sound_filter_type ? 4 : 3;
	break;
    }
    SendDlgItemMessage(hDlg, IDC_SOUNDFILTER, CB_SETCURSEL, i, 0 );

    SendDlgItemMessage(hDlg, IDC_SOUNDSTEREO, CB_RESETCONTENT, 0, 0);
    WIN32GUI_LoadUIString (IDS_SOUND_MONO, txt, sizeof (txt));
    SendDlgItemMessage(hDlg, IDC_SOUNDSTEREO, CB_ADDSTRING, 0, (LPARAM)txt);
    WIN32GUI_LoadUIString (IDS_SOUND_STEREO, txt, sizeof (txt));
    SendDlgItemMessage(hDlg, IDC_SOUNDSTEREO, CB_ADDSTRING, 0, (LPARAM)txt);
    WIN32GUI_LoadUIString (IDS_SOUND_STEREO2, txt, sizeof (txt));
    SendDlgItemMessage(hDlg, IDC_SOUNDSTEREO, CB_ADDSTRING, 0, (LPARAM)txt);
    WIN32GUI_LoadUIString (IDS_SOUND_4CHANNEL, txt, sizeof (txt));
    SendDlgItemMessage(hDlg, IDC_SOUNDSTEREO, CB_ADDSTRING, 0, (LPARAM)txt);
    SendDlgItemMessage(hDlg, IDC_SOUNDSTEREO, CB_ADDSTRING, 0, (LPARAM)"Cloned Stereo (4 Channels in 5.1)");
    SendDlgItemMessage(hDlg, IDC_SOUNDSTEREO, CB_ADDSTRING, 0, (LPARAM)"4 Channels in 5.1");
    SendDlgItemMessage (hDlg, IDC_SOUNDSTEREO, CB_SETCURSEL, workprefs.sound_stereo, 0);

    SendDlgItemMessage(hDlg, IDC_SOUNDSWAP, CB_RESETCONTENT, 0, 0);
    SendDlgItemMessage(hDlg, IDC_SOUNDSWAP, CB_ADDSTRING, 0, (LPARAM)"-");
    WIN32GUI_LoadUIString (IDS_SOUND_SWAP_PAULA, txt, sizeof (txt));
    SendDlgItemMessage(hDlg, IDC_SOUNDSWAP, CB_ADDSTRING, 0, (LPARAM)txt);
    WIN32GUI_LoadUIString (IDS_SOUND_SWAP_AHI, txt, sizeof (txt));
    SendDlgItemMessage(hDlg, IDC_SOUNDSWAP, CB_ADDSTRING, 0, (LPARAM)txt);
    WIN32GUI_LoadUIString (IDS_SOUND_SWAP_BOTH, txt, sizeof (txt));
    SendDlgItemMessage(hDlg, IDC_SOUNDSWAP, CB_ADDSTRING, 0, (LPARAM)txt);
    SendDlgItemMessage(hDlg, IDC_SOUNDSWAP, CB_SETCURSEL,
	workprefs.sound_stereo_swap_paula + workprefs.sound_stereo_swap_ahi * 2, 0);

    SendDlgItemMessage(hDlg, IDC_SOUNDSTEREOSEP, CB_RESETCONTENT, 0, 0);
    for (i = 10; i >= 0; i--) {
	sprintf (txt, "%d%%", i * 10);
	SendDlgItemMessage(hDlg, IDC_SOUNDSTEREOSEP, CB_ADDSTRING, 0, (LPARAM)txt);
    }
    SendDlgItemMessage (hDlg, IDC_SOUNDSTEREOSEP, CB_SETCURSEL, 10 - workprefs.sound_stereo_separation, 0);

    SendDlgItemMessage(hDlg, IDC_SOUNDSTEREOMIX, CB_RESETCONTENT, 0, 0);
    SendDlgItemMessage(hDlg, IDC_SOUNDSTEREOMIX, CB_ADDSTRING, 0, (LPARAM)"-");
    for (i = 0; i < 10; i++) {
	sprintf (txt, "%d", i + 1);
	SendDlgItemMessage(hDlg, IDC_SOUNDSTEREOMIX, CB_ADDSTRING, 0, (LPARAM)txt);
    }
    SendDlgItemMessage (hDlg, IDC_SOUNDSTEREOMIX, CB_SETCURSEL,
	workprefs.sound_mixed_stereo_delay > 0 ? workprefs.sound_mixed_stereo_delay : 0, 0);

    SendDlgItemMessage(hDlg, IDC_SOUNDINTERPOLATION, CB_RESETCONTENT, 0, 0);
    WIN32GUI_LoadUIString (IDS_SOUND_INTERPOL_DISABLED, txt, sizeof (txt));
    SendDlgItemMessage(hDlg, IDC_SOUNDINTERPOLATION, CB_ADDSTRING, 0, (LPARAM)txt);
    SendDlgItemMessage(hDlg, IDC_SOUNDINTERPOLATION, CB_ADDSTRING, 0, (LPARAM)"Anti");
    SendDlgItemMessage(hDlg, IDC_SOUNDINTERPOLATION, CB_ADDSTRING, 0, (LPARAM)"Sinc");
    SendDlgItemMessage(hDlg, IDC_SOUNDINTERPOLATION, CB_ADDSTRING, 0, (LPARAM)"RH");
    SendDlgItemMessage(hDlg, IDC_SOUNDINTERPOLATION, CB_ADDSTRING, 0, (LPARAM)"Crux");
    SendDlgItemMessage (hDlg, IDC_SOUNDINTERPOLATION, CB_SETCURSEL, workprefs.sound_interpol, 0);

    SendDlgItemMessage(hDlg, IDC_SOUNDFREQ, CB_RESETCONTENT, 0, 0);
    i = 0;
    selected = -1;
    while (soundfreqs[i]) {
	sprintf (txt, "%d", soundfreqs[i]);
	SendDlgItemMessage( hDlg, IDC_SOUNDFREQ, CB_ADDSTRING, 0, (LPARAM)txt);
	i++;
    }
    sprintf (txt, "%d", workprefs.sound_freq);
    SendDlgItemMessage(hDlg, IDC_SOUNDFREQ, WM_SETTEXT, 0, (LPARAM)txt);

    switch (workprefs.produce_sound) {
     case 0: which_button = IDC_SOUND0; break;
     case 1: which_button = IDC_SOUND1; break;
     case 2: which_button = IDC_SOUND2; break;
     case 3: which_button = IDC_SOUND3; break;
    }
    CheckRadioButton(hDlg, IDC_SOUND0, IDC_SOUND3, which_button);

    CheckDlgButton(hDlg, IDC_SOUND_AUTO, workprefs.sound_auto);

    workprefs.sound_maxbsiz = 1 << exact_log2 (workprefs.sound_maxbsiz);
    if (workprefs.sound_maxbsiz < 2048)
	workprefs.sound_maxbsiz = 2048;
    SendDlgItemMessage(hDlg, IDC_SOUNDBUFFERRAM, TBM_SETPOS, TRUE, exact_log2 (workprefs.sound_maxbsiz / 2048));

    SendDlgItemMessage(hDlg, IDC_SOUNDVOLUME, TBM_SETPOS, TRUE, 0);
    SendDlgItemMessage(hDlg, IDC_SOUNDDRIVEVOLUME, TBM_SETPOS, TRUE, 0);

    SendDlgItemMessage(hDlg, IDC_SOUNDCARDLIST, CB_SETCURSEL, workprefs.win32_soundcard, 0);

    idx = SendDlgItemMessage (hDlg, IDC_SOUNDDRIVE, CB_GETCURSEL, 0, 0);
    if (idx < 0)
	idx = 0;
    SendDlgItemMessage(hDlg, IDC_SOUNDDRIVE, CB_RESETCONTENT, 0, 0);
    for (i = 0; i < 4; i++) {
	sprintf (txt, "DF%d:", i);
	SendDlgItemMessage(hDlg, IDC_SOUNDDRIVE, CB_ADDSTRING, 0, (LPARAM)txt);
    }
    SendDlgItemMessage (hDlg, IDC_SOUNDDRIVE, CB_SETCURSEL, idx, 0);
    SendDlgItemMessage(hDlg, IDC_SOUNDDRIVESELECT, CB_RESETCONTENT, 0, 0);
    WIN32GUI_LoadUIString (IDS_DRIVESOUND_NONE, txt, sizeof (txt));
    SendDlgItemMessage(hDlg, IDC_SOUNDDRIVESELECT, CB_ADDSTRING, 0, (LPARAM)txt);
    WIN32GUI_LoadUIString (IDS_DRIVESOUND_DEFAULT_A500, txt, sizeof (txt));
    SendDlgItemMessage(hDlg, IDC_SOUNDDRIVESELECT, CB_ADDSTRING, 0, (LPARAM)txt);
    driveclick_fdrawcmd_detect();
    if (driveclick_pcdrivemask) {
	for (i = 0; i < 2; i++) {
	    WIN32GUI_LoadUIString (IDS_DRIVESOUND_PC_FLOPPY, txt, sizeof (txt));
	    sprintf (txt2, txt, 'A' + i);
	    SendDlgItemMessage(hDlg, IDC_SOUNDDRIVESELECT, CB_ADDSTRING, 0, (LPARAM)txt2);
	}
    }
    SendDlgItemMessage (hDlg, IDC_SOUNDDRIVESELECT, CB_SETCURSEL, 0, 0);
    p = drivesounds;
    if (p) {
	while (p[0]) {
	    SendDlgItemMessage(hDlg, IDC_SOUNDDRIVESELECT, CB_ADDSTRING, 0, (LPARAM)p);
	    p += strlen (p) + 1;
	}
    }
    if (workprefs.dfxclick[idx] < 0) {
	p = drivesounds;
	i = DS_BUILD_IN_SOUNDS + (driveclick_pcdrivemask ? 2 : 0) + 1;
	while (p && p[0]) {
	    if (!strcmpi (p, workprefs.dfxclickexternal[idx])) {
		SendDlgItemMessage (hDlg, IDC_SOUNDDRIVESELECT, CB_SETCURSEL, i, 0);
		break;
	    }
	    i++;
	    p += strlen (p) + 1;
	}

    } else {
	SendDlgItemMessage (hDlg, IDC_SOUNDDRIVESELECT, CB_SETCURSEL, workprefs.dfxclick[idx], 0);
    }

    update_soundgui (hDlg);
}

static void values_from_sounddlg (HWND hDlg)
{
    char txt[10];
    LRESULT idx;

    idx = SendDlgItemMessage (hDlg, IDC_SOUNDFREQ, CB_GETCURSEL, 0, 0);
    if (idx >= 0) {
	workprefs.sound_freq = soundfreqs[idx];
    } else {
	SendDlgItemMessage (hDlg, IDC_SOUNDFREQ, WM_GETTEXT, (WPARAM)sizeof (txt), (LPARAM)txt);
	workprefs.sound_freq = atol (txt);
    }
    if (workprefs.sound_freq < 8000)
	workprefs.sound_freq = 8000;
    if (workprefs.sound_freq > 96000)
	workprefs.sound_freq = 96000;

    workprefs.produce_sound = (IsDlgButtonChecked (hDlg, IDC_SOUND0) ? 0
			       : IsDlgButtonChecked (hDlg, IDC_SOUND1) ? 1
			       : IsDlgButtonChecked (hDlg, IDC_SOUND2) ? 2 : 3);

    workprefs.sound_auto = IsDlgButtonChecked (hDlg, IDC_SOUND_AUTO) ? 1 : 0;

    idx = SendDlgItemMessage (hDlg, IDC_SOUNDSTEREO, CB_GETCURSEL, 0, 0);
    if (idx != CB_ERR)
	workprefs.sound_stereo = idx;
    workprefs.sound_stereo_separation = 0;
    workprefs.sound_mixed_stereo_delay = 0;
    if (workprefs.sound_stereo > 0) {
	idx = SendDlgItemMessage (hDlg, IDC_SOUNDSTEREOSEP, CB_GETCURSEL, 0, 0);
	if (idx != CB_ERR) {
	    if (idx > 0)
		workprefs.sound_mixed_stereo_delay = -1;
	    workprefs.sound_stereo_separation = 10 - idx;
	}
	idx = SendDlgItemMessage (hDlg, IDC_SOUNDSTEREOMIX, CB_GETCURSEL, 0, 0);
	if (idx != CB_ERR && idx > 0)
	    workprefs.sound_mixed_stereo_delay = idx;
    }

    workprefs.sound_interpol = SendDlgItemMessage (hDlg, IDC_SOUNDINTERPOLATION, CB_GETCURSEL, 0, 0);
    workprefs.win32_soundcard = SendDlgItemMessage (hDlg, IDC_SOUNDCARDLIST, CB_GETCURSEL, 0, 0L);
    switch (SendDlgItemMessage (hDlg, IDC_SOUNDFILTER, CB_GETCURSEL, 0, 0))
    {
	case 0:
	workprefs.sound_filter = FILTER_SOUND_OFF;
	break;
	case 1:
	workprefs.sound_filter = FILTER_SOUND_EMUL;
	workprefs.sound_filter_type = 0;
	break;
	case 2:
	workprefs.sound_filter = FILTER_SOUND_EMUL;
	workprefs.sound_filter_type = 1;
	break;
	case 3:
	workprefs.sound_filter = FILTER_SOUND_ON;
	workprefs.sound_filter_type = 0;
	break;
	case 4:
	workprefs.sound_filter = FILTER_SOUND_ON;
	workprefs.sound_filter_type = 1;
	break;
    }

    workprefs.sound_stereo_swap_paula = (SendDlgItemMessage (hDlg, IDC_SOUNDSWAP, CB_GETCURSEL, 0, 0) & 1) ? 1 : 0;
    workprefs.sound_stereo_swap_ahi = (SendDlgItemMessage (hDlg, IDC_SOUNDSWAP, CB_GETCURSEL, 0, 0) & 2) ? 1 : 0;

    idx = SendDlgItemMessage (hDlg, IDC_SOUNDDRIVE, CB_GETCURSEL, 0, 0);
    if (idx != CB_ERR && idx >= 0) {
	LRESULT res = SendDlgItemMessage (hDlg, IDC_SOUNDDRIVESELECT, CB_GETCURSEL, 0, 0);
	if (res != CB_ERR && res >= 0) {
	    int xtra = driveclick_pcdrivemask ? 2 : 0;
	    if (res > DS_BUILD_IN_SOUNDS + xtra) {
		int j = res - (DS_BUILD_IN_SOUNDS + xtra + 1);
		char *p = drivesounds;
		while (j-- > 0)
		    p += strlen (p) + 1;
		workprefs.dfxclick[idx] = -1;
		strcpy (workprefs.dfxclickexternal[idx], p);
	    } else {
		workprefs.dfxclick[idx] = res;
		workprefs.dfxclickexternal[idx][0] = 0;
	    }
	}
    }
}

extern int sound_calibrate (HWND, struct uae_prefs*);

static INT_PTR CALLBACK SoundDlgProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int recursive = 0;
    int numdevs;
    int card;
    char **sounddevs;

    switch (msg) {
    case WM_INITDIALOG:
	sound_loaddrivesamples ();
	SendDlgItemMessage (hDlg, IDC_SOUNDBUFFERRAM, TBM_SETRANGE, TRUE, MAKELONG (MIN_SOUND_MEM, MAX_SOUND_MEM));
	SendDlgItemMessage (hDlg, IDC_SOUNDBUFFERRAM, TBM_SETPAGESIZE, 0, 1);

	SendDlgItemMessage (hDlg, IDC_SOUNDVOLUME, TBM_SETRANGE, TRUE, MAKELONG (0, 100));
	SendDlgItemMessage (hDlg, IDC_SOUNDVOLUME, TBM_SETPAGESIZE, 0, 1);

	SendDlgItemMessage (hDlg, IDC_SOUNDDRIVEVOLUME, TBM_SETRANGE, TRUE, MAKELONG (0, 100));
	SendDlgItemMessage (hDlg, IDC_SOUNDDRIVEVOLUME, TBM_SETPAGESIZE, 0, 1);

	SendDlgItemMessage( hDlg, IDC_SOUNDADJUST, TBM_SETRANGE, TRUE, MAKELONG (-100, +30) );
	SendDlgItemMessage( hDlg, IDC_SOUNDADJUST, TBM_SETPAGESIZE, 0, 1 );

	SendDlgItemMessage( hDlg, IDC_SOUNDCARDLIST, CB_RESETCONTENT, 0, 0L );
	sounddevs = enumerate_sound_devices (&numdevs);
	for (card = 0; card < numdevs; card++)
	    SendDlgItemMessage (hDlg, IDC_SOUNDCARDLIST, CB_ADDSTRING, 0, (LPARAM)sounddevs[card]);
	if (numdevs == 0)
	    workprefs.produce_sound = 0; /* No sound card in system, enable_for_sounddlg will accomodate this */

	pages[SOUND_ID] = hDlg;
	currentpage = SOUND_ID;

    case WM_USER:
	recursive++;
	values_to_sounddlg (hDlg);
	enable_for_sounddlg (hDlg);
	recursive--;
	return TRUE;

    case WM_COMMAND:
	if (recursive > 0)
	    break;
	recursive++;
	if(LOWORD (wParam) == IDC_SOUNDDRIVE) {
	    values_to_sounddlg (hDlg);
	}
	values_from_sounddlg (hDlg);
	enable_for_sounddlg (hDlg);
	recursive--;
	break;

     case WM_HSCROLL:
	workprefs.sound_maxbsiz = 2048 << SendMessage(GetDlgItem(hDlg, IDC_SOUNDBUFFERRAM), TBM_GETPOS, 0, 0);
	workprefs.sound_volume = 100 - SendMessage(GetDlgItem(hDlg, IDC_SOUNDVOLUME), TBM_GETPOS, 0, 0);
	workprefs.dfxclickvolume = 100 - SendMessage(GetDlgItem(hDlg, IDC_SOUNDDRIVEVOLUME), TBM_GETPOS, 0, 0);
	update_soundgui (hDlg);
	break;
    }
    return FALSE;
}

#ifdef FILESYS

struct fsvdlg_vals
{
    char volume[MAX_DPATH];
    char device[MAX_DPATH];
    char rootdir[MAX_DPATH];
    int bootpri;
    int autoboot;
    int donotmount;
    int rw;
    int rdb;
};

static struct fsvdlg_vals empty_fsvdlg = { "", "", "", 0, 1, 1, 1, 0 };
static struct fsvdlg_vals current_fsvdlg;

struct hfdlg_vals
{
    char volumename[MAX_DPATH];
    char devicename[MAX_DPATH];
    char filename[MAX_DPATH];
    char fsfilename[MAX_DPATH];
    int sectors;
    int reserved;
    int surfaces;
    int cylinders;
    int blocksize;
    int rw;
    int rdb;
    int bootpri;
    int donotmount;
    int autoboot;
    int controller;
};

static struct hfdlg_vals empty_hfdlg = { "", "", "", "", 32, 2, 1, 0, 512, 1, 0, 0, 0, 1, 0 };
static struct hfdlg_vals current_hfdlg;
static int archivehd;

static INT_PTR CALLBACK VolumeSettingsProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int recursive = 0;
    BROWSEINFO browse_info;
    char directory_path[MAX_DPATH] = "";
    LPITEMIDLIST browse;
    char szTitle[MAX_DPATH];

    switch (msg) {
	case WM_INITDIALOG:
	{
	    archivehd = -1;
	    if (my_existsfile(current_fsvdlg.rootdir))
	        archivehd = 1;
	    else if (my_existsdir(current_fsvdlg.rootdir))
	        archivehd = 0;
	    recursive++;
	    setac (hDlg, IDC_PATH_NAME);
	    SetDlgItemText (hDlg, IDC_VOLUME_NAME, current_fsvdlg.volume);
	    SetDlgItemText (hDlg, IDC_VOLUME_DEVICE, current_fsvdlg.device);
	    SetDlgItemText (hDlg, IDC_PATH_NAME, current_fsvdlg.rootdir);
	    SetDlgItemInt (hDlg, IDC_VOLUME_BOOTPRI, current_fsvdlg.bootpri >= -127 ? current_fsvdlg.bootpri : -127, TRUE);
	    if (archivehd > 0)
		current_fsvdlg.rw = 0;
	    CheckDlgButton (hDlg, IDC_FS_RW, current_fsvdlg.rw);
	    CheckDlgButton (hDlg, IDC_FS_AUTOBOOT, current_fsvdlg.autoboot);
	    current_fsvdlg.donotmount = 0;
	    ew (hDlg, IDC_FS_RW, archivehd <= 0);
	    recursive--;
	}
	return TRUE;

	case WM_COMMAND:
	    if (recursive)
		break;
	    recursive++;
	    if (HIWORD (wParam) == BN_CLICKED) {
		switch (LOWORD (wParam))
		{
		    case IDC_FS_SELECT_EJECT:
			SetDlgItemText (hDlg, IDC_PATH_NAME, "");
			SetDlgItemText (hDlg, IDC_VOLUME_NAME, "");
			CheckDlgButton (hDlg, IDC_FS_RW, TRUE);
			ew (hDlg, IDC_FS_RW, TRUE);
			archivehd = -1;
		    break;
		    case IDC_FS_SELECT_FILE:
			strcpy (directory_path, current_fsvdlg.rootdir);
			if (DiskSelection(hDlg, 0, 14, &workprefs, directory_path)) {
			    char *s = filesys_createvolname (NULL, directory_path, "Harddrive");
			    SetDlgItemText (hDlg, IDC_PATH_NAME, directory_path);
			    SetDlgItemText (hDlg, IDC_VOLUME_NAME, s);
			    xfree (s);
			    CheckDlgButton (hDlg, IDC_FS_RW, FALSE);
			    ew (hDlg, IDC_FS_RW, FALSE);
			    archivehd = 1;
			}
		    break;
		    case IDC_FS_SELECT_DIR:
			strcpy (directory_path, current_fsvdlg.rootdir);
			WIN32GUI_LoadUIString(IDS_SELECTFILESYSROOT, szTitle, MAX_DPATH);
			browse_info.hwndOwner = hDlg;
			browse_info.pidlRoot = NULL;
			browse_info.pszDisplayName = directory_path;
			browse_info.lpszTitle = "";
			browse_info.ulFlags = BIF_DONTGOBELOWDOMAIN | BIF_RETURNONLYFSDIRS;
			browse_info.lpfn = NULL;
			browse_info.iImage = 0;
			if ((browse = SHBrowseForFolder (&browse_info)) != NULL) {
			    SHGetPathFromIDList (browse, directory_path);
			    SetDlgItemText (hDlg, IDC_PATH_NAME, directory_path);
			    ew (hDlg, IDC_FS_RW, TRUE);
			    archivehd = 0;
			}
		    break;
		    case IDOK:
			EndDialog (hDlg, 1);
		    break;
		    case IDCANCEL:
			EndDialog (hDlg, 0);
		    break;
		}
	    }
	    GetDlgItemText (hDlg, IDC_PATH_NAME, current_fsvdlg.rootdir, sizeof current_fsvdlg.rootdir);
	    GetDlgItemText (hDlg, IDC_VOLUME_NAME, current_fsvdlg.volume, sizeof current_fsvdlg.volume);
	    GetDlgItemText (hDlg, IDC_VOLUME_DEVICE, current_fsvdlg.device, sizeof current_fsvdlg.device);
	    current_fsvdlg.rw = IsDlgButtonChecked (hDlg, IDC_FS_RW);
	    current_fsvdlg.bootpri = GetDlgItemInt(hDlg, IDC_VOLUME_BOOTPRI, NULL, TRUE);
	    current_fsvdlg.autoboot = IsDlgButtonChecked (hDlg, IDC_FS_AUTOBOOT);
	    recursive--;
	break;
    }
    return FALSE;
}

STATIC_INLINE int is_hdf_rdb(void)
{
    return current_hfdlg.sectors == 0 && current_hfdlg.surfaces == 0 && current_hfdlg.reserved == 0;
}

static void sethardfile (HWND hDlg)
{
    int rdb = is_hdf_rdb();
    SetDlgItemText (hDlg, IDC_PATH_NAME, current_hfdlg.filename);
    SetDlgItemText (hDlg, IDC_PATH_FILESYS, current_hfdlg.fsfilename);
    SetDlgItemText (hDlg, IDC_HARDFILE_DEVICE, current_hfdlg.devicename);
    SetDlgItemInt (hDlg, IDC_SECTORS, current_hfdlg.sectors, FALSE);
    SetDlgItemInt (hDlg, IDC_HEADS, current_hfdlg.surfaces, FALSE);
    SetDlgItemInt (hDlg, IDC_RESERVED, current_hfdlg.reserved, FALSE);
    SetDlgItemInt (hDlg, IDC_BLOCKSIZE, current_hfdlg.blocksize, FALSE);
    SetDlgItemInt (hDlg, IDC_HARDFILE_BOOTPRI, current_hfdlg.bootpri >= -127 ? current_hfdlg.bootpri : -127, TRUE);
    CheckDlgButton (hDlg, IDC_HDF_RW, current_hfdlg.rw);
    CheckDlgButton (hDlg, IDC_HDF_AUTOBOOT, current_hfdlg.autoboot);
    CheckDlgButton (hDlg, IDC_HDF_DONOTMOUNT, current_hfdlg.donotmount);
    ew (hDlg, IDC_HDF_RDB, !rdb);
    ew (hDlg, IDC_HDF_AUTOBOOT, !rdb);
    ew (hDlg, IDC_HDF_DONOTMOUNT, !rdb);
    SendDlgItemMessage (hDlg, IDC_HDF_CONTROLLER, CB_SETCURSEL, current_hfdlg.controller, 0);
}

static void inithdcontroller (HWND hDlg)
{
    SendDlgItemMessage(hDlg, IDC_HDF_CONTROLLER, CB_RESETCONTENT, 0, 0);
    SendDlgItemMessage (hDlg, IDC_HDF_CONTROLLER, CB_ADDSTRING, 0, (LPARAM)"UAE");
    SendDlgItemMessage (hDlg, IDC_HDF_CONTROLLER, CB_ADDSTRING, 0, (LPARAM)"IDE0");
    SendDlgItemMessage (hDlg, IDC_HDF_CONTROLLER, CB_ADDSTRING, 0, (LPARAM)"IDE1");
    SendDlgItemMessage (hDlg, IDC_HDF_CONTROLLER, CB_ADDSTRING, 0, (LPARAM)"IDE2");
    SendDlgItemMessage (hDlg, IDC_HDF_CONTROLLER, CB_ADDSTRING, 0, (LPARAM)"IDE3");
    SendDlgItemMessage (hDlg, IDC_HDF_CONTROLLER, CB_ADDSTRING, 0, (LPARAM)"SCSI0");
    SendDlgItemMessage (hDlg, IDC_HDF_CONTROLLER, CB_ADDSTRING, 0, (LPARAM)"SCSI1");
    SendDlgItemMessage (hDlg, IDC_HDF_CONTROLLER, CB_ADDSTRING, 0, (LPARAM)"SCSI2");
    SendDlgItemMessage (hDlg, IDC_HDF_CONTROLLER, CB_ADDSTRING, 0, (LPARAM)"SCSI3");
    SendDlgItemMessage (hDlg, IDC_HDF_CONTROLLER, CB_ADDSTRING, 0, (LPARAM)"SCSI4");
    SendDlgItemMessage (hDlg, IDC_HDF_CONTROLLER, CB_ADDSTRING, 0, (LPARAM)"SCSI5");
    SendDlgItemMessage (hDlg, IDC_HDF_CONTROLLER, CB_ADDSTRING, 0, (LPARAM)"SCSI6");
    //SendDlgItemMessage (hDlg, IDC_HDF_CONTROLLER, CB_ADDSTRING, 0, (LPARAM)"SCSRAM");
    SendDlgItemMessage (hDlg, IDC_HDF_CONTROLLER, CB_SETCURSEL, 0, 0);
}

static void inithardfile (HWND hDlg)
{
    char tmp[MAX_DPATH];

    ew (hDlg, IDC_HF_DOSTYPE, FALSE);
    ew (hDlg, IDC_HF_CREATE, FALSE);
    inithdcontroller (hDlg);
    SendDlgItemMessage(hDlg, IDC_HF_TYPE, CB_RESETCONTENT, 0, 0);
    WIN32GUI_LoadUIString (IDS_HF_FS_CUSTOM, tmp, sizeof (tmp));
    SendDlgItemMessage (hDlg, IDC_HF_TYPE, CB_ADDSTRING, 0, (LPARAM)"OFS/FFS/RDB");
    SendDlgItemMessage (hDlg, IDC_HF_TYPE, CB_ADDSTRING, 0, (LPARAM)"SFS");
    SendDlgItemMessage (hDlg, IDC_HF_TYPE, CB_ADDSTRING, 0, (LPARAM)tmp);
    SendDlgItemMessage (hDlg, IDC_HF_TYPE, CB_SETCURSEL, 0, 0);
}

static void sethfdostype (HWND hDlg, int idx)
{
    if (idx == 1)
	SetDlgItemText (hDlg, IDC_HF_DOSTYPE, "0x53465300");
    else
	SetDlgItemText (hDlg, IDC_HF_DOSTYPE, "");
}

static void hardfile_testrdb (HWND hDlg)
{
    void *f = zfile_fopen (current_hfdlg.filename, "rb");
    char tmp[8] = { 0 };
    if (!f)
	return;
    zfile_fread (tmp, 1, sizeof (tmp), f);
    zfile_fclose (f);
    if (memcmp (tmp, "RDSK\0\0\0", 7))
	return;
    current_hfdlg.sectors = 0;
    current_hfdlg.surfaces = 0;
    current_hfdlg.reserved = 0;
    current_hfdlg.fsfilename[0] = 0;
    current_hfdlg.bootpri = 0;
    current_hfdlg.autoboot = 1;
    current_hfdlg.donotmount = 0;
    current_hfdlg.devicename[0] = 0;
    sethardfile (hDlg);
}

static void updatehdfinfo (HWND hDlg, int force)
{
    static uae_u64 bsize;
    static uae_u8 id[4];
    int blocks, cyls, i;
    char tmp[200], idtmp[5], tmp2[200];

    if (force) {
	struct zfile *zf = zfile_fopen (current_hfdlg.filename, "rb");
	if (zf) {
	    memset (id, 0, sizeof (id));
	    zfile_fread (id, 1, sizeof (id), zf);
	    zfile_fseek (zf, 0, SEEK_END);
	    bsize = zfile_ftell (zf);
	    zfile_fclose (zf);
	    if (bsize == -1) { /* perhaps file is >=2G? */
		FILE *fp = fopen (current_hfdlg.filename, "rb");
		if (fp) {
		    _fseeki64 (fp, 0, SEEK_END);
		    bsize = _ftelli64 (fp);
		    fclose (fp);
		}
	    }
	}
    }
    cyls = 0;
    if (current_hfdlg.blocksize * current_hfdlg.sectors * current_hfdlg.surfaces)
        cyls = bsize / (current_hfdlg.blocksize * current_hfdlg.sectors * current_hfdlg.surfaces);
    blocks = cyls * (current_hfdlg.sectors * current_hfdlg.surfaces);
    for (i = 0; i < 4; i++) {
	unsigned char c = id[i];
	if (c < 32 || c > 126)
	    c = '.';
	idtmp[i] = c;
	idtmp[i + 1] = 0;
    }

    tmp[0] = 0;
    if (bsize) {
	sprintf (tmp2, " %s [%02X%02X%02X%02X]", idtmp, id[0], id[1], id[2], id[3]);
	if (!cyls || !blocks)
	    sprintf (tmp, "%dMB", bsize / (1024 * 1024));
	else
	    sprintf (tmp, "%d cyls, %d blocks, %.1fMB/%.1fMB",
		cyls, blocks,
		(double)blocks * 1.0 * current_hfdlg.blocksize / (1024.0 * 1024.0),
		(double)bsize / (1024.0 * 1024.0));
	if (cyls > 65535)
	    strcat (tmp, " [Too many cyls]");
	strcat (tmp, tmp2);
    }
    SetDlgItemText (hDlg, IDC_HDFINFO, tmp);
}


static INT_PTR CALLBACK HardfileSettingsProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int recursive = 0;
    LRESULT res, posn;
    char tmp[MAX_DPATH];

    switch (msg) {
    case WM_INITDIALOG:
	recursive++;
	inithardfile (hDlg);
	sethardfile (hDlg);
	sethfdostype (hDlg, 0);
	updatehdfinfo (hDlg, 1);
	setac (hDlg, IDC_PATH_NAME);
	recursive--;
	return TRUE;

    case WM_COMMAND:
	if (recursive)
	    break;
	recursive++;

	switch (LOWORD (wParam)) {
	    case IDC_HF_SIZE:
		ew (hDlg, IDC_HF_CREATE, CalculateHardfileSize (hDlg) > 0);
	    break;
	    case IDC_HF_TYPE:
		res = SendDlgItemMessage (hDlg, IDC_HF_TYPE, CB_GETCURSEL, 0, 0);
		sethfdostype (hDlg, (int)res);
		ew (hDlg, IDC_HF_DOSTYPE, res >= 2);
	    break;
	    case IDC_HF_CREATE:
	    {
		UINT setting = CalculateHardfileSize (hDlg);
		char dostype[16];
		GetDlgItemText (hDlg, IDC_HF_DOSTYPE, dostype, sizeof (dostype));
		res = SendDlgItemMessage (hDlg, IDC_HF_TYPE, CB_GETCURSEL, 0, 0);
		if (res == 0)
		    dostype[0] = 0;
		if (!CreateHardFile(hDlg, setting, dostype)) {
		    char szMessage[MAX_DPATH];
		    char szTitle[MAX_DPATH];
		    WIN32GUI_LoadUIString (IDS_FAILEDHARDFILECREATION, szMessage, MAX_DPATH);
		    WIN32GUI_LoadUIString (IDS_CREATIONERROR, szTitle, MAX_DPATH);
		    MessageBox(hDlg, szMessage, szTitle,
				MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
		}
	    }
	    break;
	    case IDC_SELECTOR:
		DiskSelection (hDlg, IDC_PATH_NAME, 2, &workprefs, 0);
		GetDlgItemText (hDlg, IDC_PATH_NAME, current_hfdlg.filename, sizeof current_hfdlg.filename);
		hardfile_testrdb (hDlg);
		inithardfile(hDlg);
		updatehdfinfo (hDlg, 1);
		break;
	    case IDC_FILESYS_SELECTOR:
		DiskSelection (hDlg, IDC_PATH_FILESYS, 12, &workprefs, 0);
		break;
	    case IDOK:
		if(strlen(current_hfdlg.filename) == 0 )
		{
		    char szMessage[MAX_DPATH];
		    char szTitle[MAX_DPATH];
		    WIN32GUI_LoadUIString(IDS_MUSTSELECTFILE, szMessage, MAX_DPATH);
		    WIN32GUI_LoadUIString(IDS_SETTINGSERROR, szTitle, MAX_DPATH);

		    MessageBox( hDlg, szMessage, szTitle,
				MB_OK | MB_ICONERROR | MB_APPLMODAL | MB_SETFOREGROUND);
		    break;
		}
		EndDialog (hDlg, 1);
		break;
	    case IDCANCEL:
		EndDialog (hDlg, 0);
		break;
	    case IDC_HDF_RW:
		current_hfdlg.rw = IsDlgButtonChecked (hDlg, IDC_HDF_RW);
		break;
	    case IDC_HDF_AUTOBOOT:
		current_hfdlg.autoboot = IsDlgButtonChecked (hDlg, IDC_HDF_AUTOBOOT);
		break;
	    case IDC_HDF_DONOTMOUNT:
		current_hfdlg.donotmount = IsDlgButtonChecked (hDlg, IDC_HDF_DONOTMOUNT);
		break;
	    case IDC_HDF_RDB:
		SetDlgItemInt (hDlg, IDC_SECTORS, 0, FALSE);
		SetDlgItemInt (hDlg, IDC_RESERVED, 0, FALSE);
		SetDlgItemInt (hDlg, IDC_HEADS, 0, FALSE);
		SetDlgItemText (hDlg, IDC_PATH_FILESYS, "");
		SetDlgItemText (hDlg, IDC_HARDFILE_DEVICE, "");
		current_hfdlg.sectors = current_hfdlg.reserved = current_hfdlg.surfaces = 0;
		current_hfdlg.autoboot = 1;
		current_hfdlg.donotmount = 0;
		sethardfile (hDlg);
		break;
	}

	current_hfdlg.sectors   = GetDlgItemInt(hDlg, IDC_SECTORS, NULL, FALSE);
	current_hfdlg.reserved  = GetDlgItemInt(hDlg, IDC_RESERVED, NULL, FALSE);
	current_hfdlg.surfaces  = GetDlgItemInt(hDlg, IDC_HEADS, NULL, FALSE);
	current_hfdlg.blocksize = GetDlgItemInt(hDlg, IDC_BLOCKSIZE, NULL, FALSE);
	current_hfdlg.bootpri = GetDlgItemInt(hDlg, IDC_HARDFILE_BOOTPRI, NULL, TRUE);
	GetDlgItemText (hDlg, IDC_PATH_NAME, tmp, sizeof tmp);
	if (strcmp (tmp, current_hfdlg.filename)) {
	    strcpy (current_hfdlg.filename, tmp);
	    updatehdfinfo (hDlg, 1);
	}
	GetDlgItemText (hDlg, IDC_PATH_FILESYS, current_hfdlg.fsfilename, sizeof current_hfdlg.fsfilename);
	GetDlgItemText (hDlg, IDC_HARDFILE_DEVICE, current_hfdlg.devicename, sizeof current_hfdlg.devicename);
	posn = SendDlgItemMessage (hDlg, IDC_HDF_CONTROLLER, CB_GETCURSEL, 0, 0);
	if (posn != CB_ERR)
	    current_hfdlg.controller = posn;
	updatehdfinfo (hDlg, 0);
	recursive--;

	break;
    }
    return FALSE;
}

extern int harddrive_to_hdf(HWND, struct uae_prefs*, int);
static INT_PTR CALLBACK HarddriveSettingsProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int recursive = 0;
    int i;
    LRESULT posn;
    static int oposn;

    switch (msg) {
    case WM_INITDIALOG:
    {
	int index;
	oposn = -1;
	hdf_init ();
	recursive++;
	inithdcontroller (hDlg);
	CheckDlgButton (hDlg, IDC_HDF_RW, current_hfdlg.rw);
	SendDlgItemMessage(hDlg, IDC_HARDDRIVE, CB_RESETCONTENT, 0, 0);
	ew (hDlg, IDC_HARDDRIVE_IMAGE, FALSE);
	ew (hDlg, IDOK, FALSE);
	ew (hDlg, IDC_HDF_RW, FALSE);
	ew (hDlg, IDC_HDF_CONTROLLER, FALSE);
	index = -1;
	for (i = 0; i < hdf_getnumharddrives(); i++) {
	    SendDlgItemMessage(hDlg, IDC_HARDDRIVE, CB_ADDSTRING, 0, (LPARAM)hdf_getnameharddrive(i, 1, NULL));
	    if (!strcmp (current_hfdlg.filename, hdf_getnameharddrive (i, 0, NULL)))
		index = i;
	}
	if (index >= 0) {
	    SendDlgItemMessage (hDlg, IDC_HARDDRIVE, CB_SETCURSEL, index, 0);
	    SendDlgItemMessage (hDlg, IDC_HDF_CONTROLLER, CB_SETCURSEL, current_hfdlg.controller, 0);
	}
	recursive--;
	return TRUE;
    }
    case WM_COMMAND:
	if (recursive)
	    break;
	recursive++;
	posn = SendDlgItemMessage (hDlg, IDC_HARDDRIVE, CB_GETCURSEL, 0, 0);
	if (oposn != posn && posn != CB_ERR) {
	    oposn = posn;
	    if (posn >= 0) {
		ew (hDlg, IDC_HARDDRIVE_IMAGE, TRUE);
		ew (hDlg, IDOK, TRUE);
		ew (hDlg, IDC_HDF_RW, TRUE);
		ew (hDlg, IDC_HDF_CONTROLLER, TRUE);
	    }
	}
	if (HIWORD (wParam) == BN_CLICKED) {
	    switch (LOWORD (wParam)) {
	    case IDOK:
		EndDialog (hDlg, 1);
		break;
	    case IDCANCEL:
		EndDialog (hDlg, 0);
		break;
	    case IDC_HARDDRIVE_IMAGE:
		if (posn != CB_ERR)
		    harddrive_to_hdf (hDlg, &workprefs, posn);
		break;
	    }
	}
	if (posn != CB_ERR)
	    strcpy (current_hfdlg.filename, hdf_getnameharddrive ((int)posn, 0, &current_hfdlg.blocksize));
	current_hfdlg.rw = IsDlgButtonChecked (hDlg, IDC_HDF_RW);
	posn = SendDlgItemMessage (hDlg, IDC_HDF_CONTROLLER, CB_GETCURSEL, 0, 0);
	if (posn != CB_ERR)
	    current_hfdlg.controller = posn;
	recursive--;
	break;
    }
    return FALSE;
}

static int tweakbootpri (int bp, int ab, int dnm)
{
    if (dnm)
	return -129;
    if (!ab)
	return -128;
    if (bp < -127)
	bp = -127;
    return bp;
}

static void new_filesys (HWND hDlg, int entry)
{
    struct uaedev_config_info *uci;
    int bp = tweakbootpri (current_fsvdlg.bootpri, current_fsvdlg.autoboot, current_fsvdlg.donotmount);

    uci = add_filesys_config (&workprefs, entry, current_fsvdlg.device, current_fsvdlg.volume,
		    current_fsvdlg.rootdir, ! current_fsvdlg.rw, 0, 0, 0, 0, bp, 0, 0, 0);
    if (uci) {
	if (uci->rootdir[0])
	    filesys_media_change (uci->rootdir, 1, uci);
	else
	    filesys_eject (uci->configoffset);
    }
}

static void new_hardfile (HWND hDlg, int entry)
{
    struct uaedev_config_info *uci;
    int bp = tweakbootpri (current_hfdlg.bootpri, current_hfdlg.autoboot, current_hfdlg.donotmount);

    uci = add_filesys_config (&workprefs, entry, current_hfdlg.devicename, 0,
				current_hfdlg.filename, ! current_hfdlg.rw,
				current_hfdlg.sectors, current_hfdlg.surfaces,
				current_hfdlg.reserved, current_hfdlg.blocksize,
				bp, current_hfdlg.fsfilename,
				current_hfdlg.controller, 0);
    if (uci)
	hardfile_do_disk_change (uci->configoffset, 1);
}

static void new_harddrive (HWND hDlg, int entry)
{
    struct uaedev_config_info *uci;

    uci = add_filesys_config (&workprefs, entry, 0, 0,
				current_hfdlg.filename, ! current_hfdlg.rw, 0, 0,
				0, current_hfdlg.blocksize, 0, 0, current_hfdlg.controller, 0);
    if (uci)
	hardfile_do_disk_change (uci->configoffset, 1);
}

static void harddisk_remove (HWND hDlg)
{
    int entry = listview_find_selected (GetDlgItem (hDlg, IDC_VOLUMELIST));
    if (entry < 0)
	return;
    kill_filesys_unitconfig (&workprefs, entry);
}

static void harddisk_move (HWND hDlg, int up)
{
    int entry = listview_find_selected (GetDlgItem (hDlg, IDC_VOLUMELIST));
    if (entry < 0)
	return;
    move_filesys_unitconfig (&workprefs, entry, up ? entry - 1 : entry + 1);
}

static void harddisk_edit (HWND hDlg)
{
    int entry = listview_find_selected (GetDlgItem (hDlg, IDC_VOLUMELIST));
    int type;
    struct uaedev_config_info *uci;
    struct mountedinfo mi;

    if (entry < 0 || entry >= workprefs.mountitems)
	return;
    uci = &workprefs.mountconfig[entry];

    type = get_filesys_unitconfig (&workprefs, entry, &mi);
    if (type < 0)
	type = uci->ishdf ? FILESYS_HARDFILE : FILESYS_VIRTUAL;

    if(type == FILESYS_HARDFILE || type == FILESYS_HARDFILE_RDB)
    {
	current_hfdlg.sectors = uci->sectors;
	current_hfdlg.surfaces = uci->surfaces;
	current_hfdlg.reserved = uci->reserved;
	current_hfdlg.cylinders = mi.nrcyls;
	current_hfdlg.blocksize = uci->blocksize;
	current_hfdlg.controller = uci->controller;

	strncpy (current_hfdlg.filename, uci->rootdir, (sizeof current_hfdlg.filename) - 1);
	current_hfdlg.filename[(sizeof current_hfdlg.filename) - 1] = '\0';
	current_hfdlg.fsfilename[0] = 0;
	if (uci->filesys) {
	    strncpy (current_hfdlg.fsfilename, uci->filesys, (sizeof current_hfdlg.fsfilename) - 1);
	    current_hfdlg.fsfilename[(sizeof current_hfdlg.fsfilename) - 1] = '\0';
	}
	current_fsvdlg.device[0] = 0;
	if (uci->devname) {
	    strncpy (current_hfdlg.devicename, uci->devname, (sizeof current_hfdlg.devicename) - 1);
	    current_hfdlg.devicename[(sizeof current_hfdlg.devicename) - 1] = '\0';
	}
	current_hfdlg.rw = !uci->readonly;
	current_hfdlg.bootpri = uci->bootpri;
	current_hfdlg.autoboot = uci->autoboot;
	current_hfdlg.donotmount = uci->donotmount;
	if (CustomDialogBox(IDD_HARDFILE, hDlg, HardfileSettingsProc))
	{
	    new_hardfile (hDlg, entry);
	}
    }
    else if (type == FILESYS_HARDDRIVE) /* harddisk */
    {
	current_hfdlg.controller = uci->controller;
	current_hfdlg.rw = !uci->readonly;
	strncpy (current_hfdlg.filename, uci->rootdir, (sizeof current_hfdlg.filename) - 1);
	current_hfdlg.filename[(sizeof current_hfdlg.filename) - 1] = '\0';
	if (CustomDialogBox(IDD_HARDDRIVE, hDlg, HarddriveSettingsProc))
	{
	    new_harddrive (hDlg, entry);
	}
    }
    else /* Filesystem */
    {
	strncpy (current_fsvdlg.rootdir, uci->rootdir, (sizeof current_fsvdlg.rootdir) - 1);
	current_fsvdlg.rootdir[(sizeof current_fsvdlg.rootdir) - 1] = '\0';
	strncpy (current_fsvdlg.volume, uci->volname, (sizeof current_fsvdlg.volume) - 1);
	current_fsvdlg.volume[(sizeof current_fsvdlg.volume) - 1] = '\0';
	current_fsvdlg.device[0] = 0;
	if (uci->devname) {
	    strncpy (current_fsvdlg.device, uci->devname, (sizeof current_fsvdlg.device) - 1);
	    current_fsvdlg.device[(sizeof current_fsvdlg.device) - 1] = '\0';
	}
	current_fsvdlg.rw = !uci->readonly;
	current_fsvdlg.bootpri = uci->bootpri;
	current_fsvdlg.autoboot = uci->autoboot;
	current_fsvdlg.donotmount = uci->donotmount;
	archivehd = -1;
	if (CustomDialogBox(IDD_FILESYS, hDlg, VolumeSettingsProc)) {
	    new_filesys (hDlg, entry);
	}
    }
}

static ACCEL HarddiskAccel[] = {
    { FVIRTKEY, VK_UP, 10001 }, { FVIRTKEY, VK_DOWN, 10002 },
    { FVIRTKEY|FSHIFT, VK_UP, IDC_UP }, { FVIRTKEY|FSHIFT, VK_DOWN, IDC_DOWN },
    { FVIRTKEY, VK_RETURN, IDC_EDIT }, { FVIRTKEY, VK_DELETE, IDC_REMOVE },
    { 0, 0, 0 }
};

static void harddiskdlg_button (HWND hDlg, int button)
{
    switch (button) {
     case IDC_NEW_FS:
	current_fsvdlg = empty_fsvdlg;
	archivehd = 0;
	if (CustomDialogBox(IDD_FILESYS, hDlg, VolumeSettingsProc))
	    new_filesys (hDlg, -1);
	break;
     case IDC_NEW_FSARCH:
	 archivehd = 1;
	current_fsvdlg = empty_fsvdlg;
	if (CustomDialogBox(IDD_FILESYS, hDlg, VolumeSettingsProc))
	    new_filesys (hDlg, -1);
	break;

     case IDC_NEW_HF:
	current_hfdlg = empty_hfdlg;
	if (CustomDialogBox (IDD_HARDFILE, hDlg, HardfileSettingsProc))
	    new_hardfile (hDlg, -1);
	break;

     case IDC_NEW_HD:
	memset (&current_hfdlg, 0, sizeof (current_hfdlg));
	if (hdf_init() == 0) {
	    char tmp[MAX_DPATH];
	    WIN32GUI_LoadUIString (IDS_NOHARDDRIVES, tmp, sizeof (tmp));
	    gui_message (tmp);
	} else {
	    if (CustomDialogBox (IDD_HARDDRIVE, hDlg, HarddriveSettingsProc))
		new_harddrive (hDlg, -1);
	}
	break;

     case IDC_EDIT:
	harddisk_edit (hDlg);
	break;

     case IDC_REMOVE:
	harddisk_remove (hDlg);
	break;

     case IDC_UP:
	harddisk_move (hDlg, 1);
	clicked_entry--;
	break;

     case IDC_DOWN:
	harddisk_move (hDlg, 0);
	clicked_entry++;
	break;

     case IDC_MAPDRIVES_AUTO:
	workprefs.win32_automount_removable = IsDlgButtonChecked(hDlg, IDC_MAPDRIVES_AUTO);
	break;

     case IDC_MAPDRIVES:
	workprefs.win32_automount_drives = IsDlgButtonChecked(hDlg, IDC_MAPDRIVES);
	break;

     case IDC_MAPDRIVES_CD:
	workprefs.win32_automount_cddrives = IsDlgButtonChecked(hDlg, IDC_MAPDRIVES_CD);
	break;

     case IDC_MAPDRIVES_NET:
	workprefs.win32_automount_netdrives = IsDlgButtonChecked(hDlg, IDC_MAPDRIVES_NET);
	break;

    case IDC_NOUAEFSDB:
	workprefs.filesys_no_uaefsdb = IsDlgButtonChecked(hDlg, IDC_NOUAEFSDB);
	break;

    case IDC_NORECYCLEBIN:
	workprefs.win32_norecyclebin = IsDlgButtonChecked(hDlg, IDC_NORECYCLEBIN);
	break;

    }
}

static void harddiskdlg_volume_notify (HWND hDlg, NM_LISTVIEW *nmlistview)
{
    HWND list = nmlistview->hdr.hwndFrom;
    int dblclick = 0;
    int entry = 0;

    switch (nmlistview->hdr.code) {
     case LVN_BEGINDRAG:
	drag_start (hDlg, cachedlist, (LPARAM)nmlistview);
	break;
     case NM_DBLCLK:
	dblclick = 1;
	/* fall through */
     case NM_CLICK:
	entry = listview_entry_from_click (list, 0);
	if (entry >= 0)
	{
	    if(dblclick)
		harddisk_edit (hDlg);
	    InitializeListView (hDlg);
	    clicked_entry = entry;
	    // Hilite the current selected item
	    ListView_SetItemState (cachedlist, clicked_entry, LVIS_SELECTED, LVIS_SELECTED);
	}
	break;
    }
}

static void hilitehd (void)
{
    int total = ListView_GetItemCount (cachedlist);
    if (total <= 0)
	return;
    if (clicked_entry < 0)
	clicked_entry = 0;
    if (clicked_entry >= total)
	clicked_entry = total;
    ListView_SetItemState( cachedlist, clicked_entry, LVIS_SELECTED | LVIS_FOCUSED, LVIS_SELECTED | LVIS_FOCUSED );
}

static INT_PTR CALLBACK HarddiskDlgProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg) {
    case WM_INITDIALOG:
	clicked_entry = 0;
	pages[HARDDISK_ID] = hDlg;
	currentpage = HARDDISK_ID;
	EnableWindow (GetDlgItem(hDlg, IDC_NEW_HD), os_winnt && os_winnt_admin > 1 ? TRUE : FALSE);

    case WM_USER:
	CheckDlgButton (hDlg, IDC_MAPDRIVES_AUTO, workprefs.win32_automount_removable);
	CheckDlgButton (hDlg, IDC_MAPDRIVES, workprefs.win32_automount_drives);
	CheckDlgButton (hDlg, IDC_MAPDRIVES_CD, workprefs.win32_automount_cddrives);
	CheckDlgButton (hDlg, IDC_MAPDRIVES_NET, workprefs.win32_automount_netdrives);
	CheckDlgButton (hDlg, IDC_NOUAEFSDB, workprefs.filesys_no_uaefsdb);
	CheckDlgButton (hDlg, IDC_NORECYCLEBIN, workprefs.win32_norecyclebin);
	InitializeListView (hDlg);
	hilitehd ();
	break;

    case WM_MOUSEMOVE:
	if (drag_move (hDlg, lParam))
	    return TRUE;
	break;
    case WM_LBUTTONUP:
    {
	int *draggeditems, item;
	if ((item = drag_end (hDlg, cachedlist, lParam, &draggeditems)) >= 0) {
	    move_filesys_unitconfig (&workprefs, draggeditems[0], item);
	    InitializeListView (hDlg);
	    clicked_entry = item;
	    hilitehd ();
	}
	xfree (draggeditems);
	break;
    }

    case WM_COMMAND:
	switch (LOWORD(wParam))
	{
	    case 10001:
	    clicked_entry--;
	    hilitehd ();
	    break;
	    case 10002:
	    clicked_entry++;
	    hilitehd ();
	    break;
	    default:
	    harddiskdlg_button (hDlg, LOWORD (wParam));
	    InitializeListView (hDlg);
	    hilitehd ();
	    break;
	}
    break;

    case WM_NOTIFY:
	if (((LPNMHDR) lParam)->idFrom == IDC_VOLUMELIST)
	    harddiskdlg_volume_notify (hDlg, (NM_LISTVIEW *) lParam);
	return TRUE;
    default:
	return FALSE;
    }

    return FALSE;
}

#endif

static void out_floppyspeed (HWND hDlg)
{
    char txt[30];
    char tmp1[MAX_DPATH];
    char tmp2[MAX_DPATH];

    WIN32GUI_LoadUIString (IDS_FLOPPY_COMPATIBLE, tmp1, sizeof (tmp1));
    WIN32GUI_LoadUIString (IDS_FLOPPY_TURBO, tmp2, sizeof (tmp2));
    if (workprefs.floppy_speed)
	sprintf (txt, "%d%%%s", workprefs.floppy_speed, workprefs.floppy_speed == 100 ? tmp1 : "");
    else
	strcpy (txt, tmp2);
    SetDlgItemText (hDlg, IDC_FLOPPYSPDTEXT, txt);
}

#define BUTTONSPERFLOPPY 7
static int floppybuttons[][BUTTONSPERFLOPPY] = {
    { IDC_DF0TEXT,IDC_DF0,IDC_EJECT0,IDC_DF0TYPE,IDC_DF0WP,IDC_SAVEIMAGE0,IDC_DF0ENABLE },
    { IDC_DF1TEXT,IDC_DF1,IDC_EJECT1,IDC_DF1TYPE,IDC_DF1WP,IDC_SAVEIMAGE1,IDC_DF1ENABLE },
    { IDC_DF2TEXT,IDC_DF2,IDC_EJECT2,IDC_DF2TYPE,IDC_DF2WP,IDC_SAVEIMAGE2,IDC_DF2ENABLE },
    { IDC_DF3TEXT,IDC_DF3,IDC_EJECT3,IDC_DF3TYPE,IDC_DF3WP,IDC_SAVEIMAGE3,IDC_DF3ENABLE }
};
static int floppybuttonsq[][BUTTONSPERFLOPPY] = {
    { IDC_DF0TEXTQ,IDC_DF0QQ,IDC_EJECT0Q,-1,IDC_DF0WPQ,-1,IDC_DF0QENABLE },
    { IDC_DF1TEXTQ,IDC_DF1QQ,IDC_EJECT1Q,-1,IDC_DF1WPQ,-1,IDC_DF1QENABLE },
    { -1,-1,-1,-1,-1,-1,-1 },
    { -1,-1,-1,-1,-1,-1,-1 }
};

static void floppytooltip (HWND hDlg, int num, uae_u32 crc32)
{
    TOOLINFO ti;
    int id;
    char tmp[100];

    if (currentpage == QUICKSTART_ID)
	id = floppybuttonsq[num][0];
    else
	id = floppybuttons[num][0];
    ti.cbSize = sizeof(TOOLINFO);
    ti.uFlags = TTF_SUBCLASS | TTF_IDISHWND;
    ti.hwnd = hDlg;
    ti.hinst = hInst;
    ti.uId = (UINT_PTR)GetDlgItem (hDlg, id);
    SendMessage (ToolTipHWND, TTM_DELTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);
    if (crc32 == 0)
	return;
    sprintf (tmp, "CRC=%08.8X", crc32);
    ti.lpszText = tmp;
    SendMessage (ToolTipHWND, TTM_ADDTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);
}

static void addfloppyhistory_2 (HWND hDlg, UAEREG *fkey, int n, int f_text)
{
    int i, j;
    char *s;
    int nn = workprefs.dfxtype[n] + 1;

    if (f_text < 0)
	return;
    SendDlgItemMessage(hDlg, f_text, CB_RESETCONTENT, 0, 0);
    SendDlgItemMessage(hDlg, f_text, WM_SETTEXT, 0, (LPARAM)workprefs.df[n]);
    i = 0;
    while (s = DISK_history_get (i)) {
	char tmpname[MAX_DPATH], tmppath[MAX_DPATH], *p, *p2;
	if (strlen(s) == 0)
	    break;
	i++;
	strcpy (tmppath, s);
	p = tmppath + strlen(tmppath) - 1;
	for (j = 0; uae_archive_extensions[j]; j++) {
	    p2 = strstr (tmppath, uae_archive_extensions[j]);
	    if (p2) {
		p = p2;
		break;
	    }
	}
	while (p > tmppath) {
	    if (*p == '\\' || *p == '/')
		break;
	    p--;
	}
	strcpy (tmpname, p + 1);
	*++p = 0;
	if (tmppath[0]) {
	    strcat (tmpname, " { ");
	    strcat (tmpname, tmppath);
	    strcat (tmpname, " }");
	}
	if (f_text >= 0)
	    SendDlgItemMessage (hDlg, f_text, CB_ADDSTRING, 0, (LPARAM)tmpname);
	if (!strcmp (workprefs.df[n], s)) {
	    if (f_text >= 0)
		SendDlgItemMessage (hDlg, f_text, CB_SETCURSEL, i - 1, 0);
	}
	if (nn <= 0)
	    break;
    }
}

static void addfloppyhistory(HWND hDlg)
{
    UAEREG *fkey;
    int f_text, max, n;

    if (currentpage == QUICKSTART_ID)
	max = 2;
    else if (currentpage == FLOPPY_ID)
	max = 4;
    else if (currentpage == DISK_ID)
	max = 1;
    else
	return;
    fkey = read_disk_history ();
    for (n = 0; n < max; n++) {
	if (currentpage == QUICKSTART_ID)
	    f_text = floppybuttonsq[n][0];
	else if (currentpage == FLOPPY_ID)
	    f_text = floppybuttons[n][0];
	else
	    f_text = IDC_DISKTEXT;
	addfloppyhistory_2 (hDlg, fkey, n, f_text);
    }
    regclosetree (fkey);
}

static void addfloppytype (HWND hDlg, int n)
{
    int state, chk;
    int nn = workprefs.dfxtype[n] + 1;

    int f_text = floppybuttons[n][0];
    int f_drive = floppybuttons[n][1];
    int f_eject = floppybuttons[n][2];
    int f_type = floppybuttons[n][3];
    int f_wp = floppybuttons[n][4];
    int f_si = floppybuttons[n][5];
    int f_enable = floppybuttons[n][6];

    if (currentpage == QUICKSTART_ID) {
	f_text = floppybuttonsq[n][0];
	f_drive = floppybuttonsq[n][1];
	f_type = -1;
	f_eject = floppybuttonsq[n][2];
	f_wp = floppybuttonsq[n][4];
	f_si = -1;
	f_enable = floppybuttonsq[n][6];
    }

    if (nn <= 0)
	state = FALSE;
    else
	state = TRUE;
    if (f_type >= 0)
	SendDlgItemMessage (hDlg, f_type, CB_SETCURSEL, nn, 0);
    if (f_si >= 0)
	ShowWindow (GetDlgItem(hDlg, f_si), zfile_exists (DISK_get_saveimagepath (workprefs.df[n])) ? SW_SHOW : SW_HIDE);

    if (f_text >= 0)
	ew (hDlg, f_text, state);
    if (f_eject >= 0)
	ew (hDlg, f_eject, TRUE);
    if (f_drive >= 0)
	ew (hDlg, f_drive, state);
    if (f_enable >= 0) {
	if (currentpage == QUICKSTART_ID)
	    ew (hDlg, f_enable, n > 0 && workprefs.nr_floppies > 0);
	else
	    ew (hDlg, f_enable, TRUE);
	CheckDlgButton(hDlg, f_enable, state ? BST_CHECKED : BST_UNCHECKED);
    }
    chk = disk_getwriteprotect (workprefs.df[n]) && state == TRUE ? BST_CHECKED : 0;
    if (f_wp >= 0)
	CheckDlgButton(hDlg, f_wp, chk);
    chk = state && DISK_validate_filename (workprefs.df[n], 0, NULL, NULL) ? TRUE : FALSE;
    if (f_wp >= 0)
	ew (hDlg, f_wp, chk);
 }

static void getfloppytype (HWND hDlg, int n)
{
    int f_type = floppybuttons[n][3];
    LRESULT val = SendDlgItemMessage (hDlg, f_type, CB_GETCURSEL, 0, 0L);

    if (val != CB_ERR && workprefs.dfxtype[n] != val - 1) {
	workprefs.dfxtype[n] = (int)val - 1;
	addfloppytype (hDlg, n);
    }
}
static void getfloppytypeq (HWND hDlg, int n)
{
    int f_enable = currentpage == QUICKSTART_ID ? floppybuttonsq[n][6] : floppybuttons[n][6];
    int chk;

    if (f_enable <= 0 || (n == 0 && currentpage == QUICKSTART_ID))
	return;
    chk = IsDlgButtonChecked (hDlg, f_enable) ? 0 : -1;
    if (chk != workprefs.dfxtype[n]) {
	workprefs.dfxtype[n] = chk;
	addfloppytype (hDlg, n);
    }
    if (currentpage == QUICKSTART_ID) {
	if (chk == 0)
	    quickstart_floppy = 2;
	else
	    quickstart_floppy = 1;
	regsetint (NULL, "QuickStartFloppies", quickstart_floppy);
    }
}

static int getfloppybox (HWND hDlg, int f_text, char *out, int maxlen)
{
    LRESULT val;
    char *p1, *p2, *p;
    char *tmp;
    int i;

    out[0] = 0;
    val = SendDlgItemMessage (hDlg, f_text, CB_GETCURSEL, 0, 0L);
    if (val != CB_ERR)
	val = SendDlgItemMessage (hDlg, f_text, CB_GETLBTEXT, (WPARAM)val, (LPARAM)out);
    else
	SendDlgItemMessage (hDlg, f_text, WM_GETTEXT, (WPARAM)maxlen, (LPARAM)out);

    tmp = xmalloc (maxlen + 1);
    strcpy (tmp, out);
    p1 = strstr(tmp, " { ");
    p2 = strstr(tmp, " }");
    if (p1 && p2 && p2 > p1) {
	*p1 = 0;
	memset (out, 0, maxlen);
	memcpy (out, p1 + 3, p2 - p1 - 3);
	strcat (out, tmp);
    }
    xfree (tmp);
    i = 0;
    while ((p = DISK_history_get(i))) {
	if (!strcmp(p, out)) {
	    DISK_history_add (out, -1);
	    break;
	}
	i++;
    }
    return out[0] ? 1 : 0;
}

static void getfloppyname (HWND hDlg, int n)
{
    int f_text = currentpage == QUICKSTART_ID ? floppybuttonsq[n][0] : floppybuttons[n][0];
    char tmp[MAX_DPATH];

    if (getfloppybox (hDlg, f_text, tmp, sizeof (tmp))) {
	disk_insert (n, tmp);
	strcpy (workprefs.df[n], tmp);
    }
}

static void addallfloppies (HWND hDlg)
{
    int i;

    for (i = 0; i < 4; i++)
	addfloppytype (hDlg, i);
    addfloppyhistory (hDlg);
}

static void floppysetwriteprotect (HWND hDlg, int n, int protect)
{
    disk_setwriteprotect (n, workprefs.df[n], protect);
    addfloppytype (hDlg, n);
}

static void deletesaveimage (HWND hDlg, int num)
{
    char *p = DISK_get_saveimagepath (workprefs.df[num]);
    if (zfile_exists (p)) {
	DeleteFile (p);
	DISK_reinsert (num);
	addfloppytype (hDlg, num);
    }
}

static void diskselect (HWND hDlg, WPARAM wParam, struct uae_prefs *p, int drv)
{
    MultiDiskSelection (hDlg, wParam, 0, &workprefs, NULL);
    disk_insert (0, p->df[0]);
    disk_insert (1, p->df[1]);
    disk_insert (2, p->df[2]);
    disk_insert (3, p->df[3]);
    addfloppytype (hDlg, drv);
    addfloppyhistory (hDlg);
}

static INT_PTR CALLBACK FloppyDlgProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int recursive = 0;
    int i;
    static char diskname[40] = { "empty" };

    switch (msg)
    {
    case WM_INITDIALOG:
    {
	char ft35dd[20], ft35hd[20], ft525sd[20], ftdis[20], ft35ddescom[20];
	int df0texts[] = { IDC_DF0TEXT, IDC_DF1TEXT, IDC_DF2TEXT, IDC_DF3TEXT, -1 };

	WIN32GUI_LoadUIString(IDS_FLOPPYTYPE35DD, ft35dd, sizeof ft35dd);
	WIN32GUI_LoadUIString(IDS_FLOPPYTYPE35HD, ft35hd, sizeof ft35hd);
	WIN32GUI_LoadUIString(IDS_FLOPPYTYPE525SD, ft525sd, sizeof ft525sd);
	WIN32GUI_LoadUIString(IDS_FLOPPYTYPE35DDESCOM, ft35ddescom, sizeof ft35ddescom);
	WIN32GUI_LoadUIString(IDS_FLOPPYTYPEDISABLED, ftdis, sizeof ftdis);
	pages[FLOPPY_ID] = hDlg;
	if (workprefs.floppy_speed > 0 && workprefs.floppy_speed < 10)
	    workprefs.floppy_speed = 100;
	currentpage = FLOPPY_ID;
	SendDlgItemMessage (hDlg, IDC_FLOPPYSPD, TBM_SETRANGE, TRUE, MAKELONG (0, 4));
	SendDlgItemMessage (hDlg, IDC_FLOPPYSPD, TBM_SETPAGESIZE, 0, 1);
	SendDlgItemMessage (hDlg, IDC_FLOPPYTYPE, CB_RESETCONTENT, 0, 0L);
	SendDlgItemMessage (hDlg, IDC_FLOPPYTYPE, CB_ADDSTRING, 0, (LPARAM)ft35dd);
	SendDlgItemMessage (hDlg, IDC_FLOPPYTYPE, CB_ADDSTRING, 0, (LPARAM)ft35hd);
	SendDlgItemMessage (hDlg, IDC_FLOPPYTYPE, CB_ADDSTRING, 0, (LPARAM)ft525sd);
	SendDlgItemMessage (hDlg, IDC_FLOPPYTYPE, CB_SETCURSEL, 0, 0);
	for (i = 0; i < 4; i++) {
	    int f_type = floppybuttons[i][3];
	    SendDlgItemMessage (hDlg, f_type, CB_RESETCONTENT, 0, 0L);
	    SendDlgItemMessage (hDlg, f_type, CB_ADDSTRING, 0, (LPARAM)ftdis);
	    SendDlgItemMessage (hDlg, f_type, CB_ADDSTRING, 0, (LPARAM)ft35dd);
	    SendDlgItemMessage (hDlg, f_type, CB_ADDSTRING, 0, (LPARAM)ft35hd);
	    SendDlgItemMessage (hDlg, f_type, CB_ADDSTRING, 0, (LPARAM)ft525sd);
	    SendDlgItemMessage (hDlg, f_type, CB_ADDSTRING, 0, (LPARAM)ft35ddescom);
	}
	setmultiautocomplete (hDlg, df0texts);
    }
    case WM_USER:
	recursive++;
	SetDlgItemText (hDlg, IDC_DF0TEXT, workprefs.df[0]);
	SetDlgItemText (hDlg, IDC_DF1TEXT, workprefs.df[1]);
	SetDlgItemText (hDlg, IDC_DF2TEXT, workprefs.df[2]);
	SetDlgItemText (hDlg, IDC_DF3TEXT, workprefs.df[3]);
	SetDlgItemText (hDlg, IDC_DF0TEXTQ, workprefs.df[0]);
	SetDlgItemText (hDlg, IDC_DF1TEXTQ, workprefs.df[1]);
	SetDlgItemText (hDlg, IDC_CREATE_NAME, diskname);
	SendDlgItemMessage (hDlg, IDC_FLOPPYSPD, TBM_SETPOS, TRUE,
	    workprefs.floppy_speed ? exact_log2 ((workprefs.floppy_speed) / 100) + 1 : 0);
	out_floppyspeed (hDlg);
	addallfloppies (hDlg);
	recursive--;
	break;

    case WM_COMMAND:
	if (recursive > 0)
	    break;
	recursive++;
	if (HIWORD (wParam) == CBN_SELCHANGE || HIWORD (wParam) == CBN_KILLFOCUS)  {
	    switch (LOWORD (wParam))
	    {
		case IDC_DF0TEXT:
		case IDC_DF0TEXTQ:
		getfloppyname (hDlg, 0);
		addfloppytype (hDlg, 0);
		addfloppyhistory (hDlg);
		break;
		case IDC_DF1TEXT:
		case IDC_DF1TEXTQ:
		getfloppyname (hDlg, 1);
		addfloppytype (hDlg, 1);
		addfloppyhistory (hDlg);
		break;
		case IDC_DF2TEXT:
		getfloppyname (hDlg, 2);
		addfloppytype (hDlg, 2);
		addfloppyhistory (hDlg);
		break;
		case IDC_DF3TEXT:
		getfloppyname (hDlg, 3);
		addfloppytype (hDlg, 3);
		addfloppyhistory (hDlg);
		break;
		case IDC_DF0TYPE:
		getfloppytype (hDlg, 0);
		break;
		case IDC_DF1TYPE:
		getfloppytype (hDlg, 1);
		break;
		case IDC_DF2TYPE:
		getfloppytype (hDlg, 2);
		break;
		case IDC_DF3TYPE:
		getfloppytype (hDlg, 3);
		break;
	    }
	}
	switch (LOWORD (wParam))
	{
	case IDC_DF0ENABLE:
	case IDC_DF0QENABLE:
	    getfloppytypeq (hDlg, 0);
	    break;
	case IDC_DF1ENABLE:
	case IDC_DF1QENABLE:
	    getfloppytypeq (hDlg, 1);
	    break;
	case IDC_DF2ENABLE:
	    getfloppytypeq (hDlg, 2);
	    break;
	case IDC_DF3ENABLE:
	    getfloppytypeq (hDlg, 3);
	    break;
	case IDC_DF0WP:
	case IDC_DF0WPQ:
	    floppysetwriteprotect (hDlg, 0, currentpage == QUICKSTART_ID ? IsDlgButtonChecked (hDlg, IDC_DF0WPQ) : IsDlgButtonChecked (hDlg, IDC_DF0WP));
	    break;
	case IDC_DF1WP:
	case IDC_DF1WPQ:
	    floppysetwriteprotect (hDlg, 1, currentpage == QUICKSTART_ID ? IsDlgButtonChecked (hDlg, IDC_DF1WPQ) : IsDlgButtonChecked (hDlg, IDC_DF1WP));
	    break;
	case IDC_DF2WP:
	    floppysetwriteprotect (hDlg, 2, IsDlgButtonChecked (hDlg, IDC_DF2WP));
	    break;
	case IDC_DF3WP:
	    floppysetwriteprotect (hDlg, 3, IsDlgButtonChecked (hDlg, IDC_DF3WP));
	    break;
	case IDC_DF0:
	case IDC_DF0QQ:
	    diskselect (hDlg, wParam, &workprefs, 0);
	    break;
	case IDC_DF1:
	case IDC_DF1QQ:
	    diskselect (hDlg, wParam, &workprefs, 1);
	    break;
	case IDC_DF2:
	    diskselect (hDlg, wParam, &workprefs, 2);
	    break;
	case IDC_DF3:
	    diskselect (hDlg, wParam, &workprefs, 3);
	    break;
	case IDC_EJECT0:
	case IDC_EJECT0Q:
	    SetDlgItemText (hDlg, IDC_DF0TEXT, "");
	    SetDlgItemText (hDlg, IDC_DF0TEXTQ, "");
	    workprefs.df[0][0] = 0;
	    addfloppytype (hDlg, 0);
	    break;
	case IDC_EJECT1:
	case IDC_EJECT1Q:
	    SetDlgItemText (hDlg, IDC_DF1TEXT, "");
	    SetDlgItemText (hDlg, IDC_DF1TEXTQ, "");
	    workprefs.df[1][0] = 0;
	    addfloppytype (hDlg, 1);
	    break;
	case IDC_EJECT2:
	    SetDlgItemText (hDlg, IDC_DF2TEXT, "");
	    workprefs.df[2][0] = 0;
	    addfloppytype (hDlg, 2);
	    break;
	case IDC_EJECT3:
	    SetDlgItemText (hDlg, IDC_DF3TEXT, "");
	    workprefs.df[3][0] = 0;
	    addfloppytype (hDlg, 3);
	    break;
	case IDC_SAVEIMAGE0:
	    deletesaveimage (hDlg, 0);
	    break;
	case IDC_SAVEIMAGE1:
	    deletesaveimage (hDlg, 1);
	    break;
	case IDC_SAVEIMAGE2:
	    deletesaveimage (hDlg, 2);
	    break;
	case IDC_SAVEIMAGE3:
	    deletesaveimage (hDlg, 3);
	    break;
	case IDC_CREATE:
	    DiskSelection(hDlg, wParam, 1, &workprefs, 0);
	    break;
	case IDC_CREATE_RAW:
	    DiskSelection(hDlg, wParam, 1, &workprefs, 0);
	    break;
	}
	recursive--;
	break;

    case WM_HSCROLL:
	workprefs.floppy_speed = (int)SendMessage (GetDlgItem (hDlg, IDC_FLOPPYSPD), TBM_GETPOS, 0, 0);
	if (workprefs.floppy_speed > 0) {
	    workprefs.floppy_speed--;
	    workprefs.floppy_speed = 1 << workprefs.floppy_speed;
	    workprefs.floppy_speed *= 100;
	}
	out_floppyspeed (hDlg);
	break;
    }

    return FALSE;
}

static ACCEL SwapperAccel[] = {
    { FALT|FVIRTKEY, '1', 10001 }, { FALT|FVIRTKEY, '2', 10002 }, { FALT|FVIRTKEY, '3', 10003 }, { FALT|FVIRTKEY, '4', 10004 }, { FALT|FVIRTKEY, '5', 10005 },
    { FALT|FVIRTKEY, '6', 10006 }, { FALT|FVIRTKEY, '7', 10007 }, { FALT|FVIRTKEY, '8', 10008 }, { FALT|FVIRTKEY, '9', 10009 }, { FALT|FVIRTKEY, '0', 10010 },
    { FALT|FSHIFT|FVIRTKEY, '1', 10011 }, { FALT|FSHIFT|FVIRTKEY, '2', 10012 }, { FALT|FSHIFT|FVIRTKEY, '3', 10013 }, { FALT|FSHIFT|FVIRTKEY, '4', 10014 }, { FALT|FSHIFT|FVIRTKEY, '5', 10015 },
    { FALT|FSHIFT|FVIRTKEY, '6', 10016 }, { FALT|FSHIFT|FVIRTKEY, '7', 10017 }, { FALT|FSHIFT|FVIRTKEY, '8', 10018 }, { FALT|FSHIFT|FVIRTKEY, '9', 10019 }, { FALT|FSHIFT|FVIRTKEY, '0', 10020 },
    { FVIRTKEY, VK_UP, 10101 }, { FVIRTKEY, VK_DOWN, 10102 }, { FVIRTKEY, VK_RIGHT, 10104 },
    { FVIRTKEY|FSHIFT, VK_UP, IDC_UP }, { FVIRTKEY|FSHIFT, VK_DOWN, IDC_DOWN },
    { FVIRTKEY|FCONTROL, '1', 10201 }, { FVIRTKEY|FCONTROL, '2', 10202 }, { FVIRTKEY|FCONTROL, '3', 10203 }, { FVIRTKEY|FCONTROL, '4', 10204 },
    { FVIRTKEY|FCONTROL|FSHIFT, '1', 10205 }, { FVIRTKEY|FCONTROL|FSHIFT, '2', 10206 }, { FVIRTKEY|FCONTROL|FSHIFT, '3', 10207 }, { FVIRTKEY|FCONTROL|FSHIFT, '4', 10208 },
    { FVIRTKEY, VK_RETURN, 10209 }, { FVIRTKEY, VK_DELETE, IDC_DISKLISTREMOVE },
    { 0, 0, 0 }
};

static void swapperhili (HWND hDlg, int entry)
{
    ListView_SetItemState (cachedlist, entry,
	LVIS_SELECTED | LVIS_FOCUSED, LVIS_SELECTED | LVIS_FOCUSED);
    SetDlgItemText (hDlg, IDC_DISKTEXT,  workprefs.dfxlist[entry]);
}

static void addswapperfile (HWND hDlg, int entry)
{
    char path[MAX_DPATH];
    int lastentry = entry;

    if (MultiDiskSelection (hDlg, -1, 0, &changed_prefs, path)) {
	char dpath[MAX_DPATH];
	loopmulti (path, NULL);
	while (loopmulti(path, dpath) && entry < MAX_SPARE_DRIVES) {
	    strcpy (workprefs.dfxlist[entry], dpath);
	    lastentry = entry;
	    entry++;
	}
	InitializeListView (hDlg);
	swapperhili (hDlg, lastentry);
    }
}

static INT_PTR CALLBACK SwapperDlgProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int recursive = 0;
    static int entry;
    char tmp[MAX_DPATH];

    switch (msg)
    {
    case WM_INITDIALOG:
	pages[DISK_ID] = hDlg;
	currentpage = DISK_ID;
	InitializeListView(hDlg);
	addfloppyhistory (hDlg);
	entry = 0;
	swapperhili (hDlg, entry);
	setautocomplete (hDlg, IDC_DISKTEXT);
    break;
    case WM_LBUTTONUP:
    {
	int *draggeditems;
	int item = drag_end (hDlg, cachedlist, lParam, &draggeditems);
	if (item >= 0) {
	    int i, item2;
	    entry = item;
	    for (i = 0; (item2 = draggeditems[i]) >= 0 && item2 < MAX_SPARE_DRIVES; i++, item++) {
		if (item != item2) {
		    char tmp[1000];
		    strcpy (tmp, workprefs.dfxlist[item]);
		    strcpy (workprefs.dfxlist[item], workprefs.dfxlist[item2]);
		    strcpy (workprefs.dfxlist[item2], tmp);
		}
	    }
	    InitializeListView(hDlg);
	    swapperhili (hDlg, entry);
	    return TRUE;
	}
	xfree (draggeditems);
    }
    break;
    case WM_MOUSEMOVE:
    if (drag_move (hDlg, lParam))
	return TRUE;
    break;
    case WM_COMMAND:
    {
	switch (LOWORD(wParam))
	{
	    case 10001:
	    case 10002:
	    case 10003:
	    case 10004:
	    case 10005:
	    case 10006:
	    case 10007:
	    case 10008:
	    case 10009:
	    case 10010:
	    case 10011:
	    case 10012:
	    case 10013:
	    case 10014:
	    case 10015:
	    case 10016:
	    case 10017:
	    case 10018:
	    case 10019:
	    case 10020:
	    entry = LOWORD (wParam) - 10001;
	    swapperhili (hDlg, entry);
	    break;
	    case 10101:
	    if (entry > 0) {
		entry--;
		swapperhili (hDlg, entry);
	    }
	    break;
	    case 10102:
	    if (entry >= 0 && entry < MAX_SPARE_DRIVES - 1) {
		entry++;
		swapperhili (hDlg, entry);
	    }
	    break;
	    case 10103:
	    case 10104:
	    disk_swap (entry, 1);
	    InitializeListView (hDlg);
	    swapperhili (hDlg, entry);
	    break;
	    case 10201:
	    case 10202:
	    case 10203:
	    case 10204:
	    {
		int drv = LOWORD (wParam) - 10201;
		int i;
		if (workprefs.dfxtype[drv] >= 0 && entry >= 0) {
		    for (i = 0; i < 4; i++) {
			if (!strcmp (workprefs.df[i], workprefs.dfxlist[entry]))
			    workprefs.df[i][0] = 0;
		    }
		    strcpy (workprefs.df[drv], workprefs.dfxlist[entry]);
		    disk_insert (drv, workprefs.df[drv]);
		    InitializeListView (hDlg);
		    swapperhili (hDlg, entry);
		}
	    }
	    break;
	    case 10205:
	    case 10206:
	    case 10207:
	    case 10208:
	    {
		int drv = LOWORD (wParam) - 10201;
		workprefs.df[drv][0] = 0;
		InitializeListView (hDlg);
		swapperhili (hDlg, entry);
	    }
	    break;
	    case 10209:
	    {
		addswapperfile (hDlg, entry);
	    }
	    break;

	    case IDC_DISKLISTINSERT:
		if (entry >= 0) {
		    if (getfloppybox (hDlg, IDC_DISKTEXT, tmp, sizeof (tmp))) {
			strcpy (workprefs.dfxlist[entry], tmp);
			addfloppyhistory (hDlg);
			InitializeListView (hDlg);
			swapperhili (hDlg, entry);
		    } else {
			addswapperfile (hDlg, entry);
		    }
		}
		break;

	    case IDC_DISKLISTREMOVE:
		if (entry >= 0) {
		    workprefs.dfxlist[entry][0] = 0;
		    InitializeListView (hDlg);
		    swapperhili (hDlg, entry);
		}
		break;
	    case IDC_UP:
		if (entry > 0) {
		    strcpy (tmp, workprefs.dfxlist[entry - 1]);
		    strcpy (workprefs.dfxlist[entry - 1], workprefs.dfxlist[entry]);
		    strcpy (workprefs.dfxlist[entry], tmp);
		    InitializeListView (hDlg);
		    entry--;
		    swapperhili (hDlg, entry);
		}
		break;
	    case IDC_DOWN:
		if (entry >= 0 && entry < MAX_SPARE_DRIVES - 1) {
		    strcpy (tmp, workprefs.dfxlist[entry + 1]);
		    strcpy (workprefs.dfxlist[entry + 1], workprefs.dfxlist[entry]);
		    strcpy (workprefs.dfxlist[entry], tmp);
		    InitializeListView (hDlg);
		    entry++;
		    swapperhili (hDlg, entry);
		}
		break;
	}
	break;
    }
    case WM_NOTIFY:
	if (((LPNMHDR) lParam)->idFrom == IDC_DISKLIST)
	{
	    int dblclick = 0, button = 0, col;
	    HWND list;
	    NM_LISTVIEW *nmlistview;
	    nmlistview = (NM_LISTVIEW *) lParam;
	    cachedlist = list = nmlistview->hdr.hwndFrom;
	    switch (nmlistview->hdr.code)
	    {
		case LVN_BEGINDRAG:
		drag_start (hDlg, cachedlist, lParam);
		break;
		case NM_RDBLCLK:
		case NM_DBLCLK:
		dblclick = 1;
		/* fall-through here too */
		case NM_RCLICK:
		if (nmlistview->hdr.code == NM_RCLICK || nmlistview->hdr.code == NM_RDBLCLK)
		    button = 2;
		case NM_CLICK:
		entry = listview_entry_from_click (list, &col);
		if (entry >= 0) {
		    if (col == 2) {
			if (button) {
			    if (!dblclick) {
				if (disk_swap (entry, -1))
				    InitializeListView (hDlg);
				swapperhili (hDlg, entry);
			    }
			} else {
			    if (!dblclick) {
				if (disk_swap (entry, 0))
				    InitializeListView (hDlg);
				swapperhili (hDlg, entry);
			    }
			}
		    } else if (col == 1) {
			if (dblclick) {
			    if (!button) {
				addswapperfile (hDlg, entry);
			    } else {
				workprefs.dfxlist[entry][0] = 0;
				InitializeListView (hDlg);
			    }
			}
		    }
		    SetDlgItemText (hDlg, IDC_DISKTEXT,  workprefs.dfxlist[entry]);
		}
		break;
	    }
	}
    }
    return FALSE;
}

static PRINTER_INFO_1 *pInfo = NULL;
static DWORD dwEnumeratedPrinters = 0;
#define MAX_PRINTERS 10
struct serialportinfo comports[MAX_SERIAL_PORTS];
static int ghostscript_available;

static int joy0previous, joy1previous;
static BOOL bNoMidiIn = FALSE;

static void enable_for_portsdlg( HWND hDlg )
{
    int v;

    v = workprefs.input_selected_setting > 0 ? FALSE : TRUE;
    ew (hDlg, IDC_SWAP, v);
#if !defined (SERIAL_PORT)
    ew (hDlg, IDC_MIDIOUTLIST, FALSE);
    ew (hDlg, IDC_MIDIINLIST, FALSE);
    ew (hDlg, IDC_SHARED, FALSE);
    ew (hDlg, IDC_SER_CTSRTS, FALSE);
    ew (hDlg, IDC_SERIAL_DIRECT, FALSE);
    ew (hDlg, IDC_SERIAL, FALSE);
    ew (hDlg, IDC_UAESERIAL, FALSE);
#else
    v = workprefs.use_serial ? TRUE : FALSE;
    ew (hDlg, IDC_SER_SHARED, v);
    ew (hDlg, IDC_SER_CTSRTS, v);
    ew (hDlg, IDC_SER_DIRECT, v);
    ew (hDlg, IDC_UAESERIAL, full_property_sheet);
#endif
#if !defined (PARALLEL_PORT)
    ew (hDlg, IDC_PRINTERLIST, FALSE);
    ew (hDlg, IDC_FLUSHPRINTER, FALSE);
    ew (hDlg, IDC_PSPRINTER, FALSE);
    ew (hDlg, IDC_PS_PARAMS, FALSE);
    ew (hDlg, IDC_PRINTER_AUTOFLUSH, FALSE);
#else
    ew (hDlg, IDC_FLUSHPRINTER, isprinteropen () ? TRUE : FALSE);
    ew (hDlg, IDC_PSPRINTER, full_property_sheet && ghostscript_available ? TRUE : FALSE);
    ew (hDlg, IDC_PSPRINTERDETECT, full_property_sheet ? TRUE : FALSE);
    ew (hDlg, IDC_PS_PARAMS, full_property_sheet && ghostscript_available);
#endif
}

static void updatejoyport (HWND hDlg)
{
    int i, j;
    char tmp[MAX_DPATH], tmp2[MAX_DPATH];

    enable_for_portsdlg (hDlg);
    if (joy0previous < 0)
	joy0previous = inputdevice_get_device_total (IDTYPE_JOYSTICK) + 1;
    if (joy1previous < 0)
	joy1previous = JSEM_LASTKBD + 1;
    for (i = 0; i < 2; i++) {
	int total = 2;
	int idx = i == 0 ? joy0previous : joy1previous;
	int id = i == 0 ? IDC_PORT0_JOYS : IDC_PORT1_JOYS;
	int v = workprefs.jports[i].id;
	char *p1, *p2;

	SendDlgItemMessage (hDlg, id, CB_RESETCONTENT, 0, 0L);
	SendDlgItemMessage (hDlg, id, CB_ADDSTRING, 0, (LPARAM)"");
	WIN32GUI_LoadUIString (IDS_NONE, tmp, sizeof (tmp) - 3);
	sprintf (tmp2, "<%s>", tmp);
	SendDlgItemMessage (hDlg, id, CB_ADDSTRING, 0, (LPARAM)tmp2);
	WIN32GUI_LoadUIString (IDS_KEYJOY, tmp, sizeof (tmp));
	strcat (tmp, "\n");
	p1 = tmp;
	for (;;) {
	    p2 = strchr (p1, '\n');
	    if (p2 && strlen (p2) > 0) {
		*p2++ = 0;
		SendDlgItemMessage (hDlg, id, CB_ADDSTRING, 0, (LPARAM)p1);
		total++;
		p1 = p2;
	    } else
		break;
	}
	for (j = 0; j < inputdevice_get_device_total (IDTYPE_JOYSTICK); j++, total++)
	    SendDlgItemMessage (hDlg, id, CB_ADDSTRING, 0, (LPARAM)inputdevice_get_device_name(IDTYPE_JOYSTICK, j));
	for (j = 0; j < inputdevice_get_device_total (IDTYPE_MOUSE); j++, total++)
	    SendDlgItemMessage (hDlg, id, CB_ADDSTRING, 0, (LPARAM)inputdevice_get_device_name(IDTYPE_MOUSE, j));
	if (v < 0) {
	    idx = -1;
	} else if (v >= JSEM_MICE) {
	    idx = v - JSEM_MICE;
	    if (idx >= inputdevice_get_device_total (IDTYPE_MOUSE))
		idx = 0;
	    else
		idx += inputdevice_get_device_total (IDTYPE_JOYSTICK);
	    idx += JSEM_LASTKBD;
	} else if (v >= JSEM_JOYS) {
	    idx = v - JSEM_JOYS;
	    if (idx >= inputdevice_get_device_total (IDTYPE_JOYSTICK))
		idx = 0;
	    idx += JSEM_LASTKBD;
	} else {
	    idx = v - JSEM_KBDLAYOUT;
	}
	idx+=2;
	if (idx >= total)
	    idx = 0;
	SendDlgItemMessage (hDlg, id, CB_SETCURSEL, idx, 0);
    }
}

static void fixjport (struct jport *port, int v)
{
    int vv = port->id;
    if (vv != v)
	return;
    if (vv >= JSEM_JOYS && vv < JSEM_MICE) {
	vv -= JSEM_JOYS;
	vv++;
	if (vv >= inputdevice_get_device_total (IDTYPE_JOYSTICK))
	    vv = 0;
	vv += JSEM_JOYS;
    }
    if (vv >= JSEM_MICE && vv < JSEM_END) {
	vv -= JSEM_MICE;
	vv++;
	if (vv >= inputdevice_get_device_total (IDTYPE_MOUSE))
	    vv = 0;
	vv += JSEM_MICE;
    }
    if (vv >= JSEM_KBDLAYOUT && vv < JSEM_LASTKBD) {
	vv -= JSEM_KBDLAYOUT;
	vv++;
	if (vv >= JSEM_LASTKBD)
	    vv = 0;
	vv += JSEM_KBDLAYOUT;
    }
    port->id = vv;
}

static void values_from_portsdlg (HWND hDlg)
{
    int i, lastside = 0, changed = 0, v;
    char tmp[256];
    BOOL success;
    LRESULT item;

    for (i = 0; i < 2; i++) {
	int idx = 0;
	int *port = &workprefs.jports[i].id;
	int prevport = *port;
	int id = i == 0 ? IDC_PORT0_JOYS : IDC_PORT1_JOYS;
	LRESULT v = SendDlgItemMessage (hDlg, id, CB_GETCURSEL, 0, 0L);
	if (v != CB_ERR && v > 0) {
	    v-=2;
	    if (v < 0)
		*port = -1;
	    else if (v < JSEM_LASTKBD)
		*port = JSEM_KBDLAYOUT + (int)v;
	    else if (v >= JSEM_LASTKBD + inputdevice_get_device_total (IDTYPE_JOYSTICK))
		*port = JSEM_MICE + (int)v - inputdevice_get_device_total (IDTYPE_JOYSTICK) - JSEM_LASTKBD;
	    else
		*port = JSEM_JOYS + (int)v - JSEM_LASTKBD;
	}
	if (*port != prevport) {
	    lastside = i;
	    changed = 1;
	}
    }
    if (changed) {
	if (lastside)
	    fixjport (&workprefs.jports[0], workprefs.jports[1].id);
	else
	    fixjport (&workprefs.jports[1], workprefs.jports[0].id);
    }

    item = SendDlgItemMessage (hDlg, IDC_PRINTERLIST, CB_GETCURSEL, 0, 0L);
    if(item != CB_ERR) {
	int got = 0;
	strcpy (tmp, workprefs.prtname);
	if (item > 0) {
	    item--;
	    if (item < dwEnumeratedPrinters) {
		strcpy (workprefs.prtname, pInfo[item].pName);
		got = 1;
	    } else {
		int i;
		item -= dwEnumeratedPrinters;
		for (i = 0; i < 4; i++) {
		    if ((paraport_mask & (1 << i)) && item == 0) {
			sprintf (workprefs.prtname, "LPT%d", i + 1);
			got = 1;
			break;
		    }
		    item--;
		}
	    }
	}
	if (!got)
	    workprefs.prtname[0] = 0;
#ifdef PARALLEL_PORT
	if (strcmp (workprefs.prtname, tmp))
	    closeprinter ();
#endif
    }

    workprefs.win32_midioutdev = SendDlgItemMessage(hDlg, IDC_MIDIOUTLIST, CB_GETCURSEL, 0, 0);
    workprefs.win32_midioutdev -= 2;

    if( bNoMidiIn) {
	workprefs.win32_midiindev = -1;
    } else {
	workprefs.win32_midiindev = SendDlgItemMessage(hDlg, IDC_MIDIINLIST, CB_GETCURSEL, 0, 0);
    }
    ew (hDlg, IDC_MIDIINLIST, workprefs.win32_midioutdev < -1 ? FALSE : TRUE);

    item = SendDlgItemMessage (hDlg, IDC_SERIAL, CB_GETCURSEL, 0, 0L);
    if (item != CB_ERR && item > 0) {
	workprefs.use_serial = 1;
	strcpy (workprefs.sername, comports[item - 1].dev);
    } else {
	workprefs.use_serial = 0;
	workprefs.sername[0] = 0;
    }
    workprefs.serial_demand = 0;
    if (IsDlgButtonChecked (hDlg, IDC_SER_SHARED))
	workprefs.serial_demand = 1;
    workprefs.serial_hwctsrts = 0;
    if (IsDlgButtonChecked (hDlg, IDC_SER_CTSRTS))
	workprefs.serial_hwctsrts = 1;
    workprefs.serial_direct = 0;
    if (IsDlgButtonChecked (hDlg, IDC_SER_DIRECT))
	workprefs.serial_direct = 1;

    workprefs.uaeserial = 0;
    if (IsDlgButtonChecked (hDlg, IDC_UAESERIAL))
	workprefs.uaeserial = 1;

    GetDlgItemText (hDlg, IDC_PS_PARAMS, workprefs.ghostscript_parameters, sizeof workprefs.ghostscript_parameters);
    v  = GetDlgItemInt (hDlg, IDC_PRINTERAUTOFLUSH, &success, FALSE);
    if (success)
	workprefs.parallel_autoflush_time = v;

}

static void values_to_portsdlg (HWND hDlg)
{
    LRESULT result = 0;

    if(workprefs.prtname[0]) {
	int i, got = 1;
	char tmp[10];
	result = SendDlgItemMessage(hDlg, IDC_PRINTERLIST, CB_FINDSTRINGEXACT, -1, (LPARAM)workprefs.prtname);
	for (i = 0; i < 4; i++) {
	    sprintf (tmp, "LPT%d", i + 1);
	    if (!strcmp (tmp, workprefs.prtname)) {
		got = 0;
		if (paraport_mask & (1 << i))
		    got = 1;
		break;
	    }
	}
	if(result < 0 || got == 0) {
	    // Warn the user that their printer-port selection is not valid on this machine
	    char szMessage[MAX_DPATH];
	    WIN32GUI_LoadUIString(IDS_INVALIDPRTPORT, szMessage, MAX_DPATH);
	    pre_gui_message (szMessage);
	    // Disable the invalid parallel-port selection
	    workprefs.prtname[0] = 0;
	    result = 0;
	}
    }
    SetDlgItemInt(hDlg, IDC_PRINTERAUTOFLUSH, workprefs.parallel_autoflush_time, FALSE);
    CheckDlgButton(hDlg, IDC_PSPRINTER, workprefs.parallel_postscript_emulation);
    CheckDlgButton(hDlg, IDC_PSPRINTERDETECT, workprefs.parallel_postscript_detection);
    SetDlgItemText(hDlg, IDC_PS_PARAMS, workprefs.ghostscript_parameters);

    SendDlgItemMessage(hDlg, IDC_PRINTERLIST, CB_SETCURSEL, result, 0);
    SendDlgItemMessage(hDlg, IDC_MIDIOUTLIST, CB_SETCURSEL, workprefs.win32_midioutdev + 2, 0);
    if (!bNoMidiIn && workprefs.win32_midiindev >= 0)
	SendDlgItemMessage(hDlg, IDC_MIDIINLIST, CB_SETCURSEL, workprefs.win32_midiindev, 0);
    else
	SendDlgItemMessage(hDlg, IDC_MIDIINLIST, CB_SETCURSEL, 0, 0);
    ew (hDlg, IDC_MIDIINLIST, workprefs.win32_midioutdev < -1 ? FALSE : TRUE);

    CheckDlgButton(hDlg, IDC_UAESERIAL, workprefs.uaeserial);
    CheckDlgButton(hDlg, IDC_SER_SHARED, workprefs.serial_demand);
    CheckDlgButton(hDlg, IDC_SER_CTSRTS, workprefs.serial_hwctsrts);
    CheckDlgButton(hDlg, IDC_SER_DIRECT, workprefs.serial_direct);

    if(!workprefs.sername[0])  {
	SendDlgItemMessage (hDlg, IDC_SERIAL, CB_SETCURSEL, 0, 0L);
	workprefs.use_serial = 0;
    } else {
	int i;
	LRESULT result = -1;
	for (i = 0; i < MAX_SERIAL_PORTS && comports[i].name; i++) {
	    if (!strcmp (comports[i].dev, workprefs.sername)) {
		result = SendDlgItemMessage(hDlg, IDC_SERIAL, CB_SETCURSEL, i + 1, 0L);
		break;
	    }
	}
	if(result < 0 && workprefs.sername[0]) {
	    // Warn the user that their COM-port selection is not valid on this machine
	    char szMessage[MAX_DPATH];
	    WIN32GUI_LoadUIString(IDS_INVALIDCOMPORT, szMessage, MAX_DPATH);
	    pre_gui_message (szMessage);
	    // Select "none" as the COM-port
	    SendDlgItemMessage(hDlg, IDC_SERIAL, CB_SETCURSEL, 0L, 0L);
	    // Disable the chosen serial-port selection
	    workprefs.sername[0] = 0;
	    workprefs.use_serial = 0;
	} else {
	    workprefs.use_serial = 1;
	}
    }
}

static void init_portsdlg( HWND hDlg )
{
    static int first;
    int port, numdevs;
    MIDIOUTCAPS midiOutCaps;
    MIDIINCAPS midiInCaps;

    if (!first) {
	first = 1;
	if (load_ghostscript () > 0) {
	    unload_ghostscript ();
	    ghostscript_available = 1;
	}
    }
    if (!ghostscript_available) {
	workprefs.parallel_postscript_emulation = 0;
    }

    joy0previous = joy1previous = -1;

    SendDlgItemMessage (hDlg, IDC_SERIAL, CB_RESETCONTENT, 0, 0L);
    SendDlgItemMessage (hDlg, IDC_SERIAL, CB_ADDSTRING, 0, (LPARAM)szNone);
    for (port = 0; port < MAX_SERIAL_PORTS && comports[port].name; port++) {
	SendDlgItemMessage(hDlg, IDC_SERIAL, CB_ADDSTRING, 0, (LPARAM)comports[port].name);
    }

    SendDlgItemMessage (hDlg, IDC_PRINTERLIST, CB_RESETCONTENT, 0, 0L);
    SendDlgItemMessage (hDlg, IDC_PRINTERLIST, CB_ADDSTRING, 0, (LPARAM)szNone);
    if(!pInfo) {
	int flags = PRINTER_ENUM_LOCAL | (os_winnt ? PRINTER_ENUM_CONNECTIONS : 0);
	DWORD needed = 0;
	EnumPrinters(flags, NULL, 1, (LPBYTE)pInfo, 0, &needed, &dwEnumeratedPrinters);
	if (needed > 0) {
	    DWORD size = needed;
	    pInfo = calloc(1, size);
	    dwEnumeratedPrinters = 0;
	    EnumPrinters(flags, NULL, 1, (LPBYTE)pInfo, size, &needed, &dwEnumeratedPrinters);
	}
	if (dwEnumeratedPrinters == 0) {
	    free (pInfo);
	    pInfo = 0;
	}
    }
    if (pInfo) {
	for(port = 0; port < (int)dwEnumeratedPrinters; port++)
	    SendDlgItemMessage(hDlg, IDC_PRINTERLIST, CB_ADDSTRING, 0, (LPARAM)pInfo[port].pName);
    } else {
	ew (hDlg, IDC_PRINTERLIST, FALSE);
    }
    if (paraport_mask) {
	int mask = paraport_mask;
	int i = 1;
	while (mask) {
	    if (mask & 1) {
		char tmp[30];
		sprintf (tmp, "LPT%d", i);
		SendDlgItemMessage (hDlg, IDC_PRINTERLIST, CB_ADDSTRING, 0, (LPARAM)tmp);
	    }
	    i++;
	    mask >>= 1;
	}
    }

    SendDlgItemMessage (hDlg, IDC_MIDIOUTLIST, CB_RESETCONTENT, 0, 0L);
    SendDlgItemMessage (hDlg, IDC_MIDIOUTLIST, CB_ADDSTRING, 0, (LPARAM)szNone);
    if((numdevs = midiOutGetNumDevs()) == 0) {
	ew (hDlg, IDC_MIDIOUTLIST, FALSE);
    } else {
	char szMidiOut[MAX_DPATH];
	WIN32GUI_LoadUIString(IDS_DEFAULTMIDIOUT, szMidiOut, MAX_DPATH);
	SendDlgItemMessage(hDlg, IDC_MIDIOUTLIST, CB_ADDSTRING, 0, (LPARAM)szMidiOut);

	for(port = 0; port < numdevs; port++) {
	    if(midiOutGetDevCaps(port, &midiOutCaps, sizeof(midiOutCaps)) == MMSYSERR_NOERROR)
		SendDlgItemMessage(hDlg, IDC_MIDIOUTLIST, CB_ADDSTRING, 0, (LPARAM)midiOutCaps.szPname);
	}
	ew (hDlg, IDC_MIDIOUTLIST, TRUE);
    }

    SendDlgItemMessage(hDlg, IDC_MIDIINLIST, CB_RESETCONTENT, 0, 0L);
    if((numdevs = midiInGetNumDevs()) == 0) {
	ew (hDlg, IDC_MIDIINLIST, FALSE);
	bNoMidiIn = TRUE;
    } else {
	for(port = 0; port < numdevs; port++) {
	    if(midiInGetDevCaps(port, &midiInCaps, sizeof(midiInCaps)) == MMSYSERR_NOERROR)
		SendDlgItemMessage(hDlg, IDC_MIDIINLIST, CB_ADDSTRING, 0, (LPARAM)midiInCaps.szPname);
	}
    }
}

/* Handle messages for the Joystick Settings page of our property-sheet */
static INT_PTR CALLBACK PortsDlgProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int recursive = 0;
    int temp;

    switch (msg)
    {
    case WM_INITDIALOG:
	recursive++;
	pages[PORTS_ID] = hDlg;
	currentpage = PORTS_ID;
	init_portsdlg(hDlg);
	enable_for_portsdlg(hDlg);
	values_to_portsdlg(hDlg);
	updatejoyport(hDlg);
	recursive--;
	break;
    case WM_USER:
	recursive++;
	enable_for_portsdlg (hDlg);
	updatejoyport (hDlg);
	recursive--;
	return TRUE;

    case WM_COMMAND:
	if (recursive > 0)
	    break;
	recursive++;
	if (wParam == IDC_SWAP) {
	    struct jport tmp;
	    memcpy (&tmp, &workprefs.jports[0], sizeof (struct jport));
	    memcpy (&workprefs.jports[0], &workprefs.jports[1], sizeof (struct jport));
	    memcpy (&workprefs.jports[1], &tmp, sizeof (struct jport));
	    temp = joy0previous;
	    joy0previous = joy1previous;
	    joy1previous = temp;
	    updatejoyport (hDlg);
	} else if (wParam == IDC_FLUSHPRINTER) {
	    if (isprinter ()) {
		closeprinter ();
	    }
	} else if (wParam == IDC_PSPRINTER) {
	    workprefs.parallel_postscript_emulation = IsDlgButtonChecked (hDlg, IDC_PSPRINTER) ? 1 : 0;
	    if (workprefs.parallel_postscript_emulation)
		CheckDlgButton (hDlg, IDC_PSPRINTERDETECT, 1);
	} else if (wParam == IDC_PSPRINTERDETECT) {
	    workprefs.parallel_postscript_detection = IsDlgButtonChecked (hDlg, IDC_PSPRINTERDETECT) ? 1 : 0;
	    if (!workprefs.parallel_postscript_detection)
		CheckDlgButton (hDlg, IDC_PSPRINTER, 0);
	} else   {
	    values_from_portsdlg (hDlg);
	    if (HIWORD (wParam) == CBN_SELCHANGE || HIWORD (wParam) == CBN_KILLFOCUS)
		updatejoyport (hDlg);
	}
	inputdevice_updateconfig (&workprefs);
	inputdevice_config_change ();
	recursive--;
	break;
    }
    return FALSE;
}

static char *eventnames[INPUTEVENT_END];

static void values_to_inputdlg (HWND hDlg)
{
    SendDlgItemMessage(hDlg, IDC_INPUTTYPE, CB_SETCURSEL, workprefs.input_selected_setting, 0);
    SendDlgItemMessage(hDlg, IDC_INPUTDEVICE, CB_SETCURSEL, input_selected_device, 0);
    SetDlgItemInt(hDlg, IDC_INPUTDEADZONE, workprefs.input_joystick_deadzone, FALSE);
    SetDlgItemInt(hDlg, IDC_INPUTAUTOFIRERATE, workprefs.input_autofire_framecnt, FALSE);
    SetDlgItemInt(hDlg, IDC_INPUTSPEEDD, workprefs.input_joymouse_speed, FALSE);
    SetDlgItemInt(hDlg, IDC_INPUTSPEEDA, workprefs.input_joymouse_multiplier, FALSE);
    SetDlgItemInt(hDlg, IDC_INPUTSPEEDM, workprefs.input_mouse_speed, FALSE);
    CheckDlgButton (hDlg, IDC_INPUTDEVICEDISABLE, (!input_total_devices || inputdevice_get_device_status (input_selected_device)) ? BST_CHECKED : BST_UNCHECKED);
}

static int stringboxdialogactive;
static INT_PTR CALLBACK StringBoxDialogProc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch(msg)
    {
	case WM_DESTROY:
	PostQuitMessage (0);
	return TRUE;
	case WM_CLOSE:
	DestroyWindow(hDlg);
	return TRUE;
	case WM_INITDIALOG:
	    return TRUE;
	case WM_COMMAND:
	    switch (LOWORD(wParam))
	    {
		case IDOK:
		    stringboxdialogactive = -1;
		    DestroyWindow (hDlg);
		return TRUE;
		case IDCANCEL:
		    stringboxdialogactive = 0;
		    DestroyWindow (hDlg);
		return TRUE;
	    }
	break;
    }
    return DefWindowProc(hDlg, msg, wParam, lParam);
}

static int askinputcustom(HWND hDlg, char *custom, int maxlen)
{
    HWND hwnd;
    char txt[MAX_DPATH];

    stringboxdialogactive = 1;
    hwnd = CustomCreateDialog (IDD_STRINGBOX, hDlg, StringBoxDialogProc);
    if (hwnd == NULL)
	return 0;
    SendMessage(GetDlgItem(hwnd, IDC_STRINGBOXEDIT), WM_SETTEXT, 0, (LPARAM)custom);
    SetFocus(GetDlgItem(hwnd, IDC_STRINGBOXEDIT));
    while (stringboxdialogactive == 1) {
	MSG msg;
	int ret;
	WaitMessage();
	while ((ret = GetMessage (&msg, NULL, 0, 0))) {
	    if (ret == -1)
		break;
	    if (!IsWindow(hwnd) || !IsDialogMessage(hwnd, &msg)) {
		TranslateMessage (&msg);
		DispatchMessage (&msg);
	    }
	    SendMessage(GetDlgItem(hwnd, IDC_STRINGBOXEDIT), WM_GETTEXT, sizeof txt, (LPARAM)txt);
	}
	if (stringboxdialogactive == -1) {
	    strcpy (custom, txt);
	    return 1;
	}
    }
    return 0;
}

static void init_inputdlg_2(HWND hDlg)
{
    char name1[256], name2[256];
    char custom1[MAX_DPATH], tmp1[MAX_DPATH];
    int cnt, index, af, aftmp;

    if (input_selected_widget < 0 || workprefs.input_selected_setting == 0) {
	EnableWindow(GetDlgItem(hDlg, IDC_INPUTAMIGA), FALSE);
    } else {
	EnableWindow(GetDlgItem(hDlg, IDC_INPUTAMIGA), TRUE);
    }
    SendDlgItemMessage (hDlg, IDC_INPUTAMIGA, CB_RESETCONTENT, 0, 0L);
    SendDlgItemMessage (hDlg, IDC_INPUTAMIGA, CB_ADDSTRING, 0, (LPARAM)szNone);
    WIN32GUI_LoadUIString (IDS_INPUT_CUSTOMEVENT, tmp1, MAX_DPATH);
    SendDlgItemMessage (hDlg, IDC_INPUTAMIGA, CB_ADDSTRING, 0, (LPARAM)tmp1);
    index = 0; af = 0;
    if (input_selected_widget >= 0) {
	inputdevice_get_mapped_name (input_selected_device, input_selected_widget, 0, name1, custom1, input_selected_sub_num);
	cnt = 2;
	while(inputdevice_iterate (input_selected_device, input_selected_widget, name2, &aftmp)) {
	    free (eventnames[cnt]);
	    eventnames[cnt] = strdup (name2);
	    if (name1 && !strcmp (name1, name2)) {
		index = cnt;
		af = aftmp;
	    }
	    cnt++;
	    SendDlgItemMessage (hDlg, IDC_INPUTAMIGA, CB_ADDSTRING, 0, (LPARAM)name2);
	}
	if (strlen(custom1) > 0)
	    index = 1;
	if (index >= 0) {
	    SendDlgItemMessage(hDlg, IDC_INPUTAMIGA, CB_SETCURSEL, index, 0);
	    SendDlgItemMessage(hDlg, IDC_INPUTAMIGACNT, CB_SETCURSEL, input_selected_sub_num, 0);
	}
    }
}

static void init_inputdlg( HWND hDlg )
{
    int i;
    DWORD size = sizeof(COMMCONFIG);
    char buf[100], txt[100];

    SendDlgItemMessage (hDlg, IDC_INPUTTYPE, CB_RESETCONTENT, 0, 0L);
    WIN32GUI_LoadUIString (IDS_INPUT_COMPATIBILITY, buf, sizeof (buf));
    SendDlgItemMessage (hDlg, IDC_INPUTTYPE, CB_ADDSTRING, 0, (LPARAM)buf);
    WIN32GUI_LoadUIString (IDS_INPUT_CUSTOM, buf, sizeof (buf));
    for (i = 0; i < 4; i++) {
	sprintf (txt, buf, i + 1);
	SendDlgItemMessage (hDlg, IDC_INPUTTYPE, CB_ADDSTRING, 0, (LPARAM)txt);
    }

    SendDlgItemMessage (hDlg, IDC_INPUTCOPYFROM, CB_RESETCONTENT, 0, 0L);
    WIN32GUI_LoadUIString (IDS_INPUT_COPY_DEFAULT, buf, sizeof (buf));
    SendDlgItemMessage (hDlg, IDC_INPUTCOPYFROM, CB_ADDSTRING, 0, (LPARAM)buf);
    WIN32GUI_LoadUIString (IDS_INPUT_COPY_CUSTOM, buf, sizeof (buf));
    for (i = 0; i < 4; i++) {
	sprintf (txt, buf, i + 1);
	SendDlgItemMessage (hDlg, IDC_INPUTCOPYFROM, CB_ADDSTRING, 0, (LPARAM)txt);
    }

    SendDlgItemMessage (hDlg, IDC_INPUTAMIGACNT, CB_RESETCONTENT, 0, 0L);
    for (i = 0; i < MAX_INPUT_SUB_EVENT; i++) {
	sprintf (buf, "%d", i + 1);
	SendDlgItemMessage (hDlg, IDC_INPUTAMIGACNT, CB_ADDSTRING, 0, (LPARAM)buf);
    }
    SendDlgItemMessage (hDlg, IDC_INPUTAMIGACNT, CB_SETCURSEL, input_selected_sub_num, 0);

    SendDlgItemMessage (hDlg, IDC_INPUTDEVICE, CB_RESETCONTENT, 0, 0L);
    for (i = 0; i < inputdevice_get_device_total (IDTYPE_JOYSTICK); i++) {
	SendDlgItemMessage (hDlg, IDC_INPUTDEVICE, CB_ADDSTRING, 0, (LPARAM)inputdevice_get_device_name(IDTYPE_JOYSTICK, i));
    }
    for (i = 0; i < inputdevice_get_device_total (IDTYPE_MOUSE); i++) {
	SendDlgItemMessage (hDlg, IDC_INPUTDEVICE, CB_ADDSTRING, 0, (LPARAM)inputdevice_get_device_name(IDTYPE_MOUSE, i));
    }
    for (i = 0; i < inputdevice_get_device_total (IDTYPE_KEYBOARD); i++) {
	SendDlgItemMessage (hDlg, IDC_INPUTDEVICE, CB_ADDSTRING, 0, (LPARAM)inputdevice_get_device_name(IDTYPE_KEYBOARD, i));
    }
    input_total_devices = inputdevice_get_device_total (IDTYPE_JOYSTICK) +
	inputdevice_get_device_total (IDTYPE_MOUSE) +
	inputdevice_get_device_total (IDTYPE_KEYBOARD);
    if (input_selected_device >= input_total_devices)
	input_selected_device = 0;
    InitializeListView(hDlg);
    init_inputdlg_2 (hDlg);
    values_to_inputdlg (hDlg);
}

static void enable_for_inputdlg (HWND hDlg)
{
    int v = workprefs.input_selected_setting == 0 ? FALSE : TRUE;
    ew (hDlg, IDC_INPUTLIST, TRUE);
    ew (hDlg, IDC_INPUTAMIGA, v);
    ew (hDlg, IDC_INPUTAMIGACNT, TRUE);
    ew (hDlg, IDC_INPUTDEADZONE, TRUE);
    ew (hDlg, IDC_INPUTAUTOFIRERATE, v);
    ew (hDlg, IDC_INPUTSPEEDA, v);
    ew (hDlg, IDC_INPUTSPEEDD, v);
    ew (hDlg, IDC_INPUTSPEEDM, TRUE);
    ew (hDlg, IDC_INPUTCOPY, v);
    ew (hDlg, IDC_INPUTCOPYFROM, v);
    ew (hDlg, IDC_INPUTSWAP, v);
    ew (hDlg, IDC_INPUTDEVICEDISABLE, v);
}

static void clearinputlistview (HWND hDlg)
{
    ListView_DeleteAllItems(GetDlgItem(hDlg, IDC_INPUTLIST));
}

static void doinputcustom(HWND hDlg, int newcustom)
{
    char custom1[MAX_DPATH];
    int flags;
    custom1[0] = 0;
    inputdevice_get_mapped_name (input_selected_device, input_selected_widget,
	&flags, 0, custom1, input_selected_sub_num);
    if (strlen(custom1) > 0 || newcustom) {
	if (askinputcustom(hDlg, custom1, sizeof custom1)) {
	    inputdevice_set_mapping (input_selected_device, input_selected_widget,
		0, custom1, (flags & IDEV_MAPPED_AUTOFIRE_SET) ? 1 : 0, input_selected_sub_num);
	}
    }
}

static void values_from_inputdlg (HWND hDlg)
{
    int doselect = 0, v;
    BOOL success;
    LRESULT item;

    v  = GetDlgItemInt(hDlg, IDC_INPUTDEADZONE, &success, FALSE);
    if (success) {
	currprefs.input_joystick_deadzone = workprefs.input_joystick_deadzone = v;
	currprefs.input_joystick_deadzone = workprefs.input_joymouse_deadzone = v;
    }
    v  = GetDlgItemInt(hDlg, IDC_INPUTAUTOFIRERATE, &success, FALSE);
    if (success)
	currprefs.input_autofire_framecnt = workprefs.input_autofire_framecnt = v;
    v  = GetDlgItemInt(hDlg, IDC_INPUTSPEEDD, &success, FALSE);
    if (success)
	currprefs.input_joymouse_speed = workprefs.input_joymouse_speed = v;
    v  = GetDlgItemInt(hDlg, IDC_INPUTSPEEDA, &success, FALSE);
    if (success)
	currprefs.input_joymouse_multiplier = workprefs.input_joymouse_multiplier = v;
    v  = GetDlgItemInt(hDlg, IDC_INPUTSPEEDM, &success, FALSE);
    if (success)
	currprefs.input_mouse_speed = workprefs.input_mouse_speed = v;

    item = SendDlgItemMessage(hDlg, IDC_INPUTAMIGACNT, CB_GETCURSEL, 0, 0L);
    if (item != CB_ERR && input_selected_sub_num != item) {
	input_selected_sub_num = (int)item;
	doselect = 0;
	init_inputdlg_2 (hDlg);
	update_listview_input (hDlg);
	return;
    }

    item = SendDlgItemMessage(hDlg, IDC_INPUTTYPE, CB_GETCURSEL, 0, 0L);
    if(item != CB_ERR) {
	if (item != workprefs.input_selected_setting) {
	    workprefs.input_selected_setting = (int)item;
	    input_selected_widget = -1;
	    inputdevice_updateconfig (&workprefs);
	    enable_for_inputdlg( hDlg );
	    InitializeListView (hDlg);
	    doselect = 1;
	}
    }
    item = SendDlgItemMessage(hDlg, IDC_INPUTDEVICE, CB_GETCURSEL, 0, 0L);
    if(item != CB_ERR) {
	if (item != input_selected_device) {
	    input_selected_device = (int)item;
	    input_selected_widget = -1;
	    input_selected_event = -1;
	    InitializeListView (hDlg);
	    init_inputdlg_2 (hDlg);
	    values_to_inputdlg (hDlg);
	    doselect = 1;
	}
    }
    item = SendDlgItemMessage(hDlg, IDC_INPUTAMIGA, CB_GETCURSEL, 0, 0L);
    if(item != CB_ERR) {
	if (item == 1) {
	    if (item != input_selected_event)
		doinputcustom(hDlg, 1);
	}
	input_selected_event = (int)item;
	doselect = 1;
    }

    if (doselect && input_selected_device >= 0 && input_selected_event >= 0) {
	int flags;
	char custom[MAX_DPATH];
	custom[0] = 0;
	inputdevice_get_mapped_name (input_selected_device, input_selected_widget,
	    &flags, 0, custom, input_selected_sub_num);
	if (input_selected_event != 1)
	    custom[0] = 0;
	inputdevice_set_mapping (input_selected_device, input_selected_widget,
	    eventnames[input_selected_event], strlen(custom) == 0 ? NULL : custom,
	    (flags & IDEV_MAPPED_AUTOFIRE_SET) ? 1 : 0, input_selected_sub_num);
	update_listview_input (hDlg);
	inputdevice_updateconfig (&workprefs);
    }
}

static void input_swap (HWND hDlg)
{
    inputdevice_swap_ports (&workprefs, input_selected_device);
    init_inputdlg (hDlg);
}

static void input_copy (HWND hDlg)
{
    int dst = workprefs.input_selected_setting;
    LRESULT src = SendDlgItemMessage(hDlg, IDC_INPUTCOPYFROM, CB_GETCURSEL, 0, 0L);
    if (src == CB_ERR)
	return;
    inputdevice_copy_single_config (&workprefs, (int)src, workprefs.input_selected_setting, input_selected_device);
    init_inputdlg (hDlg);
}

static void input_toggleautofire (void)
{
    int af, flags, evt;
    char name[256];
    char custom[MAX_DPATH];
    if (input_selected_device < 0 || input_selected_widget < 0)
	return;
    evt = inputdevice_get_mapped_name (input_selected_device, input_selected_widget,
	&flags, name, custom, input_selected_sub_num);
    if (evt <= 0)
	return;
    af = (flags & IDEV_MAPPED_AUTOFIRE_SET) ? 0 : 1;
    inputdevice_set_mapping (input_selected_device, input_selected_widget,
	name, custom, af, input_selected_sub_num);
}

static INT_PTR CALLBACK InputDlgProc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    char name_buf[MAX_DPATH] = "", desc_buf[128] = "";
    char *posn = NULL;
    HWND list;
    int dblclick = 0;
    NM_LISTVIEW *nmlistview;
    int items = 0, entry = 0;
    static int recursive;

    switch (msg)
    {
    case WM_INITDIALOG:
	pages[INPUT_ID] = hDlg;
	currentpage = INPUT_ID;
	inputdevice_updateconfig (&workprefs);
	inputdevice_config_change ();
	input_selected_widget = -1;
	init_inputdlg(hDlg);

    case WM_USER:
	recursive++;
	enable_for_inputdlg (hDlg);
	values_to_inputdlg (hDlg);
	recursive--;
	return TRUE;
    case WM_COMMAND:
	if (recursive)
	    break;
	recursive++;
	switch (wParam)
	{
	    case IDC_INPUTCOPY:
	    input_copy (hDlg);
	    break;
	    case IDC_INPUTSWAP:
	    input_swap (hDlg);
	    break;
	    case IDC_INPUTDEVICEDISABLE:
	    inputdevice_set_device_status (input_selected_device, IsDlgButtonChecked( hDlg, IDC_INPUTDEVICEDISABLE) ? 1 : 0);
	    break;
	    default:
	    values_from_inputdlg (hDlg);
	    break;
	}
	enable_for_portsdlg (hDlg);
	inputdevice_config_change ();
	recursive--;
	break;
    case WM_NOTIFY:
	if (((LPNMHDR) lParam)->idFrom == IDC_INPUTLIST)
	{
	    int row;
	    nmlistview = (NM_LISTVIEW *) lParam;
	    list = nmlistview->hdr.hwndFrom;
	    switch (nmlistview->hdr.code)
	    {
		case NM_DBLCLK:
		dblclick = 1;
		/* fall-through */
		case NM_CLICK:
		entry = listview_entry_from_click (list, &row);
		if (entry >= 0) {
		    int oldentry = input_selected_widget;
		    input_selected_widget = entry;
		    if (row == 2 && entry == oldentry)
			input_toggleautofire ();
		    if (row == 3) {
			input_selected_sub_num++;
			if (input_selected_sub_num >= MAX_INPUT_SUB_EVENT)
			    input_selected_sub_num = 0;
			SendDlgItemMessage(hDlg, IDC_INPUTAMIGACNT, CB_SETCURSEL, input_selected_sub_num, 0);
		    }
		} else {
		    input_selected_widget = -1;
		}
		if (dblclick)
		    doinputcustom(hDlg, 0);
		update_listview_input (hDlg);
		init_inputdlg_2 (hDlg);
	    }
	}
    }
    return FALSE;
}

#ifdef GFXFILTER

static int scanlineratios[] = { 1,1,1,2,1,3, 2,1,2,2,2,3, 3,1,3,2,3,3, 0,0 };
static int scanlineindexes[100];
static int filterpreset_selected = -1, filterpreset_builtin = -1;

static void enable_for_hw3ddlg (HWND hDlg)
{
    int v = workprefs.gfx_filter ? TRUE : FALSE;
    int vv = FALSE, vv2 = FALSE, vv3 = FALSE, vv4 = FALSE;
    struct uae_filter *uf;
    int i;

    uf = &uaefilters[0];
    i = 0;
    while (uaefilters[i].name) {
	if (workprefs.gfx_filter == uaefilters[i].type) {
	    uf = &uaefilters[i];
	    break;
	}
	i++;
    }
    if (v && (uf->x[0] || uf->x[1] || uf->x[2] || uf->x[3] || uf->x[4]))
	vv = TRUE;
    if (v && (uf->x[0] || uf->yuv))
	vv2 = TRUE;
    if (v && (uf->x[0] && !uf->yuv))
	vv3 = TRUE;
    if (v && uf->x[0])
	vv4 = TRUE;
    ew (hDlg, IDC_FILTERENABLE, TRUE);
    ew (hDlg, IDC_FILTERMODE, v);
    CheckDlgButton(hDlg, IDC_FILTERENABLE, v);
    ew (hDlg, IDC_FILTERHZ, v);
    ew (hDlg, IDC_FILTERVZ, v);
    ew (hDlg, IDC_FILTERHZMULT, vv && !vv4);
    ew (hDlg, IDC_FILTERVZMULT, vv && !vv4);
    ew (hDlg, IDC_FILTERHO, v);
    ew (hDlg, IDC_FILTERVO, v);
    ew (hDlg, IDC_FILTERSLR, vv3);
    ew (hDlg, IDC_FILTERXL, vv2);
    ew (hDlg, IDC_FILTERXLV, vv2);
    ew (hDlg, IDC_FILTERXTRA, vv2);
    ew (hDlg, IDC_FILTERDEFAULT, v);
    ew (hDlg, IDC_FILTERFILTER, vv);
    ew (hDlg, IDC_FILTERUPSCALE, vv && !vv2);
    ew (hDlg, IDC_FILTERAUTORES, vv && !vv2);

    ew (hDlg, IDC_FILTERPRESETSAVE, filterpreset_builtin < 0);
    ew (hDlg, IDC_FILTERPRESETLOAD, filterpreset_selected > 0);
    ew (hDlg, IDC_FILTERPRESETDELETE, filterpreset_selected > 0 && filterpreset_builtin < 0);
    CheckDlgButton(hDlg, IDC_FILTERUPSCALE, workprefs.gfx_filter_upscale);
}

static void makefilter(char *s, int x, int flags)
{
    sprintf (s, "%dx", x);
    if ((flags & (UAE_FILTER_MODE_16_16 | UAE_FILTER_MODE_32_32)) == (UAE_FILTER_MODE_16_16 | UAE_FILTER_MODE_32_32)) {
	strcat (s, " (16bit/32bit)");
	return;
    }
    if (flags & UAE_FILTER_MODE_16)
	strcat (s, " (16bit)");
    if (flags & UAE_FILTER_MODE_32)
	strcat (s, " (32bit)");
}

static char *filtermultnames[] = { "1x", "2x", "4x", "6x", "8x", NULL };
static int filtermults[] = { 1000, 500, 250, 167, 125 };
struct filterxtra {
    char *label;
    int *varw, *varc;
    int min, max, step;
};
static struct filterxtra *filter_extra, *filter_selected;

static struct filterxtra filter_pal_extra[] =
{
    "Brightness", &workprefs.gfx_filter_luminance, &currprefs.gfx_filter_luminance, -1000, 1000, 10,
    "Contrast", &workprefs.gfx_filter_contrast, &currprefs.gfx_filter_contrast, -1000, 1000, 10,
    "Saturation", &workprefs.gfx_filter_saturation, &currprefs.gfx_filter_saturation, -1000, 1000, 10,
    "Gamma", &workprefs.gfx_gamma, &currprefs.gfx_gamma, -1000, 1000, 10,
    "Scanlines", &workprefs.gfx_filter_scanlines, &currprefs.gfx_filter_scanlines, 0, 100, 1,
    "Blurriness", &workprefs.gfx_filter_blur, &currprefs.gfx_filter_blur,0, 2000, 10,
    "Noise", &workprefs.gfx_filter_noise, &currprefs.gfx_filter_noise,0, 100, 10,
    NULL
};
static struct filterxtra filter_3d_extra[] =
{
    "Scanline transparency", &workprefs.gfx_filter_scanlines, &currprefs.gfx_filter_scanlines, 0, 100, 10,
    "Scanline level", &workprefs.gfx_filter_scanlinelevel, &currprefs.gfx_filter_scanlinelevel, 0, 100, 10,
    NULL
};
static int *filtervars[] = {
	&workprefs.gfx_filter, &workprefs.gfx_filter_filtermode,
	&workprefs.gfx_filter_vert_zoom, &workprefs.gfx_filter_horiz_zoom,
	&workprefs.gfx_filter_vert_zoom_mult, &workprefs.gfx_filter_horiz_zoom_mult,
	&workprefs.gfx_filter_vert_offset, &workprefs.gfx_filter_horiz_offset,
	&workprefs.gfx_filter_scanlines, &workprefs.gfx_filter_scanlinelevel, &workprefs.gfx_filter_scanlineratio,
	&workprefs.gfx_resolution, &workprefs.gfx_linedbl, &workprefs.gfx_correct_aspect,
	&workprefs.gfx_xcenter, &workprefs.gfx_ycenter,
	&workprefs.gfx_filter_luminance, &workprefs.gfx_filter_contrast, &workprefs.gfx_filter_saturation,
	&workprefs.gfx_filter_gamma, &workprefs.gfx_filter_blur, &workprefs.gfx_filter_noise,
	&workprefs.gfx_filter_upscale,
	NULL
    };

struct filterpreset {
    char *name;
    int conf[23];
};
static struct filterpreset filterpresets[] =
{
    { "PAL example", 8, 0, 0, 0, 1000, 1000, 0, 0, 50, 0, 0, 0, 1, 0, 0, 0, 10, 0, 0, 0, 300, 30, 0 },
    { NULL }
};

static void values_to_hw3ddlg (HWND hDlg)
{
    char txt[100], tmp[100];
    int i, j, nofilter, fltnum, modenum;
    struct uae_filter *uf;
    UAEREG *fkey;

    SendDlgItemMessage(hDlg, IDC_FILTERHZ, TBM_SETRANGE, TRUE, MAKELONG (-999, +999));
    SendDlgItemMessage(hDlg, IDC_FILTERHZ, TBM_SETPAGESIZE, 0, 1);
    SendDlgItemMessage(hDlg, IDC_FILTERVZ, TBM_SETRANGE, TRUE, MAKELONG (-999, +999));
    SendDlgItemMessage(hDlg, IDC_FILTERVZ, TBM_SETPAGESIZE, 0, 1);
    SendDlgItemMessage(hDlg, IDC_FILTERHO, TBM_SETRANGE, TRUE, MAKELONG (-999, +999));
    SendDlgItemMessage(hDlg, IDC_FILTERHO, TBM_SETPAGESIZE, 0, 1);
    SendDlgItemMessage(hDlg, IDC_FILTERVO, TBM_SETRANGE, TRUE, MAKELONG (-999, +999));
    SendDlgItemMessage(hDlg, IDC_FILTERVO, TBM_SETPAGESIZE, 0, 1);

    SendDlgItemMessage (hDlg, IDC_FILTERMODE, CB_RESETCONTENT, 0, 0L);
    uf = &uaefilters[0];
    nofilter = 0; fltnum = 0;
    i = 0; j = 0;
    while (uaefilters[i].name) {
	switch (uaefilters[i].type)
	{
#ifndef D3D
	    case UAE_FILTER_DIRECT3D:
	    nofilter = 1;
	    break;
#endif
#ifndef OPENGL
	    case UAE_FILTER_OPENGL:
	    nofilter = 1;
	    break;
#endif
	    default:
	    nofilter = 0;
	    break;
	}
	if (nofilter == 0) {
	    SendDlgItemMessage (hDlg, IDC_FILTERMODE, CB_ADDSTRING, 0, (LPARAM)uaefilters[i].name);
	    if (uaefilters[i].type == workprefs.gfx_filter) {
		uf = &uaefilters[i];
		fltnum = j;
	    }
	    j++;
	}
	i++;
    }
    SendDlgItemMessage (hDlg, IDC_FILTERMODE, CB_SETCURSEL, fltnum, 0);

    filter_extra = NULL;
    SendDlgItemMessage (hDlg, IDC_FILTERFILTER, CB_RESETCONTENT, 0, 0L);
    if (uf->x[0]) {
	WIN32GUI_LoadUIString (IDS_3D_NO_FILTER, txt, sizeof (txt));
	sprintf (tmp, txt, 16);
	SendDlgItemMessage (hDlg, IDC_FILTERFILTER, CB_ADDSTRING, 0, (LPARAM)tmp);
	WIN32GUI_LoadUIString (IDS_3D_BILINEAR, txt, sizeof (txt));
	sprintf (tmp, txt, 16);
	SendDlgItemMessage (hDlg, IDC_FILTERFILTER, CB_ADDSTRING, 0, (LPARAM)tmp);
	WIN32GUI_LoadUIString (IDS_3D_NO_FILTER, txt, sizeof (txt));
	sprintf (tmp, txt, 32);
	SendDlgItemMessage (hDlg, IDC_FILTERFILTER, CB_ADDSTRING, 0, (LPARAM)tmp);
	WIN32GUI_LoadUIString (IDS_3D_BILINEAR, txt, sizeof (txt));
	sprintf (tmp, txt, 32);
	SendDlgItemMessage (hDlg, IDC_FILTERFILTER, CB_ADDSTRING, 0, (LPARAM)tmp);
	modenum = 4;
	filter_extra = filter_3d_extra;
    } else {
	modenum = 0;
	for (i = 1; i <= 4; i++) {
	    if (uf->x[i]) {
		makefilter (tmp, i, uf->x[i]);
		SendDlgItemMessage (hDlg, IDC_FILTERFILTER, CB_ADDSTRING, 0, (LPARAM)tmp);
		modenum++;
	    }
	}
	if (uf->yuv) {
	    filter_extra = filter_pal_extra;
	}
    }
    SendDlgItemMessage(hDlg, IDC_FILTERXL, TBM_SETRANGE, TRUE, MAKELONG (   0, +1000));
    SendDlgItemMessage(hDlg, IDC_FILTERXL, TBM_SETPAGESIZE, 0, 1);
    if (filter_extra) {
	int idx2 = -1;
	int idx = SendDlgItemMessage (hDlg, IDC_FILTERXTRA, CB_GETCURSEL, 0, 0L);
	if (idx == CB_ERR)
	    idx = -1;
	SendDlgItemMessage (hDlg, IDC_FILTERXTRA, CB_RESETCONTENT, 0, 0L);
	for (i = 0; filter_extra[i].label; i++) {
	    if (filter_selected == &filter_extra[i] && idx < 0)
		idx2 = i;
	    SendDlgItemMessage (hDlg, IDC_FILTERXTRA, CB_ADDSTRING, 0, (LPARAM)filter_extra[i].label);
	}
	if (idx2 >= 0)
	    filter_selected = &filter_extra[idx2];
	else if (idx >= 0)
	    filter_selected = &filter_extra[idx];
	else
	    filter_selected = &filter_extra[0];
	SendDlgItemMessage (hDlg, IDC_FILTERXTRA, CB_SETCURSEL, filter_selected - &filter_extra[0], 0);
	SendDlgItemMessage(hDlg, IDC_FILTERXL, TBM_SETRANGE, TRUE, MAKELONG (filter_selected->min, filter_selected->max));
	SendDlgItemMessage(hDlg, IDC_FILTERXL, TBM_SETPAGESIZE, 0, filter_selected->step);
	if (filter_selected) {
	    SendDlgItemMessage (hDlg, IDC_FILTERXL, TBM_SETPOS, TRUE, *(filter_selected->varw));
	    SetDlgItemInt (hDlg, IDC_FILTERXLV, *(filter_selected->varw), TRUE);
	}
    }

    if (workprefs.gfx_filter_filtermode >= modenum)
	workprefs.gfx_filter_filtermode = 0;
    SendDlgItemMessage (hDlg, IDC_FILTERFILTER, CB_SETCURSEL, workprefs.gfx_filter_filtermode, 0);

    SendDlgItemMessage (hDlg, IDC_FILTERHZMULT, CB_RESETCONTENT, 0, 0L);
    SendDlgItemMessage (hDlg, IDC_FILTERVZMULT, CB_RESETCONTENT, 0, 0L);
    for (i = 0; filtermultnames[i]; i++) {
	SendDlgItemMessage (hDlg, IDC_FILTERHZMULT, CB_ADDSTRING, 0, (LPARAM)filtermultnames[i]);
	SendDlgItemMessage (hDlg, IDC_FILTERVZMULT, CB_ADDSTRING, 0, (LPARAM)filtermultnames[i]);
    }
    SendDlgItemMessage (hDlg, IDC_FILTERHZMULT, CB_SETCURSEL, 0, 0);
    SendDlgItemMessage (hDlg, IDC_FILTERVZMULT, CB_SETCURSEL, 0, 0);
    for (i = 0; filtermultnames[i]; i++) {
	if (filtermults[i] == workprefs.gfx_filter_horiz_zoom_mult)
	    SendDlgItemMessage (hDlg, IDC_FILTERHZMULT, CB_SETCURSEL, i, 0);
	if (filtermults[i] == workprefs.gfx_filter_vert_zoom_mult)
	    SendDlgItemMessage (hDlg, IDC_FILTERVZMULT, CB_SETCURSEL, i, 0);
    }

    SendDlgItemMessage (hDlg, IDC_FILTERSLR, CB_RESETCONTENT, 0, 0L);
    i = j = 0;
    while (scanlineratios[i * 2]) {
	int sl = scanlineratios[i * 2] * 16 + scanlineratios[i * 2 + 1];
	sprintf (txt, "%d:%d", scanlineratios[i * 2], scanlineratios[i * 2 + 1]);
	if (workprefs.gfx_filter_scanlineratio == sl)
	    j = i;
	SendDlgItemMessage (hDlg, IDC_FILTERSLR, CB_ADDSTRING, 0, (LPARAM)txt);
	scanlineindexes[i] = sl;
	i++;
    }
    SendDlgItemMessage (hDlg, IDC_FILTERSLR, CB_SETCURSEL, j, 0);

    j = 0;
    SendDlgItemMessage (hDlg, IDC_FILTERPRESETS, CB_RESETCONTENT, 0, 0L);
    SendDlgItemMessage (hDlg, IDC_FILTERPRESETS, CB_ADDSTRING, 0, (LPARAM)"");
    for (i = 0; filterpresets[i].name; i++) {
	char tmp[MAX_DPATH];
	sprintf(tmp, "* %s", filterpresets[i].name);
	SendDlgItemMessage (hDlg, IDC_FILTERPRESETS, CB_ADDSTRING, 0, (LPARAM)tmp);
    }
    fkey = regcreatetree (NULL, "FilterPresets");
    if (fkey) {
        int idx = 0;
        char tmp[MAX_DPATH], tmp2[MAX_DPATH];
        DWORD size, size2;

        for (;;) {
	    size = sizeof (tmp);
	    size2 = sizeof (tmp2);
	    if (!regenumstr (fkey, idx, tmp, &size, tmp2, &size2))
		break;
	    SendDlgItemMessage (hDlg, IDC_FILTERPRESETS, CB_ADDSTRING, 0, (LPARAM)tmp);
	    idx++;
	}
	SendDlgItemMessage (hDlg, IDC_FILTERPRESETS, CB_SETCURSEL, filterpreset_selected, 0);
	regclosetree (fkey);
    }

    SendDlgItemMessage (hDlg, IDC_FILTERHZ, TBM_SETPOS, TRUE, workprefs.gfx_filter_horiz_zoom);
    SendDlgItemMessage (hDlg, IDC_FILTERVZ, TBM_SETPOS, TRUE, workprefs.gfx_filter_vert_zoom);
    SendDlgItemMessage (hDlg, IDC_FILTERHO, TBM_SETPOS, TRUE, workprefs.gfx_filter_horiz_offset);
    SendDlgItemMessage (hDlg, IDC_FILTERVO, TBM_SETPOS, TRUE, workprefs.gfx_filter_vert_offset);
    SetDlgItemInt (hDlg, IDC_FILTERHZV, workprefs.gfx_filter_horiz_zoom, TRUE);
    SetDlgItemInt (hDlg, IDC_FILTERVZV, workprefs.gfx_filter_vert_zoom, TRUE);
    SetDlgItemInt (hDlg, IDC_FILTERHOV, workprefs.gfx_filter_horiz_offset, TRUE);
    SetDlgItemInt (hDlg, IDC_FILTERVOV, workprefs.gfx_filter_vert_offset, TRUE);
}

static void values_from_hw3ddlg (HWND hDlg)
{
}

static void filter_preset (HWND hDlg, WPARAM wParam)
{
    int ok, load, i, builtin, userfilter;
    char tmp1[MAX_DPATH], tmp2[MAX_DPATH];
    DWORD outsize;
    UAEREG *fkey;
    LRESULT item;

    load = 0;
    ok = 0;
    for (builtin = 0; filterpresets[builtin].name; builtin++);
    fkey = regcreatetree (NULL, "FilterPresets");
    item = SendDlgItemMessage (hDlg, IDC_FILTERPRESETS, CB_GETCURSEL, 0, 0);
    tmp1[0] = 0;
    if (item != CB_ERR) {
	filterpreset_selected = (int)item;
	SendDlgItemMessage (hDlg, IDC_FILTERPRESETS, CB_GETLBTEXT, (WPARAM)item, (LPARAM)tmp1);
    } else {
	filterpreset_selected = -1;
	SendDlgItemMessage (hDlg, IDC_FILTERPRESETS, WM_GETTEXT, (WPARAM)sizeof tmp1, (LPARAM)tmp1);
    }
    userfilter = 0;
    filterpreset_builtin = -1;
    if (filterpreset_selected < 0 || filterpreset_selected == 0) {
	userfilter = -1;
    } else if (filterpreset_selected > builtin) {
	userfilter = filterpreset_selected - builtin;
    } else {
	filterpreset_builtin = filterpreset_selected - 1;
    }

    if (filterpreset_builtin < 0) {
	outsize = sizeof (tmp2);
	if (tmp1[0] && regquerystr (fkey, tmp1, tmp2, &outsize))
	    ok = 1;
    } else {
	char *p = tmp2;
	for (i = 0; filtervars[i]; i++) {
	    if (i > 0) {
		strcat (p, ",");
		p++;
	    }
	    sprintf (p, "%d", filterpresets[filterpreset_builtin].conf[i]);
	    p += strlen(p);
	}
	ok = 1;
    }

    if (wParam == IDC_FILTERPRESETSAVE && userfilter && fkey) {
	char *p = tmp2;
	for (i = 0; filtervars[i]; i++) {
	    if (i > 0) {
		strcat (p, ",");
		p++;
	    }
	    sprintf (p, "%d", *(filtervars[i]));
	    p += strlen(p);
	}
	if (ok == 0) {
	    tmp1[0] = 0;
	    SendDlgItemMessage (hDlg, IDC_FILTERPRESETS, WM_GETTEXT, (WPARAM)sizeof (tmp1), (LPARAM)tmp1);
	    if (tmp1[0] == 0)
		goto end;
	}
	regsetstr (fkey, tmp1, tmp2);
	values_to_hw3ddlg (hDlg);
    }
    if (ok) {
	if (wParam == IDC_FILTERPRESETDELETE && userfilter) {
	    regdelete (fkey, tmp1);
	    values_to_hw3ddlg (hDlg);
	} else if (wParam == IDC_FILTERPRESETLOAD) {
	    char *s = tmp2;
	    char *t;

	    load = 1;
	    strcat (s, ",");
	    t = strchr (s, ',');
	    *t++ = 0;
	    for (i = 0; filtervars[i]; i++) {
		*(filtervars[i]) = atol(s);
		s = t;
		t = strchr (s, ',');
		if (!t)
		    break;
		*t++ = 0;
	    }
	}
    }
end:
    regclosetree (fkey);
    if (load) {
	values_to_hw3ddlg (hDlg);
	SendMessage (hDlg, WM_HSCROLL, 0, 0);
    }
    enable_for_hw3ddlg (hDlg);
}

static void filter_handle (HWND hDlg)
{
    LRESULT item = SendDlgItemMessage (hDlg, IDC_FILTERMODE, CB_GETCURSEL, 0, 0L);
    if (item != CB_ERR) {
	int of = workprefs.gfx_filter;
	int off = workprefs.gfx_filter_filtermode;
	workprefs.gfx_filter = 0;
	if (IsDlgButtonChecked (hDlg, IDC_FILTERENABLE)) {
	    workprefs.gfx_filter = uaefilters[item].type;
	    item = SendDlgItemMessage (hDlg, IDC_FILTERFILTER, CB_GETCURSEL, 0, 0L);
	    if (item != CB_ERR)
		workprefs.gfx_filter_filtermode = (int)item;
	    if (of != workprefs.gfx_filter || off != workprefs.gfx_filter_filtermode) {
		values_to_hw3ddlg (hDlg);
		hw3d_changed = 1;
	    }
	}
    }
    enable_for_hw3ddlg (hDlg);
    updatedisplayarea ();
}

static int getfiltermult(HWND hDlg, DWORD dlg)
{
    LRESULT v = SendDlgItemMessage (hDlg, dlg, CB_GETCURSEL, 0, 0L);
    if (v == CB_ERR)
	return 1000;
    return filtermults[v];
}

static INT_PTR CALLBACK hw3dDlgProc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int recursive;
    LRESULT item;

    switch (msg)
    {
    case WM_INITDIALOG:
#if WINUAEBETA == 0
	ShowWindow (GetDlgItem(hDlg, IDC_FILTERAUTORES), SW_HIDE);
#endif
	pages[HW3D_ID] = hDlg;
	currentpage = HW3D_ID;
	enable_for_hw3ddlg (hDlg);

    case WM_USER:
	if(recursive > 0)
	    break;
	recursive++;
	enable_for_hw3ddlg( hDlg );
	values_to_hw3ddlg (hDlg);
	recursive--;
	return TRUE;
    case WM_COMMAND:
	if(recursive > 0)
	    break;
	recursive++;
	switch (wParam)
	{
	    case IDC_FILTERDEFAULT:
	    currprefs.gfx_filter_horiz_zoom = workprefs.gfx_filter_horiz_zoom = 0;
	    currprefs.gfx_filter_vert_zoom = workprefs.gfx_filter_vert_zoom = 0;
	    currprefs.gfx_filter_horiz_offset = workprefs.gfx_filter_horiz_offset = 0;
	    currprefs.gfx_filter_vert_offset = workprefs.gfx_filter_vert_offset = 0;
	    values_to_hw3ddlg (hDlg);
	    updatedisplayarea ();
	    WIN32GFX_WindowMove ();
	    break;
	    case IDC_FILTERPRESETLOAD:
	    case IDC_FILTERPRESETSAVE:
	    case IDC_FILTERPRESETDELETE:
	    filter_preset (hDlg, wParam);
	    break;
	    case IDC_FILTERENABLE:
	    filter_handle (hDlg);
	    break;
	    case IDC_FILTERAUTORES:
	    workprefs.gfx_autoresolution = IsDlgButtonChecked (hDlg, IDC_FILTERAUTORES);
	    break;
	    case IDC_FILTERUPSCALE:
	    currprefs.gfx_filter_upscale = workprefs.gfx_filter_upscale = IsDlgButtonChecked (hDlg, IDC_FILTERUPSCALE);
	    updatedisplayarea ();
	    WIN32GFX_WindowMove ();
	    break;
	    default:
	    if (HIWORD (wParam) == CBN_SELCHANGE || HIWORD (wParam) == CBN_KILLFOCUS)  {
		switch (LOWORD (wParam))
		{
		    case IDC_FILTERXTRA:
		    values_to_hw3ddlg (hDlg);
		    break;
		    case IDC_FILTERPRESETS:
		    filter_preset (hDlg, LOWORD (wParam));
		    break;
		    case IDC_FILTERSLR:
		    item = SendDlgItemMessage (hDlg, IDC_FILTERSLR, CB_GETCURSEL, 0, 0L);
		    if (item != CB_ERR) {
			currprefs.gfx_filter_scanlineratio = workprefs.gfx_filter_scanlineratio = scanlineindexes[item];
			updatedisplayarea ();
		    }
		    break;
		    case IDC_FILTERMODE:
		    case IDC_FILTERFILTER:
		    filter_handle (hDlg);
		    break;
		    case IDC_FILTERHZMULT:
		    case IDC_FILTERVZMULT:
		    currprefs.gfx_filter_horiz_zoom_mult = workprefs.gfx_filter_horiz_zoom_mult = getfiltermult(hDlg, IDC_FILTERHZMULT);
		    currprefs.gfx_filter_vert_zoom_mult = workprefs.gfx_filter_vert_zoom_mult = getfiltermult(hDlg, IDC_FILTERVZMULT);
		    updatedisplayarea ();
		    WIN32GFX_WindowMove ();
		    break;
		}
	    }
	    break;
	}
	recursive--;
	break;
    case WM_HSCROLL:
	currprefs.gfx_filter_horiz_zoom = workprefs.gfx_filter_horiz_zoom = (int)SendMessage(GetDlgItem(hDlg, IDC_FILTERHZ), TBM_GETPOS, 0, 0 );
	currprefs.gfx_filter_vert_zoom = workprefs.gfx_filter_vert_zoom = (int)SendMessage(GetDlgItem(hDlg, IDC_FILTERVZ), TBM_GETPOS, 0, 0 );
	currprefs.gfx_filter_horiz_offset = workprefs.gfx_filter_horiz_offset = (int)SendMessage(GetDlgItem(hDlg, IDC_FILTERHO), TBM_GETPOS, 0, 0);
	currprefs.gfx_filter_vert_offset = workprefs.gfx_filter_vert_offset = (int)SendMessage(GetDlgItem(hDlg, IDC_FILTERVO), TBM_GETPOS, 0, 0);
	if (filter_selected) {
	    int *pw = filter_selected->varw;
	    int *pc = filter_selected->varc;
	    int v = (int)SendMessage(GetDlgItem(hDlg, IDC_FILTERXL), TBM_GETPOS, 0, 0);
	    if (v < filter_selected->min)
		v = filter_selected->min;
	    if (v > filter_selected->max)
		v = filter_selected->max;
	    *pw = v;
	    *pc = v;
	    SetDlgItemInt (hDlg, IDC_FILTERXLV, v, TRUE);
	}
	SetDlgItemInt (hDlg, IDC_FILTERHZV, workprefs.gfx_filter_horiz_zoom, TRUE);
	SetDlgItemInt (hDlg, IDC_FILTERVZV, workprefs.gfx_filter_vert_zoom, TRUE);
	SetDlgItemInt (hDlg, IDC_FILTERHOV, workprefs.gfx_filter_horiz_offset, TRUE);
	SetDlgItemInt (hDlg, IDC_FILTERVOV, workprefs.gfx_filter_vert_offset, TRUE);
	init_colors();
	notice_new_xcolors();
	reset_drawing();
	updatedisplayarea ();
	WIN32GFX_WindowMove ();
	break;
    }
    return FALSE;
}
#endif

#ifdef AVIOUTPUT
static void values_to_avioutputdlg(HWND hDlg)
{
    char tmpstr[256];

    updatewinfsmode (&workprefs);
    SetDlgItemText(hDlg, IDC_AVIOUTPUT_FILETEXT, avioutput_filename);

    sprintf(tmpstr, "%d fps", avioutput_fps);
    SendMessage(GetDlgItem(hDlg, IDC_AVIOUTPUT_FPS_STATIC), WM_SETTEXT, (WPARAM) 0, (LPARAM) tmpstr);

    sprintf(tmpstr, "Actual: %d x %d", workprefs.gfx_size.width, workprefs.gfx_size.height);
    SendMessage(GetDlgItem(hDlg, IDC_AVIOUTPUT_DIMENSIONS_STATIC), WM_SETTEXT, (WPARAM) 0, (LPARAM) tmpstr);

    switch(avioutput_fps)
    {
	case VBLANK_HZ_PAL:
	    CheckRadioButton(hDlg, IDC_AVIOUTPUT_PAL, IDC_AVIOUTPUT_NTSC, IDC_AVIOUTPUT_PAL);
	break;

	case VBLANK_HZ_NTSC:
	    CheckRadioButton(hDlg, IDC_AVIOUTPUT_PAL, IDC_AVIOUTPUT_NTSC, IDC_AVIOUTPUT_NTSC);
	break;

	default:
	    CheckDlgButton(hDlg, IDC_AVIOUTPUT_PAL, BST_UNCHECKED);
	    CheckDlgButton(hDlg, IDC_AVIOUTPUT_NTSC, BST_UNCHECKED);
	break;
    }

    CheckDlgButton (hDlg, IDC_AVIOUTPUT_FRAMELIMITER, avioutput_framelimiter ? FALSE : TRUE);
    CheckDlgButton (hDlg, IDC_AVIOUTPUT_NOSOUNDOUTPUT, avioutput_nosoundoutput ? TRUE : FALSE);
    CheckDlgButton (hDlg, IDC_AVIOUTPUT_ACTIVATED, avioutput_requested ? BST_CHECKED : BST_UNCHECKED);
    CheckDlgButton (hDlg, IDC_SAMPLERIPPER_ACTIVATED, sampleripper_enabled ? BST_CHECKED : BST_UNCHECKED);
}

static void values_from_avioutputdlg(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
	LRESULT tmp;

	updatewinfsmode (&workprefs);
	tmp = SendMessage(GetDlgItem(hDlg, IDC_AVIOUTPUT_FPS), TBM_GETPOS, 0, 0);
	if (tmp < 1)
	    tmp = 1;
	if (tmp != avioutput_fps) {
	    avioutput_fps = (int)tmp;
	    AVIOutput_Restart ();
	}
}

static void enable_for_avioutputdlg(HWND hDlg)
{
    char tmp[1000];
#if defined (PROWIZARD)
    ew (hDlg, IDC_PROWIZARD, TRUE);
    if (full_property_sheet)
	ew (hDlg, IDC_PROWIZARD, FALSE);
#endif

    ew (hDlg, IDC_SCREENSHOT, full_property_sheet ? FALSE : TRUE);
    ew (hDlg, IDC_SAMPLERIPPER_ACTIVATED, full_property_sheet ? FALSE : TRUE);

    ew (hDlg, IDC_AVIOUTPUT_PAL, TRUE);
    ew (hDlg, IDC_AVIOUTPUT_NTSC, TRUE);
    ew (hDlg, IDC_AVIOUTPUT_FPS, TRUE);
    ew (hDlg, IDC_AVIOUTPUT_FILE, TRUE);

    if(workprefs.produce_sound < 2) {
	ew (hDlg, IDC_AVIOUTPUT_AUDIO, FALSE);
	ew (hDlg, IDC_AVIOUTPUT_AUDIO_STATIC, FALSE);
	avioutput_audio = 0;
    } else {
	ew (hDlg, IDC_AVIOUTPUT_AUDIO, TRUE);
	ew (hDlg, IDC_AVIOUTPUT_AUDIO_STATIC, TRUE);
    }

    if (avioutput_audio == AVIAUDIO_WAV) {
       strcpy (tmp, "Wave (internal)");
    } else {
	avioutput_audio = AVIOutput_GetAudioCodec(tmp, sizeof tmp);
    }
    if(!avioutput_audio) {
	CheckDlgButton(hDlg, IDC_AVIOUTPUT_AUDIO, BST_UNCHECKED);
	WIN32GUI_LoadUIString (IDS_AVIOUTPUT_NOCODEC, tmp, sizeof tmp);
    }
    SetWindowText(GetDlgItem(hDlg, IDC_AVIOUTPUT_AUDIO_STATIC), tmp);

    if (avioutput_audio != AVIAUDIO_WAV)
	avioutput_video = AVIOutput_GetVideoCodec(tmp, sizeof tmp);
    if(!avioutput_video) {
	CheckDlgButton(hDlg, IDC_AVIOUTPUT_VIDEO, BST_UNCHECKED);
	WIN32GUI_LoadUIString (IDS_AVIOUTPUT_NOCODEC, tmp, sizeof tmp);
    }
    SetWindowText(GetDlgItem(hDlg, IDC_AVIOUTPUT_VIDEO_STATIC), tmp);

    ew (hDlg, IDC_AVIOUTPUT_NOSOUNDOUTPUT, avioutput_framelimiter ? TRUE : FALSE);

    if (!avioutput_framelimiter)
	avioutput_nosoundoutput = 1;
    CheckDlgButton (hDlg, IDC_AVIOUTPUT_FRAMELIMITER, avioutput_framelimiter ? FALSE : TRUE);
    CheckDlgButton (hDlg, IDC_AVIOUTPUT_NOSOUNDOUTPUT, avioutput_nosoundoutput ? TRUE : FALSE);

    ew (hDlg, IDC_AVIOUTPUT_ACTIVATED, (!avioutput_audio && !avioutput_video) ? FALSE : TRUE);

    ew (hDlg, IDC_INPREC_RECORD, input_recording >= 0);
    CheckDlgButton(hDlg, IDC_INPREC_RECORD, input_recording > 0 ? BST_CHECKED : BST_UNCHECKED);
    ew (hDlg, IDC_INPREC_PLAY, input_recording <= 0);
    CheckDlgButton(hDlg, IDC_INPREC_PLAY, input_recording < 0 ? BST_CHECKED : BST_UNCHECKED);
    ew (hDlg, IDC_INPREC_PLAYMODE, input_recording == 0);
}

static INT_PTR CALLBACK AVIOutputDlgProc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int recursive = 0;
    char tmp[1000];

    switch(msg)
    {
	case WM_INITDIALOG:
	    pages[AVIOUTPUT_ID] = hDlg;
	    currentpage = AVIOUTPUT_ID;
	    AVIOutput_GetSettings();
	    enable_for_avioutputdlg(hDlg);
	    SendDlgItemMessage(hDlg, IDC_AVIOUTPUT_FPS, TBM_SETRANGE, TRUE, MAKELONG(1, VBLANK_HZ_NTSC));
	    SendDlgItemMessage(hDlg, IDC_AVIOUTPUT_FPS, TBM_SETPOS, TRUE, avioutput_fps);
	    SendMessage(hDlg, WM_HSCROLL, (WPARAM) NULL, (LPARAM) NULL);
	    if (!avioutput_filename[0]) {
		fetch_path ("VideoPath", avioutput_filename, sizeof (avioutput_filename));
		strcat (avioutput_filename, "output.avi");
	    }

	case WM_USER:
	    recursive++;
	    values_to_avioutputdlg(hDlg);
	    enable_for_avioutputdlg(hDlg);
	    recursive--;
	return TRUE;

	case WM_HSCROLL:
	{
	    recursive++;
	    values_from_avioutputdlg(hDlg, msg, wParam, lParam);
	    values_to_avioutputdlg(hDlg);
	    enable_for_avioutputdlg(hDlg);
	    recursive--;
	    return TRUE;
	}

	case WM_COMMAND:

	    if(recursive > 0)
		break;

	    recursive++;

	    switch(wParam)
	    {

		case IDC_AVIOUTPUT_FRAMELIMITER:
		    avioutput_framelimiter = IsDlgButtonChecked (hDlg, IDC_AVIOUTPUT_FRAMELIMITER) ? 0 : 1;
		    AVIOutput_SetSettings();
		break;
		case IDC_AVIOUTPUT_NOSOUNDOUTPUT:
		    avioutput_nosoundoutput = IsDlgButtonChecked (hDlg, IDC_AVIOUTPUT_NOSOUNDOUTPUT) ? 1 : 0;
		    AVIOutput_SetSettings();
		break;

		case IDC_INPREC_PLAYMODE:
		break;
		case IDC_INPREC_RECORD:
		    if (input_recording)
			inprec_close ();
		    else
			DiskSelection(hDlg, wParam, 16, &workprefs, NULL);
		break;
		case IDC_INPREC_PLAY:
		    if (input_recording)
			inprec_close ();
		    else
			DiskSelection(hDlg, wParam, 15, &workprefs, NULL);
		break;

#ifdef PROWIZARD
		case IDC_PROWIZARD:
		    moduleripper ();
		break;
#endif
		case IDC_SAMPLERIPPER_ACTIVATED:
		    sampleripper_enabled = !sampleripper_enabled;
		    audio_sampleripper(-1);
		break;

		case IDC_AVIOUTPUT_ACTIVATED:
		    avioutput_requested = !avioutput_requested;
		    SendMessage(hDlg, WM_HSCROLL, (WPARAM) NULL, (LPARAM) NULL);
		    if (!avioutput_requested)
			AVIOutput_End ();
		break;

		case IDC_SCREENSHOT:
		    screenshot(1, 0);
		break;

		case IDC_AVIOUTPUT_PAL:
		    SendDlgItemMessage(hDlg, IDC_AVIOUTPUT_FPS, TBM_SETPOS, TRUE, VBLANK_HZ_PAL);
		    SendMessage(hDlg, WM_HSCROLL, (WPARAM) NULL, (LPARAM) NULL);
		break;

		case IDC_AVIOUTPUT_NTSC:
		    SendDlgItemMessage(hDlg, IDC_AVIOUTPUT_FPS, TBM_SETPOS, TRUE, VBLANK_HZ_NTSC);
		    SendMessage(hDlg, WM_HSCROLL, (WPARAM) NULL, (LPARAM) NULL);
		break;

		case IDC_AVIOUTPUT_AUDIO:
		{
		    if (avioutput_enabled)
			AVIOutput_End ();
		    if(IsDlgButtonChecked(hDlg, IDC_AVIOUTPUT_AUDIO) == BST_CHECKED) {
			avioutput_audio = AVIOutput_ChooseAudioCodec(hDlg, tmp, sizeof tmp);
			enable_for_avioutputdlg(hDlg);
		    } else {
			avioutput_audio = 0;
		    }
		    break;
		}

		case IDC_AVIOUTPUT_VIDEO:
		{
		    if (avioutput_enabled)
			AVIOutput_End ();
		    if(IsDlgButtonChecked(hDlg, IDC_AVIOUTPUT_VIDEO) == BST_CHECKED) {
			avioutput_video = AVIOutput_ChooseVideoCodec(hDlg, tmp, sizeof tmp);
			if (avioutput_audio = AVIAUDIO_WAV)
			    avioutput_audio = 0;
			enable_for_avioutputdlg(hDlg);
		    } else {
			avioutput_video = 0;
		    }
		    enable_for_avioutputdlg(hDlg);
		    break;
		}

		case IDC_AVIOUTPUT_FILE:
		{
		    OPENFILENAME ofn;

		    ZeroMemory(&ofn, sizeof(OPENFILENAME));
		    ofn.lStructSize = sizeof(OPENFILENAME);
		    ofn.hwndOwner = hDlg;
		    ofn.hInstance = hInst;
		    ofn.Flags = OFN_EXTENSIONDIFFERENT | OFN_EXPLORER | OFN_PATHMUSTEXIST | OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT | OFN_NOCHANGEDIR;
		    ofn.lpstrCustomFilter = NULL;
		    ofn.nMaxCustFilter = 0;
		    ofn.nFilterIndex = 0;
		    ofn.lpstrFile = avioutput_filename;
		    ofn.nMaxFile = MAX_DPATH;
		    ofn.lpstrFileTitle = NULL;
		    ofn.nMaxFileTitle = 0;
		    ofn.lpstrInitialDir = NULL;
		    ofn.lpfnHook = NULL;
		    ofn.lpTemplateName = NULL;
		    ofn.lCustData = 0;
		    ofn.lpstrFilter = "Video Clip (*.avi)\0*.avi\0Wave Sound (*.wav)\0";

		    if(!GetSaveFileName(&ofn))
			break;
		    if (ofn.nFilterIndex == 2) {
			avioutput_audio = AVIAUDIO_WAV;
			avioutput_video = 0;
			if (strlen (avioutput_filename) > 4 && !stricmp (avioutput_filename + strlen (avioutput_filename) - 4, ".avi"))
			    strcpy (avioutput_filename + strlen (avioutput_filename) - 4, ".wav");
		    }
		    break;
		}

	    }

	    values_from_avioutputdlg(hDlg, msg, wParam, lParam);
	    values_to_avioutputdlg(hDlg);
	    enable_for_avioutputdlg(hDlg);

	    recursive--;

	    return TRUE;
	}
	return FALSE;
}
#endif

struct GUIPAGE {
    DLGPROC dlgproc;
    LPCTSTR title;
    LPCTSTR icon;
    HTREEITEM tv;
    int himg;
    int idx;
    const char *help;
    HACCEL accel;
    int fullpanel;
    struct newresource *nres;
};

static int GetPanelRect (HWND hDlg, RECT *r)
{
    RECT rect;
    if (!GetWindowRect (guiDlg, &rect))
	return 0;
    if (!GetWindowRect (hDlg, r))
	return 0;
    r->top -= rect.top;
    r->left -= rect.left;
    r->right -= rect.left;
    r->bottom -= rect.top;
    return 1;
}

static int ToolTipImageIDs[] = { 1, IDB_XARCADE, -1 };

static WNDPROC ToolTipWndProcOld;
static LRESULT FAR PASCAL ToolTipWndProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    RECT r1;
    POINT p1;
    PAINTSTRUCT ps;
    HBITMAP bm;
    BITMAP binfo;
    HDC memdc;
    int w, h, i;

    for (i = 0; ToolTipHWNDS2[i].hwnd; i++) {
	if (hwnd == ToolTipHWNDS2[i].hwnd)
	    break;
    }
    if (!ToolTipHWNDS2[i].hwnd)
	return CallWindowProc (ToolTipHWNDS2[i].proc, hwnd, message, wParam, lParam);

    switch (message)
    {
	case WM_WINDOWPOSCHANGED:
	bm = LoadBitmap (hInst, MAKEINTRESOURCE (ToolTipHWNDS2[i].imageid));
	GetObject (bm, sizeof (binfo), &binfo);
	w = binfo.bmWidth;
	h = binfo.bmHeight;
	GetWindowRect (hwnd, &r1);
	GetCursorPos (&p1);
	r1.right = r1.left + w;
	r1.bottom = r1.top + h;
	MoveWindow (hwnd, r1.left, r1.top, r1.right - r1.left, r1.bottom - r1.top, TRUE);
	DeleteObject (bm);
	return 0;

	case WM_PAINT:
	bm = LoadBitmap (hInst, MAKEINTRESOURCE (ToolTipHWNDS2[i].imageid));
	GetObject (bm, sizeof (binfo), &binfo);
	w = binfo.bmWidth;
	h = binfo.bmHeight;
	GetWindowRect (hwnd, &r1);
	GetCursorPos (&p1);
	r1.right = r1.left + w;
	r1.bottom = r1.top + h;
	BeginPaint (hwnd, &ps);
	memdc = CreateCompatibleDC (ps.hdc);
	SelectObject (memdc, bm);
	ShowWindow (hwnd, SW_SHOWNA);
	MoveWindow (hwnd, r1.left, r1.top, r1.right - r1.left, r1.bottom - r1.top, TRUE);
	SetBkMode (ps.hdc, TRANSPARENT);
	BitBlt (ps.hdc, 0, 0, w, h, memdc, 0, 0, SRCCOPY);
	DeleteObject (bm);
	EndPaint (hwnd, &ps);
	DeleteDC (memdc);
	return FALSE;
	case WM_PRINT:
	PostMessage (hwnd, WM_PAINT, 0, 0);
	return TRUE;
    }
    return CallWindowProc (ToolTipHWNDS2[i].proc, hwnd, message, wParam, lParam);
}

static int ignorewindows[] = {
    IDD_FLOPPY, IDC_DF0TEXT, IDC_DF1TEXT, IDC_DF2TEXT, IDC_DF3TEXT, IDC_CREATE_NAME,
    -1,
    IDD_QUICKSTART, IDC_DF0TEXTQ, IDC_DF1TEXTQ, IDC_QUICKSTART_HOSTCONFIG,
    -1,
    IDD_AVIOUTPUT, IDC_AVIOUTPUT_FILETEXT, IDC_AVIOUTPUT_VIDEO_STATIC, IDC_AVIOUTPUT_AUDIO_STATIC,
    -1,
    IDD_DISK, IDC_DISKTEXT, IDC_DISKLIST,
    -1,
    IDD_DISPLAY, IDC_DISPLAYSELECT,
    -1,
    IDD_FILTER, IDC_FILTERPRESETS,
    -1,
    IDD_HARDDISK, IDC_VOLUMELIST,
    -1,
    IDD_INPUT, IDC_INPUTDEVICE, IDC_INPUTLIST, IDC_INPUTAMIGA,
    -1,
    IDD_KICKSTART, IDC_ROMFILE, IDC_ROMFILE2, IDC_CARTFILE, IDC_FLASHFILE,
    -1,
    IDD_LOADSAVE, IDC_CONFIGTREE, IDC_EDITNAME, IDC_EDITDESCRIPTION, IDC_CONFIGLINK, IDC_EDITPATH,
    -1,
    IDD_MISC1, IDC_LANGUAGE,
    -1,
    IDD_PATHS, IDC_PATHS_ROM, IDC_PATHS_CONFIG, IDC_PATHS_SCREENSHOT, IDC_PATHS_SAVESTATE, IDC_PATHS_AVIOUTPUT, IDC_PATHS_SAVEIMAGE,
    -1,
    IDD_PORTS, IDC_PRINTERLIST, IDC_PS_PARAMS, IDC_SERIAL, IDC_MIDIOUTLIST, IDC_MIDIINLIST,
    -1,
    IDD_SOUND, IDC_SOUNDCARDLIST, IDC_SOUNDDRIVESELECT,
    -1,
    0
};

static BOOL CALLBACK childenumproc (HWND hwnd, LPARAM lParam)
{
    int i;
    TOOLINFO ti;
    char tmp[MAX_DPATH];
    char *p;
    LRESULT v;

    if (GetParent(hwnd) != panelDlg)
	return 1;
    i = 0;
    while (ignorewindows[i]) {
	if (ignorewindows[i] == (int)lParam) {
	    int dlgid = GetDlgCtrlID(hwnd);
	    i++;
	    while (ignorewindows[i] > 0) {
		if (dlgid == ignorewindows[i])
		    return 1;
		i++;
	    }
	    break;
	} else {
	    while (ignorewindows[i++] > 0);
	}
    }
    tmp[0] = 0;
    SendMessage (hwnd, WM_GETTEXT, (WPARAM)sizeof (tmp), (LPARAM)tmp);
    p = strchr (tmp, '[');
    if (strlen (tmp) > 0 && p && strlen(p) > 2 && p[1] == ']') {
	int imageid = 0;
	*p++ = 0;
	*p++ = 0;
	if (p[0] == ' ')
	    *p++;
	if (p[0] == '#')
	    imageid = atol (p + 1);
	tmp[strlen(tmp) - 1] = 0;
	ti.cbSize = sizeof (TOOLINFO);
	ti.uFlags = TTF_SUBCLASS | TTF_IDISHWND;
	ti.hwnd = GetParent (hwnd);
	ti.hinst = hInst;
	ti.uId = (UINT_PTR)hwnd;
	ti.lpszText = p;
	if (imageid > 0) {
	    int idx, i;

	    idx = 0;
	    while (ToolTipHWNDS2[idx].hwnd)
		idx++;
	    for (i = 0; ToolTipImageIDs[i] >= 0; i += 2) {
		if (ToolTipImageIDs[i] == imageid)
		    break;
	    }
	    if (ToolTipImageIDs[i] >= 0 && idx < MAX_IMAGETOOLTIPS) {
		HWND ToolTipHWND2 = CreateWindowEx (WS_EX_TOPMOST,
		    TOOLTIPS_CLASS, NULL,
		    WS_POPUP | TTS_NOPREFIX,
		    CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		    panelDlg, NULL, hInst, NULL);
		ToolTipHWNDS2[idx].hwnd = ToolTipHWND2;
		ToolTipHWNDS2[idx+1].hwnd = NULL;
		ToolTipHWNDS2[idx].proc = (WNDPROC)GetWindowLongPtr (ToolTipHWND2, GWLP_WNDPROC);
		ToolTipHWNDS2[idx].imageid = ToolTipImageIDs[i + 1];
		SetWindowLongPtr (ToolTipHWND2, GWLP_WNDPROC, (LONG_PTR)ToolTipWndProc);
		SetWindowPos (ToolTipHWND2, HWND_TOPMOST, 0, 0, 0, 0,
		    SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
		SendMessage(ToolTipHWND2, TTM_SETDELAYTIME, (WPARAM)TTDT_AUTOPOP, (LPARAM)MAKELONG(20000, 0));
		SendMessage(ToolTipHWND2, TTM_SETMAXTIPWIDTH, 0, 400);
		SendMessage(hwnd, WM_SETTEXT, 0, (LPARAM)tmp);
		v = SendMessage(ToolTipHWND2, TTM_ADDTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);
	    }
	} else {
	    v = SendMessage(hwnd, WM_SETTEXT, 0, (LPARAM)tmp);
	    v = SendMessage(ToolTipHWND, TTM_ADDTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);
	}
	return 1;
    }
    p = strchr (tmp, ']');
    if (strlen (tmp) > 0 && p && strlen(p) > 2 && p[1] == '[') {
	RECT r;
	*p++ = 0;
	*p++ = 0;
	if (p[0] == ' ')
	    *p++;
	tmp[strlen(tmp) - 1] = 0;
	SendMessage(hwnd, WM_SETTEXT, 0, (LPARAM)tmp);
	ti.cbSize = sizeof (TOOLINFO);
	ti.uFlags = TTF_SUBCLASS;
	ti.hwnd = GetParent (hwnd);
	ti.hinst = hInst;
	ti.uId = (UINT_PTR)hwnd;
	ti.lpszText = p;
	GetWindowRect (GetParent (hwnd), &r);
	GetWindowRect (hwnd, &ti.rect);
	ti.rect.top -= r.top;
	ti.rect.left -= r.left;
	ti.rect.right -= r.left;
	ti.rect.bottom -= r.top;
	SendMessage(ToolTipHWND, TTM_ADDTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);
    }
    return 1;
}

static struct GUIPAGE ppage[MAX_C_PAGES];
#define PANEL_X 174
#define PANEL_Y 12
#define PANEL_WIDTH 456
#define PANEL_HEIGHT 396

static HWND updatePanel (HWND hDlg, int id)
{
    static HWND hwndTT;
    RECT r1c, r1w, r2c, r2w, r3c, r3w;
    int w, h, pw, ph, x , y, i;
    int fullpanel;
    struct newresource *tres;

    ew (guiDlg, IDC_RESETAMIGA, full_property_sheet ? FALSE : TRUE);
    ew (guiDlg, IDOK, TRUE);
    if (panelDlg != NULL) {
	ShowWindow (panelDlg, FALSE);
	DestroyWindow (panelDlg);
	panelDlg = NULL;
    }
    if (ToolTipHWND != NULL) {
	DestroyWindow (ToolTipHWND);
	ToolTipHWND = NULL;
    }
    for (i = 0; ToolTipHWNDS2[i].hwnd; i++) {
	DestroyWindow (ToolTipHWNDS2[i].hwnd);
	ToolTipHWNDS2[i].hwnd = NULL;
    }
    hAccelTable = NULL;
    if (id < 0) {
	if (isfullscreen () <= 0) {
	    RECT r;
	    if (GetWindowRect (hDlg, &r)) {
		LONG left, top;
		left = r.left;
		top = r.top;
		regsetint (NULL, "GUIPosX", left);
		regsetint (NULL, "GUIPosY", top);
	    }
	}
	ew (hDlg, IDHELP, FALSE);
	return NULL;
    }

    fullpanel = ppage[id].fullpanel;
    GetWindowRect (GetDlgItem (hDlg, IDC_PANEL_FRAME), &r1w);
    GetClientRect (GetDlgItem (hDlg, IDC_PANEL_FRAME), &r1c);
    GetWindowRect (hDlg, &r2w);
    GetClientRect (hDlg, &r2c);
    gui_width = r2c.right;
    gui_height = r2c.bottom;
    tres = scaleresource(ppage[id].nres, hDlg);
    panelDlg = CreateDialogIndirectParam (tres->inst, tres->resource, hDlg, ppage[id].dlgproc, id);
    freescaleresource(tres);
    GetWindowRect (hDlg, &r3w);
    GetClientRect (panelDlg, &r3c);
    x = r1w.left - r2w.left;
    y = r1w.top - r2w.top;
    w = r3c.right - r3c.left + 1;
    h = r3c.bottom - r3c.top + 1;
    pw = r1w.right - r1w.left + 1;
    ph = r1w.bottom - r1w.top + 1;
    SetWindowPos (panelDlg, HWND_TOP, 0, 0, 0, 0,
	SWP_NOSIZE | SWP_NOOWNERZORDER);
    GetWindowRect (panelDlg, &r3w);
    GetClientRect (panelDlg, &r3c);
    x -= r3w.left - r2w.left - 1;
    y -= r3w.top - r2w.top - 1;
    if (!fullpanel) {
	SetWindowPos (panelDlg, HWND_TOP, x + (pw - w) / 2, y + (ph - h) / 2, 0, 0,
	    SWP_NOSIZE | SWP_NOOWNERZORDER);
    }
    ShowWindow(GetDlgItem(hDlg, IDC_PANEL_FRAME), SW_HIDE);
    ShowWindow(GetDlgItem(hDlg, IDC_PANEL_FRAME_OUTER), !fullpanel ? SW_SHOW : SW_HIDE);
    ShowWindow(GetDlgItem(hDlg, IDC_PANELTREE), !fullpanel ? SW_SHOW : SW_HIDE);
    ShowWindow (panelDlg, SW_SHOW);
    ew (hDlg, IDHELP, pHtmlHelp && ppage[currentpage].help ? TRUE : FALSE);

    ToolTipHWND = CreateWindowEx (WS_EX_TOPMOST,
	TOOLTIPS_CLASS, NULL,
	WS_POPUP | TTS_NOPREFIX | TTS_BALLOON,
	CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
	panelDlg, NULL, hInst, NULL);
    SetWindowPos (ToolTipHWND, HWND_TOPMOST, 0, 0, 0, 0,
	SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);

#if 0
    SendMessage(ToolTipHWND, TTM_SETDELAYTIME, (WPARAM)TTDT_INITIAL, (LPARAM)MAKELONG(100, 0));
    SendMessage(ToolTipHWND, TTM_SETDELAYTIME, (WPARAM)TTDT_RESHOW, (LPARAM)MAKELONG(0, 0));
#endif
    SendMessage(ToolTipHWND, TTM_SETDELAYTIME, (WPARAM)TTDT_AUTOPOP, (LPARAM)MAKELONG(20000, 0));
    SendMessage(ToolTipHWND, TTM_SETMAXTIPWIDTH, 0, 400);

    EnumChildWindows (panelDlg, &childenumproc, (LPARAM)ppage[currentpage].nres->tmpl);
    SendMessage (panelDlg, WM_NULL, 0, 0);

    hAccelTable = ppage[currentpage].accel;

    return panelDlg;
}

static HTREEITEM CreateFolderNode (HWND TVhDlg, int nameid, HTREEITEM parent, int nodeid, int sub)
{
    TVINSERTSTRUCT is;
    char txt[100];

    memset (&is, 0, sizeof (is));
    is.hInsertAfter = TVI_LAST;
    is.hParent = parent;
    is.itemex.mask = TVIF_TEXT | TVIF_PARAM | TVIF_STATE | TVIF_IMAGE | TVIF_SELECTEDIMAGE;
    WIN32GUI_LoadUIString (nameid, txt, sizeof (txt));
    is.itemex.pszText = txt;
    is.itemex.lParam = (LPARAM)(nodeid | (sub << 16));
    is.itemex.iImage = C_PAGES;
    is.itemex.iSelectedImage = C_PAGES;
    is.itemex.state = TVIS_BOLD | TVIS_EXPANDED;
    is.itemex.stateMask = TVIS_BOLD | TVIS_EXPANDED;
    return TreeView_InsertItem (TVhDlg, &is);
}

static void CreateNode (HWND TVhDlg, int page, HTREEITEM parent)
{
    TVINSERTSTRUCT is;
    struct GUIPAGE *p;

    if (page < 0)
	return;
    p = &ppage[page];
    memset (&is, 0, sizeof (is));
    is.hInsertAfter = TVI_LAST;
    is.hParent = parent;
    is.itemex.mask = TVIF_TEXT | TVIF_PARAM | TVIF_STATE | TVIF_IMAGE | TVIF_SELECTEDIMAGE;
    is.itemex.pszText = (char*)p->title;
    is.itemex.lParam = (LPARAM)p->idx;
    is.itemex.iImage = p->himg;
    is.itemex.iSelectedImage = is.itemex.iImage;
    p->tv = TreeView_InsertItem (TVhDlg, &is);
}
#define CN(page) CreateNode(TVhDlg, page, p);

static void createTreeView (HWND hDlg, int currentpage)
{
    HWND TVhDlg;
    int i;
    HIMAGELIST himl;
    HTREEITEM p, root;

    himl = ImageList_Create (16, 16, ILC_COLOR8 | ILC_MASK, C_PAGES + 1, 0);
    if (himl) {
	HICON icon;
	for (i = 0; i < C_PAGES; i++) {
	    icon = LoadIcon (hInst, (LPCSTR)ppage[i].icon);
	    ppage[i].himg = ImageList_AddIcon (himl, icon);
	}
	icon = LoadIcon (hInst, MAKEINTRESOURCE (IDI_ROOT));
	ImageList_AddIcon (himl, icon);
    }
    TVhDlg = GetDlgItem(hDlg, IDC_PANELTREE);
    TreeView_SetImageList (TVhDlg, himl, TVSIL_NORMAL);

    p = root = CreateFolderNode (TVhDlg, IDS_TREEVIEW_SETTINGS, NULL, ABOUT_ID, 0);
    CN(ABOUT_ID);
    CN(PATHS_ID);
    CN(QUICKSTART_ID);
    CN(LOADSAVE_ID);
#if FRONTEND == 1
    CN(FRONTEND_ID);
#endif

    p = CreateFolderNode (TVhDlg, IDS_TREEVIEW_HARDWARE, root, LOADSAVE_ID, CONFIG_TYPE_HARDWARE);
    CN(CPU_ID);
    CN(CHIPSET_ID);
    CN(CHIPSET2_ID);
    CN(KICKSTART_ID);
    CN(MEMORY_ID);
    CN(FLOPPY_ID);
    CN(HARDDISK_ID);

    p = CreateFolderNode (TVhDlg, IDS_TREEVIEW_HOST, root, LOADSAVE_ID, CONFIG_TYPE_HOST);
    CN(DISPLAY_ID);
    CN(SOUND_ID);
    CN(PORTS_ID);
    CN(INPUT_ID);
    CN(AVIOUTPUT_ID);
    CN(HW3D_ID);
    CN(DISK_ID);
    CN(MISC1_ID);
    CN(MISC2_ID);

    TreeView_SelectItem (TVhDlg, ppage[currentpage].tv);
}

static int dialog_x_offset, dialog_y_offset;

static void centerWindow (HWND hDlg)
{
    RECT rc, rcDlg, rcOwner;
    HWND owner = GetParent(hDlg);
    LONG x = 0, y = 0;
    POINT pt1, pt2;

    if (owner == NULL)
	owner = GetDesktopWindow();
    if (isfullscreen () <= 0) {
	regqueryint (NULL, "GUIPosX", &x);
	regqueryint (NULL, "GUIPosY", &y);
    } else {
	GetWindowRect (owner, &rcOwner);
	GetWindowRect (hDlg, &rcDlg);
	CopyRect (&rc, &rcOwner);
	OffsetRect(&rcDlg, -rcDlg.left, -rcDlg.top);
	OffsetRect(&rc, -rc.left, -rc.top);
	OffsetRect(&rc, -rcDlg.right, -rcDlg.bottom);
	x = rcOwner.left + (rc.right / 2);
	y = rcOwner.top + (rc.bottom / 2);
    }
    SetForegroundWindow (hDlg);
    pt1.x = x;
    pt1.y = y;
    pt2.x = x + 16;
    pt2.y = y + GetSystemMetrics (SM_CYMENU) + GetSystemMetrics (SM_CYBORDER);
    if (MonitorFromPoint (pt1, MONITOR_DEFAULTTONULL) == NULL || MonitorFromPoint (pt2, MONITOR_DEFAULTTONULL) == NULL) {
	if (isfullscreen() > 0) {
	    x = 0;
	    y = 0;
	} else {
	    x = 16;
	    y = 16;
	}
    }
    dialog_x_offset = x;
    dialog_y_offset = y;
    SetWindowPos (hDlg,  HWND_TOP, x, y, 0, 0, SWP_NOSIZE);
}

static int do_filesys_insert (const char *root)
{
    if (filesys_insert (-1, NULL, root, 0, 0) == 0)
	return filesys_media_change (root, 2, NULL);
    return 1;
}

int dragdrop (HWND hDlg, HDROP hd, struct uae_prefs *prefs, int	currentpage)
{
    int cnt, i, drv, firstdrv, list;
    char file[MAX_DPATH];
    int dfxtext[] = { IDC_DF0TEXT, IDC_DF0TEXTQ, IDC_DF1TEXT, IDC_DF1TEXTQ, IDC_DF2TEXT, -1, IDC_DF3TEXT, -1 };
    POINT pt;
    RECT r, r2;
    int ret = 0;
    DWORD flags;

    DragQueryPoint (hd, &pt);
    pt.y += GetSystemMetrics (SM_CYMENU) + GetSystemMetrics (SM_CYBORDER);
    cnt = DragQueryFile (hd, 0xffffffff, NULL, 0);
    if (!cnt)
	return 0;
    drv = 0;
    if (currentpage < 0) {
	GetClientRect(hMainWnd, &r2);
	GetClientRect(hStatusWnd, &r);
	if (pt.y >= r2.bottom && pt.y < r2.bottom + r.bottom) {
	    if (pt.x >= window_led_drives && pt.x < window_led_drives_end && window_led_drives > 0) {
		drv = pt.x - window_led_drives;
		drv /= (window_led_drives_end - window_led_drives) / 4;
		if (drv < 0 || drv > 3)
		    drv = 0;
	    }
	}
    } else if (currentpage == FLOPPY_ID || currentpage == QUICKSTART_ID) {
	for (i = 0; i < 4; i++) {
	    int id = dfxtext[i * 2 + (currentpage == QUICKSTART_ID ? 1 : 0)];
	    if (workprefs.dfxtype[i] >= 0 && id >= 0) {
		if (GetPanelRect (GetDlgItem (panelDlg, id), &r)) {
		    if (PtInRect (&r, pt)) {
			drv = i;
			break;
		    }
		}
	    }
	}
    }
    firstdrv = drv;
    for (i = 0; i < cnt; i++) {
	struct romdata *rd = NULL;
	struct zfile *z;
	int type = -1;

	DragQueryFile (hd, i, file, sizeof (file));
	flags = GetFileAttributes(file);
	if (flags & FILE_ATTRIBUTE_DIRECTORY)
	    type = ZFILE_HDF;
	if (type < 0) {
	    z = zfile_fopen (file, "rb");
	    if (z) {
		type = zfile_gettype (z);
		rd = getromdatabyzfile (z);
		zfile_fclose (z);
		z = NULL;
	    }
	}

	if (currentpage < 0 && i == 0) {
	    if (do_filesys_insert (file))
		continue;
	}

	switch (type)
	{
	    case  ZFILE_DISKIMAGE:
		if (currentpage == DISK_ID) {
		    list = 0;
		    while (list < MAX_SPARE_DRIVES) {
			if (!strcasecmp (prefs->dfxlist[list], file))
			    break;
			list++;
		    }
		    if (list == MAX_SPARE_DRIVES) {
			list = 0;
			while (list < MAX_SPARE_DRIVES) {
			    if (!prefs->dfxlist[list][0]) {
				strcpy (prefs->dfxlist[list], file);
				break;
			    }
			    list++;
			}
		    }
		} else if (currentpage == HARDDISK_ID) {
		    add_filesys_config (&workprefs, -1, NULL, "", file, 0,
			0, 0, 0, 0, 0, NULL, 0, 0);
		} else {
		    strcpy (workprefs.df[drv], file);
		    disk_insert (drv, workprefs.df[drv]);
		    drv++;
		    if (drv >= (currentpage == QUICKSTART_ID ? 2 : 4))
			drv = 0;
		    if (workprefs.dfxtype[drv] < 0)
			drv = 0;
		    if (drv == firstdrv)
			i = cnt;
		}
	    break;
	    case ZFILE_ROM:
		if (rd) {
		    if (rd->type == ROMTYPE_KICK || rd->type == ROMTYPE_KICKCD32)
			strcpy (prefs->romfile, file);
		    if (rd->type == ROMTYPE_EXTCD32 || rd->type == ROMTYPE_EXTCDTV)
			strcpy (prefs->romextfile, file);
		    if (rd->type == ROMTYPE_AR)
			strcpy (prefs->cartfile, file);
		} else {
		    strcpy (prefs->romfile, file);
		}
	    break;
	    case ZFILE_HDF:
		if (flags & FILE_ATTRIBUTE_DIRECTORY) {
		    if (!full_property_sheet && currentpage < 0)
			do_filesys_insert (file);
		    else
			add_filesys_config (&workprefs, -1, NULL, "", file, 0,
			    0, 0, 0, 0, 0, NULL, 0, 0);
		} else {
		    add_filesys_config (&workprefs, -1, NULL, NULL, file, 0,
			32, 1, 2, 512, 0, NULL, 0, 0);
		}
	    break;
	    case ZFILE_NVR:
		strcpy (prefs->flashfile, file);
	    break;
	    case ZFILE_CONFIGURATION:
		if (target_cfgfile_load (&workprefs, file, 0, 0)) {
		    if (full_property_sheet) {
			inputdevice_updateconfig (&workprefs);
			if (!workprefs.start_gui)
			    ret = 1;
		    } else {
			uae_restart (workprefs.start_gui, file);
			ret = 1;
		    }
		}
	    break;
	    case ZFILE_STATEFILE:
		savestate_state = STATE_DORESTORE;
		strcpy (savestate_fname, file);
		ret = 1;
	    break;
	    default:
		if (currentpage < 0 && !full_property_sheet) {
		    do_filesys_insert (file);
		} else if (currentpage == HARDDISK_ID) {
		    add_filesys_config (&workprefs, -1, NULL, "", file, 0,
			0, 0, 0, 0, 0, NULL, 0, 0);
		} else {
		    rd = scan_arcadia_rom (file, 0);
		    if (rd) {
			if (rd->type == ROMTYPE_ARCADIABIOS)
			    strcpy (prefs->romextfile, file);
			else if (rd->type == ROMTYPE_ARCADIAGAME)
			    strcpy (prefs->cartfile, file);
		    }
		}
	    break;
	}
    }
    DragFinish (hd);
    return ret;
}

static int dialogreturn;
static INT_PTR CALLBACK DialogProc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int recursive = 0;
    static int waitfornext;

    switch(msg)
    {
	case WM_DEVICECHANGE:
	{
	    DEV_BROADCAST_HDR *pBHdr = (DEV_BROADCAST_HDR *)lParam;
	    int doit = 0;
	    if (wParam == DBT_DEVNODES_CHANGED && lParam == 0) {
		if (waitfornext)
		    doit = 1;
	    } else if (pBHdr && pBHdr->dbch_devicetype == DBT_DEVTYP_DEVICEINTERFACE) {
		DEV_BROADCAST_DEVICEINTERFACE *dbd = (DEV_BROADCAST_DEVICEINTERFACE*)lParam;
		write_log ("%s: %s\n", wParam == DBT_DEVICEREMOVECOMPLETE ? "Removed" : "Inserted",
		    dbd->dbcc_name);
		if (wParam == DBT_DEVICEREMOVECOMPLETE)
		    doit = 1;
		else if (wParam == DBT_DEVICEARRIVAL)
		    waitfornext = 1; /* DirectInput enumeration does not yet show the new device.. */
	    }
	    if (doit) {
	        inputdevice_devicechange (&workprefs);
		updatePanel (hDlg, currentpage);
		waitfornext = 0;
	    }
	}
	return TRUE;
	case WM_DESTROY:
	PostQuitMessage (0);
	return TRUE;
	case WM_CLOSE:
        addnotifications (hDlg, 1);
	DestroyWindow(hDlg);
	if (dialogreturn < 0) {
	    dialogreturn = 0;
	    if (allow_quit) {
		quit_program = 1;
		regs.spcflags |= SPCFLAG_BRK;
	    }
	}
	return TRUE;
	case WM_INITDIALOG:
	    waitfornext = 0;
	    guiDlg = hDlg;
	    SendMessage(hDlg, WM_SETICON, ICON_SMALL, (LPARAM)LoadIcon (GetModuleHandle (NULL), MAKEINTRESOURCE(IDI_APPICON)));
	    if (full_property_sheet) {
		char tmp[100];
		WIN32GUI_LoadUIString (IDS_STARTEMULATION, tmp, sizeof (tmp));
		SetWindowText (GetDlgItem (guiDlg, IDOK), tmp);
	    }
	    ShowWindow (GetDlgItem(guiDlg, IDC_RESTARTEMU), full_property_sheet ? SW_HIDE : SW_SHOW);
	    centerWindow (hDlg);
	    createTreeView (hDlg, currentpage);
	    updatePanel (hDlg, currentpage);
	    addnotifications (hDlg, 0);
	return TRUE;
	case WM_DROPFILES:
	    if (dragdrop (hDlg, (HDROP)wParam, (gui_active || full_property_sheet) ? &workprefs : &changed_prefs, currentpage))
		SendMessage (hDlg, WM_COMMAND, IDOK, 0);
	    updatePanel (hDlg, currentpage);
	return FALSE;
	case WM_NOTIFY:
	{
	    switch (((LPNMHDR)lParam)->code)
	    {
		case TVN_SELCHANGING:
		return FALSE;
		case TVN_SELCHANGED:
		{
		    int cp, cf;
		    LPNMTREEVIEW tv = (LPNMTREEVIEW)lParam;
		    cp = (int)(tv->itemNew.lParam & 0xffff);
		    cf = (int)(tv->itemNew.lParam >> 16);
		    if (cp != currentpage || cf != configtype) {
			currentpage = cp;
			configtypepanel = configtype = cf;
			updatePanel (hDlg, currentpage);
		    }
		    return TRUE;
		}
		break;
	    }
	    break;
	}
	case WM_COMMAND:
	{
	    switch (LOWORD(wParam))
	    {
		case IDC_RESETAMIGA:
		    uae_reset (1);
		    SendMessage (hDlg, WM_COMMAND, IDOK, 0);
		    return TRUE;
		case IDC_QUITEMU:
		    uae_quit ();
		    SendMessage (hDlg, WM_COMMAND, IDCANCEL, 0);
		    return TRUE;
		case IDC_RESTARTEMU:
		    uae_restart (-1, NULL);
		    exit_gui(1);
		    return TRUE;
		case IDHELP:
		    if (pHtmlHelp && ppage[currentpage].help)
			HtmlHelp (NULL, help_file, HH_DISPLAY_TOPIC, ppage[currentpage].help);
		    return TRUE;
		case IDOK:
		    updatePanel (hDlg, -1);
		    dialogreturn = 1;
		    DestroyWindow (hDlg);
		    gui_to_prefs ();
		    guiDlg = NULL;
		    return TRUE;
		case IDCANCEL:
		    updatePanel (hDlg, -1);
		    dialogreturn = 0;
		    DestroyWindow (hDlg);
		    if (allow_quit) {
			quit_program = 1;
			regs.spcflags |= SPCFLAG_BRK;
		    }
		    guiDlg = NULL;
		    return TRUE;
	    }
	    break;
	}
    }
    return FALSE;
}

static ACCEL EmptyAccel[] = {
    { FVIRTKEY, VK_UP, 20001 }, { FVIRTKEY, VK_DOWN, 20002 },
    { 0, 0, 0 }
};

struct newresource *getresource(int tmpl)
{
    char rid[10];
    HRSRC hrsrc;
    HGLOBAL res;
    HINSTANCE inst = hUIDLL ? hUIDLL : hInst;
    void *resdata, *newres;
    struct newresource *nr;
    int size;

    sprintf(rid,"#%d", tmpl);
    hrsrc = FindResource (inst, rid, RT_DIALOG);
    if (!hrsrc) {
	inst = hInst;
	hrsrc = FindResource (inst, rid, RT_DIALOG);
    }
    if (!hrsrc)
	return NULL;
    res = LoadResource(inst, hrsrc);
    if (!res)
	return NULL;
    resdata = LockResource(res);
    size = SizeofResource(inst, hrsrc);
    nr = xcalloc (sizeof (struct newresource), 1);
    newres = xmalloc (size);
    memcpy (newres, resdata, size);
    nr->resource = newres;
    nr->size = size;
    nr->tmpl = tmpl;
    nr->inst = inst;
    return nr;
}

INT_PTR CustomDialogBox(int templ, HWND hDlg, DLGPROC proc)
{
    struct newresource *res, *r;
    INT_PTR h = -1;

    res = getresource (templ);
    if (!res)
	return h;
    r = scaleresource (res, hDlg);
    if (r) {
	h = DialogBoxIndirect (r->inst, r->resource, hDlg, proc);
	freescaleresource (r);
    }
    freescaleresource (res);
    return h;
}

HWND CustomCreateDialog (int templ, HWND hDlg, DLGPROC proc)
{
    struct newresource *res, *r;
    HWND h = NULL;

    res = getresource (templ);
    if (!res)
	return h;
    r = scaleresource (res, hDlg);
    if (r) {
	h = CreateDialogIndirect (r->inst, r->resource, hDlg, proc);
	freescaleresource (r);
    }
    freescaleresource (res);
    return h;
}

static int init_page (int tmpl, int icon, int title,
    INT_PTR (CALLBACK FAR *func) (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam), ACCEL *accels, char *help)
{
    LPTSTR lpstrTitle;
    static id = 0;
    int i = -1;
    struct newresource *res;

    res = getresource(tmpl);
    if (!res)
	return -1;
    ppage[id].nres = res;
    ppage[id].icon = MAKEINTRESOURCE (icon);
    lpstrTitle = calloc (1, MAX_DPATH);
    LoadString (hUIDLL, title, lpstrTitle, MAX_DPATH);
    ppage[id].title = lpstrTitle;
    ppage[id].dlgproc = func;
    ppage[id].help = help;
    ppage[id].idx = id;
    ppage[id].accel = NULL;
    if (!accels)
	accels = EmptyAccel;
    while (accels[++i].key);
    ppage[id].accel = CreateAcceleratorTable (accels, i);
    if (tmpl == IDD_FRONTEND)
	ppage[id].fullpanel = TRUE;
    id++;
    return id - 1;
}

static RECT dialog_rect;

static void dialogmousemove(HWND hDlg)
{
    static int newmx, newmy;
    RECT rc;
    POINT pt;
    static POINT pt2;
    int dx, dy;
    int sw, sh;

    if (full_property_sheet || isfullscreen() <= 0)
	return;
    GetCursorPos(&pt);
    if (pt.x == pt2.x && pt.y == pt2.y)
	return;
    sw = WIN32GFX_GetWidth();
    sh = WIN32GFX_GetHeight();
    dx = dialog_x_offset;
    dy = dialog_y_offset;
    GetWindowRect(hDlg, &rc);
    rc.right -= rc.left;
    rc.bottom -= rc.top;
    rc.left = 0;
    rc.top = 0;
    if (rc.right <= sw && rc.bottom <= sh)
	return;
    pt2.x = pt.x;
    pt2.y = pt.y;

    newmx = pt.x;
    newmy = pt.y;

    if (newmx >= sw - 1 && rc.right > sw)
	dx = sw - rc.right;
    if (newmx <= 1)
	dx = 0;
    if (newmy >= sh - 1 && rc.bottom > sh)
	dy = sh - rc.bottom;
    if (newmy <= 1)
	dy = 0;

    if (dx != dialog_x_offset || dy != dialog_y_offset) {
	dialog_x_offset = dx;
	dialog_y_offset = dy;
	SetWindowPos(hDlg, 0, dialog_x_offset, dialog_y_offset, 0, 0,
	    SWP_NOOWNERZORDER | SWP_NOSIZE | SWP_NOACTIVATE | SWP_DEFERERASE);
    }
}

static int GetSettings (int all_options, HWND hwnd)
{
    static int init_called = 0;
    int psresult;
    HWND dhwnd;
    int first = 0;
    static struct newresource *panelresource;
    struct newresource *tres;

    gui_active++;

    full_property_sheet = all_options;
    allow_quit = all_options;
    pguiprefs = &currprefs;
    default_prefs (&workprefs, 0);

    WIN32GUI_LoadUIString (IDS_NONE, szNone, MAX_DPATH);
    prefs_to_gui (&changed_prefs);

    if (!init_called) {
	first = 1;
	panelresource = getresource(IDD_PANEL);
	LOADSAVE_ID = init_page (IDD_LOADSAVE, IDI_CONFIGFILE, IDS_LOADSAVE, LoadSaveDlgProc, NULL, "gui/configurations.htm");
	MEMORY_ID = init_page (IDD_MEMORY, IDI_MEMORY, IDS_MEMORY, MemoryDlgProc, NULL, "gui/ram.htm");
	KICKSTART_ID = init_page (IDD_KICKSTART, IDI_MEMORY, IDS_KICKSTART, KickstartDlgProc, NULL, "gui/rom.htm");
	CPU_ID = init_page (IDD_CPU, IDI_CPU, IDS_CPU, CPUDlgProc, NULL, "gui/cpu.htm");
	DISPLAY_ID = init_page (IDD_DISPLAY, IDI_DISPLAY, IDS_DISPLAY, DisplayDlgProc, NULL, "gui/display.htm");
#if defined (GFXFILTER)
	HW3D_ID = init_page (IDD_FILTER, IDI_DISPLAY, IDS_FILTER, hw3dDlgProc, NULL, "gui/filter.htm");
#endif
	CHIPSET_ID = init_page (IDD_CHIPSET, IDI_CPU, IDS_CHIPSET, ChipsetDlgProc, NULL, "gui/chipset.htm");
	CHIPSET2_ID = init_page (IDD_CHIPSET2, IDI_CPU, IDS_CHIPSET2, ChipsetDlgProc2, NULL, "gui/chipset.htm");
	SOUND_ID = init_page (IDD_SOUND, IDI_SOUND, IDS_SOUND, SoundDlgProc, NULL, "gui/sound.htm");
	FLOPPY_ID = init_page (IDD_FLOPPY, IDI_FLOPPY, IDS_FLOPPY, FloppyDlgProc, NULL, "gui/floppies.htm");
	DISK_ID = init_page (IDD_DISK, IDI_FLOPPY, IDS_DISK, SwapperDlgProc, SwapperAccel, "gui/disk.htm");
#ifdef FILESYS
	HARDDISK_ID = init_page (IDD_HARDDISK, IDI_HARDDISK, IDS_HARDDISK, HarddiskDlgProc, HarddiskAccel, "gui/hard-drives.htm");
#endif
	PORTS_ID = init_page (IDD_PORTS, IDI_PORTS, IDS_PORTS, PortsDlgProc, NULL, "gui/ports.htm");
	INPUT_ID = init_page (IDD_INPUT, IDI_INPUT, IDS_INPUT, InputDlgProc, NULL, "gui/input.htm");
	MISC1_ID = init_page (IDD_MISC1, IDI_MISC1, IDS_MISC1, MiscDlgProc1, NULL, "gui/misc.htm");
	MISC2_ID = init_page (IDD_MISC2, IDI_MISC2, IDS_MISC2, MiscDlgProc2, NULL, "gui/misc2.htm");
#ifdef AVIOUTPUT
	AVIOUTPUT_ID = init_page (IDD_AVIOUTPUT, IDI_AVIOUTPUT, IDS_AVIOUTPUT, AVIOutputDlgProc, NULL, "gui/output.htm");
#endif
	PATHS_ID = init_page (IDD_PATHS, IDI_PATHS, IDS_PATHS, PathsDlgProc, NULL, "gui/paths.htm");
	QUICKSTART_ID = init_page (IDD_QUICKSTART, IDI_QUICKSTART, IDS_QUICKSTART, QuickstartDlgProc, NULL, "gui/quickstart.htm");
	ABOUT_ID = init_page (IDD_ABOUT, IDI_ABOUT, IDS_ABOUT, AboutDlgProc, NULL, NULL);
	FRONTEND_ID = init_page (IDD_FRONTEND, IDI_QUICKSTART, IDS_FRONTEND, AboutDlgProc, NULL, NULL);
	C_PAGES = FRONTEND_ID + 1;
	init_called = 1;
	if (quickstart && !qs_override)
	    currentpage = QUICKSTART_ID;
	else
	    currentpage = LOADSAVE_ID;
    }

    if (all_options || !configstore)
	CreateConfigStore (NULL);

    dialogreturn = -1;
    hAccelTable = NULL;
    DragAcceptFiles(hwnd, TRUE);
    if (first)
	write_log ("Entering GUI idle loop\n");

    scaleresource_setmaxsize(800, 600);
    tres = scaleresource(panelresource, hwnd);
    dhwnd = CreateDialogIndirect (tres->inst, tres->resource, hwnd, DialogProc);
    dialog_rect.top = dialog_rect.left = 0;
    dialog_rect.right = tres->width;
    dialog_rect.bottom = tres->height;
    freescaleresource(tres);
    psresult = 0;
    if (dhwnd != NULL) {
	MSG msg;
	DWORD v;

	setguititle (dhwnd);
	ShowWindow (dhwnd, SW_SHOW);
	MapDialogRect(dhwnd, &dialog_rect);

	for (;;) {
	    HANDLE IPChandle;
	    IPChandle = geteventhandleIPC();
	    if (IPChandle != INVALID_HANDLE_VALUE) {
		MsgWaitForMultipleObjects (1, &IPChandle, FALSE, INFINITE, QS_ALLINPUT);
		while (checkIPC(&workprefs));
	    } else {
		WaitMessage();
	    }
	    dialogmousemove(dhwnd);
	    while ((v = PeekMessage (&msg, NULL, 0, 0, PM_REMOVE))) {
		if (dialogreturn >= 0)
		    break;
		if (v == -1)
		    continue;
		if (!IsWindow (dhwnd))
		    continue;
		if (hAccelTable && panelDlg) {
		    if (TranslateAccelerator (panelDlg, hAccelTable, &msg))
			continue;
		}
		if (IsDialogMessage (dhwnd, &msg))
		    continue;
		TranslateMessage (&msg);
		DispatchMessage (&msg);
	    }
	    if (dialogreturn >= 0)
		break;
	}
	psresult = dialogreturn;
    }

    if (quit_program)
	psresult = -2;
    else if (qs_request_reset && quickstart)
	uae_reset (qs_request_reset == 2 ? 1 : 0);

    qs_request_reset = 0;
    full_property_sheet = 0;
    gui_active--;
    return psresult;
}

int gui_init (void)
{
    int ret;

    read_rom_list();
    for (;;) {
	ret = GetSettings(1, currprefs.win32_notaskbarbutton ? hHiddenWnd : NULL);
	if (!restart_requested)
	    break;
	restart_requested = 0;
    }
    if (ret > 0) {
#ifdef AVIOUTPUT
	AVIOutput_Begin ();
#endif
    }
    return ret;
}

int gui_update (void)
{
    return 1;
}

void gui_exit (void)
{
    int i;

    for (i = 0; i < C_PAGES; i++) {
	if (ppage[i].accel)
	    DestroyAcceleratorTable (ppage[i].accel);
	ppage[i].accel = NULL;
    }
    FreeConfigStore ();
#ifdef PARALLEL_PORT
    closeprinter(); // Bernd Roesch
#endif
}

extern HWND hStatusWnd;
struct gui_info gui_data;

void check_prefs_changed_gui( void )
{
}

void gui_hd_led (int unitnum, int led)
{
    static int resetcounter;

    int old = gui_data.hd;
    if (led == 0) {
	resetcounter--;
	if (resetcounter > 0)
	    return;
    }
#ifdef RETROPLATFORM
    rp_hd_activity (unitnum, led);
#endif
    gui_data.hd = led;
    resetcounter = 6;
    if (old != gui_data.hd)
	gui_led (5, gui_data.hd);
}

void gui_cd_led (int unitnum, int led)
{
    static int resetcounter;

    int old = gui_data.cd;
    if (led == 0) {
	resetcounter--;
	if (resetcounter > 0)
	    return;
    }
#ifdef RETROPLATFORM
    rp_cd_activity (unitnum, led);
#endif
    gui_data.cd = led;
    resetcounter = 6;
    if (old != gui_data.cd)
	gui_led (6, gui_data.cd);
}

void gui_fps (int fps, int idle)
{
    gui_data.fps = fps;
    gui_data.idle = idle;
    gui_led (7, 0);
    gui_led (8, 0);
    gui_led (9, gui_data.sndbuf_status > 1 || gui_data.sndbuf_status < 0);
}

void gui_led (int led, int on)
{
    WORD type;
    static char drive_text[NUM_LEDS * 16];
    static char dfx[4][300];
    char *ptr, *tt, *p;
    int pos = -1, j;

    indicator_leds (led, on);
#ifdef LOGITECHLCD
    lcd_update (led, on);
#endif
#ifdef RETROPLATFORM
    rp_update_leds (led, on);
#endif
    if (!hStatusWnd)
	return;
    tt = NULL;
    if (led >= 1 && led <= 4) {
	pos = 6 + (led - 1);
	ptr = drive_text + pos * 16;
	if (gui_data.drive_disabled[led - 1])
	    strcpy (ptr, "");
	else
	    sprintf (ptr , "%02d  .", gui_data.drive_track[led - 1]);
	p = gui_data.df[led - 1];
	j = strlen (p) - 1;
	if (j < 0)
	    j = 0;
	while (j > 0) {
	    if (p[j - 1] == '\\' || p[j - 1] == '/')
		break;
	    j--;
	}
	tt = dfx[led - 1];
	tt[0] = 0;
	if (strlen (p + j) > 0)
	    sprintf (tt, "%s (CRC=%08.8X)", p + j, gui_data.crc32[led - 1]);
    } else if (led == 0) {
	pos = 3;
	ptr = strcpy (drive_text + pos * 16, "Power");
    } else if (led == 5) {
	pos = 4;
	ptr = strcpy (drive_text + pos * 16, "HD");
    } else if (led == 6) {
	pos = 5;
	ptr = strcpy (drive_text + pos * 16, "CD");
    } else if (led == 7) {
	pos = 2;
	ptr = drive_text + pos * 16;
	sprintf(ptr, "FPS: %.1f", (double)(gui_data.fps  / 10.0));
    } else if (led == 8) {
	pos = 1;
	ptr = drive_text + pos * 16;
	sprintf(ptr, "CPU: %.0f%%", (double)((gui_data.idle) / 10.0));
    } else if (led == 9) {
	pos = 0;
	ptr = drive_text + pos * 16;
	if (gui_data.sndbuf_status < 3) {
	    sprintf(ptr, "SND: %+.0f%%", (double)((gui_data.sndbuf) / 10.0));
	} else {
	    strcpy (ptr, "SND: -");
	    on = 0;
	}
    }
    if (on)
	type = SBT_POPOUT;
    else
	type = 0;
    if (pos >= 0) {
	PostMessage (hStatusWnd, SB_SETTEXT, (WPARAM) ((pos + 1) | type), (LPARAM) ptr);
	if (tt != NULL)
	    PostMessage (hStatusWnd, SB_SETTIPTEXT, (WPARAM) (pos + 1), (LPARAM) tt);
    }
}

void gui_filename (int num, const char *name)
{
}


static int fsdialog (HWND *hwnd, DWORD *flags)
{
    HRESULT hr;

    if (gui_active) {
	*hwnd = guiDlg;
	*flags |= MB_SETFOREGROUND;
	return 0;
    }
    *hwnd = hAmigaWnd;
    if (isfullscreen () <= 0)
	return 0;
    hr = DirectDraw_FlipToGDISurface();
    if (FAILED(hr))
	write_log ("FlipToGDISurface failed, %s\n", DXError (hr));
    *flags |= MB_SETFOREGROUND;
    *flags |= MB_TOPMOST;
    return 0;
/*
    HRESULT hr;
    hr = DirectDraw_FlipToGDISurface();
    if (FAILED(hr)) {
	write_log ("FlipToGDISurface failed, %s\n", DXError (hr));
	return 0;
    }
    *hwnd = NULL;
    return 1;
*/
}

int gui_message_multibutton (int flags, const char *format,...)
{
    char msg[2048];
    char szTitle[MAX_DPATH];
    va_list parms;
    int flipflop = 0;
    int fullscreen = 0;
    int focuso = focus;
    int mbflags, ret;
    HWND hwnd;

    mbflags = MB_ICONWARNING | MB_TASKMODAL;
    if (flags == 0)
	mbflags |= MB_OK;
    else if (flags == 1)
	mbflags |= MB_YESNO;
    else if (flags == 2)
	mbflags |= MB_YESNOCANCEL;

    pause_sound ();
    flipflop = fsdialog (&hwnd, &mbflags);
    if (flipflop)
	ShowWindow (hAmigaWnd, SW_MINIMIZE);

    va_start (parms, format);
    vsprintf( msg, format, parms );
    va_end (parms);
    write_log ( msg );
    if (msg[strlen(msg)-1]!='\n')
	write_log ("\n");

    WIN32GUI_LoadUIString (IDS_ERRORTITLE, szTitle, MAX_DPATH);

    ret = MessageBox (hwnd, msg, szTitle, mbflags);

    if (flipflop)
	ShowWindow (hAmigaWnd, SW_RESTORE);

    resume_sound ();
    setmouseactive (focuso);
    if (ret == IDOK)
	return 0;
    if (ret == IDYES)
	return 1;
    if (ret == IDNO)
	return 2;
    if (ret == IDCANCEL)
	return -1;
    return 0;
}

void gui_message (const char *format,...)
{
    char msg[2048];
    char szTitle[MAX_DPATH];
    va_list parms;
    int flipflop = 0;
    int fullscreen = 0;
    int focuso = focus;
    DWORD flags = MB_OK | MB_TASKMODAL;
    HWND hwnd;

    va_start (parms, format);
    vsprintf(msg, format, parms);
    va_end (parms);
    if (full_property_sheet) {
	pre_gui_message (msg);
	return;
    }
    pause_sound ();
    flipflop = fsdialog (&hwnd, &flags);
    if (flipflop)
	ShowWindow (hAmigaWnd, SW_MINIMIZE);

    write_log (msg);
    if (msg[strlen(msg) - 1] != '\n')
	write_log ("\n");

    WIN32GUI_LoadUIString (IDS_ERRORTITLE, szTitle, MAX_DPATH);

    if (!MessageBox (hwnd, msg, szTitle, flags))
	write_log ("MessageBox(%s) failed, err=%d\n", msg, GetLastError());

    if (flipflop)
	ShowWindow (hAmigaWnd, SW_RESTORE);
    resume_sound();
    setmouseactive (focuso);
}

void gui_message_id (int id)
{
    char msg[MAX_DPATH];
    WIN32GUI_LoadUIString (id, msg, sizeof (msg));
    gui_message (msg);
}

void pre_gui_message (const char *format,...)
{
    char msg[2048];
    char szTitle[MAX_DPATH];
    va_list parms;

    va_start (parms, format);
    vsprintf (msg, format, parms);
    va_end (parms);
    write_log (msg);
    if (msg[strlen(msg)-1]!='\n')
	write_log ("\n");

    WIN32GUI_LoadUIString (IDS_ERRORTITLE, szTitle, MAX_DPATH);
    strcat (szTitle, BetaStr);
    MessageBox (guiDlg, msg, szTitle, MB_OK | MB_TASKMODAL | MB_SETFOREGROUND);

}

static int transla[] = {
    NUMSG_NEEDEXT2, IDS_NUMSG_NEEDEXT2,
    NUMSG_NOROMKEY,IDS_NUMSG_NOROMKEY,
    NUMSG_NOROM,IDS_NUMSG_NOROM,
    NUMSG_KSROMCRCERROR,IDS_NUMSG_KSROMCRCERROR,
    NUMSG_KSROMREADERROR,IDS_NUMSG_KSROMREADERROR,
    NUMSG_NOEXTROM,IDS_NUMSG_NOEXTROM,
    NUMSG_MODRIP_NOTFOUND,IDS_NUMSG_MODRIP_NOTFOUND,
    NUMSG_MODRIP_FINISHED,IDS_NUMSG_MODRIP_FINISHED,
    NUMSG_MODRIP_SAVE,IDS_NUMSG_MODRIP_SAVE,
    NUMSG_KS68EC020,IDS_NUMSG_KS68EC020,
    NUMSG_KS68020,IDS_NUMSG_KS68020,
    NUMSG_KS68030,IDS_NUMSG_KS68030,
    NUMSG_ROMNEED,IDS_NUMSG_ROMNEED,
    NUMSG_EXPROMNEED,IDS_NUMSG_EXPROMNEED,
    NUMSG_NOZLIB,IDS_NUMSG_NOZLIB,
    NUMSG_STATEHD,IDS_NUMSG_STATEHD,
    NUMSG_OLDCAPS, IDS_NUMSG_OLDCAPS,
    NUMSG_NOCAPS, IDS_NUMSG_NOCAPS,
    NUMSG_KICKREP, IDS_NUMSG_KICKREP,
    NUMSG_KICKREPNO, IDS_NUMSG_KICKREPNO,
    -1
};

static int gettranslation (int msg)
{
    int i;

    i = 0;
    while (transla[i] >= 0) {
	if (transla[i] == msg)
	    return transla[i + 1];
	i += 2;
    }
    return -1;
}

void notify_user (int msg)
{
    char tmp[MAX_DPATH];
    int c = 0;

    c = gettranslation (msg);
    if (c < 0)
	return;
    WIN32GUI_LoadUIString (c, tmp, MAX_DPATH);
    gui_message (tmp);
}

void notify_user_parms (int msg, const char *parms, ...)
{
    char msgtxt[MAX_DPATH];
    char tmp[MAX_DPATH];
    int c = 0;
    va_list parms2;

    c = gettranslation (msg);
    if (c < 0)
	return;
    WIN32GUI_LoadUIString (c, tmp, MAX_DPATH);
    va_start (parms2, parms);
    vsprintf (msgtxt, tmp, parms2);
    gui_message (msgtxt);
    va_end (parms);
}


int translate_message (int msg,	char *out)
{
    msg = gettranslation (msg);
    out[0] = 0;
    if (msg < 0)
	return 0;
    WIN32GUI_LoadUIString (msg, out, MAX_DPATH);
    return 1;
}

void gui_lock (void)
{
}

void gui_unlock (void)
{
}
